                             (* SLIB.PAS *)
Unit slib;

interface

uses classes, s_constants;

type
(* This type declaration has the purpose to pass functions of this type
   as arguments to procedures *)
     TRealF = function(const th: pvec): Real;// of object;
     TRealReal = function(const x: real): Real;// of object;

function Root(x: real): real;

function log(x: real): real;

function logit(x:real):real;

function expo(x:real):real;

function expit(x:real):real;

function logtrinomial(n,a,b,c: integer): real;

function logbinomial(n,s: integer): real;

function d(a,b: LongInt):real;

function divi(a,b: real): real;

function signum(a: boolean): integer;

function sgn(x: real): integer; overload; (*[s]*)

function sgn(x: shortint): integer; overload; (*[s]*)

function sgn(x: integer): integer; overload; (*[s]*)

function rchoose(x,n: integer): real;

function bytesquare(a: byte): longint;
(* Note that sqr(a) goes wrong for bytes a >= 16 *)

function logfactorial(a: integer): real;

procedure ComLineString(var s: shortstring);
(* if there is a parameter string,
   s is set to the value of the first parameter string *)

procedure Outputfiles(var apnd: boolean; writedate: boolean;
               var pn: shortstring; var f, lf: TFileStream);
(*  pn is the project name;
   output files named pn.out and pn.log are attached;
   the value of apnd determines whether output is appended
   to existing files of thes names;
   The file gets a header with filename and date. *)

procedure Outpfile(fn: shortstring; var f: textfile);
(* Makes an outputfile f with filename fn;
   The file gets a header with filename and date. *)

procedure Protocol(name,tekst : string);
(* writes text 'tekst' to file 'name', creating this file if necessary *)

procedure pprod(p: byte; var a: pmatrix; var b,c: pvec);
(* c := a*b as a px1 vector *)

procedure product(p: byte; var a,b,c: pmatrix);
(* c := a*b *)

function veclength(p: byte; var a: pvec; var s: pmatrix): real;
(* length := a'*inv(s)*a *)

PROCEDURE pinverse(p: Byte; var a: pMatrix; var y: pMatrix;
                   var errtext: shortstring);
(* calculates matrix inverse: y := inverse(a),
   where dimension is taken to be p. *)

PROCEDURE jacobi(n: integer; VAR a: pmatrix; VAR d: pvec;
       VAR v: pmatrix; VAR nrot: integer);
(* Computes all eigenvalues and eigenvectors of a real symmetric matrix A
   of size N by N.
   On output, elements of A above the diagonal are destroyed.
   D returns the eigenvalues of A.
   The columns of V return the normalized eigenvectors of A.
   NROT returns the number of Jacobi rotations required.
   Results can be sorted by EIGSRT. *)

PROCEDURE eigsrt(n: integer; VAR d: pvec; VAR v: pmatrix);
(* Sorts output of Jacobi with eigenvalues in ascending order,
   and corresponding rearrangement of eigenvectors. *)


function max(a,b: real):real;
(* maximum of a and b *)

function normalcdf(x: real): real;
(* Cumulative distribution function of chi-squared distribution;
   from Press, Teukolsky & Vetterling. *)

function chi2(df: integer; x: real): real;
(* Cumulative distribution function of chi-squared distribution *)

procedure quicksort(var a: array of integer; Lo,Hi: integer);
(* sorts vector a, elements Lo through Hi, from small to high *)

procedure quicksortr(var a: array of integer; var b: array of shortint;
                     Lo,Hi: integer);
(* sorts vector a, elements Lo through Hi, from small to high, )
   and sort b accordingly. *)
   
function st(i: integer): shortstring;
(* The string containing the number i . *)

function one(a: boolean): integer;
(* 1 if a is true, 0 otherwise *)

function st2(r: real): shortstring;
(* The string containing the number r . *)

function AlignString(s: shortstring; len: integer): shortstring;
(* returns a string of length len by either cutting down or adding spaces *)

function RightAlignString(s: shortstring; len: integer): shortstring;
(* returns a string of length len by either cutting down or adding spaces *)

procedure NiceString(x: real; var s: shortstring);

procedure NiceString1(x: real; var s: shortstring);

procedure NiceString2(y: real; var s: shortstring);

function makestring(n: integer): shortstring;
(* converts an integer number into a string *)

procedure Heading(level: byte; var f: TFileStream; s: shortstring);

procedure Headingtext(level: byte; var f: textfile; s: shortstring);

procedure parse(var s: shortstring);
(* takes away out of s all characters with ASCII numbers less than 33 *)

function parselo(var s: shortstring; mini: integer): shortstring;
(* takes away out of s all characters with ASCII numbers less than mini *)

function firstnumber(var s: string): integer;
(* converts the leftmost substring of consecutive digits 0..9 in s
   into a nonnegative integer; -1 if this is not contained in s *)

(* Help for Latex users: a picture using Pictex. *)
procedure StartPictexPlot(xmin,xmax,ymin,ymax: real; var lf: text);
procedure EndPictexPlot(var lf: text);

procedure WriteERGMInfile(nn: integer; ppname, dname: shortstring);
(* Writes .in file for project ppname with only 1 network data file dname
    and nn actors. *)

procedure Histogram(a: array of real; left,right,width: integer; var f: textfile);

procedure FsHistogram(a: array of integer;
               left,right,width,circ,exx: integer; var f: TFileStream);

procedure NRMinimize(p, mode: byte; var theta: pvec; var Cov: pmatrix;
                     var cf: textfile;  funct: TRealF);
(* Maximization of function funct *)
(* Mode 0: one Quasi-Newton Raphson step;
   Mode 1: iterated Quasi-Newton Raphson steps,
           but in each step, if NR is not indicated,
           a steepest descent step is made;
   Mode 2: iterated steepest descent steps;
   Mode 3: first multi steepest descent steps,
           then NR steps;
           in each step where NR is not indicated,
           multi steepest descent steps are made. *)
 (* theta is the input (initial) and output parameter vector *)
 (* If funct is a minus loglikelihood,
    then Cov is the estimated covariance matrix *)

function sumsquares(const th: pvec): Real;
(* example declaration for function Minimized by NRMinimize *)


PROCEDURE Plot(var outf: TFileStream;
               var XD, YD: array of real; (* plotted points *)
               fu: TRealReal;
               const M: Integer (* number of points *);
               const xlabel, ylabel: shortstring (* labels for X and Y axes *) );
(* Makes a plot of the points (xd, yd) and function fu;
   range of coordinates is determined by observed xd and yd;
   if there is no fu, use fu = nil *)
(* Adapted from Numerical Recipes procedure SCRSHO by T.S. *)

procedure reportp(var outf: textfile; p: real; dec: integer);
(* reports p-value p with dec decimals *)

procedure writevec(var outf: TFileStream; p: integer; x: array of real);
(* writes x[lo] to x[lo+p-1] to outf in format ' ',x[i]:9:4, with writeln after 10 numbers *)

procedure wwrite(var outf: TFileStream; x: string);
(* writes x to outf *)

procedure wwriteln(var outf: TFileStream; x: string);
(* writes x to outf with eoln *)

implementation

uses SysUtils;

function Root(x: real): real;
begin
  if abs(x) > 1e-36 then Root := sqrt(abs(x)) else Root := 1e-18;
end; (* Root *)

function log(x: real): real;
begin
  if ((x > 1e-300) and (x < 1e300)) then
    log := ln(x)
   else if (x <= 1e-300) then
    log := ln(1e-300) else
    log := ln(1e300);
end;

function logit(x:real):real;
begin
  logit := log(x/(1.0-x))
end;

function expo(x:real):real;
begin
  if (x < -700.0) then expo := exp(-700.0) else
  begin if (x > 700.0) then expo := exp(700.0) else expo := exp(x) end;
(* In Delphi 6, the Real type ranges to wider than 1e-308 *)
end;

function expit(x:real):real;
var y: real;
begin
  y := expo(x);
  expit := y/(1.0+y);
end;

function logtrinomial(n,a,b,c: integer): real;
var i: integer;
    x: real;
begin
  x := 0.0;
  for i:= a+1 to n do x := x + log(i);
  for i:= 1 to b do x := x - log(i);
  for i:= 1 to c do x := x - log(i);
  logtrinomial := x;
end;

function logbinomial(n,s: integer): real;
var i: integer;
    x: real;
begin
  x := 0.0;
  for i:= 1 to s do x := x + log(n-i+1) - log(i);
  logbinomial := x;
end;


function d(a,b: LongInt):real;
begin
  if (b >= 1) then d := a/b else d := 0.0;
end;

function divi(a,b: real): real;
begin
  if (abs(b) > 1e-12) then divi := a/b else divi := 0.0;
end;

function signum(a: boolean): integer;
begin
signum:=-1;
if a then signum:=1;
end;

function sgn(x: real): integer; overload;
begin
sgn:=0;
if (x > 0) then sgn:=1 else
 if (x < 0) then sgn := -1;
end;

function sgn(x: shortint): integer; overload;
begin
sgn:=0;
if (x > 0) then sgn:=1 else
 if (x < 0) then sgn := -1;
end;

function sgn(x: integer): integer; overload;
begin
sgn:=0;
if (x > 0) then sgn:=1 else
 if (x < 0) then sgn := -1;
end;

function rchoose(x,n: integer): real;
var i: integer; c: real;
begin
  c := 1.0;
  for i := n-1 downto 0 do c:= c*(x-i)/(i+1);
  rchoose := c;
end;

function bytesquare(a: byte): longint;
var aa: longint;
begin
  aa := a;
  bytesquare := sqr(aa);
end;

function logfactorial(a: integer): real;
var b: integer;
begin
  result := 0.0;
  b := a;
  while (b > 1) do begin
   result := result + log(b);
   dec(b);
  end;
end;

procedure ComLineString(var s: shortstring);
var s_exist: boolean;
begin
  s_exist := (ParamCount > 0);
  if s_exist then begin
   s := ParamStr(1);
  end else  s := '';
end; (* ComLineString *)

procedure Outputfiles(var apnd: boolean; writedate: boolean;
                      var pn: shortstring; var f, lf: TFileStream);
var i: integer;
     fno, fnl: shortstring;
     appf, apfl: boolean;
     x : string;

begin
  fno := pn+'.out';
  fnl := pn+'.log';
  appf := apnd;
  apfl := apnd;
  if (not FileExists(fno)) then begin
    appf  := false;
    f := TFileStream.Create(fno, fmCreate);
    f.Free;
    f := TFileStream.Create(fno, fmOpenWrite or fmShareDenyWrite);
  end
  else begin
    f := TFileStream.Create(fno,fmOpenReadWrite or fmShareDenyWrite);
  end;

  if (not FileExists(fnl)) then begin
    apfl := false;
    lf := TFileStream.Create(fnl, fmCreate);
    lf.Free;
    lf := TFileStream.Create(fnl, fmOpenWrite or fmShareDenyWrite);
  end
  else begin
    lf := TFileStream.Create(fnl,fmOpenReadWrite or fmShareDenyWrite);

  end;
//  assignfile(f,fno);
//  assignfile(lf,fnl);


  if apfl
  then begin
    lf.Seek(0,sofromEnd);
//    append(lf);
//    Heading(1,lf,'Attach output files.');
  end
  else begin
//   rewrite(lf);
//    write(lf,'                        ************');
//    for i:=length(fnl) downto 1 do write(lf,'*');
//    writeln(lf);
//    writeln(lf,'                              ',fnl);
//    write(lf,'                        ************');
//    for i:=length(fnl) downto 1 do write(lf,'*');
//    writeln(lf);
//    writeln(lf);
//    writeln(lf,'Filename is ',fnl,'.');
//    writeln(lf);
//    writeln(lf,'This file contains a log of the activities of the SIENA program');
//    writeln(lf,'as well as secondary output not reported in ',fno,'.');
//    writeln(lf);
//    writeln(lf,'Date and time: '+DateTimeToStr(Now));
//    writeln(lf);
    lf.size := 0;
    x := '                        ************';
    lf.write(pchar(x)^,length(x));
    for i:=length(fnl) downto 1 do lf.write(pchar(string('*'))^,1);
    lf.write(pchar(eol)^,length(eol));
    x := '                              '+fnl+eol;
    lf.write(pchar(x)^,length(x));
    x := '                        ************';
    lf.write(pchar(x)^,length(x));
    for i:=length(fnl) downto 1 do lf.write(pchar(string('*'))^,1);
    lf.write(pchar(eol)^,length(eol));
    lf.write(pchar(eol)^,length(eol));
    x := 'Filename is '+fnl+'.'+eol;
    lf.write(pchar(x)^,length(x));
    lf.write(pchar(eol)^,length(eol));
    x := 'This file contains a log of the activities of the SIENA program'+eol;
    lf.write(pchar(x)^,length(x));
    x := 'as well as secondary output not reported in '+fno+'.'+eol;
    lf.write(pchar(x)^,length(x));
    lf.write(pchar(eol)^,length(eol));
    x := 'Date and time: '+DateTimeToStr(Now)+eol;
    lf.write(pchar(x)^,length(x));
    lf.write(pchar(eol)^,length(eol));
    Heading(1,lf,'New project initialisation.');
  end;

  if appf
  then begin
    f.Seek(0,sofromEnd);
//    append(f);
//    writeln(lf,'Output is appended to file ',fno,'.');
    if writedate
    then begin
//      writeln(f);
//      writeln(f,'-----------------------------------');
//      writeln(f,'New analysis started.');
      f.write(pchar(eol)^,length(eol));
      x := '-----------------------------------'+eol;
      f.write(pchar(x)^,length(x));
      x := 'New analysis started.'+eol;
      f.write(pchar(x)^,length(x));
    end
  end
  else begin
//    rewrite(f);  ? do I need to position at start?
//    write(f,'                        ************');
//    for i:=length(fno) downto 1 do write(f,'*');
//    writeln(f);
//    writeln(f,'                              ',fno);
//    write(f,'                        ************');
//    for i:=length(fno) downto 1 do write(f,'*');
//    writeln(f);
//    writeln(f);
//    writeln(f,'Filename is ',fno,'.');
//    writeln(f);
//    writeln(f,'This file contains primary output for SIENA project <<',pn,'>>.');
//    writeln(f,'File ',fnl,' contains secondary output.');
//    writeln(f);
//    writeln(lf,'File ',fno,' created for primary output.');
    f.size := 0;
    x := '                        ************';
    f.Write(pchar(x)^,length(x));
    for i:=length(fno) downto 1 do f.write(pchar(string('*'))^,1);
    f.write(pchar(eol)^,length(eol));
    x := '                              '+fno+eol;
    f.write(pchar(x)^,length(x));
    x := '                        ************';
    f.write(pchar(x)^,length(x));
    for i:=length(fno) downto 1 do f.write(pchar(string('*'))^,1);
    f.write(pchar(eol)^,length(eol));
    f.write(pchar(eol)^,length(eol));
    x := 'Filename is '+fno+'.'+eol;
    f.write(pchar(x)^,length(x));
    f.write(pchar(eol)^,length(eol));
    x := 'This file contains primary output for SIENA project <<'+pn+'>>.'+eol;
    f.write(pchar(x)^,length(x));
    x := 'File '+fnl+' contains secondary output.'+eol;
    f.write(pchar(x)^,length(x));
    f.write(pchar(eol)^,length(eol));
    x := 'File '+fno+' created for primary output.'+eol;
    lf.write(pchar(x)^,length(x));
  end;

  if (writedate or (not appf))
  then begin
//    writeln(f,'Date and time: '+DateTimeToStr(Now));
    x := 'Date and time: '+DateTimeToStr(Now)+eol;
    f.write(pchar(x)^,length(x));
  end;

  if (appf and writedate)
  then begin
//    writeln(f,'New results follow.');
//    writeln(f,'-----------------------------------')
    x := 'New results follow.'+eol;
    f.write(pchar(x)^,length(x));
    x := '-----------------------------------'+eol;
    f.write(pchar(x)^,length(x));
  end;

//  writeln(f)
  f.write(pchar(eol)^,length(eol));

end; (* Outputfiles *)

procedure Outpfile(fn: shortstring; var f: textfile);
(* Makes an outputfile f with filename fn;
   The file gets a header with filename and date. *)
var i: integer;
begin
  assignfile(f,fn);
  rewrite(f);
   write(f,'                        ************');
    for i:=length(fn) downto 1 do write(f,'*');
    writeln(f);
    writeln(f,'                              ',fn);
    write(f,'                        ************');
    for i:=length(fn) downto 1 do write(f,'*');
  writeln(f);
  writeln(f);
  writeln(f,'Filename is ',fn,' .');
  writeln(f);
  writeln(f,'Date and time: '+DateTimeToStr(Now));
  writeln(f); writeln(f);
end;

PROCEDURE Protocol(name,tekst : string);
(* writes text 'tekst' to file 'name' *)
var protocol : textfile;
begin
  assign(protocol,name);
  if not fileexists(name)
  then begin
    rewrite(protocol);
    writeln(protocol,tekst)
  end
  else begin
    append(protocol);
    writeln(protocol,tekst)
  end;
  close(protocol)
end;

PROCEDURE pinverse(p: byte; var a: pMatrix; var y: pMatrix;
                   var errtext: shortstring);
(* berekent matrixinverse *)

CONST
   tiny=1.0e-15;
type rivec   = ARRAY[1..pmax] OF integer;
     rvec    = ARRAY[1..pmax] OF real;

VAR i,j: integer;
    d: real;
    vecy: rvec;
    indx: rivec;
    aa: pmatrix;
(* ee: pmatrix *)

(* programs using inverse must define the constant constn and the type
TYPE pMatrix = ARRAY[1..pmax,1..pmax] OF real;
   in the main program *)

PROCEDURE ludcmp(VAR a: pMatrix; n: integer;
       VAR indx: rivec; VAR d: real; VAR errtext: shortstring);
  (* used for matrix inversion *)
VAR
   k,j,imax,i: integer;
   som,dum,big: real;
   vv: rvec;
BEGIN
   errtext := '';
   d := 1.0;
   imax := 0;
   FOR i := 1 to n DO BEGIN
      big := 0.0;
      FOR j := 1 to n DO IF (abs(a[i,j]) > big) THEN big := abs(a[i,j]);
      IF (big = 0.0) THEN BEGIN
         errtext :='error in LUDCMP - singular matrix'; exit
      END;
      vv[i] := 1.0/big
   END;
   FOR j := 1 to n DO BEGIN
      IF (j > 1) THEN BEGIN
         FOR i := 1 to j-1 DO BEGIN
            som := a[i,j];
            IF (i > 1) THEN BEGIN
               FOR k := 1 to i-1 DO BEGIN
                  som := som-a[i,k]*a[k,j]
               END;
               a[i,j] := som
            END
         END
      END;
      big := 0.0;
      FOR i := j to n DO BEGIN
         som := a[i,j];
         IF (j > 1) THEN BEGIN
            FOR k := 1 to j-1 DO BEGIN
               som := som-a[i,k]*a[k,j]
            END;
            a[i,j] := som
         END;
         dum := vv[i]*abs(som);
         IF (dum > big) THEN BEGIN
            big := dum;
            imax := i
         END
      END;
      IF (j <> imax) THEN BEGIN
         FOR k := 1 to n DO BEGIN
            dum := a[imax,k];
            a[imax,k] := a[j,k];
            a[j,k] := dum
         END;
         d := -d;
         vv[imax] := vv[j]
      END;
      indx[j] := imax;
      IF (j <> n) THEN BEGIN
         IF (abs(a[j,j]) < tiny) THEN a[j,j] := tiny;
         dum := 1.0/a[j,j];
         FOR i := j+1 to n DO BEGIN
            a[i,j] := a[i,j]*dum
         END
      END
   END;
   IF (abs(a[n,n]) < tiny) THEN a[n,n] := tiny
END;

PROCEDURE lubksb(var a: pMatrix; n: integer; var indx: rivec; VAR b: rvec;
                 var errtext: shortstring);
// In Siena, not Lazarus, there have been runtime errors which were solved
// by making a and indx var parameters of this procedure.
VAR
   j,ip,ii,i: integer;
   som: real;
BEGIN
   errtext := '';
   ii := 0;
   FOR i := 1 to n DO BEGIN
      ip := indx[i];
      som := b[ip];
      b[ip] := b[i];
      IF  (ii <> 0) THEN BEGIN
         FOR j := ii to i-1 DO BEGIN
            som := som-a[i,j]*b[j]
         END
      END ELSE IF (som <> 0.0) THEN BEGIN
         ii := i
      END;
      b[i] := som
   END;
   FOR i := n DOWNTO 1 DO BEGIN
      som := b[i];
      IF (i < n) THEN BEGIN
         FOR j := i+1 to n DO BEGIN
            som := som-a[i,j]*b[j]
         END
      END;
      if (abs(a[i,i]) < tiny) then begin
        errtext :='error in routine lubksb in BLIB: a[i,i] small.';
        exit;
       end;
      b[i] := som/a[i,i]
   END
END;

BEGIN (* inverse *)
errtext := '';
aa := a;
if (p <= 0) then errtext := 'dimension not positive';
if (p > pmax) then errtext := 'dimension larger than pmax';
if (errtext > '') then exit;
FOR i := 1 TO pmax Do begin
  FOR j := 1 TO pmax DO y[i,j] := 0;
  y[i,i] := 1;
 END;
ludcmp(aa,p,indx,d,errtext);
if (errtext > '') then exit;

 FOR j := 1 TO p DO BEGIN
  FOR i := 1 TO p DO y[i,j] := 0.0;
  y[j,j] := 1.0;
 END;


 FOR j := 1 TO p DO BEGIN
  FOR i := 1 TO p DO vecy[i]:=0;
  vecy[j] := 1;
  lubksb(aa,p,indx,vecy,errtext);
  if (errtext > '') then exit;
  FOR i:= 1 TO p DO y[i,j] := vecy[i];
 END;
(* pproduct(p,a,y,ee);
for i:=1 to p do for j:=1 to p do
 if (((i=j) and (abs(ee[i,j]-1.0) > 1e-8))
   or ((i <> j) and (abs(ee[i,j]) > 1e-8)))
    then errtext := 'Matrix inversion not successful.'; *)
END; (* inverse *)



PROCEDURE jacobi(n: integer; VAR a: pmatrix; VAR d: pvec;
       VAR v: pmatrix; VAR nrot: integer);
(* Computes all eigenvalues and eigenvectors of a real symmetric matrix A
   of size N by N.
   D returns the eigenvalues of A.
   The columns of V return the normalized eigenvectors of A.
   NROT returns the number of Jacobi rotations required.
   Results can be sorted by EIGSRT. *)
LABEL 99;
VAR
   j,iq,ip,i: integer;
   tresh,theta,tau,t,sm,s,h,g,c: real;
   b,z: ARRAY OF real;
BEGIN
   SETLENGTH(b,n+1);
   SETLENGTH(z,n+1);
   FOR ip := 1 to n DO BEGIN
      FOR iq := 1 to n DO BEGIN
         v[ip,iq] := 0.0
      END;
      v[ip,ip] := 1.0
   END;
   FOR ip := 1 to n DO BEGIN
      b[ip] := a[ip,ip];
      d[ip] := b[ip];
      z[ip] := 0.0
   END;
   nrot := 0;
   FOR i := 1 to 50 DO BEGIN
      sm := 0.0;
      FOR ip := 1 to n-1 DO BEGIN
         FOR iq := ip+1 to n DO BEGIN
            sm := sm+abs(a[ip,iq])
         END
      END;
      IF (sm = 0.0) THEN GOTO 99;
      IF (i < 4) THEN tresh := 0.2*sm/sqr(n)
      ELSE tresh := 0.0;
      FOR ip := 1 to n-1 DO BEGIN
         FOR iq := ip+1 to n DO BEGIN
            g := 100.0*abs(a[ip,iq]);
            IF ((i > 4) AND ((abs(d[ip])+g) = abs(d[ip]))
               AND ((abs(d[iq])+g) = abs(d[iq]))) THEN
               a[ip,iq] := 0.0
            ELSE IF (abs(a[ip,iq]) > tresh) THEN BEGIN
               h := d[iq]-d[ip];
               IF ((abs(h)+g) = abs(h)) THEN BEGIN
                  t := a[ip,iq]/h
               END ELSE BEGIN
                  theta := 0.5*h/a[ip,iq];
                  t := 1.0/(abs(theta)+sqrt(1.0+sqr(theta)));
                  IF (theta < 0.0) THEN t := -t
               END;
               c := 1.0/sqrt(1+sqr(t));
               s := t*c;
               tau := s/(1.0+c);
               h := t*a[ip,iq];
               z[ip] := z[ip]-h;
               z[iq] := z[iq]+h;
               d[ip] := d[ip]-h;
               d[iq] := d[iq]+h;
               a[ip,iq] := 0.0;
               FOR j := 1 to ip-1 DO BEGIN
                  g := a[j,ip];
                  h := a[j,iq];
                  a[j,ip] := g-s*(h+g*tau);
                  a[j,iq] := h+s*(g-h*tau)
               END;
               FOR j := ip+1 to iq-1 DO BEGIN
                  g := a[ip,j];
                  h := a[j,iq];
                  a[ip,j] := g-s*(h+g*tau);
                  a[j,iq] := h+s*(g-h*tau)
               END;
               FOR j := iq+1 to n DO BEGIN
                  g := a[ip,j];
                  h := a[iq,j];
                  a[ip,j] := g-s*(h+g*tau);
                  a[iq,j] := h+s*(g-h*tau)
               END;
               FOR j := 1 to n DO BEGIN
                  g := v[j,ip];
                  h := v[j,iq];
                  v[j,ip] := g-s*(h+g*tau);
                  v[j,iq] := h+s*(g-h*tau)
               END;
               nrot := nrot+1
            END
         END
      END
   END;
   FOR ip := 1 to n DO BEGIN
      b[ip] := b[ip]+z[ip];
      d[ip] := b[ip];
      z[ip] := 0.0
   END;
   writeln('pause in routine JACOBI');
   writeln('50 iterations should not happen'); readln;
99:
   (* restore upper triangle of a  ; *)
   for ip:= n-1 downto 1 do for iq := ip+1 to n do a[ip,iq] := a[iq,ip];
  END;

PROCEDURE eigsrt( n: integer; VAR d: pvec; VAR v: pmatrix);
(* Sorts output of Jacobi with eigenvalues in ascending order,
   and corresponding rearrangement of eigenvectors. *)
VAR
   k,j,i: integer;
   p: real;
BEGIN
   FOR i := 1 to n-1 DO BEGIN
      k := i;
      p := d[i];
      FOR j := i+1 to n DO BEGIN
         IF (d[j] >= p) THEN BEGIN
            k := j;
            p := d[j]
         END
      END;
      IF (k <> i) THEN BEGIN
         d[k] := d[i];
         d[i] := p;
         FOR j := 1 to n DO BEGIN
            p := v[j,i];
            v[j,i] := v[j,k];
            v[j,k] := p
         END
      END
   END
END;

procedure matrixroot(p: byte; var a,v: pmatrix; plusmin: real;
                     var errtext: shortstring );
(* computes v := a^(+/- 1/2) where a is a positive(semi)definite
   symmetric matrix, and where the sign +/- is given by plusmin.
   If plusmin >= 0, all eigenvalues of a must be >= 0;
   if plusmin < 0, all eigenvalues of a must be > 0. *)
var d: pvec; nrot: integer; i,j,h: byte; v1: pmatrix; s: real;
begin
   errtext := '';
  jacobi(p,a,d,v1,nrot);
  for i:= p downto 1 do
    if ((d[i] < 0.0) or ((plusmin < 0.0) and (d[i] = 0.0))) then begin
      str(d[i],errtext);
      errtext := 'Procedure matrixroot discovered eigenvalue '+errtext;
      exit;
     end else begin
      if (plusmin >= 0.0) then d[i] := root(d[i]) else d[i] := 1/sqrt(d[i]);
     end;
  for i:=p downto 1 do for j:=i downto 1 do begin
    s := 0.0; for h:=p downto 1 do s := s+v1[i,h]*v1[j,h]*d[h];
    v[i,j] := s;
    if (j < i) then v[j,i] := s;
   end;
end; (* matrixroot *)


procedure pprod(p: byte; var a: pmatrix; var b,c: pvec);
var i,h: Byte; temp: real;
begin
  for i:=p downto 1 do begin
    temp := 0.0;
    for h:=p downto 1 do temp := temp + a[i,h]*b[h];
    c[i] := temp;
   end;
end; (* pprod *)

procedure product(p: byte; var a,b,c: pMatrix);
var i,j,h: Byte; temp: real;
begin
  for i:=p downto 1 do for j:=p downto 1 do begin
    temp := 0.0;
    for h:=p downto 1 do temp := temp + a[i,h]*b[h,j];
    c[i,j] := temp;
   end;
end; (* product *)

function veclength(p: byte; var a: pvec; var s: pmatrix): real;
var sinv: pmatrix; ec: shortstring; temp: real; i,j: byte;
begin
  pinverse(p,s,sinv,ec);
  if (ec > '') then
    for i:=p downto 1 do begin
      for j:=p downto 1 do sinv[i,j] := 0.0; sinv[i,i] := 1; end;
  temp := 0.0;
  for i:=p downto 1 do for j:= p downto 1 do
    temp := temp+a[i]*a[j]*sinv[i,j];
  veclength := temp;
end;


function max(a,b: real):real;
begin
  if (a < b) then max := b else max := a;
end;


FUNCTION gammln(xx: real): real; (* used for chi2 *)
CONST
   stp = 2.50662827465;
   half = 0.5;
   one = 1.0;
   fpf = 5.5;
VAR
   x,tmp,ser: double;
   j: integer;
   cof: ARRAY [1..6] OF double;
BEGIN
   cof[1] := 76.18009173;
   cof[2] := -86.50532033;
   cof[3] := 24.01409822;
   cof[4] := -1.231739516;
   cof[5] := 0.120858003e-2;
   cof[6] := -0.536382e-5;
   x := xx-one;
   tmp := x+fpf;
   tmp := (x+half)*ln(tmp)-tmp;
   ser := one;
   FOR j := 1 to 6 DO BEGIN
      x := x+one;
      ser := ser+cof[j]/x
   END;
   gammln := tmp+ln(stp*ser);
END;

PROCEDURE gser(a,x: real; VAR gamser,gln: real; var ertext: shortstring);
(* used for chi2 *)
LABEL 1;
CONST
   itmax=10000;
   eps=3.0e-7;
VAR
   n: integer;
   sum,del,ap: real;
BEGIN
   ertext := '';
   gln := gammln(a);
   IF (x <= 0.0) THEN BEGIN
      IF (x < 0.0) THEN BEGIN
         ertext := 'pause in GSER - x less than 0';
         gamser := 0.0;
         exit;
      END;
      gamser := 0.0
   END ELSE BEGIN
      ap := a;
      sum := 1.0/a;
      del := sum;
      FOR n := 1 to itmax DO BEGIN
         ap := ap+1.0;
         del := del*x/ap;
         sum := sum+del;
         IF (abs(del) < abs(sum)*eps) THEN GOTO 1
      END;
         ertext := 'pause in GSER - a too large, itmax too small';
         gamser := 0.0;
         exit;
1:      gamser := sum*exp(-x+a*ln(x)-gln)
   END
END;

PROCEDURE gcf(a,x: real; VAR gammcf,gln: real; var ertext: shortstring);
(* used for chi2 *)
LABEL 1;
CONST
   itmax=10000;
   eps=3.0e-7;
VAR
   n: integer;
   gold,g,fac,b1,b0,anf,ana,an,a1,a0: real;
BEGIN
   gln := gammln(a);
   gold := 0.0;
   a0 := 1.0;
   a1 := x;
   b0 := 0.0;
   b1 := 1.0;
   fac := 1.0;
   FOR n := 1 to itmax DO BEGIN
      an := 1.0*n;
      ana := an-a;
      a0 := (a1+a0*ana)*fac;
      b0 := (b1+b0*ana)*fac;
      anf := an*fac;
      a1 := x*a0+anf*a1;
      b1 := x*b0+anf*b1;
      IF (a1 <> 0.0) THEN BEGIN
         fac := 1.0/a1;
         g := b1*fac;
         IF (abs((g-gold)/g) < eps) THEN GOTO 1;
         gold := g
      END
   END;
   ertext := 'pause in GCF - a too large, itmax too small';
   gammcf := 0.0;
   exit;
1:   gammcf := exp(-x+a*ln(x)-gln)*g
END;


FUNCTION gammp(a,x: real; var ertext: shortstring): real;  (* used for chi2 *)
VAR
   gammcf,gln: real;
BEGIN
   ertext := '';
   IF ((x < 0.0) OR (a <= 0.0)) THEN BEGIN
     ertext := 'pause in GAMMP - invalid arguments';
     gammp := 0.0;
     exit;
   END;
   IF (x < (a+1.0)) THEN BEGIN
      gser(a,x,gammcf,gln,ertext);
      gammp := gammcf
   END ELSE BEGIN
      gcf(a,x,gammcf,gln,ertext);
      gammp := 1.0-gammcf
   END
END;

function normalcdf(x: real): real;
(* Cumulative distribution function of chi-squared distribution;
   from Press, Teukolsky & Vetterling. *)
var  ertext: shortstring;
begin
   IF (x > 0.0) THEN
     normalcdf := 0.5 + 0.5*gammp(0.5,0.5*sqr(x),ertext)
    ELSE normalcdf := 0.5 - 0.5*gammp(0.5,0.5*sqr(x),ertext);
end;


function chi2(df: integer; x: real): real;
(* Cumulative distribution function of chi-squared distribution;
   from Press, Teukolsky & Vetterling. *)
var a,y: real; ertext: shortstring;

begin (* chi2 *)
  chi2 := 0.0;
  if (df = 0) then begin
    if (x < 0) then chi2 := 0.0 else chi2 := 1.0;
   end;
  if (df >= 1) then begin
    a := 0.5*df;
    y := 0.5*x;
    if (x <= 0.0) then chi2 := 0.0 else chi2 := gammp(a,y, ertext);
   end;
end; (* chi2 *)


procedure quicksort(var a: array of integer; Lo,Hi: integer);

procedure sort(l,r: integer);
var
  i,j,x,y: integer;
begin
  i:=l; j:=r; x:=a[(l+r) DIV 2];
  repeat
    while a[i]<x do i:=i+1;
    while x<a[j] do j:=j-1;
    if i<=j then
    begin
      y:=a[i]; a[i]:=a[j]; a[j]:=y;
      i:=i+1; j:=j-1;
    end;
  until i>j;
  if l<j then sort(l,j);
  if i<r then sort(i,r);
end;

begin (* quicksort *);
  if (Lo < low(a)) then Lo := low(a);
  if (Hi > high(a)) then Hi := high(a);
  sort(Lo,Hi);
end;

procedure quicksortr(var a: array of integer; var b: array of shortint;
                     Lo,Hi: integer);

procedure sort(l,r: integer);
var
  i,j,x,y: integer;
  z: shortint;
begin
  i:=l; j:=r; x:=a[(l+r) DIV 2];
  repeat
    while a[i]<x do i:=i+1;
    while x<a[j] do j:=j-1;
    if i<=j then
    begin
      y:=a[i]; a[i]:=a[j]; a[j]:=y;
      z:=b[i]; b[i]:=b[j]; b[j]:=z;
      i:=i+1; j:=j-1;
    end;
  until i>j;
  if l<j then sort(l,j);
  if i<r then sort(i,r);
end;

begin (* quicksortr *);
  if (Lo < low(a)) then Lo := low(a);
  if (Hi > high(a)) then Hi := high(a);
  if (Lo < low(b)) then Lo := low(b);
  if (Hi > high(b)) then Hi := high(b);
  sort(Lo,Hi);
end;

function one(a: boolean): integer;
begin
  if a then one := 1 else one := 0;
end;

function st(i: integer): shortstring;
(* The string containing the number i . *)
var ss: shortstring;
begin
  str(i,ss);
  st := ss;
end; (* st *)

function st2(r: real): shortstring;
(* The string containing the number i . *)
var ss: shortstring;
begin
  str(r:3:2,ss);
  st2 := ss;
end; (* st *)

function AlignString(s: shortstring; len: integer): shortstring;
(* returns a string of length len by either cutting down or adding spaces *)
var j: integer;
 s1: shortstring;
begin
  if length(s)>len
  then s1:=copy(s,1,len)
  else begin
    s1:=s;
    for j := length(s) + 1 to len do s1 := s1+' '
  end;
  AlignString := s1
end;

function RightAlignString(s: shortstring; len: integer): shortstring;
(* returns a string of length len by either cutting down or adding spaces *)
var j: integer;
 s1: shortstring;
begin
  if length(s)>len
  then s1:=copy(s,1,len)
  else begin
    s1:=s;
    for j := length(s) + 1 to len do s1 := ' '+s1
  end;
  RightAlignString := s1
end;

procedure NiceString(x: real; var s: shortstring);
var j: integer;
 s1: shortstring;
begin
  str(x:10:4,s1);
  if (length(s1) < 10) then for j:=10-length(s1) downto 1 do s1 := ' '+s1;
  if (copy(s1,7,4) = '0000') then begin
    str(x:5:0,s1);
    s1 := s1+'     ';
    if (length(s1) < 10) then
      for j:=10-length(s1) downto 1 do s1 := ' '+s1;
  end;
  s := s1;
end; (* NiceString *)

procedure NiceString1(x: real; var s: shortstring);
begin
  if abs(round(x) - x)<1e-4 then s := inttostr(round(x))
   else
    begin
    str(x:10:4,s);
    end;
end; (* NiceString1 *)

procedure NiceString2(y: real; var s: shortstring);
var im: integer;
begin
  if (abs(y) <= 1e-3) then begin
   s := '0';
   exit;
  end;
  if (abs(y) > 100.0) then begin
     s := FloatToStrF(y,ffFixed,15,0);
     exit;
    end;
  if (abs(y) >= 1) then begin
    im := trunc(ln(abs(y))/ln(10.0)+1);
    if (abs(y - round(y)) < 0.05) then
      s := FloatToStrF(y,ffFixed,15,0)
      else s := FloatToStrF(y,ffFixed,15,im);
   end else begin
    im := trunc(-ln(abs(y))/ln(10.0)+1);
    s := FloatToStrF(y,ffFixed,15,im)
   end;
end;



function MakeString(n: integer): shortstring;
var s : shortstring;
begin
  str(n,s);
  MakeString := s
end;

procedure Heading(level: byte; var f: TFileStream; s: shortstring);
var i,sl: integer; x: string;
begin
  sl := length(s);
//  writeln(f);
//  writeln(f,'@',level);
//  writeln(f,s);
  f.Write(pchar(eol)^,length(eol));
  x :=  format('%s%d',['@',level])+eol;
  f.Write(pchar(x)^,length(x));
  f.Write(pchar(s+eol)^,length(eol)+length(s));

  case level of
    1: begin
         for i:=1 to sl do f.Write(pchar(string('='))^,1);//write(f,'=');
//         writeln(f);
         f.Write(pchar(eol)^,length(eol));
        end;
    2: begin
         for i:=1 to sl do f.Write(pchar(string('-'))^,1); //write(f,'-');
 //        writeln(f);
         f.Write(pchar(eol)^,length(eol));
        end;
    end; (* case *)
//  writeln(f);
  f.Write(pchar(eol)^,length(eol));
end;

 procedure Headingtext(level: byte; var f: textfile; s: shortstring);
var i,sl: integer;
begin
  sl := length(s);
  writeln(f);
  writeln(f,'@',level);
  writeln(f,s);
  case level of
    1: begin
         for i:=1 to sl do write(f,'=');
         writeln(f);
        end;
    2: begin
         for i:=1 to sl do write(f,'-');
        writeln(f);
         end;
    end; (* case *)
  writeln(f);
end;

procedure parse(var s: shortstring);
(* takes away out of s all characters with ASCII numbers less than 33 *)
var s1: shortstring;
 x: char;
 i,n: integer;
begin
 n := length(s);
 s1 := '';
 for i:=1 to n do begin
   x := s[i];
   if (ord(x) >= 33) then s1 := s1+x;
  end;
 s := s1;
end;


function parselo(var s: shortstring; mini: integer): shortstring;
(* takes away out of s all characters with ASCII numbers less than mini *)
                                                                          
var s1: shortstring;
 x: char;
 i,n: integer;
begin
 n := length(s);
 s1 := '';
 for i:=1 to n do begin
   x := s[i];
   if (ord(x) >= mini) then s1 := s1+x;
  end;
 parselo := s1;
end;

function firstnumber(var s: string): integer;
(* converts the leftmost substring of consecutive digits 0..9 in s
   into a nonnegative integer; -1 if this is not contained in s *)
var s1: shortstring;
 x: char;
 i,n,dumcode: integer;
 started, digit: boolean;
begin
 n := length(s);
 result := -1;
 if (n <= 0) then exit;
 s1 := '';
 i := 0;
 started := false;
 repeat
   inc(i);
   x := s[i];
   if (x in ['1'..'9']) or (x = '0') then
     begin
     s1 := s1+x;
     started := true;
     digit := true;
     end
    else
     digit := false;
    until (i >= n) or (started and (not digit));
  if (s1 > '') then val(s1,result,dumcode);
end;



procedure StartPictexPlot(xmin,xmax,ymin,ymax: real; var lf: text);
(* Here comes some help for Latex users: a picture using Pictex. *)
var i,dmx,dmy: integer;
    rmod: real;
begin
    if (xmin >= xmax) then xmax := xmin+1;
    if (ymin >= ymax) then ymax := ymin+1;
    writeln(lf,'Picture in Latex format:');
    writeln(lf,
      'use pictexwd and newcommands \gast and \rast (e.g., green and red asterisks) ');
    writeln(lf,'  for (colored) plot symbols.');
    writeln(lf);
    writeln(lf,'\begin{figure}[htbp]                              ');
    writeln(lf,'\begin{center}                                    ');
    writeln(lf,'\begin{minipage}[htb]{22cm}                       ');
    writeln(lf,'\beginpicture                                     ');
    write(lf,'\setcoordinatesystem units <');
    rmod := (100/(xmax-xmin));
    if (rmod < 3.0) then write(lf, rmod:5:2,'mm,') else write(lf,rmod:5:0,'mm,');
    rmod := (100/(ymax-ymin));
    if (rmod < 3.0) then writeln(lf, rmod:5:2,'mm> point at 0 0 ') else
      writeln(lf,rmod:5:0,'mm> point at 0 0   ');
    if ((xmax - xmin) >= 5) then dmx := 0 else dmx := 1;
    write(lf,'\setplotarea x from ', xmin:8:dmx,' to ',xmax:8:dmx,', ');
    if ((ymax - ymin) >= 5) then dmy := 0 else dmy := 1;
    writeln(lf,'y from ',ymin:8:dmy,' to ',ymax:8:dmy);
    writeln(lf,'\axis bottom                                      ');
    writeln(lf,'   ticks out short                                ');
    rmod := (xmax-xmin)/5;
    write(lf,'   numbered at ');
    i := 0;
    repeat
      write(lf,' ',xmin + i*rmod:8:dmx);
      inc(i);
     until (i*rmod > (xmax-xmin)+0.001);
    writeln(lf,' / ');
    writeln(lf,'     /                                            ');
    writeln(lf,'\axis left                                        ');
    writeln(lf,'   ticks out short                                ');
    write(lf,'   numbered at ');
    rmod := (ymax-ymin)/5;
    i := 0;
    repeat
      write(lf,' ',ymin + i*rmod:8:dmy);
      inc(i);
     until (i*rmod > (ymax-ymin)+0.001);
    writeln(lf,' / ');
    writeln(lf,'     /                                            ');
    write(lf,'\put{$\theta_j$} at ');
    writeln(lf,(1.1*xmax + 0.1*xmin):10:2,' ',(1.05*ymin - 0.05*ymax):10:2);
    write(lf,'\put{$u_j(y)$} at ');
    writeln(lf,(1.05*xmin - 0.05*xmax):10:2,' ', (1.1*ymax-0.1*ymin):10:2);

(* Then plot symbols e.g. by:

    writeln(lf,'\multiput{\gast} at                              ');
    writeln(lf,ix:4,' ',iy:4);
    writeln(lf,'  /                                               ');

    writeln(lf,'\multiput{\rast} at                              ');
    writeln(lf,ix:4,' ',iy:4);
    writeln(lf,'  /                                               ');
*)
end; (* StartPictexPlot *)



procedure EndPictexPlot(var lf: text);
begin
    writeln(lf,'\endpicture                                       ');
    writeln(lf,'\end{minipage}                                    ');
    writeln(lf,'\end{center}                                      ');
    writeln(lf,'\end{figure}                                      ');
end; (* EndPictexPlot *)


procedure WriteERGMInfile(nn: integer; ppname, dname: shortstring);
(* Writes .in file for project ppname with only 1 network data file dname
    and nn actors. *)
var f: textfile;
begin
  assign(f,ppname+'.in');
  rewrite(f);
  writeln(f,'@1[general information about SIENA project <',ppname,'>:]');
  writeln(f,'1 [number of waves] ');
  writeln(f,nn,' [number of actors]');
  writeln(f,'1 [number of dependent network variables]');
  writeln(f,'0 [number of dependent actor variables]');
  writeln(f,'0 [number of files with constant actor covariates]  ');
  writeln(f,'0 [number of exogenous changing actor covariates]');
  writeln(f,'0 [number of constant dyadic covariates]');
  writeln(f,'0 [number of exogenous changing dyadic covariates]          ');
  writeln(f,'0 [indicator for file with composition change directives]');
  writeln(f,'');
  writeln(f,'@2[network files in temporal order; names follow:]');
  writeln(f,dname);
  writeln(f,'1 [code for tie]    ');
  writeln(f,'9 [code for missing]');
  closefile(f);
end;

procedure Histogram(a: array of real; left,right, width: integer; var f: textfile);
(* copied from Numerical Recipes *)
const
  llen = 50;
var lo,hi,j,k,klim,iscal: integer;
   txt : PACKED ARRAY [1..llen] OF char;
   maxi: real;
begin
  if (left >= low(a)) then lo := left else lo := low(a);
  if (right <= high(a)) then hi := right else hi := high(a);
  if (lo > hi) then begin
    writeln(f,'error in Histogram bounds');
    exit;
   end;
  while (lo+1 < hi) and (a[hi] <= 0.0) do dec(hi);
  if (hi < right) and (hi < high(a)) then inc(hi);
  if (width > llen) then width := llen;
  maxi := 0.0;
  for j:=lo to hi do if (a[j] > maxi) then maxi := a[j];
  if (maxi > 0.0) then iscal := trunc(width/maxi) else iscal := 10;
  writeln(f,'x':5,'p(x)':8,'graph:':10);
  FOR j := lo to hi DO BEGIN
      FOR k := 1 to 50 DO BEGIN
         txt[k] := ' '
      END;
      klim := trunc(iscal*a[j]);
      IF (klim > width) THEN klim := width;
      FOR k := 1 to klim DO BEGIN
         txt[k] := '*'
      END;
      writeln (f,1.0*j:6:2,a[j]:8:4,'   ',txt)
   END;
end; (* histogram *)

procedure FsHistogram(a: array of integer;
                 left,right, width, circ, exx: integer; var f: TFileStream);
(* copied from Numerical Recipes *)
const
  llen = 50;
var lo,hi,j,k,klim: integer;
   txt : PACKED ARRAY [1..llen] OF char;
   maxi: integer;
   rscal: real;
   x: string;
begin
  if (left >= low(a)) then lo := left else lo := low(a);
  if (right <= high(a)) then hi := right else hi := high(a);
  if (lo > hi) then begin
    wwriteln(f,'error in Histogram bounds');
    exit;
   end;
  while (lo+1 < hi) and (a[hi] <= 0.0) do dec(hi);
  if (hi < right) and (hi < high(a)) then inc(hi);
  if (width > llen) then width := llen;
  maxi := 0;
  for j:=lo to hi do if (a[j] > maxi) then maxi := a[j];
  if (maxi > 0) then rscal := width/maxi else rscal := 10.0;
  wwriteln(f,'     x    n(x)    graph:    ');
  FOR j := lo to hi DO BEGIN
      FOR k := 1 to 50 DO BEGIN
         txt[k] := ' '
      END;
      klim := trunc(rscal*a[j]);
      IF (klim > width) THEN klim := width;
      FOR k := 1 to klim DO BEGIN
         txt[k] := '*'
      END;
      if (j = circ) then
        x := format('%s%2d%s%5d%s',['    ',j,' o',round(a[j]),'   '+txt])+eol
       else
        x := format('%s%2d%s%5d%s',['    ',j,'  ',round(a[j]),'   '+txt])+eol;
      f.Write(pchar(x)^,length(x));
   END;
      if (exx >= 1) then case exx of
        1..9:   wwriteln(f,'Outside:    '+inttostr(exx));
        10..99:  wwriteln(f,'Outside:   '+inttostr(exx));
        100..999: wwriteln(f,'Outside:  '+inttostr(exx));
        else       wwriteln(f,'Outside: '+inttostr(exx));
        end; (* case *)     
end; (* Fshistogram *)

procedure NRMinimize(p, mode: byte; var theta: pvec; var Cov: pmatrix;
                     var cf: textfile;  funct: TRealF);
(* Maximization of function funct *)
(* Mode 0: one Quasi-Newton Raphson step;
   Mode 1: iterated Quasi-Newton Raphson steps,
           but in each step, if NR is not indicated,
           a steepest descent step is made;
   Mode 2: iterated steepest descent steps;
   Mode 3: first multi steepest descent steps,
           then NR steps;
           in each step where NR is not indicated,
           multi steepest descent steps are made. *)
 (* theta is the input (initial) and output parameter vector *)
const epsi = 1e-12; (* Convergence criterion *)
      eps = 0.0001;  (* Argument difference for approximating derivatives *)
      nstepmax = 100; (* maximum number of iteration steps *)
type
    pvecplus = array[1..pmax] of double;
    pmatrixplus = array[1..pmax,1..pmax] of double;
var funct0,thefunction: double;
  ste: Boolean;
  i,j,multi,nsteps: integer;
  thet: pvec;     (* the working parameter *)
  insensitive: array[1..pmax] of Boolean;
procedure ThetaIncrement(i: byte; r: real);
begin
  thet[i] := thet[i] + r;
end; (* ThetaIncrement *)
procedure VecSum(a,b: pvecplus; var c: pVec; r: real);
var i: byte;
begin
  for i:=1 to p do c[i] := a[i]+r*b[i];
end; (* VecSum *)
procedure NormLength(var a: pVec; var s: real);
var i: byte;
begin
  s := 0.0;
  for i:=1 to p do s := s + sqr(a[i]);
  s := root(s);
  if (s > 0.0) then for i:=1 to p do a[i]:=a[i]/s;
end; (* NormLength *)
procedure NewtonRaphsonStep(option: byte; var stepp: boolean);
(* stepp indicates whether a step has been made *)
var i,j: byte;
    theta0: pVecplus;
    d1,step: pvecplus;
    d2: pmatrixplus;
    d2r: pmatrix;
    d0: double;
    s: real;
    ert: shortstring;
    outs: boolean;
procedure NoStepNoCov;
var i,j: byte;
begin
  stepp := false;
  for i:=1 to p do for j:=1 to p do Cov[i,j] := 9999.0;
end;
begin (* NewtonRaphsonStep *)
  inc(nsteps);
  stepp := false;
  for i:=1 to pmax do theta0[i] := thet[i];
  d0 := funct(thet);
(* approximate first and second derivatives by difference quotients *)
  for i:=1 to p do //if (not fix[i]) then
    begin
    ThetaIncrement(i,eps);
    d1[i] := funct(thet);
    ThetaIncrement(i,-eps);
   end;           //  else d1[i] := d0;
  for i:=1 to p do for j:=1 to i do
//   if (not(fix[i] or fix[j])) then
    begin
    ThetaIncrement(i,eps); ThetaIncrement(j,eps);
    d2[i,j] := (funct(thet) - d1[i] - d1[j] + d0)/sqr(eps);
    if (j < i) then d2[j,i] := d2[i,j];
    ThetaIncrement(i,-eps); ThetaIncrement(j,-eps);
    end (* for i, j if *);
//     else begin
//     d2[i,j] := 0.0;
//     if (j < i) then d2[j,i] := 0.0 else d2[i,i] := 1/9999.0;
//    end;
  for i:=1 to p do d1[i] := (d1[i]-d0)/eps;
(* Now d1 is vector of 1st, d2 matrix of 2nd derivatives. *)
  for i:=1 to pmax do for j:=1 to pmax do d2r[i,j] := d2[i,j];
  for i:=1 to p do if (d2[i,i] <= 0.0) then begin
    insensitive[i] := true;
    writeln(cf,'Function insensitive for parameter ',i,'.');
    d2r[i,i] := 0.1;
   end (* for i if *) else insensitive[i] := false;
  pinverse(p,d2r,Cov,ert);
  if (ert > '') then begin
(*   Matrix inversion error in Quasi-Newton Raphson procedure. *)
    (* add a ridge *)
    writeln(cf,'Error in matrix inversion: ',ert);
    for j:=1 to p do d2r[j,j] := d2r[j,j] + 1.0;
    pinverse(p,d2r,Cov,ert);
    if (ert > '') then begin
      writeln(cf,'Still error in matrix inversion: ',ert);
      NoStepNoCov; exit;
     end;
   end; (* if *)
(* If one of the diagonal elements of Cov is negative, the thefunction function
   is not locally convex and a Quasi-Newton Raphson step will not be made: *)
  outs := false;
  for i:=1 to p do if (//(not fix[i]) and
                     (Cov[i,i] <= 0)) then
    begin outs := true; end;
  if outs then begin (* NoStepNoCov; *) exit; end;

(* Calculate the step to be made and make it: *)
  for i:=1 to p do //if fix[i] then step[i] := 0.0 else
    begin
    step[i] := 0.0;
    for j:=1 to p do step[i] := step[i] - Cov[i,j]*d1[j];
    end; (* for i if *)
  VecSum(theta0,step,thet,1.0);
  thefunction := funct(thet);

(* Check whether the step is an improvement;
   if not, take action depending on option. *)
  case option of
  1: begin
       s := 1.0;
       while ((thefunction >= d0) and (s > 1e-20)) do begin
    (* halve the step *)
         s := 0.5*s;
         VecSum(theta0,step,thet,s);
         thefunction := funct(thet);
        end; (* while *)
       stepp := (thefunction < d0);
     end; (* case 1 *)
   2: if ((thefunction-d0) > 0.0) then begin
(* skip the step that was set *)
        for i:=1 to pmax do thet[i] := theta0[i];
        stepp := false;
        thefunction := d0;
       end else stepp := true;
    end; (* case *)
  writeln(cf,'End Quasi-Newton Raphson step, function value = ', thefunction:20:10);
end; (* NewtonRaphsonStep *)

procedure Descentstep(var step: boolean);
(* step indicates whether a step has been made *)
var i: byte;
  d0,d1,d2: double;
  length: real;
  de1,theta0: pVecplus;
  de1r: pvec;
begin
  inc(nsteps);
  step := false;
  d0 := funct(thet);
(* approximate first derivatives *)
  for i:=1 to p do //if not fix[i] then
    begin
    ThetaIncrement(i,eps);
    de1[i] := (funct(thet)-d0)/eps;
    ThetaIncrement(i,-eps);
    end;          // else de1[i] := 0.0;
  for i:=1 to p do de1r[i] := de1[i];
  NormLength(de1r,length);
  if (length < 1e-10) then exit;
(* de1 is the direction of steepest ascent.
   Calculate first and second derivatives in this direction,
   and then calculate length := stepsize according to
   onedimensional Quasi-Newton Raphson method. *)
  for i:=1 to pmax do theta0[i] := thet[i];
  VecSum(theta0,de1,thet,eps);
  d1 := funct(thet);
  VecSum(theta0,de1,thet,2.0*eps);
  d2 := funct(thet);
  if (abs(d0 - 2.0*d1 + d2) < 1e-20) then exit else
    length := (eps*(d1-d0))/(d0 - 2.0*d1 + d2);
(* if length < 0.0 then the 1-dimensional function is not locally convex,
   and a Newton-Raphson step size makes no sense. *)
  if (length <= 0.0) then begin length := 1.0; end;
  i:= 0;
  repeat
(* Make the step and halve the step size if the step yields no improvement. *)
    inc(i);
    VecSum(theta0,de1,thet,-length);
    length := 0.5*length;
    thefunction := funct(thet);
   until ((thefunction <= d0) or (i >= 20));
  if (thefunction >= d0) then begin
    for i:=1 to pmax do thet[i] := theta0[i];
    thefunction := d0; step := false;
   end else step := true;
  writeln(cf,'Descent step, function value = ', thefunction:20:10);
end; (* Descentstep *)

begin (* NRMinimize *)
  nsteps := 0;
  thet := theta;
  case mode of
  0: NewtonRaphsonstep(1,ste);
  1: begin
      thefunction := funct(thet);
        writeln(cf,'Initial function value = ', thefunction:20:10);
      for i:=1 to 5 do NewtonRaphsonstep(1,ste);
      repeat
        funct0 := thefunction;
        NewtonRaphsonstep(1,ste);
        if (not ste) then DescentStep(ste);
        thefunction := funct(thet);
      until ((abs(thefunction - funct0) < epsi) or (nsteps > nstepmax));
      NewtonRaphsonstep(2,ste);
     end; (* case 1 *)
  2: begin
      thefunction := funct(thet);
      repeat
        funct0 := thefunction;
        DescentStep(ste);
        thefunction := funct(thet);
      until ((abs(thefunction - funct0) < epsi) or (nsteps > nstepmax));
      NewtonRaphsonstep(2,ste);
     end; (* case 2 *)
   3: begin
       multi := p;
       if (multi < 5) then multi := 5;
       thefunction := funct(thet);
       for i:=1 to multi do Descentstep(ste);
       repeat
         funct0 := thefunction;
         NewtonRaphsonStep(1,ste);
         thefunction := funct(thet);
         if (not ste) then begin
         (* no NR step was set, therefore some DC steps are made. *)
           i := 0;
           repeat
             funct0 := thefunction;
             inc(i);
             Descentstep(ste);
             thefunction := funct(thet);
            until (i >= multi);
          end; (* if not ste *)
        until ((abs(thefunction - funct0) < epsi) or (nsteps > nstepmax));
      end; (* case 3 *)
  end; (* case *)
  for i:=1 to p do for j:=1 to p do
    if (insensitive[i] or insensitive[j]) then Cov[i,j] := -1.0;
  theta := thet;
  if (mode >= 1) then writeln(cf,'Number of steps = ',nsteps,'.');
end; (* NRMinimize *)

function sumsquares(const th: pvec): Real;
var i: integer;
s: real;
begin
  s := 0.0;
  for i:= pmax downto 1 do s := s + sqr(th[i]);
  sumsquares := s;
end;


PROCEDURE Plot(var outf: TFileStream;
               var XD, YD: array of real;
               fu: TRealReal;
               const M: Integer { number of points };
               const xlabel, ylabel: shortstring {labels for X and Y axes } );
(* Makes a plot of the points (xd, yd) and function fu;
   range of coordinates is determined by observed xd and yd;
   if there is no fu, use fu = nil *)
(* Strongly adapted from Numerical Recipes procedure SCRSHO *)
(* Correspondence between plot and values:
   for plot coordinates (ix,iy) we have the point
   (x1 + (ix-1)*(x2-x1)/(iscr-1), (ysml + (iy-1)*(ybig-ysml)/(jscr-1)) *)

CONST
   iscr=61;  (* horizontal size *)
   jscr=21;  (* vertical size *)
(* symbols for plotting: *)
   blank=' ';
   plu='+';
   yy='|';
   xx='-';
   ff='x';
   mult='X';
   funct='o';
   truncint = 1.1;  (* factor for extending axis end points beyond observed extremes *)
VAR
   jz,j,i,l1,l2,ix,iy,lo,outside,errcode,nspa: integer;
   ysml,ybig,x2,x1,x,dyj,dx,ee,d1,minx,maxx,miny,maxy: real;
   scr: ARRAY [1..iscr,1..jscr] OF char;
   scr1: string[1];
   sx: string;
BEGIN
   if (m <=1) then
     begin
     sx := format('%s%d%s',['Number of points to be plotted given as ',m,'. No plot.'])+eol;
     outf.Write(pchar(sx)^,length(sx));
     exit;
     end;

   if (low(xd) <> low(yd)) then
     begin
     sx := 'The two arrays to be plotted do not have the same lower index bound.'+eol;
     outf.Write(pchar(sx)^,length(sx));
     sx := 'No plot.'+eol;
     outf.Write(pchar(sx)^,length(sx));
     exit;
     end;

   lo := low(xd);
   if (high(xd) < lo + m - 1) or (high(yd) < lo + m - 1) then
     begin
     sx := 'The arrays to be plotted do not have enough elements.'+eol;
     outf.Write(pchar(sx)^,length(sx));
     sx := 'No plot.'+eol;
     outf.Write(pchar(sx)^,length(sx));
     exit;
     end;

(* frame: *)
   FOR j := 1 TO jscr DO BEGIN
      scr[1,j] := yy;
      scr[iscr,j] := yy
   END;
   FOR i := 2 TO iscr-1 DO BEGIN
      scr[i,1] := xx;
      scr[i,jscr] := xx;
      FOR j := 2 TO jscr-1 DO BEGIN
         scr[i,j] := blank
      END
   END;
   scr[1 + (iscr div 2),1] := plu;
   scr[1 + (iscr div 2),jscr] := plu;
   scr[1,jscr div 2  + 1] := plu;
   scr[iscr,jscr div 2 + 1] := plu;
   scr[1,1] := plu;
   scr[iscr,1] := plu;
   scr[1,jscr] := plu;
   scr[iscr,jscr] := plu;

(* extremes for x: *)
   x1 := xd[lo];
   x2 := x1;
   for i:= lo+1 to lo+m-1 do
     begin
     if xd[i] < x1 then x1 := xd[i];
     if xd[i] > x2 then x2 := xd[i];
     end;

   if (x1 = x2) then
     begin
     sx := 'There is only one x value. No plot.'+eol;
     outf.Write(pchar(sx)^,length(sx));
     exit;
     end;

  minx := x1;
  maxx := x2;

  l1 := trunc((ln(x2-x1))/ln(truncint))+1;
  ee := exp(l1*ln(truncint));
  if (x1 > 0.0) then x1 := ee*trunc(x1/ee)
   else x1 := -ee*trunc(-x1/ee+0.99);
  if (x2 > 0.0) then x2 := ee*trunc(x2/ee+0.99)
   else x2 := -ee*trunc(-x2/ee);

  if (abs(x1) >= 1.0) then l1 := trunc(ln(abs(x1)+1e-20)/ln(10.0))
    else l1 := trunc(ln(abs(x1)+1e-20)/ln(10.0))-1;
  if (abs(x2) >= 1.0) then l2 := trunc(ln(abs(x2)+1e-20)/ln(10.0))
    else l2 := trunc(ln(abs(x2)+1e-20)/ln(10.0))-1;
  if (l2 > l1) then l1 := l2;
  d1 := exp(l1*ln(10.0));
//  writeln(outf,'::: ',x1:8:4,' ',x2:8:4,' ',d1:6:2,' ', x1/d1:20:4,' ',x2/d1:20:4);
  x1 := d1*(trunc(x1/d1)+1);
  x2 := d1*(trunc(x2/d1)-1);
  if (x1 > minx) then x1 := x1-d1;
  if (x1 > minx) then x1 := x1-d1;
  if (x2 < maxx) then x2 := x2+d1;
  if (x2 < maxx) then x2 := x2+d1;



(* extremes for y : *)
   ybig := yd[lo];
   ysml := ybig;
   for i:= lo+1 to lo+m-1 do
     begin
     if yd[i] < ysml then ysml := yd[i];
     if yd[i] > ybig then ybig := yd[i];
     end;
  IF (ybig = ysml) THEN  ybig := ysml+1.0;

  miny := ysml;
  maxy := ybig;

  l1 := trunc((ln(ybig-ysml))/ln(truncint))+1;
  ee := exp(l1*ln(truncint));
  if (ysml > 0.0) then ysml := ee*trunc(ysml/ee)
   else ysml := -ee*trunc(-ysml/ee+0.99);
  if (ybig > 0.0) then ybig := ee*trunc(ybig/ee+0.99)
   else ybig := -ee*trunc(-ybig/ee);

  if (abs(ysml) >= 1.0) then l1 := trunc(ln(abs(ysml)+1e-20)/ln(10.0))
    else l1 := trunc(ln(abs(ysml)+1e-20)/ln(10.0))-1;
  if (abs(ybig) >= 1.0) then l2 := trunc(ln(abs(ybig)+1e-20)/ln(10.0))
    else l2 := trunc(ln(abs(ybig)+1e-20)/ln(10.0))-1;
  if (l2 > l1) then l1 := l2;
  d1 := exp(l1*ln(10.0));
//  writeln(outf,'::: ',ysml:8:4,' ',ybig:8:4,' ',d1:6:2,' ', ysml/d1:20:4,' ',ybig/d1:20:4);
  ysml := d1*(trunc(ysml/d1)+1);
  ybig := d1*(trunc(ybig/d1)-1);
  if (ysml > miny) then ysml := ysml-d1;
  if (ysml > miny) then ysml := ysml-d1;
  if (ybig < maxy) then ybig := ybig+d1;
  if (ybig < maxy) then ybig := ybig+d1;


(* define plotting positions for function *)
   dyj := (jscr-1)/(ybig-ysml);
   dx := (iscr-1)/(x2-x1);
   if assigned(fu) then for i:=2 to iscr-1 do
     begin
      iy := 1+trunc((fu(x1 + ((i-1)/dx))-ysml)*dyj);
      if (iy > 1) and (iy < jscr) then scr[i,iy] := funct;
      end;

(* define plotting positions for points *)
   outside := 0;
   for i:=1 to m do
     begin
     ix := 1 + trunc((xd[lo+i-1]-x1)*dx);
     iy := 1 + trunc((yd[lo+i-1]-ysml)*dyj);
     if (ix >= 1) and (ix <= iscr) and (iy >= 1) and (iy <= jscr)
       then
       case scr[ix,iy] of
         blank:  scr[ix,iy] := ff;
         funct:  scr[ix,iy] := ff;
         xx:     scr[ix,iy] := ff;
         yy:     scr[ix,iy] := ff;
         ff:     scr[ix,iy] := '2';
         mult:   scr[ix,iy] := mult;
         else begin
              val(scr[ix,iy],l1,errcode);
              if (errcode = 0) then
                begin
                if (l1 <= 8) then
                  begin
                  str(l1+1,scr1);
                  scr[ix,iy] := scr1[1];
                  end else scr[ix,iy] := mult;
                end else scr[ix,iy] := '?';
              end;
       end (* case *)
      else inc(outside);
     end;


   sx :=' '+ylabel+eol;
   outf.Write(pchar(sx)^,length(sx));
   sx:= format('%s%10.3f%s',[' ',ybig,' ']);
   outf.Write(pchar(sx)^,length(sx));
   FOR i := 1 TO iscr DO begin
    sx := scr[i,jscr];
    outf.Write(pchar(sx)^,length(sx));
   end;
   outf.Write(pchar(eol)^,length(eol));//writeln(outf);
   FOR j := jscr-1 DOWNTO 2 DO BEGIN
      if (j*2-1 = jscr) then begin
        sx := format('%s%10.3f%s',[' ',(ybig+ysml)/2,' ']);
        outf.Write(pchar(sx)^,length(sx));
      end
       else begin
        sx := format('%12s',[' ']);
        outf.Write(pchar(sx)^,length(sx));
        end;
      FOR i := 1 TO iscr DO begin
        sx := scr[i,j];
        outf.Write(pchar(sx)^,length(sx));
      end;
      outf.Write(pchar(eol)^,length(eol));//writeln(outf);
   END;
   sx := format('%s%10.3f%s',[' ',ysml,' ']);
   outf.Write(pchar(sx)^,length(sx));
   FOR i := 1 TO iscr DO begin
      sx := scr[i,1];
      outf.Write(pchar(sx)^,length(sx));
   end;
   outf.Write(pchar(eol)^,length(eol));//writeln(outf);
   nspa := (iscr-26) div 2;
   sx := format('%8s%10.3f%*s$10.3f%*s%10.3f%s%s',[' ',x1,nspa,' ',(x1+x2)/2,nspa,' ',x2,' ',xlabel])+eol;
   outf.Write(pchar(sx)^,length(sx));

   if (outside > 0) then
     begin
     outf.Write(pchar(eol)^,length(eol));//writeln(outf);
     if (outside <= 1) then begin
      sx := 'There was one point ';
      outf.Write(pchar(sx)^,length(sx));
     end
      else begin
        sx := format('%s%d%s',['There were ',outside,' points ']);
        outf.Write(pchar(sx)^,length(sx));
      end;
     sx := 'outside the plot window.'+eol;
     outf.Write(pchar(sx)^,length(sx));
     end;
end; (* plot *)

procedure reportp(var outf: textfile; p: real; dec: integer);
(* reports p-value p with dec decimals *)
var pp, thr: real;
begin
  pp := p;
  thr := expo(-dec*ln(10.0));
  if (pp > thr) then write(outf,'p = ',pp:dec+2:dec)
   else write(outf,'p < ',thr:dec+2:dec);
end;

procedure writevec(var outf: TFilestream; p: integer; x: array of real);
(* writes x[lo] to x[lo+p-1] to outf in format ' ',x[i]:9:4, with writeln after 10 numbers *)
var i,j: integer;
xx: string;
begin
  j := low(x)+p-1;
  for i:=low(x) to j do
    begin
    xx := format('%s%9.4f',[' ',x[i]])+eol;
    outf.Write(pchar(xx)^,length(xx));
    if (i-low(x) div 10 = 9) then outf.Write(pchar(eol)^,length(eol));//writeln(outf);
  end;
  if (j-low(x) div 10 <> 9) then outf.Write(pchar(eol)^,length(eol));//writeln(outf);
end;

procedure wwrite(var outf: TFileStream; x: string);
(* writes x to outf *)
begin
  outf.Write(pchar(x)^,length(x));
end;

procedure wwriteln(var outf: TFileStream; x: string);
(* writes x to outf with eoln *)
begin
  x := x+eol;
  outf.Write(pchar(x)^,length(x));
end;

begin
end.
