{$if not defined(lazarus)}
{$A+,B-,C+,D+,E-,F-,G+,H+,I+,J+,K-,L+,M-,N+,O+,P+,Q-,R+,S-,T-,U-,V+,W-,X+,Y+,Z1}
{$MINSTACKSIZE $00004000}
{$MAXSTACKSIZE $00100000}
{$IMAGEBASE $00400000}
{$ifend}
                           (* S_EST.PAS*)


(* This unit is for the estimation part of the Siena program. *)

unit s_est;

interface

uses
{$if not defined(linux)}
Windows,
{$ifend}
SysUtils, Controls, Forms, Graphics, Buttons, Dialogs,
     s_Constants, slib, eight, digraph, rangen, s_dat, s_base, s_re, s_theta, s_test,
     s_matrices, s_ml, s_hysteresis, s_mcmc, s_mcmc_II, s_ergm_est, amrandomx, Classes;

type
darray = array of array of byte;
Activity = object(MLModel)
  procedure SimulateData(var outf, lf: TFileStream);
  (*MS: Simulates data according to the probability model specified in <pname.mo>.*)
  procedure Estimate(var halter: Boolean;
                     var estimates, standard_errors, conv_t: pvec; var outf, lf: TFileStream);
  (* Estimation according to the Robbins Monro / Polyak Ruppert algorithm. *)
 end; (* Activities *)

var nexto,      (* CS: flag stopo is moved to S_Base, for availability reasons. *)
    {stopo,} starto, (* flags used for user interruption in estimation process through form buttons. *)
    nextt, apf,      (* CS: fullstop and apf are necessary for running under SIENA08, do not delete! *)
    fullstop    : boolean;
    h           : byte;
    n2partsum   : array[0..10] of integer;
    (*used for the progress bar in EstForm*)
    stud        : textfile;
    cov_covRE,
    cov2        : pmatrix;

implementation

uses siena_7;

var histograms: boolean;


procedure Activity.SimulateData(var outf, lf: TFileStream);
(*MS: Simulates data according to the probability model specified in <pname.mo>.*)
var
  seed, i, j, h, k, count, simtimepoints, c00, c01, c10, c11: integer;
  error: shortstring;
  s, s1, s2: shortstring;
  hdim: byte;
  maximum: real;
  poisson_parameter: extended;
  file1, file2, file3: textfile;
  u, sum, covariate, original: real;
  bool1, bool2, bool3, condition, initial_state_input: boolean;
  vec: pvec;
  varcov, cholfactor: pmatrix;
  fixing: pbvec;
  copy, dx, dy: rrarray;
  densities, scale: R_Ar;
  blackbox: blackarray;
  x : string;
begin
n := nmax;
SetArray(dx,integer(n+1),integer(n+1));
SetArray(dy,integer(n+1),integer(n+1));
SetArray(densities,integer(n*n));
(*MS
Please note:
-------------------------------------------------------------------------------------------------------------
The present procedure simulates digraphs according to the probability model specified in the <.mo> file and
stores the generated digraphs and behavior in the Siena input files, which in the sequel are read by Siena.
However, if you want to include other data (covariates etc.), you must generate them youselves and store them
- internally: in the form of variables that are used by Siena.
- and externally: in the form of the ASCII files used by Siena to read input.
Some code exists (below) to generate covariates, but most of the code is disabled.
-------------------------------------------------------------------------------------------------------------
*)
Heading(1,outf,'Monte Carlo simulation');
x := 'Monte Carlo sample from probability model specified in <'+CurrentProjectName+'.mo> file.'+eol;
outf.Write(pchar(x)^,length(x));
outf.Write(pchar(eol)^,length(eol));
(*
writeln(outf);
writeln(outf,'Pseudo-random number generator fed with seed ',seed,'.');
*)
(*********************************************)
(*A. Actor-dependent time-invariant covariate*)
(*********************************************)
condition := true; (*Set <condition> to true when covariates are desired.*)
if condition then
  begin
  nz := 1;
  s := CurrentProjectName + '.dav';
  OpenFile(file1,s); (*Actor-dependent time-invariant covariate.*)
  x := 'Actor-dependent time-invariant covariate obtained by generating independent observations from the Bernoulli(1/2) distribution.'+eol;
  outf.Write(pchar(x)^,length(x));
  outf.Write(pchar(eol)^,length(eol));//writeln(outf);
  for i := 1 to n do
    begin
    z[i,1] := random(2); (*Internal storage as actor-dependent covariate.*)
    writeln(file1,z[i,1]:10:6); (*External storage as actor-dependent covariate.*)
    x := format('%10.6f',[z[i,1]]);
    outf.Write(pchar(x)^,length(x));
    end;
  //writeln(outf); writeln(outf);
  outf.Write(pchar(eol)^,length(eol));
  outf.Write(pchar(eol)^,length(eol));
  closefile(file1);
  end;
(************************************)
(*B. Dyadic time-invariant covariate*)
(************************************)
condition := true; (*Set <condition> to true when covariates are desired.*)
if condition then
  begin
  x := format('%s%4.4f%s',['Dyadic time-invariant covariate obtained by generating independent observations from the Poisson(',poisson_parameter,') distribution.'])+eol;
  outf.Write(pchar(x)^,length(x));
  outf.Write(pchar(eol)^,length(eol));
  //writeln(outf);
  nzz := 1;
  zz := nil;
  SetLength(zz,n+1,n+1,nzz+1);
  poisson_parameter := 1; (*Parameter of Poisson distribution.*)
  maximum := 0;
  for i := 2 to n do
    begin
    for j := 1 to (i - 1) do
      begin
      zz[i,j,1] := Random_Poisson(poisson_parameter); (*Internal storage as dyadic covariate.*)
      zz[j,i,1] := zz[i,j,1]; (*Symmetrize.*)
      x := format('10.6f',[zz[i,j,1]]);
      outf.Write(pchar(x)^,length(x));
      if (zz[i,j,1] > maximum) then maximum := zz[i,j,1];
      end;
    //writeln(outf);
    outf.Write(pchar(eol)^,length(eol));
    end;
  //writeln(outf);
  outf.Write(pchar(eol)^,length(eol));
  s := CurrentProjectName + '.z01';
  OpenFile(file1,s);
  for i := 1 to n do (*External storage as dyadic covariate.*)
    begin
    for j := 1 to n do write(file1,zz[i,j,1]:10:6);
    writeln(file1);
    end;
  closefile(file1);
  zcmin[1] := 0;
  zcmax[1] := maximum;
  range[1] := maximum;
  end;
(**********************************************************************)
(*C. Heterogeneous actors: generate outlier(s) with parameter value...*)
(**********************************************************************)
condition := false; (*Set <condition> to true when outliers are desired.*)
if condition then
  begin
  alternative := 1; (*The outliers themselves need to be specified in the <.mo> file.*)
  x := format('%s%4.4f',['True data-generating value of the specified actor(s) / period is ',alternative])+eol;
  outf.Write(pchar(x)^,length(x));
  //writeln(outf);
  outf.Write(pchar(eol)^,length(eol));
  end;
(*******************)
(*D. Random effects*)
(*******************)
IndicateFixedThetaAndRC(fixing,thetarandom);
if randomeffects then
  begin
  s := 'CovRE.in';
  ReadCov(s,nre,varcov); (*Exploited in Monte Carlo simulations: read true variance-covariance matrix of the random effects.*)
  x := 'True variance-covariance matrix of the random effects:'+eol;
  outf.Write(pchar(x)^,length(x));
  (*Write the vector of non-redundant elements of variance-covariance matrix <varcov> to the file,
  where the vector was obtained by applying the vec operator to <varcov> and then deleting the above-diagonal elements:*)
  MatrixWrite(outf,nre,nre,varcov);
  SetArray(RE,n,nre);
  SetArray(scale,n);
  Cholesky(nre,cholfactor,varcov,u); (*Cholesky decomposition of positive definite variance-covariance-matrix.*)
  for i := 1 to n do scale[i] := 1;
  SampleMVN(densities,n,nre,cholfactor,scale,RE); (*Samples random effects from the multivariate normal distribution.*)
  (*Save and output random effects.*)
  SaveRE(n,nre,re); (*Write the sampled actor-dependent random effects to be used in future runs as starting point.*)
  s := 'RE.m';
  OpenFile(file1,s);
  x := 'Monte Carlo sample of random effects written to file <'+s+'>.'+eol;
  outf.Write(pchar(x)^,length(x));
  outf.Write(pchar(eol)^,length(eol));//writeln(outf);
  for i := 1 to n do
    begin
    for j := 1 to nre do
      begin
      write(file1,' ',re[i,j]:10:6);
      x := format('%s%10.6f',[' ',re[i,j]]);
      outf.Write(pchar(x)^,length(x));
      end;
    writeln(file1);
    //writeln(outf);
    outf.Write(pchar(eol)^,length(eol));
    end;
  closefile(file1);
 // writeln(outf);
  outf.Write(pchar(eol)^,length(eol));
  end;
(******************************************************************************)
bool1 := false;
simtimepoints := observations; (*Original number of observations is stored.*)
(*Simulate digraphs according to the probability model specified in the <.mo> file, conditional on empty digraph at time t_0.*)
(*
writeln(outf,'Simulated digraphs: can be viewed by <Netdraw> (Borgatti, 2002):'); writeln(outf);
*)
u := 1; (*Generate initial state.*)
bool1 := false;
bool2 := false;
bool3 := false;
(*Note: actually, <Initialise_y> must be directly preceded by <Initialise_actors>;
however, in simulations it is usually pointless to have inactive actors, hence:*)
for i := 1 to n do StructActive[i] := true; (*In the simulation (but not necessarily in the data set), all actors are assumed to be structurally active.*)
n_act := n;
exo := false;
DefineFunctions(outf,lf);
ReadPData(bool1,outf,lf);
if (nza > 0) then (*First behavioral outcome variable: determine range, for all time points.*)
  begin
  zcmin[1] := 1;
  zcmax[1] := 5;
  range[1+nz]:= zcmax[1] - zcmin[1];
  end;
if signed then DefineSignedNetworkEffects(NewNetworkEffects)
else DefineNetworkEffects(NewNetworkEffects);
observations := 2; (*Otherwise <Siena> attempts to reset the simulated digraph to the observed digraph, which do not exist.*)
cconditional := false; (*We have no digraphs at time points <t_1, t_2, t_3>, hence the distance between observed digraphs is undefined --> possible runtime error.*)
d_estAB := false;
d_estA := false;
d_estB := false;
initial_state_input := true; (*Whether an observed digraph is used as initial digraph: may be preferable, since otherwise one may end up in the stationary distribution soon.*)
if initial_state_input then
  begin
  (*Input: the digraph observed at the first time point, which remains constant across simulations runs.*)
   ReadDigraph(CurrentProjectName+'.d01',n,1); (*Read digraph in <file.d01>.*)
  (*Input: the outcome of the behavioral variable observed at the first time point, which remains constant across simulations runs.*)
  if (nza > 0) then ReadBehavior(1); (*On the assumption that there is at most one behavioral outcome variable.*)
  (*Observe that the two procedures above initialise the current digraph and behavior as well.*)
  x := 'Initial state: user input.'+eol;
  outf.Write(pchar(x)^,length(x));
  //writeln(outf);
  outf.Write(pchar(eol)^,length(eol));
  end
else (*The option of generating the initial digraph by simulation is restricted to network (without behavior) models.*)
  begin
  initialise_y(1);
  Runepoch(u,bool1,false,bool2,error,bool3,outf,lf); (*Simulate random digraph in time interval (0,u).*)
  for i := 1 to n do (*Store the digraph at time point <u>.*)
    begin
    for j := 1 to n do
      begin
      if yn(i,j) then setym(i,j,1,true) else setym(i,j,1,false);
      end;
    end;
  end;
{----------------}
{-Write to files-}
{----------------}
k := 1; (*First observation point.*)
x := format('%s%d%s',['network time point ',k,':'])+eol;
outf.Write(pchar(x)^,length(x));
count := 0; (*Compute density defined as #ties / #possible ties.*)
for i := 1 to n do
  begin
  for j := 1 to n do
    begin
    if yn(i,j) then
      begin
      dx[i,j] := 1;
      count := count + 1;
      outf.write(pchar(string('1'))^,1);
      end
    else
      begin
      dx[i,j] := 0;
      outf.write(pchar(string('0'))^,1);
      end;
    end;
 // writeln(outf);
  outf.Write(pchar(eol)^,length(eol));
  end;
SetMatrix(n,n,dy,dx);
//writeln(outf);
outf.Write(pchar(eol)^,length(eol));
x := format('%s%4d',['- number of ties = ',count])+eol;
outf.Write(pchar(x)^,length(x));
//writeln(outf);
outf.Write(pchar(eol)^,length(eol));
(*
writeln(outf,'- density = ',count / (n * (n - 1)):4:4); writeln(outf);
*)
x := format('%s%4.4f',['- average out-degree = ',count / n])+eol;
outf.Write(pchar(x)^,length(x));
outf.Write(pchar(eol)^,length(eol));//writeln(outf);
if not(initial_state_input) then (*Write digraph to Siena input file.*)
  begin
  s := CurrentProjectName + '.d0' + inttostr(k);
  OpenFile(file1,s);
  writeln(file1,'Siena data file: edge list');
  for i := 1 to n do
    begin
    for j := 1 to n do
      begin
      if yn(i,j) then writeln(file1,i:4,' ',j:4,' ',1:4);
      end;
    end;
  closefile(file1);
  end;
if (nza > 0) then
  begin
  (*Note: the first observation of the behavioral outcome variable is not simulated, but is required to be user input:
  therefore, writing the first observation to the Siena input file is pointless;
  in contrast, the first observation of the network may be simulated, which is why the network must written to the input file (see above).*)
  x := format('%s%d%s',['behavioral outcome variable time point ',k,':'])+eol;
  outf.Write(pchar(x)^,length(x));
  outf.Write(pchar(x)^,length(x));//writeln(outf);
  sum := 0;
  u := 0;
  for i := 1 to n do
    begin
    sum := sum + zc[i,1,k]; (*First moement.*)
    u := u + sqr(zc[i,1,k]); (*Second moment.*)
    x := format('%10.4f',[zc[i,1,k]]);
    outf.Write(pchar(x)^,length(x));
    end;
 // writeln(outf); writeln(outf);
  outf.Write(pchar(eol)^,length(eol));
  outf.Write(pchar(eol)^,length(eol));
  u := u - (sqr(sum) / n);
  u := u / n; (*Note: the aim is purely descriptive, hence the bias does not matters.*)
  u := sqrt(u);
  sum := sum / n;
  x := format('%s%10.4f',['- average = ',sum])+eol;
  outf.Write(pchar(x)^,length(x));
  outf.Write(pchar(x)^,length(x));//writeln(outf);
  x := format('%s%1-.4f',['- standard deviation = ',u])+eol;
  outf.Write(pchar(x)^,length(x));
  outf.Write(pchar(x)^,length(x));//writeln(outf);
  end;
{---------------------}
{-Start simulating...-}
{---------------------}
changealtern; (*Changes, for the case of outlier tests, the value of the global variable <altern>.*)
for k := 2 to simtimepoints do
  begin
  (*Note:
  The following procedure sets the digraph at present <yn> to <ynm[k]>
  and the behavioral variable at present <ya> to <z[k]>.*)
  Initialise_y(k-1);
  u := 1; (*Simulate Markov chain in time interval [0,1].*)
  bool1 := false; bool2 := false; bool3 := false;
  observations := 2; (*Otherwise <Siena> attempts to reset the simulated digraph to the observed digraph, which do not exist.*)
  cconditional := false; (*We have no digraphs at time points <t_1, t_2, t_3>, hence the distance NR observed digraphs is undefined --> possible runtime error.*)
  if ((homogeneous > 20) and (k = period)) then (*Should be right, but I changed something, so it could be wrong.*)
    begin
    original := alpa_f[1];
    alpa_f[1] := alternative;
    end;
  Runepoch(u,bool1,false,bool2,error,bool3,outf,lf);
  if ((homogeneous > 20) and (k = period)) then alpa_f[1] := original; (*Should be right, but I changed something, so it could be wrong.*)
  s := CurrentProjectName + '.d0' + inttostr(k);
  OpenFile(file1,s);
  writeln(file1,'Siena data file: edge list');
  {
  writeln(file2,'dl');
  writeln(file2,'n = ',n);
  writeln(file2,'format = fullmatrix');
  }
  x:= format('%s%d%s',['network time point ',k,':'])+eol;
  outf.Write(pchar(x)^,length(x));
  count := 0; (*Compute density defined as #ties / #possible ties.*)
  for i := 1 to n do
    begin
    for j := 1 to n do
      begin
      if (ynv(i,j) = 1) then (*On the assumption that the digraph data are binary.*)
        begin
        dx[i,j] := 1;
        h := 1;
        setym(i,j,k,true); (*Network at beginning of the period.*)
        setyvm(i,j,k,h); (*Network at beginning of the period.*)
        setyn(i,j,true); (*Network at present.*)
        setynv(i,j,h); (*Network at present.*)
        count := count + 1;
        writeln(file1,i:4,' ',j:4,' ',1:4);
        outf.write(pchar(string('1'))^,1);
        end
      else (*On the assumption that the digraph data are binary.*)
        begin
        dx[i,j] := 0;
        h := 0;
        setym(i,j,k,false); (*Network at beginning of the period.*)
        setyvm(i,j,k,h); (*Network at beginning of the period.*)
        setyn(i,j,false); (*Network at present.*)
        setynv(i,j,h); (*Network at present.*)
        outf.write(pchar(string('0'))^,1);
        end;
      end;
   // writeln(outf);
    outf.Write(pchar(eol)^,length(eol));
    if (nza > 0) then zc[i,1,k] := ya(i,1); (*Plug-in the simulated value of the first behavioral outcome variable.*)
    end;
  closefile(file1);
  c00 := 0; c01 := 0; c10 := 0; c11 := 0;
  if (k > 0) then
    begin
    for i := 1 to n do
      begin
      for j := 1 to n do
        begin
        if (i <> j) then
          begin
          if (dy[i,j] = 0) then
            begin
            if (dx[i,j] = 0) then c00 := c00 + 1 else c01 := c01 + 1;
            end
          else (*dy[i,j] <> 0*)
            begin
            if (dx[i,j] = 0) then c10 := c10 + 1 else c11 := c11 + 1;
            end;
          end;
        end;
      end;
    end;
  SetMatrix(n,n,dy,dx);
  outf.Write(pchar(eol)^,length(eol));//writeln(outf);
  x := format('%s%4d',['- number of ties = ',count])+eol;
  outf.Write(pchar(x)^,length(x));
  outf.Write(pchar(eol)^,length(eol));//writeln(outf);
  (*
  writeln(outf,'- density = ',count / (n * (n - 1)):4:4); writeln(outf);
  *)
  x := format('%s%4.4f',['- average out-degree = ',count / n])+eol;
  outf.Write(pchar(x)^,length(x));
  outf.Write(pchar(eol)^,length(eol));//writeln(outf);
  x := format('%s%d%s%d%s',['Transitions (time t_',k-1,' to t_',k,'):'])+eol;
  outf.Write(pchar(x)^,length(x));
  x := '------------------------------'+eol;
  outf.Write(pchar(x)^,length(x));
  x := format('%s%d',['0 --> 0   ',c00])+eol;
  outf.Write(pchar(x)^,length(x));
  x := format('%s%d',['0 --> 1   ',c01])+eol;
  outf.Write(pchar(x)^,length(x));
  x := format('%s%d',['1 --> 0   ',c10])+eol;
  outf.Write(pchar(x)^,length(x));
  x := format('%s%d',['1 --> 1   ',c11])+eol;
  outf.Write(pchar(x)^,length(x));
  x := '------------------------------'+eol;
  outf.Write(pchar(x)^,length(x));
  x := format('%s%d',['distance  ',c01+c10])+eol;
  outf.Write(pchar(x)^,length(x));
  outf.Write(pchar(x)^,length(x));//writeln(outf);
  if (nza > 0) then
    begin
    (*Information: output file.*)
    x := format('%s%d%s',['behavioral outcome variable time point ',k,':'])+eol;
    outf.Write(pchar(x)^,length(x));
    outf.Write(pchar(x)^,length(x));//writeln(outf);
    sum := 0;
    u := 0;
    for i := 1 to n do
      begin
      sum := sum + zc[i,1,k]; (*First moement.*)
      u := u + sqr(zc[i,1,k]); (*Second moment.*)
      x := format('%10.4f',[zc[i,1,k]]);
      outf.Write(pchar(x)^,length(x));
      end;
    //writeln(outf); writeln(outf);
    outf.Write(pchar(eol)^,length(eol));outf.Write(pchar(eol)^,length(eol));
    u := u - (sqr(sum) / n);
    u := u / n; (*Note: the aim is purely descriptive, hence the bias does not matters.*)
    u := sqrt(u);
    sum := sum / n;
    x := format('%s%10.4f',['- average = ',sum])+eol;
    outf.Write(pchar(x)^,length(x));
    outf.Write(pchar(eol)^,length(eol));//writeln(outf);
    x:= format('%s%10.4f',['- standard deviation = ',u])+eol;
    outf.Write(pchar(x)^,length(x));
    outf.Write(pchar(eol)^,length(eol));//writeln(outf);
    count := 0;
    c01 := 0;
    c10 := 0;
    for i := 1 to n do
      begin
      if (zc[i,1,k-1] <> zc[i,1,k]) then
        begin
        inc(count); (*Total number of changed behaviors.*)
        if (zc[i,1,k-1] < zc[i,1,k]) then inc(c01) (*Step up...*)
        else inc(c10); (*...step down.*)
        end;
      end;
    x := format('%s%d%s%d%s',['Transitions (time t_',k-1,' to t_',k,'):'])+eol;
    outf.Write(pchar(x)^,length(x));
    x := '------------------------------'+eol;
    outf.Write(pchar(x)^,length(x));
    x := format('%s%4d',['down     ',c10])+eol;
    outf.Write(pchar(x)^,length(x));
    x := format('%s%4d',['constant ',n - count])+eol;
    outf.Write(pchar(x)^,length(x));
    x := format('%s%4d',['up       ',c01])+eol;
    outf.Write(pchar(x)^,length(x));
    x := '------------------------------'+eol;
    outf.Write(pchar(x)^,length(x));
    x := format('%s%4d',['distance ',count])+eol;
    outf.Write(pchar(x)^,length(x));
    outf.Write(pchar(eol)^,length(eol));//writeln(outf);
    end;
  assignfile(file3,CurrentProjectName+'dens.m');
  if FileExists(CurrentProjectName+'dens.m') then append(file3) else rewrite(file3);
  writeln(file3,count/n:4:4);
  closefile(file3);
  end;
if (nza > 0) then (*External storage.*)
  begin
  (*On the assumption that there is
  - one endogenous actor-bound outcome variable,
  - no non-endogeneous changing actor-bound covariate.*)
  s := CurrentProjectName + '.dac';
  ReadFile(file1,s);
  SetArray(copy,integer(n+1),integer((nzc*simtimepoints)+1));
  for i := 1 to n do (*Read the original file: row...*)
    begin
    for j := 1 to (nzc * simtimepoints) do (*...column.*)
      begin
      read(file1,u);
      copy[i,j] := u;
      end;
    readln(file1);
    end;
  closefile(file1);
  for i := 1 to n do (*Plug-in the simulated values.*)
    begin
    for k := 1 to simtimepoints do copy[i,k] := zc[i,1,k]; (*Concerns the first behavioral outcome variable only.*)
    end;
  OpenFile(file1,s);
  for i := 1 to n do (*Rewrite the original file.*)
    begin
    for j := 1 to (nzc * observations) do write(file1,copy[i,j]:10:4);
    writeln(file1);
    end;
  closefile(file1);
  end;
(*Free memory by dereferencing dynamic arrays:*)
copy := nil;
densities := nil;
dx := nil;
dy := nil;
scale := nil;
alternative := 999999999; (*Default*)
changealtern; (*Changes, for the case of outlier tests, the value of the global variable <altern>.*)
end; (*MS: SimulateData*)


procedure Visibility(ph: byte);
begin
  EstForm.BEnd2.Enabled := (ph=2);
  EstForm.LPhase.visible := (ph >= 1);
  EstForm.EPhase.visible := (ph >= 1);
  EstForm.LIteration.visible := (ph >= 1);
  EstForm.EIteration.visible := (ph >= 1);
  EstForm.ESubphase.visible := ((ph=2) or (ph=4));
  EstForm.LSubphase.visible := ((ph=2) or (ph=4));
end;



procedure Activity.Estimate(var halter: Boolean;
                     var estimates, standard_errors, conv_t: pvec; var outf, lf: TFileStream);
var
  mtime,vmtime: array of real;
  m, theta, theta1, thetabase, initialtheta, standardtheta,
    scale, resis, Targets, sta, tconv: pVec;
  ordgs: intvec;
  fix, posi: pBvec;
  nactors: integer; (*MS: number of actors, needed in <S_est> when random effects models are to be estimated.*)
  pp,p,ErrorCode,nsubphs,rr1,nr1,rr2,nr2,LRcode: Byte;
  i,j,sin,indexi, ActiveNetworkEffects,nsteps,
  RecalledConditioningVariable : Integer; (* needed to remember correct conditioning when temporarily swtiching to unconditional estimation *)
  expon, somepos, somefix, OldConditional, maxlike_o, Bayesian_o: boolean;
  n3,niters,n4extra,nrun: Cardinal;
  fs: integer;
  c,dd,sx: pMatrix;
  messa, baselinemodeltext: string;
  errstr: shortstring;
  cf: TFileStream;
  brieffile: TFileStream;
f09, f11: textfile;  
  cfn: string;
  us, uss, alf1, alf2, tmaxx: real;
  ntrials, ntrialmax, extraeffects, whichtime: integer;
  alpdum: CRarray;
  var_theta: R_Ar;
  time0: TDateTime;
  ef: TEffect;
  high_se, high_est: boolean;
  x: string;
  ccc: boolean;
  Label Starthere;


  (* the following procedure carries out the
   Polyak-Ruppert-Pflug-Snijders algorithm for stochastic approximation *)

procedure PolRup(p: Byte; exponential: boolean; var theta, tc: pVec;
          thbase, thfirst, m, Scale, resist: pvec;
          var posj, fixed: pBVec; r1, rn1, r2, rn2: byte;
          var errorcode: Byte; var MessageText: string;
          var Cov,D,S: pMatrix; var md, vmd: array of real;
          var tmax: real;
          var n, nits: Cardinal; n3: Cardinal; nsub: byte; firstg: real);
(* p: dimension;
   A procedure FRAN(var outf: Text; p: Byte; theta: pVec; seed: Cardinal;
                    var y1: pVec; var dist: array of real; var success: boolean;
                    var Error: Boolean; var errstr: string; var outf,lf: textfile);
   must be available, that gives outcomes of random variables y1 and dist
   of which the distribution is parametrized by theta.
   if success = false after calling FRAN, the interpretation is that
    theta is outside the feasible region, and the p'th variable is
    temporarily taken as the distance from this feasible region;
    and first the feasible region is sought again.
    If this has been achieved - i.e., success = true again -
    then the original statistic is used again.
   The procedure FRAN must start with initialising
   the random number generator with seed
   and at the end put seed to its new current value.
   The aim of POLRUP is to find the solution of E[Y|theta] = m .
   Scale is a scaling vector for the theta[i]:
     it is used to define the values of epsilon
     and to limit the size of the step after phase 1.
     In the exponential case it is used only for truncating
     the single step after phase 1.
   Derivatives are estimated as simulated difference quotients
     at distances epsilon[i].
   The definition of Scale for i with (not posj[i]) should be such that:
     replacing theta[i] by theta[i]+ scale[i],
     while using the same random number stream,
     will with probability close to 1
     increase the generated value y1[i] generated by FRAN;
     and scale[i] should not be much bigger than the
     standard error of theta[i], because of its use
     in estimating the matrix of derivatives.
     It seems ideal to me (TS) to choose scale[i] to be
     somewhere between 0.1 and 10 times the standard error of theta[i].
     (Note that these are the standard errors of the statistical estimation,
      not of the MCMC procedure!)
   If posj[j], then theta[j] is restricted to positive values,
     and epsilon[j] for the difference quotients is defined as
     0.1*theta[j] (the initial values!).
   If fixed[j], then theta[j] is fixed to its initial value,
     and not included in the estimation algorithm.
   If rn >= 1, it is assumed that the
   Parameters numbered from r1 to (including) r2 are
   ri are the first and rn the nu
   Matrix Cov is the estimated covariance matrix of the solution theta;
   matrix D is the estimated partial derivative matrix;
   matrix S is the estimated covariance matrix of Y.
   md and vmd are the estimated mean and variance of dist,
     expect that in condition exponential, they are the
     estimated mean and variance of exp((theta0-theta)'*(y1-m)),
     where theta0 is the parameter for a standard baseline model
     and also for another externally given model;
     the expected value is the likelihood ratio.
   These last two dynamic arrays must have been dimensioned correctly
     (i.e., with length at most length(dist),
     where dist is the output dynamic array of FRAN)
     before the execution of PolRup.
   The rn1*high(md) parameter estimates starting at index r1
     are multiplied by the corresponding values md[1]..md[high(md)]
     at the end of PolRup.
     This is because they have a special role of rate parameters.
     The same for rn2 and r2.
     The analogous division takes place at the start,
     by the elements of the array md for which the input values are used.
   The total number of iterations is called n.

   If exponential, then the matrix D of derivatives is assumed to be equal
     to the covariance matrix S, and the algorithm is simplified
     accordingly.
   Exponential is used in SIENA together with another way of generating
     the random FRAN; some corresponding commands are made that communicate
     with the unit S_BASE in which FRAN is located - sorry for this.

   There are phases 1, 2, 3; and phase 0 for "observations = 1".
   Phase 2 is divided into subphases.
   Within each phase, or subphase for phase 2, iterations are numbered nit.
   Overall, iterations are numbered n.
   Default / maximum lengths of (sub)phases are
     n0, n1, n2min to n2max (internal), n3 (input).
   nsub is the number of subphases.
   On output, n is the number of steps taken,
   nits is number of steps taken in phase 3.
   firstg is the initial value of the gain parameter in the update steps.
   theta as defined when starting POLRUP1 is the starting value of
   the estimation algorithm;
   theta as defined when leaving PolRup is the estimated value.
   tmax is the largest absolute convergence t-ratio
   among all parameters except the basic rate parameters.
   the values of Cov, D, n when starting PolRup are of no consequence.
   On output, fixed is set to true for those elements for which it is
   suggested to fix them.
   *)

var sdf,df,msf,msfc: pmatrix;
    dfra,dfrac,dinv, msfinv: pmatrix;
    sff: array[1..pmax,1..pmax] of extended;
    th,thav,prev_th,fr1,fra,fral,prev_fra,fchange,prod0,prod1,epsilon,sd,
    sf, sfl, statistics: pvec;
    npos: array[1..pmax] of integer;
    prob,maxrat,maxmaxrat,gain,maxacor,minacor,ac,tstat,devi, devia,devib, extramult: real;
    time1: TDateTime;
    dist,distfd,mdist,mdist2: array of real;
    phase,subphase,nsubphas,repeatsubphase,
      maxrepeatsubphase,pp,maxrep2,rep2,nrep2,reptotal: byte;
    j,j0,h,k,l,it1,it2,it3,it4,it5: integer;
    repforep, Phasemax: byte;
    truncated,positivized,distDimension: integer;
    n2min0,n2min,n2max,n0,n1,n1pos,n03,nit,writefreq: Cardinal;
    jacwarning, seed0, ranoseed, cj: integer;
    feasible,feas,smaller,
    error, SearchFeasibleRegion, SomeFixed, AllFixed, sea, anyposj,
    restart, restarted, EarlyEndPhase2,extra,checkj,tpc: Boolean;
 (* extra indicates whether phase 4 is carried out *)
    ertext, nitstring: shortstring;
    diver, newfixed: pBvec;
    checktime, Force_Fin_Dif_phase1, burning, burning2: boolean;
    ctime: TDateTime;
    x: string;
    mins,maxs,mhs: shortstring;
    minim, maxim: pvec;
    hist: array of boolean;
    cnts: array of array of integer;
    cntsh: array[0..20] of integer;
    outside: array of integer;
 label StartAgain, EndPhase2, StartPhase34, EndPhase3, EndofEnd;

  procedure minus(p: byte; var a,b,c: pvec);
  (* for update: c := a-b *)
  var i: byte;
  begin
   for i:=p downto 1 do if (not fixed[i]) then c[i] := a[i]-resist[i]*b[i];
  end;

  procedure tripleproduct(p: byte; var a,b:pmatrix;
       var c: pmatrix);
  (* c := a*b*a' *)
  var i,j,h,k: byte; temp: real;
  begin
    for i:=p downto 1 do for j:=i downto 1
    do begin
      temp := 0.0;
      for h:=p downto 1
      do for k:=p downto 1
      do temp := temp+a[i,h]*b[h,k]*a[j,k];
      c[i,j] := temp;
      if (j < i) then c[j,i] := temp;
    end;
  end;

  procedure HSetTheta;
  (*MS: Sets the parameter value of one specified actor/period equal to corresponding parameter value of all actors/periods.*)
  begin
    if (homogeneous > 0) then theta[p] := theta[coordinate];
  end;

  procedure writeouttheta;
  var i: byte;
      s: shortstring;
      x: string;
  begin
//exit;
    x := 'Current parameter values: '+eol;
    cf.Write(pchar(x)^,length(x));

    HSetTheta; (*MS: Sets the parameter value of one specified actor/period equal to corresponding parameter value of all actors/periods.*)

    for i:=1 to pp
    do begin
      x := format('%10.6f%s',[theta[i],' ']);
      cf.Write(pchar(x)^,length(x));
      NiceString(theta[i],s);
      if (i <= 9) then s := ' '+s;
      if (i <= 30)
      then EstForm.ParameterValues.lines[i-1] := st(i)+'. '+s;
      if ((i mod 8) = 0) then cf.Write(pchar(eol)^,length(eol));//writeln(cf);
    end;
    if ((pp mod 8) <> 0) then cf.Write(pchar(eol)^,length(eol));//writeln(cf);
         Estform.parametervalues.selstart:=0;
  end;

  procedure writeouttheta2;
(* This procedure is rather time-consuming *)
  var i: byte;
    x: real;
    condition: boolean;
    s: shortstring;
  begin
    HSetTheta; (*MS: Sets the parameter value of one specified actor/period equal to corresponding parameter value of all actors/periods.*)

    for i:=1 to pp do
      begin
      if ((phase = 2) or Bayesian or (nit = 1)) and (i <= 30) then
        begin
        NiceString(theta[i],s);
        if (i <= 9) then s := ' '+s;
        EstForm.ParameterValues.lines[i-1] := st(i)+'. '+s;
        end;
      if (i <= 30) then
        begin
        if (nit = 1) then s := '    --    '
         else if (phase = 2) and (nit >= 10) then
           begin
           if (prod0[i]> 1e-20) then
             begin
             x :=  prod1[i]/prod0[i];
             NiceString(x,s);
             if (i <= 9) then s := ' '+s;
             end
            else s := '    --    ';
           if fixed[i]
           then EstForm.QAutocor.lines[i-1] := st(i)+'      --    '
           else EstForm.QAutocor.lines[i-1] := st(i)+'. '+s;
           end;
        NiceString(fra[i],s);
        if (i <= 9) then s := ' '+s;
        condition := false;
        if ((homogeneous > 0) and (i = pp) and not(derivatives)) or Bayesian
          then  s := '    --    ';
        EstForm.Statistics.lines[i-1] := st(i)+'. '+s;
        end;
      end; (* i *)
    estform.parametervalues.selstart:=0;
    estform.QAutocor.selstart:=0;
    estform.Statistics.selstart:=0;
  end;

  procedure writeoutmat(var mm: pmatrix);
  var i,j: byte;
  x : string;
  begin
    for i:=1 to pp
    do begin
      for j:=1 to pp
      do begin
        x := format('%12.4f%s',[mm[i,j],' ']);
        cf.Write(pchar(x)^,length(x));
        (* if ((j mod 6) = 0) then writeln(cf); *)
      end;
      (* if ((pp mod 6) <> 0) then writeln(cf); *)
      cf.Write(pchar(eol)^,length(eol));//writeln(cf);
    end;
    cf.Write(pchar(eol)^,length(eol));//writeln(cf);
  end;

 procedure roundfreq(var w: cardinal);
   begin
   if (w < 1) then w := 1;
{$if defined(batch)}
   if (w > 500) then w := 500;
{$else}
   if (w > 100) then w := 100;
{$ifend}
    case w of
      3..19 : w := 10;
     20..30 : w := 20;
     31..65 : w := 50;
     66..99 : w := 100;
{$if defined(batch)}
     100..299: w := 200;
     300..499: w := 500;
{$ifend}
    end; (* case *)
   end; (* roundfreq *)

  function errorcodestring(k: integer): string;
(* TS 18-jun-05; brought within Polrup, with phase indication *)
  begin
    case k of
      0 : errorcodestring := 'No error.';
      1 : errorcodestring := 'User stop in phase '+inttostr(phase)+' or input error.';
      2 : errorcodestring := 'Parameter outside feasible region for conditional estimation.';
      3 : errorcodestring := 'Current parameter is attempt to find feasible region for conditional estimation.';
      4 : errorcodestring := 'Convergence seems to be poor (see t-ratios).';
      5 : errorcodestring := 'Error in unit S_ML for likelihood-based estimation.';
      6 : errorcodestring := 'Non-invertible covariance matrix.';
      else errorcodestring := 'No explanation available for this error code.';
    end;
  end;


  procedure StartFeasibilitySearch;
  var x: string;
  begin
    ErrorCode := 2;
    SearchFeasibleRegion := true;
    x := 'Parameter is outside feasible region. First find feasible values.'+eol;
    cf.Write(pchar(x)^,length(x));
    (* The current parameter value is outside the feasible region.
    The algorithm temporarily switches to unconditional estimation
    to find feasible parameter values.
    *)
  end; (* StartfeasibilitySearch *)

  procedure EndFeasibilitySearch;
  var x : string;
  begin
     ErrorCode := 3;
     SearchFeasibleRegion := false;
     MessageText := 'End of feasibility Search';
     x := MessageText+eol;
     cf.Write(pchar(x)^,length(x));
    (*  After a temporary switch to unconditional estimation
    to find feasible parameter values,
    the algorithm now has returned to conditional estimation.
    *)
  end; (* EndfeasibilitySearch *)


  procedure CheckBreak;
  var i, code :integer;
      s : string;
      value : real;
      x : string;

  begin
    (* stopo: interrupt algorithm;
       starto: start algorithm again from current parameter values;
       nexto: end phase 2;
       nextt: restart algorithm with a new parameter value; *)

    Application.ProcessMessages; (* this may return a changed stopo etc.;
                                    see Siena_7 StopKeyClick etc. *)
    if stopo
    then begin
      x := format('%s%d%s',['User stop after ',nit,' iterations.'])+eol;
      cf.Write(pchar(x)^,length(x));
      if (phase <= 2) then errorcode := 1;
      exit;
    end;

    if nextt then with ThetaForm do
    begin
      x := format('%s%d%s',['New user choice of parameter values after ',nit,' iterations.'])+eol;
      cf.Write(pchar(x)^,length(x));
      //writeln(outf);
      outf.Write(pchar(eol)^,length(eol));
      x := 'User interrupted and asked for restart.'+eol;
      outf.Write(pchar(x)^,length(x));
      writeouttheta;
      starto := true;
      with SGTheta
      do begin
        rowcount := pp;
        for i:=1 to pp
        do begin
          str(theta[i]:10:4,s);
          cells[1,i-1]:=s;
          cells[0,i-1]:=Description[i];
        end;
      end; (* with SGTheta *)
      visible := false;
      ShowModal;
      if ModalResult = mrOK
      then with SGTheta
      do for i:=1 to pp
      do begin
        val(cells[1,i-1], value, code);
        if code = 0 then theta[i]:=value
      end;
      (* reset fixed : *)
      for i:=1 to pp do if newfixed[i] then fixed[i] := false;
    end;

    if (nexto and (phase=2))
    then begin
      x := format ('%s%d%s',['User termination of phase 2 after ',n,' iterations.'])+eol;
      cf.Write(pchar(x)^,length(x));
      end;

    if starto
    then begin
      x := format('%s%d%s%d%s',['Restart because of user command in phase ',phase,' after ',n,' iterations.'])+eol;
      cf.Write(pchar(x)^,length(x));
      (* reset fixed : *)
      for i:=1 to pp do if newfixed[i] then fixed[i] := false;
    end;

    if (starto or stopo or nexto or nextt)
    then  EstForm.EIteration.Text:=' ...  ';

    Application.ProcessMessages;
  end; (* CheckBreak *)

  procedure NullChecks;
  begin
    stopo := false; starto := false; nexto := false; nextt := false;
  end; (* NullChecks *)

  procedure DefineError;
  begin
    theta := th;
    Errorcode := 1;
    halter := true;
    MessageText := 'Error in function evaluation: '+ertext;
  end; (* DefineError *)

  procedure AdaptDfra;
  var i,j: byte;
  begin
    for i:=pp downto 1
    do begin
      for j := pp downto 1
      do if (fixed[i] or fixed[j]) then dfra[i,j] := 0.0;
      if Fixed[i]
      then begin
        fra[i] := 0.0;
        sf[i] := 0.0;
      end;
    end;
    for i:=pp downto 1 do if fixed[i] then dfra[i,i] := 1.0;
  end; (* AdaptDfra *)

  procedure AnnouncePhase;
  var propo, i: integer;
    n2min0, n2sum, n2maxx : integer;
    x : string;
  begin
  if (phase = 0) then
    begin
    if ((observations = 1) or maxlike or (not FinDiff_method)) then
      n1pos := n1
     else
      n1pos := n1*(p+1);
    n2min0 := 7+p;
    n2partsum[0] := 0;
    n2sum := 0;
    (* if there are more than 10 subphases the progressbar goes wrong *)
    for i:= 1 to nsubphase do
      begin
      n2min0 := trunc(n2min0*2.52);
      n2maxx := n2min0 + 200;
      if (i <= 10) then n2partsum[i] := n2partsum[i-1]+n2maxx;
      n2sum := n2sum+n2maxx;
      end; (* i *)
    if (nsubphase <= 10) then n2sum := n2partsum[nsubphase];
    with EstForm.ProgressBar do
      begin
      Min := 0;
      Max := n1pos + n2sum + n3;
      if (observations >= 2) and (not maxlike) and (d_estimator = 0)
        then Max := Max + n3*p;
      Position := 0;
      Hint := 'fraction of computations carried out for parameter estimation';
      end;
    end; (* if phase = 0 *)
    Visibility(phase);
    Application.ProcessMessages;
    if (phase = 1) then cf.Write(pchar(eol)^,length(eol));//writeln(cf);
    if ((phase = 0) and extra)
    then EstForm.ProgressBar.Max :=  EstForm.ProgressBar.Max + 10*n4extra;
    if ((phase = 3) and AllFixed)
    then EstForm.ProgressBar.Max :=  n3;
    EstForm.EPhase.Text:=' '+inttostr(phase);
    EstForm.ESubphase.Text:=' '+inttostr(subphase);
    EstForm.QAutocor.Visible := ((phase = 2) or (phase = 3)) and (observations >= 2)
                                and (not AllFixed);
    EstForm.Label1.Visible := EstForm.QAutocor.Visible;
    if extra then EstForm.EPhase.hint := 'phases go from 1 to 4';
    case phase of
      0: begin
           EstForm.EIteration.hint := 'burn-in iterations in phase 0 run from 1 to '+st(n0);
         end;
      1: begin
           EstForm.EIteration.hint := 'iterations in phase 1 run from 1 to '+st(n1);
           EstForm.QAutocor.hint := 'convergence diagnostics: negative values suggest convergence';
         end;
      2: begin
           if (observations = 1)
           then EstForm.EIteration.hint := 'total number of iterations in this subphase is '+inttostr(n2max)
           else EstForm.EIteration.hint := 'total number of iterations in this subphase between '
                                           +inttostr(n2min)+' and '+inttostr(n2max);
           EstForm.ESubphase.hint := 'phase 2 has subphases 1-'+inttostr(nsubphas);


         if ((observations = 1) or maxlike or Bayesian or (not FinDiff_method)) then
            propo := (7 + 3*p) + n2partsum[subphase-1]
           else propo := (7 + 3*p)*(p+1) + n2partsum[subphase-1];

           if (propo > EstForm.ProgressBar.Position)
             then EstForm.ProgressBar.Position := propo;
         end;
      3: begin
           EstForm.EIteration.Text := '...';
           EstForm.EIteration.hint := 'iterations in phase 3 run from 1 to '+inttostr(n3);
           if (nsubphase <= 0) then
              EstForm.QAutocor.hint := 'no estimation - no quasi-autocorrelations'
            else
             begin
             if ((maxacor > sqrt(2.0/(nit+1))) and (maxacor > 0.1)) then
               begin
               if (maxlike or Bayesian) then
               EstForm.QAutocor.hint := 'some quasi-a.c. values are positive; for likelihood estimation this is usual'
                else
               EstForm.QAutocor.hint := 'some quasi-a.c. values are positive; you may consider restarting ("Set N=0")';
               end;
             if (maxacor <= 0.0) then
               EstForm.QAutocor.hint := 'all quasi-a.c. values are negative; this suggests good convergence';
             if ((maxacor > 0.0) and ((maxacor <= sqrt(2.0/(nit+1))) or (maxacor < 0.1)))
               then EstForm.QAutocor.hint := 'the largest quasi-a.c. value is positive but small; no reason for concern';
             end;
           EstForm.StopKey.hint := 'Early termination of estimation (possible only after 10 phase 3 iterations)';
           if ((observations = 1) or maxlike or Bayesian or (not FinDiff_method)) then
            EstForm.ProgressBar.Position := (7 + 3*p) + n2partsum[nsub]
           else EstForm.ProgressBar.Position := (7 + 3*p)*(p+1) + n2partsum[nsub];
           if AllFixed then EstForm.ProgressBar.Position := 0;
         end;
      4: begin
           EstForm.EIteration.hint := 'total number of iterations in this subphase is '+inttostr(n4extra);
           EstForm.ESubphase.hint := 'phase 4 has subphases 1-'+inttostr(nsubphas);
           if ((observations = 1) or maxlike or Bayesian or (not FinDiff_method)) then
             begin
             if d_estAB then
           EstForm.ProgressBar.Position := (7 + 3*p) + n2partsum[nsub] + n3 + (subphase-1)*n4extra
           else EstForm.ProgressBar.Position := (7 + 3*p) + n2partsum[nsub] + n3*(p+1) + (subphase-1)*n4extra;
         end else
             begin
             if d_estAB then
           EstForm.ProgressBar.Position := (7 + 3*p)*(p+1) + n2partsum[nsub] + n3 + (subphase-1)*n4extra
           else EstForm.ProgressBar.Position := (7 + 3*p)*(p+1) + n2partsum[nsub] + n3*(p+1) + (subphase-1)*n4extra;
             end;
           (* note that nsub is the parameter of POLRUP indicating
              the number of subphases in phase 2;
              nsubphas is the number of subphases in the current phase,
              which may be 2 or 4 . *)
         end;
    end; (* case *)
    cf.Write(pchar(eol)^,length(eol));//writeln(cf);
    x := format('%s%d%s',['Start phase ',phase,'.']);
    cf.Write(pchar(x)^,length(x));
    if ((phase = 2) or (phase = 4))
    then begin
      x := format('%d%s',[subphase,'.'])+eol;
      cf.Write(pchar(x)^,length(x));
      end
    else cf.Write(pchar(eol)^,length(eol));//writeln(cf);
    Application.ProcessMessages;

//    if (phase >= 2) then for propo := 1 to observations-1 do
//      begin
//      writeln(cf,'period ',propo);
//      writefullperiod(propo,cf,cf);
//      writeln(cf);
//      end;
  end; (* Announce *)

  procedure FiniteDifferences;
  (*MS: Finite Differences Monte Carlo method to estimate the partial derivatives;
    now superseded by the new default likelihood ratio method.
    The present procedure carries out the additional Monte Carlo simulations required.*)
  var j, h: byte;
  fpc: boolean;
  begin
  for j:=1 to pp do
    begin
    if (Phase >= 3) or (not fixed[j]) then
      begin
      th[j] := th[j] + epsilon[j];
      if (phase >= 3) then
        begin
        alternative := 999999999; (*MS: Default, which is essential: it (a) indicates whether or not
        the present partial differential to be estimated concerns the parameter on which the outlier
        test is to be executed, and (b) if indeed this is the case, then it contains the value <th[pp] + epsilon[pp]>.*)
        if ((homogeneous > 0) and (j = pp)) then alternative := th[pp];
        end;
      ranoseed := seed0;
      fpc := ((estage > 30) and PPC);
      FRAN(pp,th,false,false,ranoseed,fr1,distfd,df,
           false,feas,error,ertext,cf,lf,estage,fpc);
       (* If feasible = true and feas = false, this is considered not serious. *)
      if ((phase >= 3) and (homogeneous > 0) and (j = pp)) then alternative := 999999999;
      th[j] := th[j] - epsilon[j];
      if error then
        begin
        DefineError;
        exit;
        end;
      for h:=1 to pp do
        begin
        sdf[h,j] := sdf[h,j] + (fr1[h]-fra[h])/epsilon[j];
        end;
      if (phase <= 1) and (abs(fr1[j]-fra[j]) > 1e-6) then inc(npos[j]);
      (* sdf is the summed realized matrix of partial derivatives *)
      end; (* if (phase >= 3 or not fixed[j]) *)
    end; (* j *)
   end; (* FiniteDifferences *)


  procedure Potential_NR(MakeStep: boolean);
  (*MS: Potential Newton-Raphson step: what would be the increment in case...
  If MakeStep, then this step is also carried out.
  Important note: as side-product, the inverse of the derivatives matrix is evaluated here.*)
  var j, h: byte;
  x : string;
  begin
   d := dfra;
   dfrac := dfra;
    if (phase = 3)
    then begin
      thav := theta; (* to retain it, because it will be changed in phase 4 *)

      (* 3.4 One partial Quasi-Newton Raphson update. *)
      error := false;
      (* dfrac and msfc are copies of dfra and msf that will possibly
         be modified for calculation of cov *)
      dfrac := dfra;
      msfc := msf;

      for j:= 1 to pp do for h:= 1 to pp do
        begin
        if ((fixed[j] or fixed[h] or diver[j] or diver[h]) and (not AllFixed)) then
          begin
          if (j = h) then
            begin
            dfrac[j,h] := 1.0; msfc[j,h] := 1.0
            end
          else
            begin
            dfrac[j,h] := 0.0; msfc[j,h] := 0.0
            end;
          end; (*if fixed or diver*)
        end; (*for j, h*)

      pinverse(pp,dfrac,dinv,ertext);
      (* ertext = '' iff the matrix dfra is invertible . *)
      if (ertext > '')
      then begin
        (* add a ridge *)
        for j:=pp downto 1 do dfrac[j,j] := dfrac[j,j] + 0.1*scale[j];
        pinverse(pp,dfrac,dinv,ertext);
        x := 'Intervention 3.4: ridge added after phase 3.'+eol;
        cf.Write(pchar(x)^,length(x));
      end;
      (* fchange is the change in theta : *)
      if (ertext = '')
      then pprod(pp,dinv,sf,fchange)
      else begin
        x := 'Warning. After phase 3, derivative matrix non-invertible even with a ridge.'+eol;
        cf.Write(pchar(x)^,length(x));
        for j:=pp downto 1 do fchange[j] := 0.0;
      end;
      x := 'dfrac :'+eol;
      cf.Write(pchar(x)^,length(x));
      writeoutmat(dfrac);
      x := 'inverse of dfra :'+eol;
      cf.Write(pchar(x)^,length(x));
      writeoutmat(dinv);

      x := 'A full Quasi-Newton Raphson step after phase 3'+eol;
      cf.Write(pchar(x)^,length(x));
      x :='would add the following numbers to the parameters, yielding the following results:'+eol;
      cf.Write(pchar(x)^,length(x));
      x := '         change     new value '+eol;
      cf.Write(pchar(x)^,length(x));
      for j:=1 to pp do begin
        x := format('%3d%s%12.6f%s%12.6f',[j,'. ',-fchange[j],' ',theta[j]-fchange[j]])+eol;
        cf.Write(pchar(x)^,length(x));
      end;
      cf.Write(pchar(eol)^,length(eol));//writeln(cf);
      if (extra or MakeStep) then
        (* this indicates that Phase 4 is carried out, or the step is carried out *)
        begin
        outf.Write(pchar(eol)^,length(eol));//writeln(outf);
        x := format('%s%d%s',['At the end of phase ',phase,', parameter values are'])+eol;
        outf.Write(pchar(x)^,length(x));
        for j:= 1 to pp do begin
          x := format('%3d%s%18.6f',[j,'. ',theta[j]])+eol;
          outf.Write(pchar(x)^,length(x));
        end;
        x := 'A full Quasi-Newton Raphson step after phase 3'+eol;
        outf.Write(pchar(x)^,length(x));
        x :='would add the following numbers to the parameters:'+eol;
        outf.Write(pchar(x)^,length(x));
        for j:=1 to pp do begin
          x := format('%3d%s%12.6f',[j,'. ',-fchange[j]])+eol;
          outf.Write(pchar(x)^,length(x));
        end;
        outf.Write(pchar(eol)^,length(eol));//writeln(outf);
        if extra then for j:=1 to pp do fchange[j] := extramult*fchange[j];
        if SomeFixed then for j:=1 to pp do
          if Fixed[j] then fchange[j] := 0.0;
        (* Check positivity *)
        for j:=1 to pp do if (posj[j] and (fchange[j] >= theta[j])) then
          begin
          fchange[j]:= 0.5*theta[j];
          x := format('%s%s%d%s',['Intervention 3.4.3: ','positivity restriction after phase 1, coordinate ',j,' .'])+eol;
          cf.Write(pchar(x)^,length(x));
          end;
        end; (* if extra or MakeStep *)
      if extra then
        (* this indicates that Phase 4 is carried out *)
        begin
        x := format('%s%6.3f%s',['This step is made 10 times, multiplied by the factor ',extramult,' .'])+eol;
        outf.Write(pchar(x)^,length(x));
        x := format('%s%d%s',['Each time defines a subphase with ',n4extra,' runs.'])+eol;
        outf.Write(pchar(x)^,length(x));
        end; (* if extra *)
      if MakeStep then (* then the step is actually carried out *)
        begin
        x := 'This step is carried out, leading to new parameter values:'+eol;
        outf.Write(pchar(x)^,length(x));
        for j:=1 to pp do
          begin
          theta[j] := theta[j] - fchange[j];
          x := format('%3d%s%18.6f',[j,'. ',theta[j]])+eol;
          outf.Write(pchar(x)^,length(x));
          end;
      end; (* if MakeStep *)
    end; (* if phase = 3 NR step *)

  end; (*Potential_NR*)


begin (* polrup main procedure *)
  (* Phase 0: initialisation. *)

  (*MS*)
  derivatives := false; (*Needed estimation of partial derivatives for the actor / period coordinate: <theta[p] + epsilon[p]>.*)
  if simulate then cconditional := false; (*More convenient and stable for simulations.*)
  (*MS*)

  checktime := true;  (* Gives reports of time duration of iterations in each (sub)phase *)
  extra := false;      (* extra defines whether phase 4 is carried out *)
  extramult := 0.025;  (* multiplier for Quasi-Newton Raphson step in phase 4 *)
  n4extra := 1000;     (* number of iterations in subphases of phase 4 *)
  maxmaxrat := 10.0;   (* truncation value for normalized statistics in phase 2 *)
  MessageText := '';
  phase := 0;
  if (n3 < 10) then n3 := 10;
  Visibility(phase);
  restarted := false;
  EarlyEndPhase2 := false;
  writefreq := 10;
  pp := p;
  outf.Write(pchar(x)^,length(x));//writeln(outf);
  x := 'Stochastic approximation algorithm.'+eol;
  cf.Write(pchar(x)^,length(x));

  if ((pp <= 0) and not VertexPar)
  then begin
    lf.Write(pchar(eol)^,length(eol));//writeln(lf);
    x := 'Wrong parameters in PolRup.'+eol;
    lf.Write(pchar(x)^,length(x));
    x := format('%s%d%s',['pp = ',pp,'.'])+eol;
    lf.Write(pchar(x)^,length(x));
    MessageText := 'Number of dimensions in PolRup is '+st(p)+'.';
    ErrorCode := 1;
    errorstring := MessageText;
    halter := true;
    exit;
  end;

  if (pp >= pmax)
  then begin
    errorcode := 1;
    lf.Write(pchar(eol)^,length(eol));//writeln(lf);
    x := format('%s%d%s%d%s',['Number of parameters p = ',p,
               ', but this should not be larger than pmax = ',pmax,'.'])+eol;
    lf.Write(pchar(x)^,length(x));
    MessageText := 'pp is too large for pmax.';
    ErrorCode := 1;
    errorstring := MessageText;
    halter := true;
    exit;
  end;

  distDimension := High(md);
  if (distDimension <= 0)
  then begin
    errorcode := 1;
    MessageText :=
      'Polrup started before dynamic array <<md>> was initialised.';
    errorstring := MessageText;
    Errorcode := 1;
    halter := true;
    exit;
  end;
  SetLength(dist,distDimension+1);
  SetLength(distfd,distDimension+1);
  SetLength(mdist,distDimension+1);
  SetLength(mdist2,distDimension+1);
  SetLength(cnts,pp+1,21);
  SetLength(hist,pp+1);
  SetLength(outside,pp+1);
  for j:=1 to pp do
    begin
    minim[j] := 0.0;
    maxim[j] := 0.0;
    end;

(* TS 04-jan-05. The following is the counterpart of the
   necessary division by mdist[h] at the end of PolRup. *)
  if (rn1 >= 1) then for j:=1 to rn1 do
    for h:=1 to distDimension do begin
      if (md[h] <= 1e-20) then begin
        MessageText :=
         'Polrup started with rn1>0 and md['+inttostr(h)+'] <= 0.';
        errorstring := MessageText;
        halter := true;
        Errorcode := 1;
        exit;
       end;
      theta[r1-1+(j-1)*DistDimension+h] :=
        theta[r1-1+(j-1)*DistDimension+h]/md[h];
     end;
  if (rn2 >= 1) then for j:=1 to rn2 do
    for h:=1 to distDimension do begin
      if (md[h] <= 1e-20) then begin
        MessageText :=
         'Polrup started with rn2>0 and md['+inttostr(h)+'] <= 0.';
        errorstring := MessageText;
        halter := true;
        exit;
       end;
      theta[r2-1+(j-1)*DistDimension+h] := theta[r2-1+(j-1)*DistDimension+h]/md[h];
     end;

  if (firstg <= 0.0)
  then begin
    x := format('%s%20.10f%s',['Initial value for gain parameter is ',firstg,'.'])+eol;
    outf.Write(pchar(x)^,length(x));
    x := 'This is not allowed; changed to 0.0001.'+eol;
    outf.Write(pchar(x)^,length(x));
    firstg := 0.0001;
  end;
  x := format('%s%12.6f%s',['Initial value for gain parameter = ',firstg,' .'])+eol;
  cf.Write(pchar(x)^,length(x));
  x :='Start of the algorithm.'+eol;
  cf.Write(pchar(x)^,length(x));
  x := 'Target function values are '+eol;
  cf.Write(pchar(x)^,length(x));
  for j:=1 to pp do
    begin
    x := format('%3d%s%10.4f%s',[j,'. ',m[j],' ']);
    cf.Write(pchar(x)^,length(x));
    if ((j mod 6) = 0) then cf.Write(pchar(eol)^,length(eol));//writeln(cf);
    end;
  if ((pp mod 6) <> 0) then cf.Write(pchar(eol)^,length(eol));//writeln(cf);
  cf.Write(pchar(eol)^,length(eol));//writeln(cf);
  SearchFeasibleRegion := (ErrorCode = 2);
  n := 0;
  maxrep2 := 1; (* maximum number of repetitions of phase 1 + 2 *)
  rep2 := 0;
  (* The number of iterations in phase 1 is an increasing function of p
     to get a reasonable covariance matrix: *)
  if (Bayesian or randomeffects) then nno := 0; (*MS*)
  if randomeffects then n1 := 5 * pp (*MS: random effects models imply more model complexity, hence more iterations are desirable.*)
   else n1 := 7 + 3*pp;
  repforep := 1;
  anyposj := false;
  for j:=1 to pmax do
    begin
    if (scale[j] > 0.1) then epsilon[j] := scale[j] else epsilon[j] := 0.1;
    if posj[j] then
      begin
      epsilon[j] := 0.1*theta[j];
      anyposj := true;
      end;
    end;

  reptotal := 0;
  Force_Fin_Dif_phase1 := false;
  burning := false; (* this indicates increased rstep *)
  burning2 := false; (* this indicates increased rstep *)

startagain:
  NullChecks;
  inc(reptotal);
  phase := 0;
  if (not Bayesian) then Initialise_d_est(lf,1,n1,pp,theta,Force_Fin_Dif_phase1);
  AnnouncePhase;
  Visibility(phase);
  if Continuous then
    begin
    (* start burn-in; the first (nit=1) iteration in phase 1 serves as burn in. *)
    rstep := rstep * 20.0;
    burning := true;
    EstForm.PLtext.visible := true;
    EstForm.PLtext.caption := 'Burn-in iterations';
    Application.ProcessMessages;
    end;
  Errorcode := 0;
  MessageText := '';
  if continuous then begin
      x := 'continuous'+eol;
      cf.Write(pchar(x)^,length(x));
      end
    else begin
      x := 'discontinuous'+eol;
      cf.Write(pchar(x)^,length(x));
      end;

  if anyposj then for j:=1 to pp do
    begin
    if ((posj[j]) and (theta[j] <= 0)) then
      begin
      lf.Write(pchar(eol)^,length(eol));//writeln(lf);
      x := format('%s%d%s',['For index ',j,' there is a positivity restriction,'])+eol;
      lf.Write(pchar(x)^,length(x));
      x := format('%s%f%s',['but a non-positive starting value ',theta[j],' .'])+eol;
      lf.Write(pchar(x)^,length(x));
      x := 'Impossible task. Algorithm stops.'+eol;
      lf.Write(pchar(x)^,length(x));
      MessageText := 'Unsatisfied positivity restriction in starting values.';
      errorcode := 1;
      errorstring := MessageText;
      halter := true;
      exit;
      end; (* if *)
    end; (* for j *)

  SomeFixed := false;
  AllFixed := true;
  for j:=1 to pp do
    if Fixed[j] then SomeFixed := true else AllFixed := false;

  for j:=1 to pmax do
    begin
    npos[j] := 0;
    newfixed[j] := false;
    end;
  writeouttheta;

  ranoseed := readstate;

  if (AllFixed or (nsub = 0)) then
    begin
    if AllFixed then begin
      x := 'All parameters are fixed.'+eol;
      outf.Write(pchar(x)^,length(x));
      end
     else begin
      x := 'Number of subphases is specified as 0.'+eol;
      outf.Write(pchar(x)^,length(x));
      end;
    x := 'Therefore, the estimation phase is skipped'+eol;
    outf.Write(pchar(x)^,length(x));
    x := 'and the program passes on immediately to phase 3'+eol;
    outf.Write(pchar(x)^,length(x));
    x := 'for checking the current parameter values and calculating standard errors.'+eol;
    outf.Write(pchar(x)^,length(x));
    if AllFixed then begin
      x := 'All parameters fixed; no estimation; only Phase 3.'+eol;
      lf.Write(pchar(x)^,length(x));
      end
     else begin
      x :='0 subphases; no estimation; only Phase 3.'+eol;
      outf.Write(pchar(x)^,length(x));
      end;
    repforep := 1;
    Phase := 2;
    goto StartPhase34;
    end;

  (* Phase 1: n1 starting iterations
              to have an initial estimate of the derivative matrix,
              finished by one simulation-based Quasi-Newton Raphson step
              and choice of standardizations for the coordinates. *)
  (* 1.1 Initialisation of phase 1. *)
  phase := 1;
  AnnouncePhase;
  Errorcode := 0;
  MessageText := '';
  for j:=1 to pmax
  do begin
    for h:=1 to pmax
    do begin
      sdf[j,h] := 0.0;
      sff[j,h] := 0.0;
      (* stt[j,h] := 0.0; stx[j,h] := 0.0; *)
    end;
    sf[j] := 0.0;
  end;

  (*MS:
  Bayesian inference: skip phase 1, since phase 2.1 serves as burn-in.
  Hierarchical Bayesian inference: exploit both Phase 1 and 2 to improve the candidate-generating density of actor-dependent random effects.
  *)
  estage := 11; (*MS: Estimation phase: phase 1, iteration 1.*)
  if (Bayesian or randomeffects) then
    begin
    it2 := nrunMH_batches; (*MS: Total number of iterations in <S_ml>.*)
    end;
  if Bayesian then (*MS: Bayesian model estimation: precautionary measures to avoid runtime errors in phase 2.*)
    begin
    SetIdentityMatrix(pp,dfra);
    for h := 1 to pp do sd[h] := 1;
    end;

  (* 1.2 n1 iterations *)

  if not(Bayesian) then (* phase 1 is redundant in Bayesian estimation.*)
    begin
//(* TS: LR method for estimating derivatives is less stable, requires more iterations;
//    the following condition is used in Initialise_d_est, must be changed to make the programming clearer. *)
// --- skipped again ---
//    if not (SoLiS or (d_estimator = 0)) then n1 := 2*n1;
    if checktime then ctime := now;
    for nit:=1 to n1 do
      begin
      if (nit = 2) then
        begin
        time1 := now;
        if checktime then ctime := now;
        end;
      if (nit = 6) then
        begin
        time1 := 86400*(now - time1);
      (* number of seconds elapsed in 4 FRANs,
         except if midnight intervened *)
          if (time1 > 1e-5)
          then writefreq := round(5.0/time1)
          else writefreq := 5;
{$if defined(batch)}
          writefreq := 20*writefreq;
{$ifend}
          roundfreq(writefreq);
        end;
      (* iteration 1 not included in timing to exclude burn-in *)
      phaseI_III := nit;
      inc(n);
      th := theta;
      seed0 := ranoseed;
      if maxlike then sea := (nit <= 1) else sea := SearchFeasibleRegion;
      feasible := true;

      tpc := false;

      FRAN(pp,th,false,MaxLike,ranoseed,fra,dist,df,
           false,feasible, error,ertext,cf,lf,estage,tpc);
ConnectedCheck(cf); 
      SearchFeasibleRegion := sea;
      estage := 12; (*MS: Estimation phase: phase 1, iteration > 1.*)
      if (errortext > '') then
        begin
        errorcode := 5;
        exit;
        end;
      if error then
        begin
        DefineError;
        exit;
        end;

      if FinDiff_method then
        begin
        FiniteDifferences; (* Finite differences method
                              to estimate partial derivatives: additional simulations.*)
        end;
      if MaxLike then
        begin
        for j:= pp downto 1 do for h:=pp downto 1 do
        sdf[j,h] := sdf[j,h] + df[j,h];
        end; (* if MaxLike *)

      if (not feasible) then begin
        x := format('%s%d%s',['Iteration ',nit,': not feasible.'])+eol;
        outf.Write(pchar(x)^,length(x));
        end;

      if ((not feasible) and (not SearchFeasibleRegion) and
        (not (maxlike or bayesian))
        and (nit >= 20)  
        ) then
        begin
        StartfeasibilitySearch;
        goto endofend;
        end;

      (* updates *)
      if exponential then IndependentStart := (not continuous); (* starting from the second iteration *)
      for h:=1 to pp do
        begin
        fra[h] := fra[h] - m[h];
        sf[h] := sf[h] + fra[h];
        if (fra[h] < minim[h]) then minim[h] := fra[h];
        if (fra[h] > maxim[h]) then maxim[h] := fra[h];
        end;
      for h:=pp downto 1 do for j:=pp downto 1
      do sff[j,h] := sff[j,h] + fra[j]*fra[h];
      (* fra is the current realized value of y-m *)


//wwriteln(outf,'');  abcda
//for h:=1 to pp do wwriteln(outf,inttostr(h)+' :: '+floattostr(fra[h]));

      (* checks and output  *)
      CheckBreak;
      if stopo then goto Endofend;
      if (starto or nextt) then goto startagain;
      NullChecks;
      if ((nit <= 5) or ((nit mod writefreq) = 0) or ((nit mod 5) = 0) or Maxlike or FinDiff_method)
       then
       begin
       EstForm.EIteration.Text:='    '+inttostr(nit);
       writeouttheta2;
{$if defined(batch)}
       write('Phase ',phase,' Iteration ',nit:4);
       writeln(' Progress ', (EstForm.ProgressBar.Position/EstForm.ProgressBar.Max*100):6:2,'%');
{$ifend}
       end;
      if FinDiff_method then EstForm.ProgressBar.StepBy(pp+1)
        else EstForm.ProgressBar.Stepit;

     if ((nit = 1) and Continuous) then
        begin (* end of burn-in *)
        rstep := rstep/20.0;
        burning := false;
        EstForm.PLtext.visible := false;
        Application.ProcessMessages;
        end;

      (* Check if the epsilons are big enough: *)
      restart := false;
      if ((nit = 10) and FinDiff_method) then
        begin
        error := false;
        j0 := 0;
        for j:=1 to pp do if ((not fixed[j]) and (npos[j] < 5)) then
          begin
          x := format('%s%d%s%d%s',['After 10 iterations, only ',npos[j],' differences in coordinate ',j,'.'])+eol;
          cf.Write(pchar(x)^,length(x));
          x := format('%s%8.2f%s',['with epsilon = ',epsilon[j],'.'])+eol;
          cf.Write(pchar(x)^,length(x));
          if (npos[j] <= 2) then
            begin
(* TS 20-jun-05: ik heb dit maal 10 cq sqrt(10) ipv 100 cq 10 gedaan, met maximaal repforep = 3 *)
            if posj[j] then epsilon[j] := 3.0*epsilon[j] else epsilon[j] := 10*epsilon[j];
            end
           else
            begin
            if posj[j] then epsilon[j] := 2.0*epsilon[j] else epsilon[j] := sqrt(10.0)*epsilon[j];
            end;
          if (epsilon[j] > 101.0*scale[j]) then epsilon[j] := 100.0*scale[j];
          if (epsilon[j] <= 0.1*scale[j]) then epsilon[j] := 0.1*scale[j];
          error := true;
          x := format('%s%8.2f%s',['New epsilon = ',epsilon[j],'.'])+eol;
          cf.Write(pchar(x)^,length(x));
          if (npos[j] <= 1) then j0 := j;
          end;
        if (error and (repforep <= 4)) then
          begin
          x := 'Change value of epsilon and restart Phase 1.'+eol;
          cf.Write(pchar(x)^,length(x));
          inc(repforep);
          restart := not(restarted);
          (* restart makes the algorithm restart once after subphase 2.2,
             if it comes through this place. *)
          goto startagain;
          end
         else
          begin
          if (j0 > 0) then
            begin
            x := format('%s%d%s',['Difficulties with parameter ',j0,'.'])+eol;
            outf.Write(pchar(x)^,length(x));
            x := format('%s%d%s',['After 10 iterations, only ',npos[j0],' differences in this coordinate.'])+eol;
            outf.Write(pchar(x)^,length(x));
            x := 'This parameter is fixed and not estimated.'+eol;
            outf.Write(pchar(x)^,length(x));
            x := format('%s%d%s',['Fix parameter ',j0,' in phase 1.'])+eol;
            cf.Write(pchar(x)^,length(x));
            fixed[j0] := true;
            newfixed[j0] := true;
            SomeFixed := true;
            end; (* if *)
          end;
        end; (* if nit = 10 and not exponential *)
      end; (* for nit *)

    if (Bayesian or randomeffects) then nno := n1 * it2; (*MS: Siena may restart multiple times; but in such cases everything should be initialised again. Value must be set here since otherwise the Bayesian algorithm will cause runtime errors.*)
    if checktime then begin
    x := format('%s%d%s%8.4f',['Time per iteration in phase ',phase, '   = ',86400*(now - ctime)/(nit-1)])+eol;
    lf.Write(pchar(x)^,length(x));
    end;
    (* 1.3 Evaluation of estimated mean, covariance matrix, derivative matrix *)
    for j:=pp downto 1
    do begin
      fra[j] := sf[j]/n1;
    end;
    x := 'Average deviations NR generated statistics and targets'+eol;
    cf.Write(pchar(x)^,length(x));
    x := 'after phase 1 :'+eol;
    cf.Write(pchar(x)^,length(x));
    for j:=1 to pp do begin
      x := format('%12.4f',[fra[j]])+eol;
      cf.Write(pchar(x)^,length(x));
      end;
    (* Calculate estimated derivative matrix *)
    if Force_Fin_Dif_phase1 then jacwarning := 1 else jacwarning := 0;
    if exponential then (* derivative = covariance matrix *)
      begin
      for j:=1 to pp do
        begin
        for h:=1 to pp do dfra[j,h] := (sff[j,h]/n1) - (fra[j]*fra[h]);
        sd[j] := root(dfra[j,j]);
        end;
      end
    else
      begin (* not exponential *)
      if d_estAB then
        begin
//writeoutmat(dfra); // abcda
        JacobianMatrix(pp,cf,outf,dfra,jacwarning); (*MS: moment estimation.*)
        end
      else if (maxlike or FinDiff_method) then
        begin
        jacwarning := 0;
        for j:=1 to pp do
          begin
          for h:=1 to pp do dfra[j,h] := sdf[j,h]/n1;
          sd[j] := root(sff[j,j]/n1 - fra[j]*fra[j]);
          end
        end;
      end; (* if not exponential *)

    d_estAB := false;
    d_estA := false;
    d_estB := false;
    (* d_est must be set to false after phase 1
                       in order to save computation time in phase 2.*)

    if (jacwarning >= 1) then
      begin
      if (not Force_Fin_Dif_phase1) then
(* This means that the definition of one or more of the diagonal elements of the
   matrix dfra led to the value zero.
   This means that the d_est (LR) method for estimating derivatives
   does not give information for scaling from "fra" to "theta", which
   leads to undesirable behavior of the Robbins Monro process.
   Recourse now is taken to the usually less precise and more time-intensive,
   but also more robust finite differences method.*)
        begin
        if (n1 < 200) then
          begin
          n1 := 2*n1;
          x := 'New phase 1 of increased length because of unreliable derivative matrix.'+eol;
          cf.Write(pchar(x)^,length(x));
          //writeln(lf,'New phase 1 of increased length because of unreliable derivative matrix.');
          lf.Write(pchar(x)^,length(x));
          wwriteln(cf,'Matrix dfra:');
          writeoutmat(dfra);
          end
         else
          begin
          Force_Fin_Dif_phase1 := true;
          x := 'New phase 1 with finite difference method because of unreliable derivative matrix.'+eol;
          cf.Write(pchar(x)^,length(x));
          //writeln(lf,'New phase 1 with finite difference method because of unreliable derivative matrix.');
          lf.Write(pchar(x)^,length(x));
          end;
        goto StartAgain;
        end
        else
        x := format('%s%d%s%s%s%s',['Warning: standard error for estimate ',jacwarning,' below is unreliable...see file ',CurrentProjectName + '.cck','.'])+eol;
        outf.Write(pchar(x)^,length(x));
      end;

    (* sd is the standard deviation of fra[j] *)
    for j:=1 to pp
    do begin
      for h:=1 to pp do if (fixed[j] or fixed[h]) then dfra[j,h] := 0.0;
      if fixed[j] then dfra[j,j] := 1.000;
    end;
    (* dfra is the estimated derivative matrix d_E[fra|theta]/d_theta; *)
    halter := false;
    x := 'Diagonal values of derivative matrix :'+eol;
    cf.Write(pchar(x)^,length(x));
    for j:=1 to pp do
      begin
      x := format('%8.4f%s',[dfra[j,j],' ']);
      cf.Write(pchar(x)^,length(x));
      if ((dfra[j,j] <= 0.0) and (not(fixed[j]))) then
        begin
        fixed[j] := true;
        newfixed[j] := true;
        somefixed := true;
        cf.Write(pchar(eol)^,length(eol));//writeln(cf);
        x := format('%s%d%s',['*** nonpositive diagonal value: parameter ',j,' is fixed.   ***'])+eol;
        cf.Write(pchar(x)^,length(x));
        x := format('%s%d%s',['Estimation problems with parameter ',j,': this parameter is fixed.'])+eol;
        outf.Write(pchar(x)^,length(x));
(*
        writeln(cf);
        writeln(cf,  '***       warning: nonpositive diagonal value      ***');
        writeln(cf,  '***         the algorithm will not converge        ***');
        writeln(cf,  '***           SIENA execution interrupted          ***');
        writeln(cf);
        writeln(outf,'***  Algorithm problem.                      ***');
        writeln(outf,'***  Try again from a better starting value. ***');
        writeln(outf,'***  The file <',pname,'.cck> may contain diagnostic information. ***');
        halter := true;
        if (MessageText = '')
        then MessageText := 'Simulated value not positive sensitive for parameter '+st(j)
        else MessageText := MessageText+', '+st(j);
*)
        end;
      if ((j mod 8) = 0) then cf.Write(pchar(eol)^,length(eol));//writeln(cf);
      end;

    if ((pp mod 8) <> 0) then cf.Write(pchar(eol)^,length(eol));//writeln(cf);
    if halter then
      begin
      ErrorCode := 1;
      errorstring := MessageText+'.';
      halter := true;
      exit;
      end;
    if SomeFixed
    then begin
    x := '(Values for fixed parameters are meaningless.)'+eol;
    cf.Write(pchar(x)^,length(x));
    end;

    (* 1.4 One first partial Quasi-Newton Raphson update. *)
    error := false;
    if SomeFixed then AdaptDfra;
    pinverse(pp,dfra,dinv,ertext);

    if (maxlike and randomeffects) then (*MS: rescale <dinv>.*)
      begin
      (*Remark: premultiply <dinv> by a diagonal matrix with diagonal elements
      - corresponding to fixed effects: ones, which changes nothing.
      - corresponding to (co)variance parameters: <scale[j]>, which rescales row <j> of <dinv>.*)
      x := 'Rescale the rows of the inverse derivatives matrix corresponding to the (co)variance parameters:'+eol;
      lf.Write(pchar(x)^,length(x));
      h := p - covariance_parameters;
      for j := (h + 1) to p do
        begin
        for k := 1 to p do
          begin
          dinv[j,k] := dinv[j,k] * scale[j];
          x := format('%10.6f',[dinv[j,k]]);
          lf.Write(pchar(x)^,length(x));
          end;
        lf.Write(pchar(eol)^,length(eol));//writeln(lf);
        end;
      outf.Write(pchar(eol)^,length(eol));//writeln(lf);
      end;

    (* ertext = '' iff the matrix dfra is invertible . *)
    if (ertext > '')
    then begin
      (* add a ridge *)
      x := 'Error message for inversion of dfra: '+ertext+eol;
      cf.Write(pchar(x)^,length(x));
      for j:=pp downto 1 do dfra[j,j] := dfra[j,j] + 1.0;
      pinverse(pp,dfra,dinv,ertext);
      x := 'Intervention 1.4.1: ridge added after phase 1.'+eol;
      cf.Write(pchar(x)^,length(x));
    end;
    for j:=1 to pp
    do if (abs(dfra[j,j]) < 1e-8)
    then begin
      x := format('%s%d%s',['Diagonal element ',j,' of derivative matrix amended in phase 1.'])+eol;
      cf.Write(pchar(x)^,length(x));
      dfra[j,j] := 1e-3;
    end;
    (* fchange is the change in theta : *)
    if (ertext = '')
    then pprod(pp,dinv,fra,fchange)
    else begin
      x := 'Warning. After phase 1, derivative matrix non-invertible even with a ridge.'+eol;
      cf.Write(pchar(x)^,length(x));
      for j:=pp downto 1 do fchange[j] := 0.0;
    end;
    x := 'dfra :'+eol;
    cf.Write(pchar(x)^,length(x));
    writeoutmat(dfra);
    x :='inverse of dfra :'+eol;
    cf.Write(pchar(x)^,length(x));
    writeoutmat(dinv);

    x := 'Full Quasi-Newton Raphson step after phase 1:'+eol;
    cf.Write(pchar(x)^,length(x));
    for j:=1 to pp do begin
      x := format('%3d%s%12.6f',[j,'. ',-fchange[j]])+eol;
      cf.Write(pchar(x)^,length(x));
      end;
    (* Quasi-Newton Raphson step * firstg *)
(* TS: used to be 0.1, experiment with 0.5; 31-jan-06 *)
    x := format('%s%8.5f%s',['This step is multiplied by the factor ',0.5*firstg,' .'])+eol;
    cf.Write(pchar(x)^,length(x));
    for j:=1 to pp do fchange[j] := 0.5*firstg*fchange[j];
    if somefixed
    then begin
      for j:=1 to pp
      do begin
        if fixed[j] then fchange[j] := 0.0;
      end;
    end;

    if randomeffects then
      begin (*MS: The Newton-Raphson step ignores actor-dependent random effects, hence it is sensible only for fixed coefficients, not for random coefficients.*)
      for j := 1 to pp do
        begin
        if thetarandom[j] then fchange[j] := 0.0;
        end;
      end;
    (* Check if jump is too large:
       truncate the jump proportionally so that on no coordinate
       it is larger than 10*scale[j] *)
    maxrat := 1.0;
    for j:=1 to pp
    do if (abs(fchange[j])/scale[j] > maxrat)
    then maxrat := abs(fchange[j])/scale[j];
    if (maxrat > 10.0)
    then begin
      for j:=1 to pp do fchange[j] := 10.0*fchange[j]/maxrat;
      x := format('%s%12.2f%s',['Intervention 1.4.2: jump after phase 1 decreased by factor ',maxrat,'.'])+eol;
      cf.Write(pchar(x)^,length(x));
    end;
    (* Check positivity *)

    if anyposj then for j:=1 to pp do
      if (posj[j] and (fchange[j] >= theta[j]))  then
        begin
        fchange[j]:= 0.5*theta[j];
        x := format('%s%s%d%s',['Intervention 1.4.3: ','positivity restriction after phase 1, coordinate ',j,' .'])+eol;
        cf.Write(pchar(x)^,length(x));
        end;

    if (nsub >= 1) then minus(pp,theta,fchange,theta);  (*Update parameter coordinates.*)

    x := format('%s%d%s',['Phase 1 achieved after ', n1,' iterations.'])+eol;
    cf.Write(pchar(x)^,length(x));
    writeouttheta;
  end; (*MS: if not(Bayesian)*)

  it1 := n; (*MS: The present value of <n> is used at the end of Phase 2.*)

  (* Phase 2:  Main iterations: *)
  (* 2.1 Initialisation of phase 2. *)
  EstForm.PLtext.visible := false;
  n2min0 := 7+pp;
  gain := 2.0*firstg;
  maxacor := 999.0;
  minacor := 1.0;
  inc(rep2);
  Phase := 2;
  if Bayesian then nsub := 1; (*MS: burn-in.*)
  nsubphas := nsub;
  time1 := now;
  maxrepeatsubphase := 4;
  n2partsum[0] := 0;
  cf.Write(pchar(eol)^,length(eol));//writeln(cf);
  x := format('%s%d%s',['Phase 2 has ',nsubphas,' subphases.'])+eol;
  cf.Write(pchar(x)^,length(x));
  if (nsubphas <=0) then begin
    x :='With 0 subphases, you decided to skip phase 2!!'+eol;
    cf.Write(pchar(x)^,length(x));
  end;
  if (maxrepeatsubphase >= 2)
  then begin
    x := format('%s%d%s',['Each subphase can be repeated up to ',maxrepeatsubphase,' times.'])+eol;
    cf.Write(pchar(x)^,length(x));
  end;
  for subphase := 1 to nsubphas
  do begin
    estage := 21; (*MS: Estimation phase: phase 2, iteration 1.*)
    if maxlike then gain := 0.25*gain else gain := 0.5*gain;
    if ((observations = 1) and (subphase = 3)) then gain := 0.5*gain;
    n2min0 := trunc(n2min0 * 2.52);
    (* 2.5198421 = 2^(4/3); this gives a gain parameter of order n^(-3/4) *)
    if Bayesian then n2min := 100 (*Burn-in.*)
    else n2min := n2min0;
    n2min := n2min0;
    n2max := n2min + 200;
    if (subphase <= 10) then n2partsum[subphase] := n2partsum[subphase-1] + n2max;
    repeatsubphase := 0;
    if maxlike then burnextra(outf,lf);
    repeat
      inc(repeatsubphase);
      AnnouncePhase;
      for j:=1 to pp
      do begin
        prod0[j] := 0.0;
        prod1[j] := 0.0;
      end;
      truncated := 0;
      positivized := 0;
      thav := theta;
      if VertexPar then InitialiseVertexAverage;
      (* thav is the sum, later average,
         of theta for the iterations in the current subphase. *)
      if checktime then ctime := now;
      nit := 0;
      time1 := now;

     if Continuous then
        begin (* start of short burnin because of change of parameter value *)
        rstep := rstep*5.0;
        burning2 := true;
        EstForm.PLtext.caption := 'Extra burn-in (start subphase)';
        EstForm.PLtext.visible := true;
        Application.ProcessMessages;
        end;

      (* 2.2 Iterations *)
      repeat
        Inc(n);
        Inc(nit);
        if Continuous and (nit = 2) and (burning2) then
          begin (* end of short burnin  *)
          rstep := rstep/5.0;
          burning2 := false;
          EstForm.PLtext.visible := false;
          Application.ProcessMessages;
          end;
        if ((subphase = 1) and (nit = 11))
        then begin
          time1 := 86400*(now - time1);
          (* number of seconds elapsed in 10 FRANs,
             except if midnight intervened *)
          if (time1 > 1e-5)
          then writefreq := round(20.0/time1)
          else writefreq := 20;
{$if defined(batch)}
          writefreq := 10*writefreq;
{$ifend}
          roundfreq(writefreq);
        end; (* nit = 11 *)
        if ((nit <= 10) or ((nit mod writefreq) = 0)) then
          begin
          nitstring := '  '+inttostr(nit);
          if (nit < 10) then nitstring := '  '+nitstring;
          if (nit < 100) then nitstring := '  '+nitstring;
          EstForm.EIteration.Text:= nitstring;
          end;
        if ((nit mod writefreq) = 0) then EstForm.ProgressBar.StepBy(writefreq);
{$if defined(batch)}
        if ((nit mod writefreq) = 0) then begin
          Write('Phase ',Phase,' Subphase ', subphase,' Iteration ', nit:4);
          Writeln(' Progress ',(EstForm.ProgressBar.Position/EstForm.ProgressBar.Max*100):4:0,'%');
        end;
{$ifend}

        (* 2.2.1 Generation of random variables *)
        th := theta;
        prev_th := theta;
        if maxlike then sea := true else sea := SearchFeasibleRegion;
        repeat
          feasible := true;
          tpc := false;
          FRAN(pp,th,false,false,ranoseed,fra,dist,df,
               false,feasible, error,ertext,cf,lf,estage,tpc); // @@@@

ConnectedCheck(cf); 
          SearchFeasibleRegion := sea;   // @@@@@
          estage := 22; (*MS: Estimation phase: phase 2, iteration > 1.*)
          if error
            then begin
            DefineError;
            exit;
          end;
          if ((not feasible) and (not SearchFeasibleRegion) and (not (maxlike or bayesian)))
          then begin
            if (nit = 1)
            then begin
              StartfeasibilitySearch;
              goto endofend;
            end
            else begin
                th := prev_th;
              (* In this case, the previous theta did lead to a feasible outcome.
                 Therefore we may continue just until another feasible outcome
                 occurs, and then continue our way. *)
            end;
          end;
        until (feasible or stopo);

        for h:=1 to pp do
          begin
          fra[h] := fra[h] - m[h];
          if (fra[h] < minim[h]) then minim[h] := fra[h];
          if (fra[h] > maxim[h]) then maxim[h] := fra[h];
          end;
//writeln(cf);
//for h:=1 to pp do write(cf,fra[h]:8:4,' ');
//writeln(cf);
        if ((nit <= 10) or ((nit mod writefreq) = 0)) then writeouttheta2;
        (* fra is the current realized value of y-m *)

        (* 2.2.2 Updates. Note that dfra[j,j] remains constant in phases 2-3. *)
        if odd(nit)
        then prev_fra := fra
        else begin
          maxacor := -99.0;
          minacor := 1.0;
          for j:=1 to pp
          do begin
            prod0[j] := prod0[j] + sqr(fra[j]);
            prod1[j] := prod1[j] + fra[j]*prev_fra[j];
            if (prod0[j]>1e-12)
            then ac := divi(prod1[j],prod0[j])
            else ac := -2.0;
            if ((not fixed[j]) and (ac > maxacor)) then maxacor := ac;
            if ((not fixed[j]) and (ac > -1.0) and (ac < minacor))
              then minacor := ac;
          end;
        end; (* if not odd(nit) *)

        (* Check if jump is too large; if necessary, rescale by maxrat *)
        maxrat := 1.0;
        for j:=1 to pp
        do if (divi(abs(fra[j]),sd[j]) > maxrat)
          then maxrat := divi(abs(fra[j]),sd[j]);
        if (maxrat > maxmaxrat)
        then begin
          maxrat := maxmaxrat/maxrat;
           (* maxrat used to decrease step size, see definition fchange below *)
          if (truncated > 0)
          then begin
            x := format('%d%s',[n,'; ']);
            cf.Write(pchar(x)^,length(x));
            if ((truncated mod 10) = 0) then cf.Write(pchar(eol)^,length(eol));//writeln(cf);
          end
          else begin
            if (positivized > 0) then cf.Write(pchar(eol)^,length(eol));//writeln(cf);
            x := format('%s%d%s%d%s%d%s',['Intervention ',phase,'.',subphase,'.1: change truncated, iteration ',n,';'])+eol;
            cf.Write(pchar(x)^,length(x));
          end;
          inc(truncated);
          positivized := 0;
        end
        else maxrat := 1.0; (* i.e., when maxrat <= maxmaxrat *)

        (* Define jump *)
        if (exponential or MaxLike)                         // @@@@@
        then begin      // changed version 2.0m
                        // note this has no maxrat - truncation; if necessary, add later.
          pprod(pp,dinv,fra,fchange);
          for j:=1 to pp
          do if fixed[j]
          then fchange[j] := 0.0
          else fchange[j] := gain*fchange[j];
        end else
(* We could study if alo for the MoM with observations >= 2,
   it is more efficient to use the full matrix dinv than only the
   diagonal elements 1/dfra[j,j].  *)
          for j:=1 to pp do begin
          fchange[j] := (gain*fra[j]*maxrat)/dfra[j,j];
          if fixed[j] then fchange[j] := 0.0;
        end;
        (* Check positivity restriction : *)
        if anyposj then for j:=1 to pp do
          if (posj[j] and (fchange[j] >= theta[j])) then
            begin
            fchange[j]:= 0.5*theta[j];
            if (positivized > 0) then
              begin
              x := format('%d%s',[n,'; ']);
              cf.Write(pchar(x)^,length(x));
              if ((positivized mod 10) = 0) then cf.Write(pchar(eol)^,length(eol));//writeln(cf);
              end
             else
              begin
              if (truncated > 0) then cf.Write(pchar(eol)^,length(eol));//writeln(cf);
              x := format('%s%d%s%d%s%d%s',['Intervention ',phase,'.',subphase,'.2: positivity restriction, coordinate ',j,' .'])+eol;
              cf.Write(pchar(x)^,length(x));
              end;
            truncated := 0;
            inc(positivized);
            end;

        (* update for matrices, theta and thav:
           updatematrices;                   *)
        (*MS*)
        if Bayesian then
          begin
          theta := th;
          prev_th := th;
          end
         else minus(pp,theta,fchange,theta);
        (*MS*)
        if VertexPar then VertexStep(gain,outf);
        for j:=1 to pp do thav[j] := thav[j] + theta[j];
        prev_th := theta;

        (* 2.2.3 checks *)
        CheckBreak;
        if stopo then goto EndOfEnd;
        if (starto or nextt) then goto startagain;
        if nexto
        then begin
          EarlyEndPhase2 := true;
          goto EndPhase2;
        end;
        NullChecks;
        if (Bayesian and (nit = 100)) then goto Startphase34;
(* ERGM case:
   Check if there is an error in the calculation of statistics
   by means of updating by change statistics .
        if ((observations <= 1) and (Gibbs in [1,4,5,6,7,9,10,11,12])) then
          begin
          if (not CheckRunningStatistics(outf,lf)) then
            begin
            ertext := 'Error in use of change statistics';
            DefineError;
            exit;
            end;
          end;
*)

        (* 2.2.4 Detection of divergence - canceled from the procedure.   *)

        (* stop criterion for subphase iterations : *)
      until (((nit >= n2min) and (maxacor < 1e-10)) or (nit >= n2max)
             or ((nit >= 50) and (minacor < -0.8) and (repeatsubphase < maxrepeatsubphase)));
// added TS minacor / repeatsubphase condition 8-2-06

      if (minacor >= -0.8) then repeatsubphase := maxrepeatsubphase;

      if (maxacor >= sqrt(2.0/(nit+1)))
      then begin
        x := 'Warning: an autocorrelation is positive at the end of this subphase.'+eol;
        cf.Write(pchar(x)^,length(x));
        x := 'Autocorrelations: '+eol;
        cf.Write(pchar(x)^,length(x));
        for j:=1 to pp
        do begin
          if (prod0[j] > 1e-20)
          then begin
            x := format('%8.4f%s',[prod1[j]/prod0[j],'  ']);
            cf.Write(pchar(x)^,length(x));
            end
          else begin
            x := '   --   ';
            cf.Write(pchar(x)^,length(x));
            end;
          if ((j mod 10) = 0) then cf.Write(pchar(eol)^,length(eol));//writeln(cf);
        end;
        if ((pp mod 10) <> 0) then cf.Write(pchar(eol)^,length(eol));//writeln(cf);
        if (repeatsubphase < maxrepeatsubphase)
        then begin
          x := format('%s%d%s',['Restart subphase 2.',subphase,'.'])+eol;
          cf.Write(pchar(x)^,length(x));
          end;
        if (minacor < -0.8) then begin
            x := format('%s%8.4f%s%d%s',['Minimum autocorrelation = ',
              minacor,' after ',nit, ' iterations in this subphase.'])+eol;
            cf.Write(pchar(x)^,length(x));
            end;
      end; (* if *)
    until ((repeatsubphase >= maxrepeatsubphase) or (maxacor < 1e-10));

    EndPhase2:
    if EarlyEndPhase2
    then begin
      x := 'The user asked for early termination of phase 2.'+eol;
      outf.Write(pchar(x)^,length(x));
      end;

    (* 2.2.5 Last actions for this subphase *)
    if checktime then begin
      x := format('%s%d%s%d%s%8.4f',[ 'Time per iteration in phase ',phase,'.',subphase, ' = ',86400*(now - ctime)/nit])+eol;
      lf.Write(pchar(x)^,length(x));
      end;
    if not(Bayesian) then (*MS*)
      begin
      for j:=1 to pp do thav[j] := thav[j]/(nit+1.0);
      theta := thav;
      end; (*not(Bayesian)*)
    if VertexPar then VertexAverage(nit);
    writeouttheta2;
    maxacor := -99.0;
    for j:=1 to pp
    do begin
      if (prod0[j]>1e-20) then ac := prod1[j]/prod0[j] else ac := -1.0;
      if ((not fixed[j]) and (ac > maxacor)) then maxacor := ac;
    end;
    if (maxacor < sqrt(2.0/(nit+1)))
    then begin
      if (positivized + truncated > 0) then cf.Write(pchar(eol)^,length(eol));//writeln(cf);
      x := format('%s%d%s%d',['Phase ',phase,'.',subphase]);
      cf.Write(pchar(x)^,length(x));
      x := format('%s%d%s',[' ended after ',nit,' iterations.'])+eol;
      cf.Write(pchar(x)^,length(x));
      writeouttheta;
      if (SearchFeasibleRegion  and (not (maxlike or bayesian)))
      then begin
        EndFeasibilitySearch;
        goto endofend;
      end;
    end
    else begin
      x := 'Autocorrelation criterion not satisfied.'+eol;
      cf.Write(pchar(x)^,length(x));
      x := 'Autocorrelations: '+eol;
      cf.Write(pchar(x)^,length(x));
      for j:=1 to pp
      do begin
        if (prod0[j] > 1e-20)
        then begin
          x := format('%8.4f%s',[prod1[j]/prod0[j],'  ']);
          cf.Write(pchar(x)^,length(x));
          end
        else begin
          x := '   --   ';
          cf.Write(pchar(x)^,length(x));
        end;
        if ((j mod 10) = 0) then cf.Write(pchar(eol)^,length(eol));//writeln(cf);
      end;
      if ((pp mod 10) <> 0) then cf.Write(pchar(eol)^,length(eol));//writeln(cf);
      if ((rep2 < maxrep2) and (not nexto))
      then begin
        writeouttheta;
        x := 'Restart from this parameter value.'+eol;
        cf.Write(pchar(x)^,length(x));
        goto startagain;
      end
      else begin
         x := format('%s%d%s%d',['Phase ',phase,'.',subphase]);
         cf.Write(pchar(x)^,length(x));
         x := format('%s%d%s',[' ended after ',nit,' iterations.'])+eol;
         cf.Write(pchar(x)^,length(x));
         x := 'Warning. Autocorrelation criterion not satisfied';
         cf.Write(pchar(x)^,length(x));
         if (maxrep2 >= 2)
         then begin
           cf.Write(pchar(eol)^,length(eol));//writeln(cf);
           x := format('%s%d%s',['after ',maxrep2,' repetitions of phases 1-2.'])+eol;
           cf.Write(pchar(x)^,length(x));
         end
         else begin
          x := '.'+eol;
          cf.Write(pchar(x)^,length(x));
          end;
         writeouttheta;
      end;
    end; (* if (maxacor >= sqrt(2/(nit+1))) *)
    if nexto
    then begin
      repforep := 1;
      goto Startphase34;
    end;
    (* termination subphase 2.x: *)
    if ((subphase = 2) and restart)
    then begin
      x := 'Restart after subphase 2.2 from current parameter values'+eol;
      cf.Write(pchar(x)^,length(x));
      x := 'because the initial values used led to questionable epsilon values.'+eol;
      cf.Write(pchar(x)^,length(x));
      for j:=1 to pp do if newfixed[j] then fixed[j] := false;
      restarted := true;
      goto Startagain;
    end;
  end; (* subphase *)

  if (SearchFeasibleRegion  and (not (maxlike or bayesian)))
  then begin
    EndFeasibilitySearch;
    goto endofend;
  end;
  (* end of Phase 2 *)

  (* Phase 3/4 : Final iterations with constant theta for estimating
                 rate parameter, derivative matrix and covariance matrix : *)
  (* 3.1 Initialisation Phase 3. *)
  repforep := 1;

  StartPhase34:
  if checktime then ctime := now;
  for j:=0 to 100 do begin
    entropy[j]:= 0;
    entropynumber[j] := 0;
   end;
  for h:=1 to pp do
    begin
    for j := 0 to 20 do cnts[h,j] := 0;
    outside[h] := 0;
    if (nsubphas >= 1) and (maxim[h] > minim[h]+1e-8) then hist[h] := true
      else hist[h] := false;
    end;
//  if maxlike then writechain := true;
  if burning then rstep := rstep/20.0; (* this occurs if nsubphas = 0 *)


  (*MS*)
  estage := 31; (*Estimation phase: phase 3, iteration 1.*)
  Initialise_d_est(lf,3,n3,pp,theta,false); (*MoM: initialise derivatives estimation.*)
  if d_estAB then
    begin
    x := 'Estimation of derivatives by the LR method';
    outf.Write(pchar(x)^,length(x));
    if (d_estimator <= 2) then begin
      x := format('%s%d%s',[' (type ',d_estimator,').'])+eol;
      outf.Write(pchar(x)^,length(x));
      end
      else begin
        x:= '.'+eol;
        outf.Write(pchar(x)^,length(x));
        end;
    end;
   if FinDiff_method then begin
    x := 'Estimation of derivatives by the finite differences method.'+eol;
    outf.Write(pchar(x)^,length(x));
    end;
   if (homogeneous > 0) then (*Actor / period homogeneity test.*)
    begin
    derivatives := true; (*Needed for numerical estimation of derivatives: <theta[p] + epsilon[p]> for the tested actor / period coordinate.*)
    PrepareHTests(outf);(*Prepare homogeneity tests.*)
    x := Description[coordinate]+eol;
    outf.Write(pchar(x)^,length(x));
    if (homogeneous < 20) then
      begin
      x := 'with respect to the identified actor(s):'+eol;
      outf.Write(pchar(x)^,length(x));
      for j := 1 to n do
        begin
        if outlier[j] then begin
          x := format('%s%d',[' ',j]);
          outf.Write(pchar(x)^,length(x));
          end;
        end;
      outf.Write(pchar(eol)^,length(eol));//writeln(outf);
      ObservedMomentActors(observations,homogeneous,k); (*Calculate observed moment.*)
      m[pp] := k;
      end
    else
      begin
      x := format('%s%d%s',['with respect to period ',period,'.']);
      outf.Write(pchar(x)^,length(x));
      ObservedMomentPeriod(statistics,outf);
      m[pp] := statistics[coordinate];
      end;
    theta[pp] := theta[coordinate];
    fix[pp] := true;
    fixed[pp] := true;
    somefixed := true;
    alternative := 999999999;
    end; (* if homogeneous > 0 *)
  if (Bayesian or randomeffects) then
    begin
    nno := (n - it1) * it2; (*<nno> is the number of iterations in Phase 2,
    which is the number of Phase 1 and 2 iterations
    minus the number of Phase 1 iterations
    times the number of iterations in <S_ml>.*)
    it5 := 10; (*The batches correspond to samples of size <it2 * it5> iterations.*)
    ldim := trunc(n3 / it5);
    SetArray(sumposre,nactors,nre);
    SetArray(mcI_theta,pp,ldim);
    SetArray(mcII_theta,pp,ldim);
    SetArray(mc_re,nactors,nre,ldim);
    SetArray(mc_posre,nactors,nre,ldim);
    SetArray(mc_covre,nre,ldim);
    SetVectorValue(pp,MI_posterior_theta,0); (*Used by Bayesian algorithm to sum over the first moment of the sampled parameter vectors.*)
    SetMatrixValue(pp,MII_posterior_theta,real1); (*Used by Bayesian algorithm to sum over the second moment of the sampled parameter vectors.*)
    end;
  if initialFEestimate then goto Endofend;
  (*MS*)

  inc(Phase);
  if (phase = 3) then nsubphas := 1 else nsubphas := 10;
  if (not (exponential or maxlike or Bayesian))
  then begin
    if d_estAB then writefreq := writefreq div 2
     else writefreq := writefreq div (pp);
    roundfreq(writefreq);
  end; (* if *)

  for subphase :=1 to nsubphas
  do begin
    AnnouncePhase;
    if (phase = 4)
    then begin
      outf.Write(pchar(eol)^,length(eol));//writeln(outf);
      x := format('%s%d%s',['For subphase ',subphase,', new parameter values:'])+eol;
      outf.Write(pchar(x)^,length(x));
      for j:= 1 to pp do begin
        x := format('%d%s%18.6f',[j,'. ',theta[j]]);
        outf.Write(pchar(x)^,length(x));
      end;
      outf.Write(pchar(eol)^,length(eol));//writeln(outf);
    end;

    if exponential then IndependentStart := false;
    (* a check for phase 3, so that if the algorithm can get stuck,
       it indeed has the chance to do so *)
    if (phase = 3)
    then for j:=1 to distDimension
    do begin
      mdist[j] := 0.0;
      mdist2[j] := 0.0;
    end;

    for j:=1 to pp
    do begin
      for h:=1 to pp
      do begin
        sdf[j,h] := 0.0; sff[j,h] := 0.0;
      end;
      sf[j] := 0.0;
      sfl[j] := 0.0;
    end;

    (* 3.2 Iterations *)
    nit := 0;
    repeat
      inc(nit);
      Inc(n);
      phaseI_III := nit;
      if ((nit <= 5) or (nit = 10) or ((nit mod writefreq) = 0))
      then begin
        nitstring := '  '+inttostr(nit);
        if (nit < 10) then nitstring := '  '+nitstring;
        if (nit < 100) then nitstring := '  '+nitstring;
        EstForm.EIteration.Text:= nitstring;

      end;
      if ((nit mod writefreq) = 0)
      then begin
        if (exponential or d_estAB or maxlike)
        then EstForm.ProgressBar.StepBy(writefreq)
        else
        EstForm.ProgressBar.StepBy(writefreq*(pp+1));
{$if defined(batch)}
        write('Phase ',phase,' Iteration ',nit:4);
        writeln(' Progress ',(EstForm.ProgressBar.Position/EstForm.ProgressBar.Max*100):4:0,'%');
{$ifend}
      end;

      (* 3.2.1 Generation of random variables and estimated derivatives *)
      th := theta;
      seed0 := ranoseed; (*MS: only needed by the Crude Monte Carlo method to estimate partial derivatives.*)
      feasible := true;
      if (nit = n3) then estage := 33; (*Final iteration.*)
      if maxlike then sea := (nit <= 1) else sea := SearchFeasibleRegion;
      tpc := ((estage > 30) and PPC);
      FRAN(pp,th,false,MaxLike,ranoseed,fra,dist,df,
        true,feasible, error,ertext,cf,lf,estage,tpc);

ConnectedCheck(cf);         
      estage := 32; (*MS: Estimation phase: phase 3, iteration > 1.*)
      if PPC then PPChecks(nit); (*MS: posterior predictive checks.*)
      if Bayesian then theta := th; (*MS: Last parameter drawn from the posterior.*)
      it1 := nit; (*MS: Iteration number.*)
      Store_MCMC_Sample(Bayesian,randomeffects,it1,it2,nactors,pp,nre,MI_posterior_theta,MII_posterior_theta,sumre,sumposre,mcI_theta,mcII_theta,sumsigma,mc_re,mc_posre,mc_covre); (*MS: Average and store MCMC samples, and clean up.*)
      if FinDiff_method then FiniteDifferences; (* Finite differences method to estimate partial derivatives: additional simulations.*)
      if error
      then begin
        DefineError;
        exit;
      end;
      (* 3.2.2 Updates *)
      for h:= pp downto 1 do
        begin
        fra[h] := fra[h] - m[h];
(* Prepare data for the histogram:
   if there was no phase 2, then the first 100 iterations of phase 3
   are used to get the minimum and maximum values. *)
        if (nsub >= 1) or (nit > 100) then
          begin
          if hist[h] then
            begin
            j := round(20*(fra[h]-minim[h])/(maxim[h]-minim[h]));
            if (j >= 0) and (j <= 20) then cnts[h,j] := cnts[h,j]+1
             else inc(outside[h]);
            end;
          end
         else
          begin
          if (fra[h] < minim[h]) then minim[h] := fra[h];
          if (fra[h] > maxim[h]) then maxim[h] := fra[h];
          if (nit = 100) and (maxim[h] > minim[h]+1e-8) then hist[h] := true;
          end;
        end;
      (* fra is the current realized value of y-m *)
      if (phase = 3) then
        begin
        if not(exponential) then
          begin
          for j:=1 to distDimension do
            begin
            mdist[j] := mdist[j] + dist[j];
            mdist2[j] := mdist2[j] + sqr(dist[j]);
            end;
          end;
        end; (* if phase = 3 *)

      for h:=1 to pp
      do begin
        sf[h] := sf[h] + fra[h];
      end;
      for h:=pp downto 1 do for j:=pp downto 1
      do begin
        sff[j,h] := sff[j,h] + fra[j]*fra[h];
        if maxlike then sdf[j,h] := sdf[j,h] + df[j,h];
      end;
      for j:=pp downto 1
      do begin
        if (nit >= 2) then sfl[j] := sfl[j] + fra[j]*fral[j];
        fral[j] := fra[j];
      end;

      (* msf is the estimated covariance matrix of fra *)

      (* 3.2.3 checks and output *)
      if (nit = 1) then begin
        cf.Write(pchar(eol)^,length(eol));//writeln(cf);
        x := 'Simulated values, phase 3. '+eol;
        cf.Write(pchar(x)^,length(x));
       end;
      if ((nit mod 25) = 0) then begin
        x := format('%4d%s',[nit,' ']);
        cf.Write(pchar(x)^,length(x));
        for h:=1 to pp do begin
          x := format('%10.4f%s',[fra[h],' ']);
          cf.Write(pchar(x)^,length(x));
        end;
        cf.Write(pchar(eol)^,length(eol));//writeln(cf);
       end;

      if ((nit <= 3) or (nit mod writefreq = 0)) then writeouttheta2;

      if (nit >= 10)
      then begin
        CheckBreak;
        if (nit = 10) then EstForm.StopKey.hint := 'Early termination of estimation';
        if stopo
        then begin
          x := 'The user asked for an early stop of the algorithm during phase 3.'+eol;
          outf.Write(pchar(x)^,length(x));
          if (nit < 500)
          then begin
            if (EarlyEndPhase2)
            then begin
              x := 'This implies that ';
              outf.Write(pchar(x)^,length(x));
              end
            else begin
              x := 'This implies that the estimates are as usual,'+eol;
              outf.Write(pchar(x)^,length(x));
              x := 'but ';
              outf.Write(pchar(x)^,length(x));
            end;
            x := 'the diagnostic checks, covariance matrices, and t-values '+eol;
            outf.Write(pchar(x)^,length(x));
            x := format('%s%d%s',['are less reliable, because they are now based on only ',nit,
                         ' phase-3 iterations.'])+eol;
            outf.Write(pchar(x)^,length(x));
          end; (* if nit < 500 *)
          goto EndPhase3;
        end; (* if stopo *)
        if starto then goto startagain;
        NullChecks;
      end;

    until (((phase=3) and (nit >= n3)) or ((phase=4) and (nit >= n4extra)));
    (* repeat .. until loop for nit; end of phase 3 iterations *)

    (* 3.3 End calculations phase 3. *)
    EndPhase3:
    if maxlike then writechain := false;
    if checktime then begin
      x := format('%s%d%s%8.4f',['Time per iteration in phase ',phase, '   = ',86400*(now - ctime)/nit])+eol;
      lf.Write(pchar(x)^,length(x));
      end;
    for j:=pp downto 1 do
      begin
      sf[j] := sf[j]/nit;
      end;

    if (nit >= 2)
    then for j:=pp downto 1
    do for h:=pp downto 1
    do msf[j,h] := (nit/(nit-1.0))*(sff[j,h]/nit - sf[j]*sf[h]);
    if (nit >= 3) then for j:=pp downto 1 do
      if (msf[j,j] > 1e-20)
       then sfl[j] := (sfl[j]/(nit-1) - sqr(sf[j]))/msf[j,j]
       else sfl[j] := 999.0;
    if exponential then (*ERGM*)
      begin
      dfra := msf;
      for j := pp downto 1 do diver[j] := false;
      end
    else (*non-ERGM*)
      begin
      if d_estAB then JacobianMatrix(pp,cf,outf,dfra,jacwarning) (*MS: moment estimation.*)
      else if (maxlike or FinDiff_method) then
        begin
        for j := pp downto 1 do
          begin
          for h := pp downto 1 do dfra[j,h] := sdf[j,h]/nit;
          end;
        end;
      for j := pp downto 1 do diver[j] := ((AllFixed and (abs(dfra[j,j]) < 1e-6)));
      end;


    Potential_NR(MakeStep3); (*MS: Potential Newton-Raphson step: what would be the increment in case...
    Important note: as side-product, the inverse of the derivatives matrix is evaluated here.
    The normal argument of Potential_NR is false;
    if true, then the step is actually carried out.
    This is experimental to see if it speeds up convergence. *)


    for j:= 1 to pp
    do if newfixed[j]
    then begin
      outf.Write(pchar(eol)^,length(eol));//writeln(outf);
      x := 'There was a problem in obtaining convergence.'+eol;
      outf.Write(pchar(x)^,length(x));
      x := format('%s%d%s',
      ['Therefore, the program decided tentatively to fix parameter ',j,'.'])+eol;
      outf.Write(pchar(x)^,length(x));
      x := 'It may be better to start all over again, '+eol;
      outf.Write(pchar(x)^,length(x));
      x := 'with better initial values or a reduced model.'+eol;
      outf.Write(pchar(x)^,length(x));
      x := '(Check that you entered the data properly!)'+eol;
      outf.Write(pchar(x)^,length(x));
    end;

    if (subphase = nsubphas) then
      begin
      if Bayesian then
        begin
        Heading(2,outf,'End of MCMC algorithm.'); 
        x :=format('%d%s%d%s',[nit,' x ',it2,' samples from posterior distribution generated.'])+eol;
        outf.Write(pchar(x)^,length(x));
        end
      else (*not(Bayesian)*)
        begin
        Heading(2,outf,'End of stochastic approximation algorithm, phase '+inttostr(phase)+'.');
        x := format('%s%d%s',['Total of ',n,' iterations.'])+eol;
        outf.Write(pchar(x)^,length(x));
        if (nsubphas >= 1) then
          begin
          x := format('%s%d%s',['Parameter estimates based on ', n-nit,' iterations,'])+eol;
          outf.Write(pchar(x)^,length(x));
          if (cconditional and (observations >=2 )) then
            begin
            x := 'basic rate parameter';
            outf.Write(pchar(x)^,length(x));
            if (observations >= 3) then outf.write(pchar(string('s'))^,1);
            x := ' as well as '+eol;
            outf.Write(pchar(x)^,length(x));
            end;
          end;
        x := format('%s%d%s',['convergence diagnostics, covariance and derivative matrices based on ',nit,' iterations.'])+eol;
        outf.Write(pchar(x)^,length(x));
        outf.Write(pchar(eol)^,length(eol));//writeln(outf);
        x := 'Information for convergence diagnosis.'+eol;
        outf.Write(pchar(x)^,length(x));
        x := 'Averages, standard deviations, and t-ratios for deviations from targets:'+eol;
        outf.Write(pchar(x)^,length(x));
        tmax := 0.0;
        // assignfile(brieffile,pname+'.bof');
        // if fileexists(pname+'.bof') then append(brieffile) else rewrite(brieffile);
{
assignfile(f09,pname+'.f09');
if fileexists(pname+'.f09') then append(f09) else rewrite(f09);
assignfile(f11,pname+'.f11');
if fileexists(pname+'.f11') then append(f11) else rewrite(f11);
}

        if not fileexists(pname+'.bof') then
        begin
          brieffile:= TFileStream.create(pname+'.bof',fmCreate);
          brieffile.free;
          brieffile := TFileStream.Create(pname+'.bof',fmOpenWrite or fmSHareDenyWrite);
        end
        else  begin
           brieffile:= TFileStream.Create(pname+'.bof',fmOpenReadWrite or fmShareDenyWrite);
           brieffile.seek(0,soFromEnd);
        end;
        x := DateTimeToStr(Now)+eol;
        brieffile.Write(pchar(x)^,length(x));
        brieffile.Write(pchar(eol)^,length(eol));//writeln(brieffile);
        if cconditional then
          begin
          x := 'conditional moment estimation.'+eol;
          brieffile.Write(pchar(x)^,length(x));
          end else
          begin
          if MaxLike then
            begin
            x := 'Maximum Likelihood estimation.'+eol;
            brieffile.Write(pchar(x)^,length(x));
            end
           else if (not Bayesian) then
           begin
           x := 'unconditional moment estimation.'+eol;
           brieffile.Write(pchar(x)^,length(x));
           end;
         end;
        x := 'Information for convergence diagnosis.'+eol;
        brieffile.Write(pchar(x)^,length(x));
        x :='Averages, standard deviations, and t-ratios for deviations from targets:'+eol;
        brieffile.Write(pchar(x)^,length(x));
        for j:=1 to pp do
          begin
          if (msf[j,j] < 1e-20*sqr(scale[j])) then
            begin
            msf[j,j] := 1e-20*sqr(scale[j]);
            if (abs(sf[j]) < 1e-10*scale[j]) then tstat := 0.0 else tstat := 999.0;
            end
           else tstat := sf[j]/sqrt(msf[j,j]);
          x := format('%3d%s%10.3f%s%8.3f%s%8.3f',[j,'. ',sf[j],' ',sqrt(msf[j,j]),' ', tstat]);
          outf.Write(pchar(x)^,length(x));
          if fixed[j] then begin
            x := '   (fixed parameter)'+eol;
            outf.Write(pchar(x)^,length(x));
            end
          else outf.Write(pchar(eol)^,length(eol));//writeln(outf);
          x := format('%3d%s%10.3f%s%8.3f%s%8.3f',[j,'. ',sf[j],' ',sqrt(msf[j,j]),' ', tstat])+eol;
          brieffile.Write(pchar(x)^,length(x));
          if ((not fixed[j]) and (not BasicRateFunction[j])
             and (abs(tstat) > tmax) and (resist[j] > 0.9)) then tmax := abs(tstat);
          tconv[j] := tstat;
          end; (* j *)
        error := ((abs(tmax) > 4.0/sqrt(nit)) and (abs(tmax)> 0.3));
        if ((tmax >= 0.4) and (Errorcode = 0)) then Errorcode := 4;
       // closefile(brieffile);
        brieffile.Free;

{if (p >= 11) then
begin
//write(f09,' & ',sqrt(msf[9,9])/20:6:2);        // between runs
//write(f11,' & ',(20/19)*((sf[11]+62.4327)/20 - sqr((sf[9]+29)/20)):6:2); // within runs
write(f09,' & ',sqrt(msf[9,9])/26:6:2);        // between runs
write(f11,' & ',(26/25)*((sf[11]+62.4327)/26 - sqr((sf[9]+29)/26)):6:2); // within runs
end;
closefile(f09);
closefile(f11);
}
        outf.Write(pchar(eol)^,length(eol));//writeln(outf);
        x := 'Good convergence is indicated by the t-ratios ';
        outf.Write(pchar(x)^,length(x));
        j:=0;
        repeat
          inc(j);
          if fixed[j] then begin
            x := 'of non-fixed parameters ';
            outf.Write(pchar(x)^,length(x));
            end;
        until ((j=pp) or fixed[j]);
        x := 'being close to zero.'+eol;
        outf.Write(pchar(x)^,length(x));
        if (nit < 100) then
          begin
          x := format('%s%d%s',['(Since the diagnostic checks now are based only on ',nit, ' iterations,'])+eol;
          outf.Write(pchar(x)^,length(x));
          x := 'they are not reliable.)'+eol;
          outf.Write(pchar(x)^,length(x));
          end;

        if (error and (subphase=nsubphas)) then
          begin
          x :='One or more of the t-statistics are rather large.'+eol;
          outf.Write(pchar(x)^,length(x));
          if (tmax > 0.5) then
            begin
            x := 'Convergence of the algorithm is doubtful.'+eol;
            outf.Write(pchar(x)^,length(x));
            if (reptotal <= 0) then
              begin
              x := 'Try again from this initial value.'+eol;
              outf.Write(pchar(x)^,length(x));
              x := 'Current parameter values: '+eol;
              outf.Write(pchar(x)^,length(x));
              for j:=1 to pp do begin
                x := format('%10.4f%s',[theta[j],' '])+eol;
                outf.Write(pchar(x)^,length(x));
                end;
              goto startagain;
              end;
            if ((observations = 1) and (reptotal <= 0)) then
              begin
              x := 'Try again from this initial value and smaller initial stepsize a_1.'+eol;
              outf.Write(pchar(x)^,length(x));
              x := 'Current parameter values: '+eol;
              outf.Write(pchar(x)^,length(x));
              for j:=1 to pp do begin
                x := format('%10.4f',[theta[j],' '])+eol;
                outf.Write(pchar(x)^,length(x));
                end;
              firstg := 0.5*firstg;
              x := format('%s%10.6f',['Step size ',firstg])+eol;
              outf.Write(pchar(x)^,length(x));
              x := format('%s%d%s',['Number of subphases increased by 1 to ',nsub,'.'])+eol;
              outf.Write(pchar(x)^,length(x));
              inc(nsub);
              writeouttheta;
              goto startagain;
              end;
            end;
          if (nit <= 50)
          then begin
            x := 'However, the standard deviations are based on few simulations.'+eol;
            outf.Write(pchar(x)^,length(x));
            end;
          end; (* if error *)
      end; (*not(Bayesian)*)
    end;

    if ((observations = 1) or MaxLike)
    then begin
      outf.Write(pchar(eol)^,length(eol));//writeln(outf);
      x := 'Autocorrelations during phase 3 : '+eol;
      outf.Write(pchar(x)^,length(x));
      for j:=1 to pp do begin
        x := format('%3d%s%8.4f',[j,'. ',sfl[j]])+eol;
        outf.Write(pchar(x)^,length(x));
        end;
      outf.Write(pchar(eol)^,length(eol));//writeln(outf);
    end; (* if observations = 1 *)

    if (subphase = nsubphas)
    then begin
      for j:=1 to pp
      do begin
        if diver[j]
        then begin
          x := format('%s%d%s',['Warning. Extremely large standard error of parameter ',j,'.'])+eol;
          outf.Write(pchar(x)^,length(x));
          if (sf[j] < 0.5*sqrt(msf[j,j]))
          then begin
            x := 'Presumably this parameter must be fixed.'+eol;
            outf.Write(pchar(x)^,length(x));
            end
          else begin
            x := 'Maybe the algorithm diverged.'+eol;
            outf.Write(pchar(x)^,length(x));
            end;
//          writeln(outf,'Standard errors of other parameters were calculated');
//          writeln(outf,'as if this parameter was fixed.');
// Dit heb ik weggehaald omdat dfra al eerder is geinverteerd tot dinv.
          for h:=pp downto 1
          do begin
            dfra[j,h] := 0.0;
            dfra[h,j] := 0.0;
          end;
          dfra[j,j] := 1e-4;
        end; (* if *)
      end; (* for j *)
    end; (* if subphase = nsubphas *)

    if (phase = 3)
    then begin
      for j:=1 to distDimension
      do begin
        mdist[j] := mdist[j]/(nit-1.0);
        md[j] := mdist[j];
        vmd[j] := ((nit+2.0)/nit)*(mdist2[j]/(nit-1.0) - sqr(mdist[j]));
      end;

      if MaxLike then
        begin
        cf.Write(pchar(eol)^,length(eol));//writeln(cf);
        x := 'Estimated complete data information matrix: '+eol;
        cf.Write(pchar(x)^,length(x));
        writeoutmat(dfra);
        x :=
'Estimated conditional covariance matrix score function (unobserved information): '+eol;
        cf.Write(pchar(x)^,length(x));
        writeoutmat(msf);
        cf.Write(pchar(eol)^,length(eol));//writeln(cf);
        for j:=1 to pp do for h:=1 to pp do
          begin
          dfrac[j,h] := dfra[j,h] - msf[j,h];
          if (fixed[j] or fixed[h]) then
            begin
            if (j=h) then dfrac[j,h] := 1.0 else dfrac[j,h] := 0.0;
            end;
          end;
        pinverse(pp,dfrac,cov,ertext);
        if (ertext > '') then begin
          x := 'Noninvertible estimated covariance matrix : '+ertext+eol;
          outf.Write(pchar(x)^,length(x));
          errorcode := 6;
         end;
        end (* if maxlike *)
         else tripleproduct(pp,dinv,msfc,cov);

      error := false;

      if not(Bayesian) then pinverse(pp,msfc,msfinv,ertext);
      if (ertext > '')
      then begin
        x := 'Covariance matrix not positive definite: '+ertext+eol;
        outf.Write(pchar(x)^,length(x));
        if somefix then begin
          x := '(This may be unimportant, and related to the fact'+eol;
          outf.Write(pchar(x)^,length(x));
          x := ' that some parameters are fixed.)'+eol;
          outf.Write(pchar(x)^,length(x));
         end else
        begin
        x := 'This may mean that the reported standard errors are invalid.'+eol;
        outf.Write(pchar(x)^,length(x));
        end;
      end;
      for j:=pp downto 1
      do begin
        diver[j] := (fixed[j] or diver[j] or (cov[j,j] < 1e-9)) and (not AllFixed);
        if diver[j]
        then begin
          for h:=pp downto 1
          do begin
            cov[j,h] := root(cov[h,h])*33.0;
            cov[h,j] := root(cov[h,h])*33.0;
          end;  (* h *)
          cov[j,j] := 999.0;
        end (* if diver[j] *);
      end; (* j *)
      D := dfra; (* not dfrac !! *)
      s := msf;  (* not msfc !! *)
      nits := nit;
     end; (* if phase = 3 *)

    (*MS*)
    InitialiseTest(test,pp,delta,dfra,omikron,msf); (*Prepares arrays that are used in case score tests are carried out.*)
    ScoreTest(Bayesian,pp,test,test_CovRE,sf,delta,omikron,maxlike,run,testresult,test_n,one_step_est);
    if not(Bayesian) then InstabilityAnalysis(pp,test,diver,cov,lf,msf); (*Check instability.*)
    ldim := ActiveNetworkeffects_l + ActiveActionEffects_l;  // TS: Michael, wat doet dit statement?
    (*MS*)
    if extra
    then begin
      devi := 0.0;
      for j:=1 to pp
      do for h:=1 to pp
      do devi := devi+sf[j]*sf[h]*msfinv[j,h];
      x := 'weighted sum of deviations after ';
      outf.Write(pchar(x)^,length(x));

      if (phase = 3)
      then begin
        x := format('%s%d',['phase ',phase]);
        outf.Write(pchar(x)^,length(x));
        end
      else begin
        x := format('%s%d',['subphase ',subphase]);
        outf.Write(pchar(x)^,length(x));
      end;
      x := format('%s%16.6f',[' = ',devi])+eol;
      outf.Write(pchar(x)^,length(x));

      if (subphase < nsubphas)
      then begin
        minus(pp,theta,fchange,theta);  (* update for theta *)
      end;
    end; (* if extra *)

  end; (* for subphase := 1 to nsubphas *)

  if ((Phase=3) and extra) then goto StartPhase34;
  theta := thav;

(* TS 04-jan-05. The following is necessary if there are several
   parallel rate parameters, of which one is used for conditioning.
   r1, rn1, r2, rn2 are input parameters set before entering PolRup. *)
  if (rn1 >= 1) then for j:=1 to rn1 do
    for h:=1 to distDimension do begin
      theta[r1-1+(j-1)*DistDimension+h] :=
        mdist[h]*theta[r1-1+(j-1)*DistDimension+h];
      for k:=pp downto 1 do begin
        cov[r1-1+(j-1)*DistDimension+h,k] := mdist[h]*cov[r1-1+(j-1)*DistDimension+h,k];
        cov[k,r1-1+(j-1)*DistDimension+h] := mdist[h]*cov[k,r1-1+(j-1)*DistDimension+h];
       end;
    end; (* if rn1 >= 1 then for j for h *)
  if (rn2 >= 1) then for j:=1 to rn2 do
    for h:=1 to distDimension do begin
      theta[r2-1+(j-1)*DistDimension+h] := mdist[h]*theta[r2-1+(j-1)*DistDimension+h];
      for k:=pp downto 1 do begin
        cov[r2-1+(j-1)*DistDimension+h,k] := mdist[h]*cov[r2-1+(j-1)*DistDimension+h,k];
        cov[k,r2-1+(j-1)*DistDimension+h] := mdist[h]*cov[k,r2-1+(j-1)*DistDimension+h];
       end;
    end; (* if rn2 >= 1 then for j for h *)

  for j:=pp downto 1
  do begin
    diver[j] := (fixed[j] or diver[j] or (cov[j,j] < 1e-9)) and (not AllFixed);
    if diver[j]
    then begin
      for h:=pp downto 1
      do begin
        cov[j,h] := root(cov[h,h])*33.0;
        cov[h,j] := root(cov[h,h])*33.0;
      end;  (* h *)
      cov[j,j] := sqr(99.0);
    end (* if diver[j] *);
  end; (* j *)

  MessageText := 'Regular end of estimation algorithm.';
  if (Errorcode <> 4) then ErrorCode := 0;

  Endofend:

  (*MS*)
  if (Bayesian or randomeffects) then
    begin
    CloseFiles;
    it1 := nit; (*These numbers must not be manipulated!*)
    it3 := trunc(nit / it5); (*Number of batches.*)
    it4 := it2 * it5; (*Batch size: total number of iterations in <S_ml> times <it5>.*)
    MCMC_final_report(Bayesian,lf,randomeffects); (*Autocorrelations of functionals whose expectation is to be approximated: basic information.*)
    end;
  if randomeffects then
    begin
    input := 'batchmeans_RE';
    MC_Error_III(eb,nactors,nre,it3,mc_re,mcre,it4,lf,input); (*Compute mean, set <eb> to the mean, and compute the Monte Carlo s.e. corresponding to the approximations of the desired integrals by ergodic averages: variance and covariances.*)
    input := 'batchmeans_posRE';
    MC_Error_III(posre,nactors,nre,it3,mc_posre,mcposre,it4,lf,input); (*Compute mean, set <posre> to the mean, and compute the Monte Carlo s.e. corresponding to the approximations of the desired integrals by ergodic averages: variance and covariances.*)
    if Bayesian then
      begin
      input := 'batchmeans_CovRE';
      MC_Error_II(nre,covRE,it3,mc_covre,cov_covRE,it4,lf,input); (*Compute mean, set <covRE> to the mean, and compute the Monte Carlo s.e. corresponding to the approximations of the desired integrals by ergodic averages: variance and covariances.*)
      end;
    end;
  if Bayesian then
    begin
    input := 'batchmeans_MI_theta';
    MC_Error_I(pp,theta,it3,c,mcI_theta,it4,lf,input); (*Compute first moment, set <theta> to the first moment, and compute the Monte Carlo s.e. corresponding to the approximations of the desired integrals by ergodic averages: variance and covariances. Standard deviations will be computed in the output procedure.*)
    SetAlpha(theta); (*Sets <alpa> to first moment: required for rewriting the <.mo>-file.*)
    input := 'batchmeans_MII_theta';
    MC_Error_II(pp,MII_posterior_theta,it3,mcII_theta,cov2,it4,lf,input);  (*Compute second moment, set <mcII_posterior_theta> to the second moment, and compute the Monte Carlo s.e. corresponding to the approximations of the desired integrals by ergodic averages: variance and covariances.*)
    ldim := nactors * nre;
    Get_MCMC_Output(Bayesian,ldim); (*Obtain (and transform) MCMC output.*)
    end;
   if (randomeffects or Bayesian) then Choose_MC_SE_Estimate(lf,pp,nit=n3,mcRE,cov_covRE,c,var_theta); (*There are two estimates of the Monte Carlo variance: choose and prepare arrays.*)
  (*MS*)

  if stopo then MessageText := 'User stop of estimation process.';
  outf.Write(pchar(eol)^,length(eol));//writeln(outf);
  if not(simulate or initialFEestimate)
  then begin
    finalize(dist);
    finalize(distfd);
    finalize(mdist);
    finalize(mdist2);
  end;

  NullChecks;

  CompleteSFiles(run,cconditional,observations,nza,pp,fixed,theta); (*MS: multiple sequences: write parameter estimates to Matlab file.*)
  if (homogeneous > 0) then derivatives := false;

  if histograms and (ErrorCode in [0,4]) then
    begin
    if (nsub = 0) then
      begin
      wwriteln(outf,'Histograms of frequencies in phase 3,');
      wwriteln(outf,'first 100 iterations used only to get minima and maxima.');
      end
     else
      begin
      wwriteln(outf,'Histograms of frequencies in phase 3,');
      wwriteln(outf,'minima and maxima over phases 1-2.');
      end;
    wwriteln(outf,'');
    for h:=1 to pp do if hist[h] then
      begin
      for j := 0 to 20 do cntsh[j] := cnts[h,j];
      wwriteln(outf,'Statistic '+inttostr(h)+': '+StatsDescriptions[h]);
      nicestring1(minim[h]+m[h],mins);
      nicestring1(maxim[h]+m[h],maxs);
      nicestring1(m[h],mhs);
      wwriteln(outf,'minimum (x=0) = '+mins+
                    ', maximum (x=20) = '+maxs+
                    ', observed (o) = '+mhs);
      wwriteln(outf,'');
      j := round(20*(0.0-minim[h])/(maxim[h]-minim[h]));
      fshistogram(cntsh,0,20,50,j,outside[h],outf);
      wwriteln(outf,'');
//      wwriteln(outf,'Histograms in logfile.');
      end;
    end;

  x := 'End of procedure PolRup in S_Est.'+eol;
  lf.Write(pchar(x)^,length(x));
  x := format('%s%d%s%s',['Error code ',Errorcode,': ',ErrorCodeString(errorcode)])+eol;
  lf.Write(pchar(x)^,length(x));
  if (errorcode <> 0) then
    begin
    if (errorcode = 4) then
      begin
      x := format('%s%d%s%s',['Warning code ',Errorcode,': ',ErrorCodeString(errorcode)])+eol;
      outf.Write(pchar(x)^,length(x));
      //writeln(lf,'Warning code ',Errorcode,': ',ErrorCodeString(errorcode));
      lf.Write(pchar(x)^,length(x));
      end
    else
      begin
      x := format('%s%d%s%s',['Error code ',Errorcode,': ',ErrorCodeString(errorcode)])+eol;
      outf.Write(pchar(x)^,length(x));
      //writeln(lf,'Error code ',Errorcode,': ',ErrorCodeString(errorcode));
      lf.Write(pchar(x)^,length(x));
      end;
    end;

  if (sequences > 1) then (*MS: multiple sequences: make sure that the same analysis is not repeated over and over again.*)
    begin
    modelseed := ranoseed; (*Store the last seed; to be used in the following analysis.*)
    seedsown := false; (*Seed not used before.*)
    end;

end; (* POLRUP *)

procedure MLEstimateParameter(p, mode: byte; var theta: pvec; var Cov: pmatrix);
(* Maximum Likelihood estimation *)
(* Used now only for pseudo-ml estimates in observations=1 case *)
(* Mode 0: one Quasi-Newton Raphson step;
   Mode 1: iterated Quasi-Newton Raphson steps,
           but in each step, if NR is not indicated,
           a steepest descent step is made;
   Mode 2: iterated steepest descent steps;
   Mode 3: first multistep steepest descent steps,
           then NR steps;
           in each step where NR is not indicated,
           multistep steepest descent steps are made. *)
 (* theta is the input (initial) and output parameter vector *)
const epsi = 1e-12; (* Convergence criterion *)
      eps = 0.0001;  (* Argument difference for approximating derivatives *)
      nstepmax = 100; (* maximum number of iteration steps *)
type
    pvecplus = array[1..pmax] of double;
    pmatrixplus = array[1..pmax,1..pmax] of double;
var dev0,deviance: double;
  ste: Boolean;
  i,j,multistep,nsteps: integer;
  thet: pvec;     (* the working parameter *)
  insensitive: array[1..pmax] of Boolean;
  x:string;
procedure ThetaIncrement(i: byte; r: real);
begin
  thet[i] := thet[i] + r;
end; (* ThetaIncrement *)
procedure VecSum(a,b: pvecplus; var c: pVec; r: real);
var i: byte;
begin
  for i:=1 to p do c[i] := a[i]+r*b[i];
end; (* VecSum *)
procedure NormLength(var a: pVec; var s: real);
var i: byte;
begin
  s := 0.0;
  for i:=1 to p do s := s + sqr(a[i]);
  s := root(s);
  if (s > 0.0) then for i:=1 to p do a[i]:=a[i]/s;
end; (* NormLength *)
procedure NewtonRaphsonStep(option: byte; var stepp: boolean);
(* stepp indicates whether a step has been made *)
var i,j: byte;
    theta0: pVecplus;
    d1,step: pvecplus;
    d2: pmatrixplus;
    d2r: pmatrix;
    d0: double;
    s: real;
    ert: shortstring;
    outs: boolean;
    x: string;
procedure NoStepNoCov;
var i,j: byte;
begin
  stepp := false;
  for i:=1 to p do for j:=1 to p do Cov[i,j] := 9999.0;
end;
begin (* NewtonRaphsonStep *)
  inc(nsteps);
  halter := false;
  stepp := false;
  for i:=1 to pmax do theta0[i] := thet[i];
  d0 := Pseudodeviance(thet,halter,outf,lf);
  if halter then exit;
(* approximate first and second derivatives by difference quotients *)
  for i:=1 to p do if (not fix[i]) then begin
    ThetaIncrement(i,eps);
    d1[i] := Pseudodeviance(thet,halter,outf,lf);
    ThetaIncrement(i,-eps);
   end else d1[i] := d0;
  for i:=1 to p do for j:=1 to i do
   if (not(fix[i] or fix[j])) then begin
    ThetaIncrement(i,eps); ThetaIncrement(j,eps);
    d2[i,j] := (Pseudodeviance(thet,halter,outf,lf) - d1[i] - d1[j] + d0)/sqr(eps);
    if (j < i) then d2[j,i] := d2[i,j];
    ThetaIncrement(i,-eps); ThetaIncrement(j,-eps);
   end (* for i, j if *) else begin
     d2[i,j] := 0.0;
     if (j < i) then d2[j,i] := 0.0 else d2[i,i] := 1/9999.0;
    end;
  if halter then exit;
  for i:=1 to p do d1[i] := (d1[i]-d0)/eps;
(* Now d1 is vector of 1st, d2 matrix of 2nd derivatives. *)
  for i:=1 to pmax do for j:=1 to pmax do d2r[i,j] := d2[i,j];
  for i:=1 to p do if (d2[i,i] <= 0.0) then begin
    insensitive[i] := true;
    x := format('%s%d%s',['Pseudolikelihood insensitive for parameter ',i,'.'])+eol;
    cf.Write(pchar(x)^,length(x));
   end (* for i if *) else insensitive[i] := false;
  pinverse(p,d2r,Cov,ert);
  if (ert > '') then begin
(*   Matrix inversion error in Quasi-Newton Raphson procedure. *)
    (* add a ridge *)
    x := 'Error in matrix inversion: '+ert+eol;
    cf.Write(pchar(x)^,length(x));
    for j:=1 to p do d2r[j,j] := d2r[j,j] + 1.0;
    pinverse(p,d2r,Cov,ert);
    if (ert > '') then begin
      x:= 'Error in matrix inversion: '+ert+eol;
      cf.Write(pchar(x)^,length(x));
      NoStepNoCov; exit;
     end;
   end; (* if *)
(* If one of the diagonal elements of Cov is negative, the deviance function
   is not locally convex and a Quasi-Newton Raphson step will not be made: *)
  outs := false;
  for i:=1 to p do if ((not fix[i]) and (Cov[i,i] <= 0)) then
    begin outs := true; end;
  if outs then begin (* NoStepNoCov; *) exit; end;

(* Calculate the step to be made and make it: *)
  for i:=1 to p do if fix[i] then step[i] := 0.0 else begin
    step[i] := 0.0;
    for j:=1 to p do step[i] := step[i] - Cov[i,j]*d1[j];
   end; (* for i if *)
  VecSum(theta0,step,thet,1.0);
  deviance := Pseudodeviance(thet,halter,outf,lf);

(* Check whether the step is an improvement;
   if not, take action depending on option. *)
  case option of
  1: begin
       s := 1.0;
       while ((deviance >= d0) and (s > 1e-20)) do begin
    (* halve the step *)
         s := 0.5*s;
         VecSum(theta0,step,thet,s);
         deviance := Pseudodeviance(thet,halter,outf,lf);
        end; (* while *)
       stepp := (deviance < d0);
     end; (* case 1 *)
   2: if ((deviance-d0) > 0.0) then begin
(* skip the step that was set *)
        for i:=1 to pmax do thet[i] := theta0[i];
        stepp := false;
        deviance := d0;
       end else stepp := true;
    end; (* case *)
  if halter then exit;
  x := format('%s%20.10f',['End Quasi-Newton Raphson step, deviance = ', deviance])+eol;
  cf.Write(pchar(x)^,length(x));
end; (* NewtonRaphsonStep *)

procedure Descentstep(var step: boolean);
(* step indicates whether a step has been made *)
var i: byte;
  d0,d1,d2: double;
  length: real;
  de1,theta0: pVecplus;
  de1r: pvec;
  x : string;
begin
  inc(nsteps);
  step := false;
  d0 := Pseudodeviance(thet,halter,outf,lf);
  if halter then exit;
(* approximate first derivatives *)
  for i:=1 to p do if not fix[i] then begin
    ThetaIncrement(i,eps);
    de1[i] := (Pseudodeviance(thet,halter,outf,lf)-d0)/eps;
    ThetaIncrement(i,-eps);
   end else de1[i] := 0.0;
  for i:=1 to p do de1r[i] := de1[i];
  NormLength(de1r,length);
  if (length < 1e-10) then exit;
(* de1 is the direction of steepest ascent.
   Calculate first and second derivatives in this direction,
   and then calculate length := stepsize according to
   onedimensional Quasi-Newton Raphson method. *)
  for i:=1 to pmax do theta0[i] := thet[i];
  VecSum(theta0,de1,thet,eps);
  d1 := Pseudodeviance(thet,halter,outf,lf);
  VecSum(theta0,de1,thet,2.0*eps);
  d2 := Pseudodeviance(thet,halter,outf,lf);
  if (abs(d0 - 2.0*d1 + d2) < 1e-20) then exit else
    length := (eps*(d1-d0))/(d0 - 2.0*d1 + d2);
(* if length < 0.0 then the 1-dimensional function is not locally convex,
   and a Newton-Raphson step size makes no sense. *)
  if (length <= 0.0) then begin length := 1.0; end;
  i:= 0;
  repeat
(* Make the step and halve the step size if the step yields no improvement. *)
    inc(i);
    VecSum(theta0,de1,thet,-length);
    length := 0.5*length;
    deviance := Pseudodeviance(thet,halter,outf,lf);
   until ((deviance <= d0) or (i >= 20));
  if (deviance >= d0) then begin
    for i:=1 to pmax do thet[i] := theta0[i];
    deviance := d0; step := false;
   end else step := true;
  x := format('%s%20.10f',['Descent step, dev = ', deviance])+eol;
  cf.Write(pchar(x)^,system.length(x));
end; (* Descentstep *)

begin (* MLEstimate *)
  nsteps := 0;
  if (observations = 1) then begin
    if Continuous then IndependentStart := false else IndependentStart := true;
   end;
  initialise_actors(1);
  initialise_y(1);
  if ((Gibbs = 8) and (observations = 1)) then
    for i:= 1 to n do for j:=1 to n do
      if ynm(i,j,1) then setyn(i,j,true) else setyn(i,j,false);
  thet := theta;
  case mode of
  0: NewtonRaphsonstep(1,ste);
  1: begin
      deviance := Pseudodeviance(thet,halter,outf,lf);
      x := format('%s%20.10f',['Initial pseudodeviance = ', deviance])+eol;
      cf.Write(pchar(x)^,length(x));
      if halter then exit;
      for i:=1 to 5 do NewtonRaphsonstep(1,ste);
      repeat
        dev0 := deviance;
        NewtonRaphsonstep(1,ste);
        if (not ste) then DescentStep(ste);
        deviance := Pseudodeviance(thet,halter,outf,lf);
      until ((abs(deviance - dev0) < epsi) or (nsteps > nstepmax));
      NewtonRaphsonstep(2,ste);
     end; (* case 1 *)
  2: begin
      deviance := Pseudodeviance(thet,halter,outf,lf);
      if halter then exit;
      repeat
        dev0 := deviance;
        DescentStep(ste);
        deviance := Pseudodeviance(thet,halter,outf,lf);
      until ((abs(deviance - dev0) < epsi) or (nsteps > nstepmax));
      NewtonRaphsonstep(2,ste);
     end; (* case 2 *)
   3: begin
       multistep := p;
       if (multistep < 5) then multistep := 5;
       deviance := Pseudodeviance(thet,halter,outf,lf);
       if halter then exit;
       for i:=1 to multistep do Descentstep(ste);
       repeat
         dev0 := deviance;
         NewtonRaphsonStep(1,ste);
         deviance := Pseudodeviance(thet,halter,outf,lf);
         if (not ste) then begin
         (* no NR step was set, therefore some DC steps are made. *)
           i := 0;
           repeat
             dev0 := deviance;
             inc(i);
             Descentstep(ste);
             deviance := Pseudodeviance(thet,halter,outf,lf);
            until (i >= multistep);
          end; (* if not ste *)
        until ((abs(deviance - dev0) < epsi) or (nsteps > nstepmax));
      end; (* case 3 *)
  end; (* case *)
  for i:=1 to p do if insensitive[i] then Cov[i,i] := -1.0;
  for i:=1 to p do for j:=1 to p do
    if (insensitive[i] or insensitive[j]) then Cov[i,j] := -1.0
    else Cov[i,j] := Cov[i,j]*2.0;
(* The factor 2 is necessary because the minimized function
  is minus twice the loglikelihood. *)
  theta := thet;
end; (* MLEstimateParameter *)

procedure InitialEstimate;
(* For the initial value of the parameter,
   the user has the option of using the value calculated in this procedure,
   or (default) using the current parameter value *)
var i: integer;
x: string;
begin (* InitialEstimate *)
  x := 'Reset of initial parameter values.'+eol;
  lf.Write(pchar(x)^,length(x));
  for i:= 1 to ccmax do begin
    alpa_l[i] := 0.0;
    alpa_f[i] := 0.0;
    alpa_g[i] := 0.0;
   end;
  InitialAlphas(alpa_l,alpa_f,0,outf);
end; (* InitialEstimate*)


procedure results(var ouf,lf: TFileStream; matrs, limited, brief: boolean);

var thesame: boolean;

  procedure wm(var m: pmatrix; var pf: TFileStream);
  var i,j: byte;
  x: string;

    function cor(var m: pmatrix; i,j: byte):real;
    begin
      if (m[i,i]*m[j,j] > 1e-12)
      then cor := m[i,j]/sqrt(m[i,i]*m[j,j])
      else cor := 0.0;
    end; (* cor *)

    procedure wx(x: real);
    var sx: string;
    begin
      if (abs(x) >= 999.995)
      then begin
        sx:= format('%10.0f%s',[x,' ']);
        pf.Write(pchar(sx)^,length(sx));
      end
      else begin
        sx := format('%10.3f%s',[x,' ']);
        pf.Write(pchar(sx)^,length(sx));
      end;
    end; (* wx *)

  begin (* wm *)
   for i:=1 to p do begin
     for j:=1 to i-1 do begin
       if ((c[i,i] >= 0.0) and (c[j,j] >= 0.0)) then  begin
         x := format('%10.3f%s',[cor(m,i,j),' ']);
         pf.Write(pchar(x)^,length(x));
         end
         else begin
         x := '       --- ';
         pf.Write(pchar(x)^,length(x));
         end;
  (*     if ((j mod 6) = 0) then writeln(pf); *)
      end;
      if (c[i,i] >= 0.0)
        then wx(m[i,i]) else begin
          x := '       --- ';
          pf.Write(pchar(x)^,length(x));
          end;
  (*   if ((i mod 6) = 0) then writeln(pf); *)
     for j:=i+1 to p do begin
       if ((c[i,i] >= 0.0) and (c[j,j] >= 0.0))
        then wx(m[i,j]) else begin
        x := '       --- ';
        pf.Write(pchar(x)^,length(x));
        end;
  (*     if ((j mod 6) = 0) then writeln(pf); *)
      end;
     pf.Write(pchar(eol)^,length(eol));//writeln(pf);
    end; (* i *)
  end; (* wm *)

var i,j,ns,varh: integer;
    allfix: boolean;
    pr,s: real;
    degmax, demod: integer;
    perc5,perc50,perc95: array of integer;
    x: string;

begin (* results *)
 n := nni[observations];
 Setlength(perc5,n+1);
 Setlength(perc50,n+1);
 Setlength(perc95,n+1);
 if (not limited) then begin
  x := format('%s%d%s',['Total of ',nrun,' iteration steps.'])+eol;
  ouf.Write(pchar(x)^,length(x));
  end;

 if Bayesian then
   begin
   Heading(3,ouf,'Posterior moments I and II (including Monte Carlo s.e.) ');
   x := 'Posterior moment I of theta:'+eol;
   ouf.Write(pchar(x)^,length(x));
   ouf.Write(pchar(eol)^,length(eol));// writeln(ouf);
   end
 else Heading(3,ouf,'Estimates and standard errors ');

 (* Report extra rate parameters in case of conditional estimation *)
 if (CConditional and (observations >= 2)) then
   begin
   x := 'Rate parameters:'+eol;
   ouf.Write(pchar(x)^,length(x));
   if (observations = 2) then
     begin
     if ((nxa = 1) and (nza = 0)) then begin
      x := ' 0. Rate parameter                                    ';
      ouf.Write(pchar(x)^,length(x));
      end
      else begin
      x := ' 0. Rate parameter of conditioning variable         ';
      ouf.Write(pchar(x)^,length(x));
      end;
     if (symmetric and (ModelType in [4..6])) then begin
(* TS Ook hier log bij gezet *)
       x := format('%9.4f%s%9.4f%s',[root(mtime[1]),'  (',0.5*root(vmtime[1]),')'])+eol; // controleren ???
       ouf.Write(pchar(x)^,length(x));
       end
      else begin
      x := format('%9.4f%s%9.4f%s',[mtime[1],'  (',root(vmtime[1]),')'])+eol;
      ouf.Write(pchar(x)^,length(x));
      end;
     end
    else (* observations >= 3 *)
     begin
     h := 0;
     for i:=1 to observations-1 do
       if not skipi[i] then
       begin
       inc(h);
       if (nza <= 0) then
         begin
         if (i <= 9)
         then begin
          x := format('%s%d%s%2d%s',[' 0.',i,'   Rate parameter period ',i,'                       ']);
          ouf.Write(pchar(x)^,length(x));
          end
         else begin
          x := format('%s%d%s%2d%s',[' 0.',i,'  Rate parameter period ',i,'                       ']);
          ouf.Write(pchar(x)^,length(x));
          end;
         end
        else
         begin
         if (i <= 9)
         then begin
         x := format('%s%d%s%2d%s',[' 0.',i,'   Rate parameter cond. variable period ',i,'        ']);
         ouf.Write(pchar(x)^,length(x));
         end
         else begin
         x := format('%s%d%s%2d%s',[' 0.',i,'  Rate parameter cond. variable period ',i,'        ']);
         ouf.Write(pchar(x)^,length(x));
         end;
         end;
       if (symmetric and (ModelType in [4..6])) then begin
        x := format('%9.4f%s%9.4f%s',[root(mtime[h]),'  (',0.5*root(vmtime[h]),')']) +eol; // controleren ???
        ouf.Write(pchar(x)^,length(x));
        end
        else begin
        x := format('%9.4f%s%9.4f%s',[mtime[h],'  (',root(vmtime[h]),')'])+eol;
        ouf.Write(pchar(x)^,length(x));
        end;
       end; (* for i *)
     end;
   ouf.Write(pchar(eol)^,length(eol));//writeln(ouf);
   x := 'Other parameters:'+eol;
   ouf.Write(pchar(x)^,length(x));
   end; (* if (CConditional and (observations >= 2)) *)

 somefix := false;
 allfix := true;

 (*MS: Matlab file to generate histogram for Test statistic estimate / s.e. test statistic when multiple runs are executed.*)
 if (sequences > 1) then
   begin
   assignfile(stud,pname+'T.m');
   if (run = 1) then rewrite(stud)
   else append(stud);
   end;
 (*MS*)

 if (maxlike and randomeffects) then h := covariance_parameters (*MS: do not list estimates of (co)variance parameters, stored in <theta>.*)
 else h := 0;

 for i:=1 to p do if (not fix[i]) then allfix := false;
 for i:=1 to (p - h) do if ((c[i,i] >= 0.0) or fix[i]) (*MS: <h> is the number of (co)variance parameters in <theta>.*)
 then
   begin
   if (not brief) then
     begin
     if ((i = 1) and (observations >= 2) and (nza >= 1)) then begin
      x := 'Network Dynamics'+eol;
      outf.Write(pchar(x)^,length(x));
      end;
     if (nza >= 1) and (i = ActiveNetworkEffects_l +  ActiveNetworkEffects_f + ActiveNetworkEffects_g + 1)
       then
       begin
       outf.Write(pchar(eol)^,length(eol));//writeln(outf);
       x := 'Behavior Dynamics'+eol;
       outf.Write(pchar(x)^,length(x));
       end;
     end;
   if ((nza <= 1) or (DependentVariable[i] <= 0)) then  begin
     x := format('%2d%s%s%9.4f',[i,'. ',Alignstring(Description[i],50),theta[i]]);
     ouf.Write(pchar(x)^,length(x));
     end
    else begin
     x := format('%2d%s%s%9.4f',[i,'. ',Alignstring('<'+inttostr(DependentVariable[i])+'> '+Description[i],50),
                  theta[i]]) ;
     ouf.Write(pchar(x)^,length(x));
     end;
   if (fix[i] and (not allfix)) then
     begin
     x:= '  (  fixed  )'+eol;
     ouf.Write(pchar(x)^,length(x));
     somefix := true;
     end
   else
     begin
     if Bayesian then begin
      x := format('%s%9.4f%s',['  (',root(var_theta[i]),')'])+eol;
      ouf.Write(pchar(x)^,length(x));
      end
     (*MS: Bayesian: <c[i,i]> contains an batch mean estimate of the Monte Carlo variance,
     while <var_theta[i]> may contain an Initial Sequence Estimate of the Monte Carlo variance.*)
     else begin
      x := format('%s%9.4f%s',['  (',root(c[i,i]),')'])+eol;
      ouf.Write(pchar(x)^,length(x));
      end;
     (*MS*)
      if (sequences > 1) then
        begin
        if cconditional or (i >= observations) then write(stud,theta[i]/root(c[i,i]):20:10,' '); (*The rate estimates only swamp the Matlab histograms.*)
        end;
     (*MS*)
     end
     end
  else begin
  x := format('%2d%s%s%s%s',[i,'. ',Alignstring(Description[i],50),'       --- ','       --- '])+eol;
  ouf.Write(pchar(x)^,length(x));
  end;
 ouf.Write(pchar(eol)^,length(eol));//writeln(ouf);
 if (Cconditional and (nxa + nza >= 2)) then
   begin
   x := 'For conditional estimation, the standard errors of rate parameters'+eol;
   ouf.Write(pchar(x)^,length(x));
   x := 'not used for conditioning, are unreliable.'+eol;
   ouf.Write(pchar(x)^,length(x));
   end;

 CompleteTFiles(sequences,stud); (*MS: Complete Matlab source files containing code to plot the sampling distribution of test statistics.*) 

 if brief then exit;

 if (VertexPar and (not limited)) then ReportVertexParameters(ouf);

 thesame := true;
 for i:=1 to p do if (abs(thetabase[i] - initialtheta[i]) > 2e-2) then thesame := false;
// Following was for observations = 1, can be skipped by setting
//  condition to observations  = 0
 if ((observations = 0) and (not limited) and (baselinemodeltext > ''))
 then begin
   Heading(3,ouf,'Deviances ');
   if (not thesame) then begin
    x := 'Two deviance differences are given.'+eol;
    ouf.Write(pchar(x)^,length(x));
   end;
   x :='The ';
   ouf.Write(pchar(x)^,length(x));
   if (not thesame) then begin
      x := 'first ';
      ouf.Write(pchar(x)^,length(x));
      end;
   x := 'deviance difference is given with respect to the following null model:'+eol;
   ouf.Write(pchar(x)^,length(x));
   x := baselinemodeltext+eol;
   ouf.Write(pchar(x)^,length(x));
   x := 'Parameters of the null model:'+eol;
   ouf.Write(pchar(x)^,length(x));
   for i:=1 to p do begin
     x := format('%2d%s%s%9.4f',[i,'. ',Alignstring(Description[i],50),thetabase[i]])+eol;
     ouf.Write(pchar(x)^,length(x));
     end;
   ouf.Write(pchar(eol)^,length(eol));//writeln(ouf);
(*   mtime[1,2] in this case are the expected value of exp((thetabase-theta)'(y-target)),
     see POLRUP  *)
   x := 'Deviance difference               ';
   ouf.Write(pchar(x)^,length(x));
   x := format('%11.4f%s%11.4f%s',[ 2.0*log(mtime[1]),'    (s.e. ',
               2.0*divi(root(vmtime[1]/niters),mtime[1]),')'])+eol;
   ouf.Write(pchar(x)^,length(x));
 end; (* if *)

 if ((observations = 0) and (not limited) and (not thesame))
 then begin
   ouf.Write(pchar(eol)^,length(eol));//writeln(ouf);
   if (baselinemodeltext = '')
   then begin
     Heading(3,ouf,'Deviances ');
     x := 'The ';
     ouf.Write(pchar(x)^,length(x));
   end
   else begin
    x := 'The second ';
    ouf.Write(pchar(x)^,length(x));
   end;
   x := 'deviance difference is given with respect to the initial parameters:'+eol;
   ouf.Write(pchar(x)^,length(x));
   for i:=1 to p
   do begin
    x := format('%2s%s%9.4f',[i,'. ',Alignstring(Description[i],50),initialtheta[i]])+eol;
    ouf.Write(pchar(x)^,length(x));
    end;
   ouf.Write(pchar(eol)^,length(eol));//writeln(ouf);
   x := 'Deviance difference               ';
   ouf.Write(pchar(x)^,length(x));
   x := format('%11.4f%s%11.4f%s',[ 2.0*log(mtime[2]),'    (s.e. ',
                2.0*divi(root(vmtime[2]/niters),mtime[2]),')'])+eol;
   ouf.Write(pchar(x)^,length(x));
 end; (* if *)

 if matrs
 then begin
   if limited
   then Heading(3,ouf,'Covariance matrix')
   else Heading(3,ouf,'Covariance matrices');
   if somefix
   then begin
     x := '(Values of the covariance matrix of estimates '+eol;
     ouf.Write(pchar(x)^,length(x));
     x := ' are meaningless for the fixed parameters.)'+eol;
     ouf.Write(pchar(x)^,length(x));
     ouf.Write(pchar(eol)^,length(eol));//writeln(ouf);
   end;
   if not(Bayesian) then begin
     x := 'Covariance matrix of estimates (correlations below diagonal): '+eol;
     ouf.Write(pchar(x)^,length(x));
     wm(c,ouf);
    end;

   if bayesian then Output_MII_theta(Bayesian,ouf,p,MII_posterior_theta,cov2,theta,lf);
     (*MS: Output the second moment of the parameter vector, considered as vector-valued random variable.*)
   if randomeffects then OutputCovRE(ouf,nre,p,covariance_parameters,covRE,cov_covRE,c,Bayesian,Description);
     (*MS: Output the variance-covariance matrix of actor-dependent random effects.*)
   if (randomeffects or bayesian) then SaveOutput(maxlike,Bayesian,p,c,nre,covRE);
     (*MS: Save output that may be used in future runs to tune MCMC algorithms.*)

   if (not limited)
   then begin
     if (not (MaxLike or Bayesian)) and (observations >= 2)
     then begin
       ouf.Write(pchar(eol)^,length(eol));//writeln(ouf);
       x := 'Derivative matrix of expected statistics X by parameters and'+eol;
       ouf.Write(pchar(x)^,length(x));
       lf.Write(pchar(eol)^,length(eol));//writeln(lf);
       x := 'Derivative matrix of expected statistics X by parameters: '+eol;
       lf.Write(pchar(x)^,length(x));
       for i:=1 to p
       do begin
         for j:=1 to p
         do begin
           x:= format('%11.3f%s',[dd[i,j],' ']);
           lf.Write(pchar(x)^,length(x));
(*         if ((j mod 6) = 0) then writeln(lf); *)
         end;
         lf.Write(pchar(eol)^,length(eol));//writeln(lf);
       end; (* i *)
      end; (* if not MaxLike or Bayesian *)
     if not(Bayesian or randomeffects) then begin
       if (observations <= 1) then begin
         x := 'Covariance/correlation matrix of X is written to file '+pname+'.log.'+eol;
         ouf.Write(pchar(x)^,length(x));
         end
        else begin
        x := 'covariance/correlation matrix of X are written to file '+pname+'.log.'+eol;
         ouf.Write(pchar(x)^,length(x));
         end;
       lf.Write(pchar(eol)^,length(eol));//writeln(lf);
       x := 'Covariance matrix of X (correlations below diagonal): '+eol;
       lf.Write(pchar(x)^,length(x));
       wm(sx,lf);
       ouf.Write(pchar(eol)^,length(eol));//writeln(ouf);
       lf.Write(pchar(eol)^,length(eol));//writeln(lf);
       end;
     (*MS*)
     if randomeffects then OutputEB_RE(ouf,nactors,nre,eb,mcre,posre,mcposre); (*Write (Empirical) Bayes estimates of actor-dependent random effects (including posterior odds and corresponding Monte Carlo s.e.) to output file.*)
     if not (Bayesian) then TestOutput(test_n,ouf,maxlike,pp,test,Description,homogeneous,theta,one_step_est); (*Checks whether a test was executed and---if yes---outputs the results.*)
     (*MS*)
   end;
 end; (* if matrs *)

 if (observations <= 1) and (nrun >= 10) then  begin
    x := format('%s%7.4f%s',['Number of Gibbs/MH steps with a change = ',
                SumGibbsChangeFraction/nrun,'.'])+eol;
 lf.Write(pchar(x)^,length(x));
 end;

 if (OutputType >= 2)
 then begin
   for i:=1 to n
   do begin
     j:= -1; demod := 0;
     while ((demod < 0.05*niters) and (j < n))
     do begin
       Inc(j);
       demod := demod+SimDegrees[i,j];
     end;
     perc5[i] := j;
     while ((demod < 0.5*niters) and (j < n))
     do begin
       Inc(j);
       demod := demod+SimDegrees[i,j];
     end;
     perc50[i] := j;
     repeat
       Inc(j);
       demod := demod+SimDegrees[i,j];
     until ((demod >= 0.95*niters) or (j >= n));
     perc95[i] := j;
   end; (* i *)

   x :='Model check: comparison of observed with expected out-degrees'+eol;
   ouf.Write(pchar(x)^,length(x));
   if (observations = 1)
   then Heading(3,ouf,'Ordered out-degrees')
   else Heading(3,ouf,'Ordered out-degrees at last observation moment');
   x := '   i'+'  obs'+'  median '+' 5 perc'+' 95 perc'+eol;
   ouf.Write(pchar(x)^,length(x));
   for i:=1 to n
   do begin
    x := format('%4d%s%4d%s%4d%s%4d%s%4d',[i,' ',ordgs[i],'  ',perc5[i],'   ',perc50[i],'    ',perc95[i]])+eol;
    ouf.Write(pchar(x)^,length(x));
    end;
   ouf.Write(pchar(eol)^,length(eol));//writeln(ouf);

(* Here comes some help for Latex users: a picture using Pictex. *)
    degmax := perc95[n];
    if (ordgs[n] > degmax) then degmax := ordgs[n];
    x := 'Picture of ordered degrees in Latex format:'+eol;
    lf.Write(pchar(x)^,length(x));
    x :=
      'use PicTex and newcommands \gast and \rast (e.g., green and red asterisks) '+eol;
    lf.Write(pchar(x)^,length(x));
    x := '  for (colored) plot symbols.'+eol;
    lf.Write(pchar(x)^,length(x));
    lf.Write(pchar(eol)^,length(eol));//writeln(lf);
    x := '\begin{center}                                    '+eol;
    lf.Write(pchar(x)^,length(x));
    x := '\begin{minipage}[htb]{22cm}                       '+eol;
    lf.Write(pchar(x)^,length(x));
    x := '\beginpicture                                     '+eol;
    lf.Write(pchar(x)^,length(x));
    x := '\setcoordinatesystem units <'+eol;
    lf.Write(pchar(x)^,length(x));
    demod := 200 div (n+1);
    if (demod < 1) then demod := 1;
    x := format('%d%s',[ demod,'mm,']);
    lf.Write(pchar(x)^,length(x));
    demod := 150 div (degmax+1);
    if (demod < 1) then demod := 1;
    x := format('%d%s',[ demod,'mm> point at 0 0   '])+eol;
    lf.Write(pchar(x)^,length(x));
    x := '\setplotarea x from 0 to ';
    lf.Write(pchar(x)^,length(x));
    x := format('%d%s',[n+1,', y from 0 to ']);
    lf.Write(pchar(x)^,length(x));
    x := format('%d',[degmax])+eol;
    lf.Write(pchar(x)^,length(x));
    x := '\axis bottom                                      '+eol;
    lf.Write(pchar(x)^,length(x));
    x := '   ticks out short                                '+eol;
    lf.Write(pchar(x)^,length(x));
    if (n > 40) then demod := 10 else demod := 5;
    x := '   numbered at ';
    lf.Write(pchar(x)^,length(x));
    i := 1;
    repeat
      x := format('%s%d',[' ',i*demod]);
      lf.Write(pchar(x)^,length(x));
      inc(i);
    until (i*demod >= n);
    x := ' / '+eol;
    lf.Write(pchar(x)^,length(x));
    x := '     /                                            '+eol;
    lf.Write(pchar(x)^,length(x));
    x := '\axis left                                        '+eol;
    lf.Write(pchar(x)^,length(x));
    x := '   ticks out short                                '+eol;
    lf.Write(pchar(x)^,length(x));
    x := '   numbered at ';
    lf.Write(pchar(x)^,length(x));
    case degmax of
      1..8: demod := 1;
      9..19: demod := 2;
      20..40: demod := 5;
      else demod := 10;
    end; (* case *)
    i := 1;
    repeat
      x := format('%s%d',[' ',i*demod]);
      lf.Write(pchar(x)^,length(x));
      inc(i);
    until (i*demod >= degmax);
    x := ' / '+eol;
    lf.Write(pchar(x)^,length(x));
    x := '     /                                            '+eol;
    lf.Write(pchar(x)^,length(x));
    x := '\put{\textit{n}} at ';
    lf.Write(pchar(x)^,length(x));
    x := format('%d%s',[n+1,' -2                        '])+eol;
    lf.Write(pchar(x)^,length(x));
    x := '\put{degrees} at ';
    lf.Write(pchar(x)^,length(x));
    x := format('%s%d%s',[' -2 ',degmax+2,'                 '])+eol;
    lf.Write(pchar(x)^,length(x));
    x := '\multiput{\gast} at                              '+eol;
    lf.Write(pchar(x)^,length(x));
    for i:=1 to n
    do if ((perc5[i] <= ordgs[i]) and (perc95[i] >= ordgs[i]))
    then begin
      x := format('%4d%s%4d',[i,' ',ordgs[i]])+eol;
      lf.Write(pchar(x)^,length(x));
    end;
    x := '  /                                               '+eol;
    lf.Write(pchar(x)^,length(x));
    x := '\multiput{\rast} at                              '+eol;
    lf.Write(pchar(x)^,length(x));
    for i:=1 to n
    do if ((perc5[i] > ordgs[i]) or (perc95[i] < ordgs[i]))
    then begin
      x := format('%4d%s%4d',[i,' ',ordgs[i]])+eol;
      lf.Write(pchar(x)^,length(x));
      end;
    x := '  /                                               '+eol;
    lf.Write(pchar(x)^,length(x));
    x := '\multiput{-} at                              '+eol;
    lf.Write(pchar(x)^,length(x));
    for i:=1 to n
    do  begin
      x := format('%s%4d%s%4d%s%4d%s%4d',[' ',i,' ',perc5[i],' ',i,' ',perc95[i]])+eol;
      lf.Write(pchar(x)^,length(x));
      end;
    x := '  /                                               '+eol;
    lf.Write(pchar(x)^,length(x));
    for i:=1 to n
    do begin
      x := format('%s%4d%s%4d%s%4d%s%4d',['\putrule from ',i,' ',perc5[i],' to ',i,' ',perc95[i]])+eol;
      lf.Write(pchar(x)^,length(x));
      end;
    x := '\endpicture                                       '+eol;
    lf.Write(pchar(x)^,length(x));
    x := '\end{minipage}                                    '+eol;
    lf.Write(pchar(x)^,length(x));
    x := '\end{center}                                      '+eol;
    lf.Write(pchar(x)^,length(x));

  if ((observations >= 2) and (not ExtendedReciprocityModel)) then begin
    Heading(3,outf,'Explained variation');
    x := 'If the highest absolute t-statistic for convergence diagnosis is more than 0.05,'+eol;
    outf.Write(pchar(x)^,length(x));
    x :='these measures for explained variation will have an upward bias.'+eol;
    outf.Write(pchar(x)^,length(x));
    x := 'Columns give times, number of used simulated values, ';
    outf.Write(pchar(x)^,length(x));
    x := 'and 1.0 - scaled entropies.'+eol;
    outf.Write(pchar(x)^,length(x));
    outf.Write(pchar(eol)^,length(eol));//writeln(outf);
    s := 0.0;
    ns := 0;
    for i:=0 to 100 do
      if (entropynumber[i] >= 1) then begin
      x := format('%8.4f%s%8d%s%8.4f',[(observations-1)*i/100,' ',
           entropynumber[i],'  ', 1.0- divi(entropy[i],entropynumber[i])])+eol;
           outf.Write(pchar(x)^,length(x));
      s := s + entropy[i];
      ns := ns + entropynumber[i];
     end;
    outf.Write(pchar(eol)^,length(eol));//writeln(outf);
    x := format('%s%8.4f',['Average of 1.0 - entropy = ',1.0 - s/ns])+eol;
    outf.Write(pchar(x)^,length(x));
    outf.Write(pchar(eol)^,length(eol));//writeln(outf);
    x := 'Picture of entropies in Latex format:'+eol;
    lf.Write(pchar(x)^,length(x));
    lf.Write(pchar(eol)^,length(eol));//writeln(lf);
    x := '\begin{center}                                    '+eol;
    lf.Write(pchar(x)^,length(x));
    x := '\begin{minipage}[htb]{22cm}                       '+eol;
    lf.Write(pchar(x)^,length(x));
    x := '\beginpicture                                     '+eol;
    lf.Write(pchar(x)^,length(x));
    x := format('%s%4.1f%s',['\setcoordinatesystem units <',12/(observations-1),
                    'cm,10cm> point at 0 0   '])+eol;
    lf.Write(pchar(x)^,length(x));
    x := format('%s%d%s',['\setplotarea x from 0 to ',observations-1,' ']);
    lf.Write(pchar(x)^,length(x));
    x := ', y from 0 to 1'+eol;
    lf.Write(pchar(x)^,length(x));
    x := '\axis bottom                                      '+eol;
    lf.Write(pchar(x)^,length(x));
    x := '   ticks out short                                '+eol;
    lf.Write(pchar(x)^,length(x));
    x := '   numbered at ';
    lf.Write(pchar(x)^,length(x));
    case observations of
     2: begin
      x := '  0.0 0.2 0.4 0.6 0.8 1.0 ';
      lf.Write(pchar(x)^,length(x));
      end;
     3: begin
      x := '  0.0 0.5 1.0 1.5 2.0 ';
      lf.Write(pchar(x)^,length(x));
     end;
     4: begin
      x := '  0.0 0.5 1.0 1.5 2.0 2.5 3.0 ';
      lf.Write(pchar(x)^,length(x));
      end;
     else for i:=1 to (observations -1) do begin
      x:= format('%s%3d%s',[' ',i,' ']);
      lf.Write(pchar(x)^,length(x));
      end;
     end; (* case *)
    x := ' / '+eol;
    lf.Write(pchar(x)^,length(x));
    x := ' / '+eol;
    lf.Write(pchar(x)^,length(x));
    x := '\axis left                                        '+eol;
    lf.Write(pchar(x)^,length(x));
    x := '   ticks out short                                '+eol;
    lf.Write(pchar(x)^,length(x));
    x := '   numbered at 0.0 0.2 0.4 0.6 0.8 1.0  /'+eol;
    lf.Write(pchar(x)^,length(x));
    x := ' / '+eol;
    lf.Write(pchar(x)^,length(x));
    x := '\multiput{*} at '+eol;
    lf.Write(pchar(x)^,length(x));
    for i:=0 to 100 do if (entropynumber[i] > 0) then begin
      x := format('%8.4f%s%8.4f',[(observations-1)*i/100,'  ',
            1.0 - entropy[i]/entropynumber[i]])+eol;
      lf.Write(pchar(x)^,length(x));
      end;
    x := ' / '+eol;
    lf.Write(pchar(x)^,length(x));
    x := '\endpicture                                       '+eol;
    lf.Write(pchar(x)^,length(x));
    x := '\end{minipage}                                    '+eol;
    lf.Write(pchar(x)^,length(x));
    x := '\end{center}                                      '+eol;
    lf.Write(pchar(x)^,length(x));
   end; (* if observations >= 2 *)

  if ((observations >=2) and (ModelType = 2) and (not symmetric)) then begin
  (* The following initialiseparameter is for recalculating xi . *)
    InitialiseAlpha(outf,theta,fs);
    pr := 1.0; s := 1.0;
    for i:=0 to n-2
    do begin
      pr := xi(i)*pr;
      s := s + pr;
    end; (* i *)
    x := 'Estimated stationary distribution for out-degrees'+eol;
    ouf.Write(pchar(x)^,length(x));
    x := 'values:'+eol;
    ouf.Write(pchar(x)^,length(x));
    for i:=0 to n-1
    do begin
      x := format('%5d',[i]);
      ouf.Write(pchar(x)^,length(x));
      if ((i mod 15) = 0) then ouf.Write(pchar(eol)^,length(eol));//writeln(ouf);
    end;
    if ((n-1 mod 15) <> 0) then ouf.Write(pchar(eol)^,length(eol));//writeln(ouf);
    x := 'probabilities:'+eol;
    ouf.Write(pchar(x)^,length(x));
    pr := 1.0/s;
    for i:=0 to n-1
    do begin
      x := format('%s%4.2f',[' ',pr]);
      ouf.Write(pchar(x)^,length(x));
      pr := xi(i)*pr;
      if ((i mod 15) = 0) then ouf.Write(pchar(eol)^,length(eol));//writeln(ouf);
    end; (* i *)
    if ((n-1 mod 15) <> 0) then ouf.Write(pchar(eol)^,length(eol));//writeln(ouf);
    x := 'cumulative probabilities:'+eol;
    ouf.Write(pchar(x)^,length(x));
    pr := 1.0/s;
    s := pr;
    for i:=0 to n-1
    do begin
      x := format('%s%4.2f',[' ',s]);
      ouf.Write(pchar(x)^,length(x));
      pr := xi(i)*pr;
      s := s+pr;
      if ((i mod 15) = 0) then ouf.Write(pchar(eol)^,length(eol));//writeln(ouf);
    end; (* i *)
    if ((n-1 mod 15) <> 0) then ouf.Write(pchar(eol)^,length(eol));//writeln(ouf);
    ouf.Write(pchar(eol)^,length(eol));//writeln(ouf);
   end; (* if observations >=2 and ModelType=2 and not symmetric *)
  end; (* if OutputType >= 2 *)

end; (* results *)


begin (* Activity.Estimate *)
  somefix := false;
  halter := false;
  n := nmax;
  setlength(ordgs,n+1);
  estimation := true;
  if (Bayesian and not(initialFEestimate))
  then Heading(1,outf,'Estimation by MCMC kernel cycle.')
  else Heading(1,outf,'Estimation by stochastic approximation algorithm.');
  lf.Write(pchar(eol)^,length(eol));//writeln(lf);
  x := 'SIENA estimation algorithm'+eol;
  lf.Write(pchar(x)^,length(x));
  x := 'Date and time: '+DateTimeToStr(Now)+eol;
  lf.Write(pchar(x)^,length(x));
  time0 := now;
  Heading(1,lf,'Estimation.');
  cfn := pname+'.cck';
 // assignfile(cf,cfn);
  if ((run = 1) or (not fileExists(cfn)))
  then begin
  //rewrite(cf)
    cf := TFileStream.Create(cfn,fmCreate);
    cf.Free;
    cf := TFileStream.Create(cfn,fmOpenWrite or fmShareDenyWrite);
  end
  else
  begin
    cf:= TFileStream.Create(cfn,fmOpenReadWrite or fmShareDenyWrite);
    cf.seek(0,soFromEnd);
    cf.size := 0;
  end;
  //append(cf); (*MS, changed by TS *)
  x := 'This file serves to check what happens within the'+eol;
  cf.Write(pchar(x)^,length(x));
  x := 'stochastic simulation algorithm of SIENA.'+eol;
  cf.Write(pchar(x)^,length(x));
  x := 'It is rewritten in each estimation procedure,'+eol;
  cf.Write(pchar(x)^,length(x));
  x := 'so it contains only the checks for the most recent procedure.'+eol;
  cf.Write(pchar(x)^,length(x));
  x := 'Project '+pname+'; estimation algorithm started '+ DateTimeToStr(Now)+' .'+eol;
  cf.Write(pchar(x)^,length(x));
  cf.Write(pchar(eol)^,length(eol));//writeln(cf);

  if ((observations = 1) and estimation and (not cconditional) and (firstgain > 0.002))
  then begin
    outf.Write(pchar(eol)^,length(eol));//writeln(outf);
    x := 'Warning:'+eol;
    outf.Write(pchar(x)^,length(x));
    x := '--------'+eol;
    outf.Write(pchar(x)^,length(x));
    x := 'For the estimation of ERGM parameters in the non-conditional option,'+eol;
    outf.Write(pchar(x)^,length(x));
    x := 'it often is advisable to use a small value of the initial gain parameter,'+eol;
    outf.Write(pchar(x)^,length(x));
    x := 'such as 0.001.'+eol;
    outf.Write(pchar(x)^,length(x));
    x := format('%s%10.6f%s',['(The current value is ',firstgain,'.)'])+eol;
    outf.Write(pchar(x)^,length(x));
    outf.Write(pchar(eol)^,length(eol));//writeln(outf);
  end;
  whichtime := 1;

Starthere:
  if (MaxLike or Bayesian or randomeffects) then Cconditional := false;
  Oldconditional := cconditional;

  for i:=1 to pmax
  do begin
    estimates[i] := 999.0;
    standard_errors[i] := 99999.0;
    conv_t[i] := 99.0;
  end;


  (* TS het volgende iets veranderd 05-jan-05
     CS logarithmes toegevoegd ivm niewe rate parametrisatie, april 2005
     TS alles gedeeld door ObservedNetworkData[1,i] i.p.v. n, 18-jun-05*)  // moet weer teruggezet TS
  if InitialStandard then InitialEstimate;

  for i:=1 to observations-1
  do begin
    if (symmetric and (ModelType in [4..6]))
    then begin
      if (alpa_l[i] < log(0.0001+root(n*ObservedNetworkDistance[1,i])/ObservedNetworkData[1,i]))
      then alpa_l[i] := log(0.0001+root(n*ObservedNetworkDistance[1,i])/ObservedNetworkData[1,i]);
    end
    else begin
      (* TS conditie toegevoegd 17-3-5 *)
//      if (ModelType <> 3) and (alpa_l[i] < log(0.1+n*ObservedNetworkDistance[1,i]/ObservedNetworkData[1,i]))
//      then alpa_l[i] := log(0.1+n*ObservedNetworkDistance[1,i]/ObservedNetworkData[1,i]);
//      for j:=1 to nza
//      do if (alpa_l[NetworkEffects_l+(j-1)*ActionEffects_l+i] < log(0.1 + ObservedActionDistance[j,i]/ObservedActionData[j,i]))
//      then alpa_l[NetworkEffects_l+(j-1)*ActionEffects_l+i] := log(0.1 + ObservedActionDistance[j,i]/ObservedActionData[j,i])
    end
   end; (* for i *)

  (* default choice of statistics for the chosen model: *)
  DefineFunctions(outf,lf);
  InitialiseBase2(theta,outf,lf);
  if (NumberParameters <= 0) then begin
    x := format('%s%d%s',['The number of parameters to be estimated is ',NumberParameters,'.'])+eol;
    outf.Write(pchar(x)^,length(x));
    x := 'This is not meaningful - the program stops.'+eol;
    outf.Write(pchar(x)^,length(x));
    halter := true;
    exit;
   end;

  initialtheta := theta;
  case (nrun3 mod 100) of
  (* This is the way for selecting LR tests. *)
    1: begin
       LRcode := 1;
       Initialestimate;   (* changes alpa *)
       SetTheta(standardtheta);
       SetAlpha(theta);   (* changes alpa back again *)
       end;
    2: begin
       LRcode := 2;
       end;
    else LRcode := 0;
  end; (* case *)

  FunctionEvaluations(ordgs,halter,outf,lf,targets);


  if halter then exit;
  if VertexPar then InitialiseVertexArrays;

  (* Give arrays mtime and vmtime the correct dimension:
     these are used for the periods,
     so the elements used have indices 1, ..., observations-1.
     For observations = 1, mtime[1]  and vmtime[1] are used
     for estimating the deviance.
     Note that indices for dynamic arrays always start with 0. *)
  if (observations = 1)
  then begin
    (* for the deviances : *)
    SetLength(mtime,3);
    SetLength(vmtime,3);
    //    assignfile(d1f,pname+'.dv1');  TS: For ERGM case; may be reused later.
    //    assignfile(d2f,pname+'.dv2');
    //    rewrite(d1f);
    //    rewrite(d2f);
  end
  else begin
    SetLength(mtime,ActivePeriods+1);
    SetLength(vmtime,ActivePeriods+1);
(* Watch out: the dimension of mtime is also used in Polrup
       as a hack to determine the number of periods
       in the conditional case. *)
  end;

  IndicateFixedThetaAndRC(fix,thetarandom);

  (*MS*)
  if initialFEestimate
  then begin (*Random effects model: initial estimate of fixed effects by unconditional moment estimation.*)
    maxlike_o := maxlike;
    Bayesian_o := Bayesian;
    maxlike := false;
    Bayesian := false;
    cconditional := false;
    randomeffects := false;
    nre := 0;
    for j := 0 to 2 do res[j,0] := 0;
    for i := 1 to pmax
    do begin
      for j := 0 to 2 do res[j,i] := 0;
      thetarandom[i] := false;
    end;
  end;
  if (homogeneous > 0)
  then begin (*Homogeneity test: actor or period homogeneity test.*)
    fix[Numberparameters+1] := true;
    if (homogeneous = 11) then Description[Numberparameters+1] := 'f: out-degree (specified actor(s))          '
    else if (homogeneous = 12) then Description[Numberparameters+1] := 'f: reciprocity (specified actor(s))           '
    else if (homogeneous > 20)
    then begin
      i := homogeneous - 20;
      i := i + ActiveNetworkeffects_l + ActiveActionEffects_l;
      Description[NumberParameters+1] := Description[i];
      Delete(Description[NumberParameters+1],31,18);
      Description[NumberParameters+1] := Description[NumberParameters+1] + '(specified period)'
    end;
  end;
  (*MS*)


  if (modelseed = 0) then
    begin
    x := 'Random initialization of random number stream.'+eol;
    outf.Write(pchar(x)^,length(x));
    fs := readstate;
    x := format('%s%d%s',['Current random number seed is ',fs,'.'])+eol;
    outf.Write(pchar(x)^,length(x));
    end
   else
    begin
    x := format('%s%d',['Random number seed = ',modelseed])+eol;
    outf.Write(pchar(x)^,length(x));
    fs := readstate;
    end;
  makestate(fs);

  (*MS*)
  if (homogeneous > 0)
  then begin
    if (homogeneous < 20)
    then coordinate := homogeneous - 10
    else coordinate := homogeneous - 20;
    coordinate := coordinate + ActiveNetworkeffects_l + ActiveActionEffects_l;
    theta[NumberParameters+1] := theta[coordinate];
  end;
  (*MS*)

  if VertexPar then ReportVertexParameters(outf);
  if (observations <= 1)
  then
    begin
    wwrite(outf,'The observed network is ');
    ccc := ConnectedNetwork;
    if not ccc then
      begin
      wwrite(outf,'not ');
      end;
    wwriteln(outf,'connected.');
    if (not ccc) and (Gibbscode = 26) then
      begin
      wwriteln(outf,'The restriction to connected networks is dropped.');
      Gibbscode := 16;
      GibbsConnected := false;
      Gibbsstring := 'M.H. nondirected, continuous'; (* hack *)
      end;
    if CConditional
    then begin
     x := 'Conditional moment estimation.'+eol;
      outf.Write(pchar(x)^,length(x));
      x := format('%s%7.4f',['Observed density = ',densty])+eol;
      outf.Write(pchar(x)^,length(x));
(* AA change *)      x := format('%s%4d',
                ['Number of ties   = ',round(densty*(n*(n-1)- nstruct(1)))])+eol;
      outf.Write(pchar(x)^,length(x));
    end
    else begin
      x := 'Unconditional moment estimation.'+eol;
      outf.Write(pchar(x)^,length(x));
    end;
    x := format('%s%s%s%d%s',['Simulation method: ',Gibbsstring,' (code ',gibbscode,').'])+eol;
    outf.Write(pchar(x)^,length(x));
(* AA change *)    x := format('%d%s',[ERGMnrun(gibbs),' steps per simulation; ']);
    outf.Write(pchar(x)^,length(x));
    x := format('%s%5.3f%s',['multiplication factor = ',rstep,'.'])+eol;
    outf.Write(pchar(x)^,length(x));
  end (* observations <=1 *)
  else
    begin (* observations >= 2 *)
    if (maxlike or Bayesian) then
      begin
      nrunMH := trunc(rstep*ObservedNetworkDistance[1,1]);
      if (nrunMH < 2) then nrunMH := 2;
      if (nrunMH > 100) then nrunMH := round(nrunMH / 100) * 100; (*MS: make the number nice.*)
      if (Bayesian or randomeffects) then (*MS: the <nrunMH> iterations are divided into <nrunMH_batches> batches.*)
        begin
        nrunMH_batches := 20;
        nrunMH := round(nrunMH / nrunMH_batches); (*Thus the number of iterations is <nrunMH_batches * nrunMH>, where <nrunMH> is given by the left-hand side.*)
        if (nrunMH < 1) then nrunMH := 1;
        end;
      if maxlike then (*MLE with or without random effects.*)
        begin
        if randomeffects then begin
          x := format('%d%s%d%s',[nrunMH_batches,' batches per RM step with ',nrunMH,' iterations.']) +eol;
          outf.Write(pchar(x)^,length(x));
        end
        else (*MLE without random effects.*)
          begin
          x := format('%d%s',[nrunMH,' MCMC steps per RM step (multiplication factor = ']);
          outf.Write(pchar(x)^,length(x));
          if (rstep < 0.5) then begin
            x := format('%6.4f%s',[rstep,').'])+eol;
            outf.Write(pchar(x)^,length(x));
          end
          else begin
            x := format('%4.1f%s',[rstep,').'])+eol;
            outf.Write(pchar(x)^,length(x));
          end;
          end;
        end;
      end;
    if randomeffects then (*Moment estimation not implemented.*)
      begin
      if cconditional then (*If user chose conditional moment estimation.*)
        begin
        cconditional := false;
        Bayesian := true;
        end
      else if not(maxlike or Bayesian) then Bayesian := true; (*If user chose unconditional moment estimation.*)
      modeltype := 1; (*Other model types not implemented.*)
      end; (* if randomeffects *)
    x := format('%s%d%s%s%s',['Model Type ',Modeltype,': ',ModelTypeString,'.'])+eol;
    outf.Write(pchar(x)^,length(x));
    if ReciprocityModel then begin
      x := 'Reciprocity Model (special rate function depending on ';
      outf.Write(pchar(x)^,length(x));
      if ExtendedReciprocityModel then begin
        x := 'evaluation function).'+eol;
        outf.Write(pchar(x)^,length(x));
      end
       else begin
        x := 'degrees).'+eol;
         outf.Write(pchar(x)^,length(x));
       end;
      extraeffects := 0;
      ef := NewNetworkEffects.FirstMod;
      if Assigned(ef) then if not ef.Dyadic then extraeffects := 1;
      while ef.HasNextMod do begin
        ef := ef.NextMod;
        if not ef.Dyadic then inc(extraeffects);
       end;
      if (extraeffects > 0) then begin
        x := 'However, the following included ';
        outf.Write(pchar(x)^,length(x));
        if (extraeffects = 1) then begin
            x :='effect implies, for a nonzero parameter value,'+eol;
           outf.Write(pchar(x)^,length(x));
        end
         else begin
            x  := 'effects imply, for nonzero parameter values,'+eol;
            outf.Write(pchar(x)^,length(x));
         end;
        x := 'a departure from the reciprocity model:';
        outf.Write(pchar(x)^,length(x));
        ef := NewNetworkEffects.FirstMod;
        if not ef.Dyadic then
          begin outf.Write(pchar(eol)^,length(eol));//writeln(outf);
          x :=ef.EName;
           outf.Write(pchar(x)^,length(x));
           end;
        while ef.HasNextMod do begin
          ef := ef.NextMod;
          if not ef.Dyadic then begin outf.Write(pchar(eol)^,length(eol));//writeln(outf);
          x := ef.EName;
          outf.Write(pchar(x)^,length(x));
          end;
         end;
        outf.Write(pchar(eol)^,length(eol));outf.Write(pchar(eol)^,length(eol));//writeln(outf,'.'); writeln(outf);
       end;
     end;
    x := 'Estimation method: ';
    outf.Write(pchar(x)^,length(x));
    if CConditional
    then begin
      x := 'conditional moment estimation.'+eol;
      outf.Write(pchar(x)^,length(x));
      x := 'Conditioning variable is the total number of observed changes ("distance") '+eol;
      outf.Write(pchar(x)^,length(x));
      if ConditioningNetworkVariable>=1
      then begin
        x := 'in the network variable.'+eol ;
        outf.Write(pchar(x)^,length(x));
      end
      else begin
        x := 'in behavioral variable '+znames[nz+ConditioningActionVariable]+eol;
        outf.Write(pchar(x)^,length(x));
      end;
      if (observations = 2)
      then begin
        x := 'Distance for simulations is ';
        outf.Write(pchar(x)^,length(x));
        if ConditioningNetworkVariable>=1
        then begin
          x := format('%4d%s',[ObservedNetworkDistance[1,1],'.'])+eol;
          outf.Write(pchar(x)^,length(x));
        end
        else begin
          x := format('%4d%s',[ObservedActionDistance[ConditioningActionVariable,1],'.'])+eol;
          outf.Write(pchar(x)^,length(x));
        end;
      end
      else begin
        x := 'Distances for simulations are'+eol;
        outf.Write(pchar(x)^,length(x));
        x := 'period   : ';
        outf.Write(pchar(x)^,length(x));
        for i:=1 to observations-1 do
        if not skipi[i] then
          begin
          x := format('%s%4d',[' ',i]);
          outf.Write(pchar(x)^,length(x));
          end;
        outf.Write(pchar(eol)^,length(eol));
        x := 'distance : ';
        outf.Write(pchar(x)^,length(x));
        for i:=1 to observations-1
        do if not skipi[i] then
        begin
        if ConditioningNetworkVariable>=1
        then begin
          x := format('%s%4d',[' ',ObservedNetworkDistance[1,i]]) ;
          outf.Write(pchar(x)^,length(x));
        end
        else begin
          x := format('%s%4d',[' ',ObservedActionDistance[ConditioningActionVariable,i]]);
          outf.Write(pchar(x)^,length(x));
        end;
        end;
        x := ' .' +eol;
        outf.Write(pchar(x)^,length(x));
      end;
    end (* if CConditional *)
    else begin
      if MaxLike then
        begin
        x := 'Maximum Likelihood estimation.'+eol;
        outf.Write(pchar(x)^,length(x));
        end
      else if Bayesian then
        begin
        x := 'Bayes estimation.'+eol;
        outf.Write(pchar(x)^,length(x));
        end
      else
        begin
        x := 'unconditional moment estimation.'+eol;
        outf.Write(pchar(x)^,length(x));
        end;
      (* MH: Unconditional estimation is the only possibility
             for implementing exogenous events *)
      if (exo) then
        begin
        x := 'Changing composition: no conditional moment estimation.'+eol;
        outf.Write(pchar(x)^,length(x));
        x := format('%s%d%s',['Joines/leavers option ',nzx,'.'])+eol;
        outf.Write(pchar(x)^,length(x));
        end;
      outf.Write(pchar(eol)^,length(eol));//writeln(outf);
      x := 'Time duration for simulations ';
      outf.Write(pchar(x)^,length(x));
      if (observations >= 3) then begin
        x := 'in each period ';
        outf.Write(pchar(x)^,length(x));
      end;
      x := 'is 1.0.'+eol; (* see FRAN in S_Base *)
      outf.Write(pchar(x)^,length(x));
    end; (* not Cconditional *)
  end; (* observations >= 2 *)

  if (observations >= 2) then
  case d_estimator of
   0: begin
      wwriteln(outf,'Standard errors are estimated with the finite difference method.');
      end;
   1: begin
      wwriteln(outf,'Standard errors are estimated with the likelihood ratio method.');
      end;
    end; (*case *)

  if ((MaxDegree < n) and ((Observations >= 2) or (Gibbs in [1,2,4,5,6])))
         then begin
   if symmetric then begin
    x := format('%s%d%s',
   [ 'All graphs are constrained to having degrees not larger than ',
    MaxDegree,'.']) +eol;
    outf.Write(pchar(x)^,length(x));
   end
   else x := format('%s%d%s', [
    'All digraphs are constrained to having outdegrees not larger than ',
    MaxDegree,'.'])+eol;
    outf.Write(pchar(x)^,length(x));
   end;
  x := format('%s%10.7f%s',['Initial value of gain parameter is ',firstgain,' .'])+eol;
  outf.Write(pchar(x)^,length(x));
  x := format('%s%d%s',['Number of subphases in Phase 2 is ',nsubphase,'.'])+eol;
  outf.Write(pchar(x)^,length(x));
  outf.Write(pchar(eol)^,length(eol));//writeln(outf);
  x := 'Initial parameter values are '+eol;
  outf.Write(pchar(x)^,length(x));
  if Cconditional then for i:=1 to observations-1 do if (not skipi[i]) then
    begin
     if (observations = 2) then begin
       if ((nxa = 1) and (nza = 0)) then begin
         x := '  0.    Rate parameter                                  ';
         outf.Write(pchar(x)^,length(x));
       end
         else begin
          x := '  0.    Rate parameter of conditioning variable       ';
           outf.Write(pchar(x)^,length(x));
         end;
         end else begin
       if ((nxa = 1) and (nza = 0)) then begin
           if (i <= 9) then
           x := format('%s%d%s%2d%s',['  0.',i,'   Rate parameter period ',i,
                          '                        '])
          else
           x := format('%s%d%s%2d%s',['  0.',i,'  Rate parameter period ',i,
                          '                        ']);
          outf.Write(pchar(x)^,length(x));
       end
         else begin
          if (i <= 9) then
          x := format('%s%d%s%2d%s',['  0.',i,'   Rate parameter cond. variable period ',i,'         '])
          else
          x := format('%s%d%s%2d%s',['  0.',i,'  Rate parameter cond. variable period ',i,'         ']) ;
          outf.Write(pchar(x)^,length(x));
         end;
        end;
     if (ConditioningNetworkVariable=1)
        then begin
          x := format('%10.4f',[alpa_l[i]]) +eol;
          outf.Write(pchar(x)^,length(x));
        end
        else begin
          x := format('%10.4f',[alpa_l[NetworkEffects_l+(ConditioningActionVariable-1)*ActionEffects_l+i]])+eol;
          outf.Write(pchar(x)^,length(x));
        end;
   end; (* for i *)
  for i:=1 to NumberParameters
  do begin
    x := format('%3d%s%s%s%10.4f',[i,'. ',Alignstring(Description[i],50),' ',theta[i]]);
    outf.Write(pchar(x)^,length(x));
    if fix[i]
    then begin
      x := '  (fixed)'+eol;
      outf.Write(pchar(x)^,length(x));
    end
    else outf.Write(pchar(eol)^,length(eol));
//    x := format('%s%d',['        corresponding to statistic ',Functions[i]])+eol;
//    outf.Write(pchar(x)^,length(x));
  end;
  if (homogeneous > 0) then begin
    x := format('%3d%s%s%s%10.4f',[NumberParameters+1,'. ',Description[NumberParameters+1],' ',theta[NumberParameters+1]]);
    outf.Write(pchar(x)^,length(x));
  end;
  outf.Write(pchar(eol)^,length(eol));outf.Write(pchar(eol)^,length(eol));//writeln(outf);
 // writeln(outf);


  (*MS*)
  if (homogeneous > 0)
  then begin
    if (homogeneous < 20)
    then begin
      ObservedMomentActors(observations,homogeneous,sin); (*Calculates observed moment for specified actor/period.*)
      us := sin;
      if (homogeneous = 11)
      then StatsDescriptions[NumberFunctions+1] := 'Number of ties (specified actor(s))             '
      else if (homogeneous = 12)
      then StatsDescriptions[NumberFunctions+1] := 'Number of reciprocated ties (specified actor(s))';
    end
    else begin
      ObservedMomentPeriod(sta,outf);
      us := sta[coordinate]; (*Note that right now Siena does not know that one parameter coordinate is added: hence we <abuse> the last parameter coordinate.*)
      StatsDescriptions[NumberFunctions+1] := StatsDescriptions[coordinate];
      Delete(StatsDescriptions[NumberFunctions+1],31,18);
      StatsDescriptions[NumberFunctions+1] := StatsDescriptions[NumberFunctions+1] + '(specified period)';
    end;
  end;
  (*MS*)

  x := 'Observed values of ';
  outf.Write(pchar(x)^,length(x));
  if not(MaxLike or Bayesian) then begin
    x := 'target ';
    outf.Write(pchar(x)^,length(x));
  end;
  x := 'statistics are '+eol;
  outf.Write(pchar(x)^,length(x));
  for i:=1 to NumberFunctions do begin
    x := format('%3d%s%s',[i,'. ',StatsDescriptions[i]]);
    outf.Write(pchar(x)^,length(x));
    if endowmentfunction[i] then begin
      x := format('%15.4f',[-Targets[i]])+eol;
      outf.Write(pchar(x)^,length(x));
    end
     else begin
      x := format('%15.4f',[Targets[i]])+eol;
      outf.Write(pchar(x)^,length(x));
     end;
   end;
  if (homogeneous > 0) then begin
    x := format('%3d%s%s%15.4f',[NumberFunctions+1,'. ',StatsDescriptions[NumberFunctions+1],us])+eol;
    outf.Write(pchar(x)^,length(x));
  end;

  outf.Write(pchar(eol)^,length(eol));//writeln(outf);
  n3 := nrun3;
  p := NumberParameters;
  pp := NumberFunctions;

  (*MS*)
  if (homogeneous > 0) then (*Homogeneity tests.*)
    begin
    p := p + 1;
    pp := p;
    NumberFunctions := p;
    NumberParameters := p;
    test[p] := true;
    end;
  if randomeffects then (*Random effects models.*)
    begin
    nactors := NumberActors; (*Get the number of actors: in the present unit unknown.*)
    InitialActionsRE(randomeffects,outf,thetarandom,Description,covrefile,sequences,run,covfile2,re,mvn,nactors,nre,exo); (*Initial actions concerning random effects.*)
    if maxlike then covariance_parameters := (nre * (nre + 1)) div 2; (*Number of (co)variance parameters to be estimated.*)
    end;
  (*MS*)

  if not(Bayesian) then
    begin
    if randomeffects then begin
      x := format('%s%d%s%d%s',[' parameters (fixed effects), ',pp,' statistics, ', covariance_parameters,' (co)variance parameters '])+eol; (*MS: random effects models estimated by ML method.*)
      outf.Write(pchar(x)^,length(x));
    end
    else begin
      x := format('%d%s%d%s',[p,' parameters, ',pp,' statistics '])+eol;
      outf.Write(pchar(x)^,length(x));
    end;
    end;
  if not (NumberParameters = NumberFunctions) then
    begin
    x := format('%s%d',['NumberParameters = ',NumberParameters])+eol;
    lf.Write(pchar(x)^,length(x));
    x := format('%d%s',['NumberFunctions = ',NumberFunctions])+eol;
    lf.Write(pchar(x)^,length(x));
    x := 'This is an error.'+eol;
    lf.Write(pchar(x)^,length(x));
    Halterror('Unit S_EST: Numberparameters not equal to NumberFunctions.',outf,lf);
    halter := true;
    exit;
    end;

  if (observations <= 1) then
    begin
    IndependentStart := (not continuous);
    if (gibbs >= 8) then SetDigraph(1);
    end;

  if CConditional then
    begin
    if ConditioningNetworkVariable=1
    then begin
      for i:=1 to (observations-1)
      do begin
        if fix_l[i]
        then begin
          x := format('%s%d',['You requested to fix the basic network rate parameter for period ',i])+eol;
          outf.Write(pchar(x)^,length(x));
          x := 'in the network-conditional estimation procedure.'+eol;
          outf.Write(pchar(x)^,length(x));
          x := 'This is a meaningless combination.'+eol;
          outf.Write(pchar(x)^,length(x));
          x := 'The basic rate parameter will be estimated anyway,'+eol;
          outf.Write(pchar(x)^,length(x));
          x := 'and then replaced by its estimated value.'+eol;
          outf.Write(pchar(x)^,length(x));
        end (*if fix_l[i] *)
      end (* for i *)
    end
    else begin (* conditioning on Action variable *)
      for i:=1 to (observations-1)
      do begin
        indexi := NetworkEffects_l+(ConditioningActionVariable-1)*ActionEffects_l+i;
        if fix_l[indexi]
        then begin
          x :=format('%s%d',['You requested to fix the basic behavioral rate parameter for period ',i])+eol;
          outf.Write(pchar(x)^,length(x));
          x := 'for the conditioning variable in the conditional estimation procedure.'+eol;
          outf.Write(pchar(x)^,length(x));
          x := 'This is a meaningless combination.';
          outf.Write(pchar(x)^,length(x));
          x := 'The basic rate parameter will be estimated anyway,'+eol;
          outf.Write(pchar(x)^,length(x));
          x := 'and then replaced by its estimated value.'+eol;
          outf.Write(pchar(x)^,length(x));
        end; (*if fix_l *)
      end (* for i *)
    end(* conditioning on Action variable *)
  end; (* if cconditional *)

(* Determine positivity restrictions for Polrup *)
   for i:=1 to p do posi[i] := BasicRateFunction[i];
(*  for i:=p downto 1 do posi[i] := false;

  for i:=(observations-1) downto 1 do
    begin
    if (Conditioningnetworkvariable <= 0) then
      for j:=nxa downto 1 do posi[(j-1)*(observations-1) + i] :=
        BasicRateFunction[(j-1)*(observations-1) + i]
     else
      for j:=nxa-1 downto 1 do posi[(j-1)*(observations-1) + i] :=
        BasicRateFunction[(j-1)*(observations-1) + i];
    ActiveNetworkEffects := ActiveNetworkEffects_l + ActiveNetworkEffects_f +
                            ActiveNetworkEffects_g;
    if (ConditioningActionVariable <= 0) then
      for j:=nza downto 1 do posi[(j-1)*(observations-1) + ActiveNetworkEffects + i] := true
     else
      for j:=nza-1 downto 1 do posi[(j-1)*(observations-1) + ActiveNetworkEffects + i] := true;
    end;
*)  somepos := false;
  for i:= p downto 1 do if posi[i] then somepos := true;
  if somepos then
    begin
    x := 'Positivity restrictions for variables:'+eol;
    lf.Write(pchar(x)^,length(x));
    for i:=1 to p do if posi[i] then begin
      x := format('%3d%s%s',[i,'. ',Description[i]])+eol;
      lf.Write(pchar(x)^,length(x));
    end;
    lf.Write(pchar(eol)^,length(eol));//writeln(lf);
    end;

  for i:=1 to pmax do thetabase[i] := 0.0;
  baselinemodeltext := '';
  (* If baselinemodeltext is not modified in the following,
     then the deviance statistic will not be reported. *)

  if ((observations = 1) and (Gibbs <> 8))
  then begin
    InitialAlphas(alpdum,alpa_l,0,outf); // alpdum only for temporary storage
    alf1 := alpa_l[1];
    alf2 := alpa_l[2];
    if symmetric
    then begin
      if ((not Cconditional) and effects_f[1])
      then begin
        baselinemodeltext := 'Independent edges model.';
        thetabase[1] := alf1;
      end;
      if Cconditional then baselinemodeltext := 'Uniform model with given number of edges.';
    end (* if symmetric *)
    else begin
      if effects_f[1]
      then begin
        thetabase[1] := alf1;
        baselinemodeltext := 'Independent arcs model.';
        if effects_f[2]
        then begin
          baselinemodeltext := 'Reciprocity model.';
          thetabase[2] := alf2;
        end;
      end; (* if effects_f[1] *)
    end; (* not symmetric *)
  end; (* observations = 1 *)


  (*MS: in the ML (but not: Bayesian) framework, the (co)variance parameters are contained in <theta>.*)
  if (maxlike and randomeffects) then h := p + covariance_parameters
  else h := p;
  EstForm.Startup(h);
  if (MaxLike or Bayesian) then
    begin
    if MaxLike then EstForm.hint := 'SIENA carries out maximum likelihood estimation'
      else EstForm.hint := 'SIENA carries out Bayesian estimation';
    EstForm.PLtext.visible := true;
    EstForm.Label3.Caption := 'Simulated scores';
    EstForm.Statistics.Hint := 'average of simulated scores over runs is 0 for maximum likelihood estimate';
    EstForm.PLtext.caption := 'MCMC burn-in';

    Application.ProcessMessages;
    MLInitialise(p,theta,outf,lf);
//writeln(outf,'.. end of initialization ..');

    if (randomeffects or bayesian) then
      begin
      assignfile(chainfile,pname+'.chs');
      rewrite(chainfile);
      end;
//    EstForm.PLtext.visible := false;
    end
   else
    begin
    EstForm.hint := 'SIENA carries out Method of Moments estimation';
    end;
   writechain := false;
   if (observations <= 1) then
     EstForm.hint := 'SIENA carries out maximum likelihood estimation';
//   describecliques(outf,1,2,10);  // cde
{$if defined(batch)}
        writeln;
        writeln(' Siena07 is executing project ',pname,'.');
        writeln;
{$ifend}


  if ((observations = 1) and (Gibbs <= 6) and effects_f[1] and (not ExistStruct[1]))
   and false
  then begin
    visibility(0);
    Heading(2,outf,'Maximum Pseudolikelihood estimates ');
    EstForm.PLtext.visible := true;
    theta1 := theta;
    Application.ProcessMessages;
    MLEstimateParameter(p,1,theta,c);  // pseudolikelihood
    results(outf,lf,true,true,false);
    theta := theta1;
    EstForm.PLtext.visible := false;
  end; (* observations = 1 *)
  Application.ProcessMessages;

  for i:=1 to p
  do begin
    if (MaxLike or Bayesian)
    then m[i] := 0.0
    else m[i] := Targets[i]
  end;

(* TS 04-jan-05: scale values smaller than in earlier versions;
   for parameters with positivity restriction, this is superseded in PolRup. *)
(* TS 22-jun-05: scale values still smaller.
   maximum repforep increased. *)
  for i := 1 to pmax do scale[i] := 0.1; (*MS: adapted.*)

  if (homogeneous > 0) then scale[p] := 0.1;
  if (maxlike and randomeffects) then
    begin
    h := p - covariance_parameters;
    for i := (h + 1) to p do scale[i] := scale[i] * 4;
    end;
  if FinDiff_method then
    begin
    lf.Write(pchar(eol)^,length(eol));//writeln(lf);
    x := 'Epsilon values finite differences method:'+eol; (*MS: write <epsilon>'s to log-file: important for (1) finite differences method and (2) step size.*)
    lf.Write(pchar(x)^,length(x));
    for i := 1 to p do
      begin
      x := format('%3d%s%6.3f%s',[i,'. ',scale[i],'; ']);
      lf.Write(pchar(x)^,length(x));
      if ((i mod 5) = 0) then lf.Write(pchar(eol)^,length(eol));//writeln(lf);
      end;
    if ((p mod 5) <> 0) then lf.Write(pchar(eol)^,length(eol));//writeln(lf);
    lf.Write(pchar(eol)^,length(eol));//writeln(lf);
    end;

  Application.ProcessMessages;
  errorcode := 0;
  x := 'Stochastic approximation algorithm.'+eol;
  lf.Write(pchar(x)^,length(x));
  x := 'The algorithm is logged in the file <<'+pname+'.cck>>.'+eol;
  lf.Write(pchar(x)^,length(x));
  lf.Write(pchar(eol)^,length(eol));//writeln(lf);
  ntrials := 0;
  ntrialmax := 2;
  nsubphs := nsubphase;
  repeat
(* if Errorcode = 2, 3 is excluded from version 3.13, because it was unimportant
   and it gave problems. *)
    if (Errorcode = 4)
    (* This means that POLRUP has obtained a result that is far
       from satisfactory.
       Hoping this was merely bad luck, we try again. *)
    then begin
      inc(ntrials);
      if (ntrials < ntrialmax)
      then begin
        x := 'Unsatisfactory convergence. Algorithm restart.'+eol;
        cf.Write(pchar(x)^,length(x));
        x := 'Because of the unsatisfactory result, the algorithm restarts from the current value.'+eol;
        outf.Write(pchar(x)^,length(x));
      end;
    end; (* if errorcode = 4*)

    (* Now the errorcode conditions are finished, and the normal execution starts. *)
    expon := (observations = 1);
    (* Added by TS 04-jan-05 to allow the use of several
       parallel rate parameters, of which one is used for conditioning: *)
    rr1 := 0; nr1 := 0;
    rr2 := 0; nr2 := 0;
    if (Cconditional and (nxa + nza >= 2)) then
      begin
       (* CS: exp-functies toegevoegd ivm niewe rate-parametrisatie: teruggezet TS*)
      nr1 := nxa+nza-1; (* The number of variables for which there is a rate parameter *)
      if (ConditioningNetworkVariable =1) then
        begin
        rr1 := FirstActionRateParameter;
        nr1 := nza;
        rr2 := 0; nr2 := 0;
(*         if (symmetric and (ModelType in [4..6])) then
           for i:=1 to observations-1 do mtime[i] := sqr(alpa_l[i])
          else
           for i:=1 to observations-1 do mtime[i] := alpa_l[i]; *)
        i := 0;
        h := 0;
        repeat
          inc(i);
          if (not skipi[i]) then
           begin
           inc(h);
           if (symmetric and (ModelType in [4..6])) then mtime[h] := sqr(alpa_l[i])
              else  mtime[h] := alpa_l[i];
           end;
        until (i >= observations-1);
        end
       else (* ConditioningActionVariable >= 1 *)
        begin
        rr1 :=1; nr1 := nxa;
        rr2 := FirstActionRateParameter; nr2 := nza-1;
        i := 0;
        h := 0;
        repeat
          inc(i);
          if (not skipi[i]) then
           begin
           inc(h);
           mtime[h] :=
            alpa_l[NetworkEffects_l+(ConditioningActionVariable-1)*ActionEffects_l+i];
           end;
        until (i >= observations-1);
        end;
      end; (* if *)
    SetVector(p,moments,m);
(* TS: resis added against instability *)
    for i:=1 to p do resis[i] := 1.0;
{    if MaxLike then
      begin
        for i:= nza*(observations-1) downto 1 do
         resis[ActiveNetworkEffects_l + ActiveNetworkEFfects_f + ActiveNetworkEffects_g +i ] := 0.1;
        for i:=1 to observations-1 do resis[i] := 0.1;
      end;
}

    POLRUP(p,expon,theta,tconv,thetabase,initialtheta,m,scale,resis,posi,fix,rr1,nr1,rr2,nr2,
           errorcode,messa,c,dd,sx, mtime, vmtime, tmaxx, nrun,niters,n3,nsubphs,firstgain);
    (* To leave this procedure with the current value for the alpa_x parameters: *)
    
    fs := readstate;
    InitialiseAlpha(outf,theta,fs);
    if CConditional
    then
      begin
      if ((errorcode = 0) or (errorcode = 4)) then
        begin
        i := 0;
        h := 0;
        repeat
          inc(i);
          if (not skipi[i]) then
           begin
           inc(h);
           if (ConditioningNetworkVariable=1) then
             begin
             if (symmetric and (ModelType in [4..6])) then alpa_l[i] := root(mtime[h])
                else alpa_l[i] := mtime[h];
             end
           else
  alpa_l[NetworkEffects_l+(ConditioningActionVariable-1)*ActionEffects_l+i]
               := mtime[h];
           end;
        until (i >= observations-1);
        end;
      end;

  if (ErrorCode = 2) then
    begin
    x := 'These parameter values are not suitable for conditional estimation.'+eol;
    outf.Write(pchar(x)^,length(x));
    x := 'Try again with other parameter values (e.g., standard initial values)'+eol;
    outf.Write(pchar(x)^,length(x));
    x := 'or with non-conditional estimation.'+eol;
    outf.Write(pchar(x)^,length(x));
    end;

  until ((ErrorCode < 3) or (Errorcode in [4,5]) or (ntrials >= ntrialmax));

  if not(simulate) then WriteDefinitionModel(lf);



  Heading(2,outf,'Estimation results.');
  Heading(2,lf,'Estimation results.');

  if (nsubphase <= 0)
  then begin
    x := 'Number of subphases = 0: no estimation, '+eol;
    outf.Write(pchar(x)^,length(x));
    x := 'only check of parameters and calculation of standard errors.'+eol;
    outf.Write(pchar(x)^,length(x));
  end;

  if simulate then WriteSeed(pname,modelseed); (*If <simulate> is enabled, write <seed> to the <pname.mo> file.*)
  (*Parameter estimates written to <pname.mo> file; undesirable if the aim is to generate multiple data sets according to some specified probability model.*)

  if (ErrorCode = 1) then
    begin
    x := 'Non-standard end of the estimation algorithm : '+eol;
    lf.Write(pchar(x)^,length(x));
    if (errortext > '') then begin
      x := errortext+eol;
      lf.Write(pchar(x)^,length(x));
    end;
    x := messa+eol;
    lf.Write(pchar(x)^,length(x));
    x := 'Non-standard end of the estimation algorithm : '+eol;
    outf.Write(pchar(x)^,length(x));
    if (errortext > '') then begin
      x := errortext+eol;
      outf.Write(pchar(x)^,length(x));
    end;
    x := messa+eol;
    outf.Write(pchar(x)^,length(x));
    x := 'Final parameter estimates:'+eol;
    outf.Write(pchar(x)^,length(x));
    for i:=1 to p do begin
      x := format('%s%s%20.12f',[Alignstring(Description[i],50),' ',theta[i]])+eol;
      outf.Write(pchar(x)^,length(x));
    end;
    outf.Write(pchar(eol)^,length(eol));
    x := 'These are not passed on to the SIENA session.'+eol;
    outf.Write(pchar(x)^,length(x));
    outf.Write(pchar(eol)^,length(eol));
    end
   else
    begin
    x := 'Normal end of the estimation algorithm.'+eol;
    lf.Write(pchar(x)^,length(x));
    x := messa+eol;
    outf.Write(pchar(x)^,length(x));
    results(outf,lf,true,false,false);
(* Output of brief results: *)
    if not Fileexists(pname+'.bof') then
     begin
        brieffile := TFileStream.Create(pname+'.bof',fmCreate);
        brieffile.free;
        brieffile := TFileStream.Create(pname+'.bof',fmOpenWrite or fmShareDenyWrite);
     end
    else
    begin
      brieffile := TFileStream.Create(pname+'.bof',fmOpenReadWrite or fmShareDenyWrite);
      brieffile.seek(0, soFromEnd);
    end;
    results(brieffile,lf,false,false,true);
    brieffile.free;

(* The following is to be activated only in special cases, e.g. for repeated simulations: *)
  if (nextra >= 1) then
    begin
    high_se := false;
    high_est := false;
    for i:=1 to p do
      begin
      if (root(c[i,i]) > 5.0) then high_se := true;
      if (abs(theta[i]) > 50.0) then high_est := true;
      end;
     if (((tmaxx >= 0.15) or high_se) and (whichtime<= nextra)) then
       begin
       case whichtime of
         1: initialstandard := high_est;
         2,3: begin
            initialstandard := high_est;
  //          if (not high_est) then firstgain := 0.2*firstgain;
            end;
         end;
       inc(whichtime);
       wwriteln(outf,'Inadequate convergence leads to automatic restart with standard initial values');
       if (tmaxx >= 0.15) then wwriteln(outf,'(t-ratio too high).')
        else if high_se then wwriteln(outf,'(standard error too high).');
       wwriteln(outf,'');
       brieffile := TFileStream.Create(pname+'.bof',fmOpenReadWrite or fmShareDenyWrite);
       brieffile.seek(0, soFromEnd);
       wwriteln(brieffile,'These results are not used (inadequate convergence, new try with these data).');
       brieffile.free;
       goto StartHere;
       end;
     end;

(* Output for simulation purposes : *)
   // assignfile(brieffile,pname+'.bos');
//    if (not fileexists(pname+'.bos')) then
      if (not fileexists('sisim.bos')) then
      begin
      //rewrite(brieffile);
//        brieffile := TFileStream.Create(pname+'.bos',fmCreate);
          brieffile := TFileStream.Create('sisim.bos',fmCreate);
        brieffile.free;
//        brieffile := TFileStream.Create(pname+'.bos',fmOpenWrite or fmShareDenyWrite);
        brieffile := TFileStream.Create('sisim.bos',fmOpenWrite or fmShareDenyWrite);
(* header *)
      x := format('%d',[p])+eol;;
      brieffile.Write(pchar(x)^,length(x));
      for i:=1 to p do
        begin
          if ((nza <= 1) or (DependentVariable[i] <= 0)) then begin
            x := Description[i]+' '+eol;
            brieffile.Write(pchar(x)^,length(x));
            end
          else begin
          x := format('%s%d%s%s%s',['<',DependentVariable[i],'> ',Description[i],' '])+eol;
          brieffile.Write(pchar(x)^,length(x));
          end;
      end;
      for i:=1 to p do
        begin
        if ((nza <= 1) or (DependentVariable[i] <= 0)) then begin
          x := format('%s%s%s',['sd-',Description[i],' '])+eol;
          brieffile.Write(pchar(x)^,length(x));
        end
          else begin
          x := format('%s%d%s%s%s',['sd-<',DependentVariable[i],'> ',Description[i],' '])+eol;
          brieffile.Write(pchar(x)^,length(x));
          end;
        end;
      end
     else begin
      //append(brieffile);
//      brieffile := TFileStream.Create(pname+'.bos',fmOpenReadWrite or fmShareDenyWrite);
        brieffile := TFileStream.Create('sisim.bos',fmOpenReadWrite or fmShareDenyWrite);
      brieffile.seek(0,soFromEnd);
     end;

    if not (maxlike or Bayesian) then begin
      x := format('%s%d%s',['MoM ',d_estimator,' ']);
      brieffile.Write(pchar(x)^,length(x));
    end
      else
        begin
      if maxlike then begin
        x := 'ML    ';
        brieffile.Write(pchar(x)^,length(x));
      end;
      if Bayesian then begin
        x := 'Bay   ';
        brieffile.Write(pchar(x)^,length(x));
      end;
      end;
    if ((tmaxx >= 0.1) or high_se) then
      begin
      if (tmaxx >= 0.1) then brieffile.write(pchar(string(' t'))^,2) else brieffile.write(pchar(string(' *'))^,2);
      if high_se then brieffile.write(pchar(string('s '))^,2) else brieffile.write(pchar(string('* '))^,2);
      end   else brieffile.write(pchar(string('    '))^,4);
    x := format('%6.2f%s',[tmaxx,' ']);
    brieffile.Write(pchar(x)^,length(x));

    if Cconditional then begin
      x := format('%9.5f%s',[0.0,' ']);
      brieffile.Write(pchar(x)^,length(x));
    end;

    for i := 1 to p do begin
      x := format('%9.5f%s',[theta[i],' ']);
      brieffile.Write(pchar(x)^,length(x));
    end;

    if Cconditional then begin
      x := format('%9.5f%s',[0.0,' ']);
      brieffile.Write(pchar(x)^,length(x));
    end;

    for i := 1 to p do begin
      x := format('%9.5f%s',[root(c[i,i]),' ']);
      brieffile.Write(pchar(x)^,length(x));
    end;
    wwriteln(brieffile,inttostr(n));
    brieffile.free;

    if not(simulate) then
      begin
      if initialFEestimate then
        begin
        maxlike := maxlike_o;
        Bayesian := Bayesian_o;
        end;
      if (not (ErrorCode in [1,2])) then WriteDefinitionModel(lf);
      end;

{    if (MaxLike or Bayesian) then
      begin
      EstForm.PLtext.visible := true;
      EstForm.PLtext.caption := 'calculate extra MCMC results';
      Application.ProcessMessages;
      ExtraMLResults(lf);
      CheckFiles(maxlike,sequences,lf,run,randomeffects,Bayesian);
        (*Write information to output file concerning Matlab source files
          generated by Siena to detect non-convergence / study sampling distributions.*)
      EstForm.PLtext.visible := false;
      end;
 }
    if (MaxLike) or (observations <= 1)
    then
      begin
       if (LRCode >= 1) then
        begin
        nsteps := 500;
          case LRCode of
          1 : begin
                x := format('%s%10.2f',['estimated twice log likelihood ratio = ',
              logLR(p,theta,standardtheta,m,nsteps,errstr,halter,outf,lf)])+eol;
              outf.Write(pchar(x)^,length(x));
              end;
          2 : if (nsubphase >= 1) then  begin
              x := format('%s%10.2f',['estimated twice log likelihood ratio = ',
              logLR(p,theta,initialtheta,m,nsteps,errstr,halter,outf,lf)])+eol;
              outf.Write(pchar(x)^,length(x));
              end;
          end; (* case *)
        end;
      end;
    end; (* not Errorcode = 1 *)

  if (MaxLike or Bayesian) then
    begin
    MLFinalizePeriods(outf,lf);
    if (randomeffects or bayesian) then closefile(chainfile);
    end;

  CheckMaxDegree(outf);
  if ((observations <= 1) and (not (Gibbs in [2,3,8])))
    then CheckRunningStatistics(outf,lf);

  if (Errorcode in [0,4])
  then begin
    for i:=1 to p
    do begin
      estimates[i] := theta[i];
      if (not fix[i]) then standard_errors[i] := root(c[i,i]);
      conv_t[i] := tconv[i];
    end;
  end;

  if ((errorcode in [0,4]) and (observations = 1))
    then HysteresisAnalysis(lf, outf, p, pp, theta, Targets, sx);


  theta := estimates;
  Cconditional := OldConditional;
  x := Messa+eol;
  cf.Write(pchar(x)^,length(x));
  //closefile(cf);
  cf.free;
  time0 := 86400*(now - time0);
  x := format('%s%8.2f%s',['Total computation time ',time0,' seconds.'])+eol;
  outf.Write(pchar(x)^,length(x));
  finalize(mtime);
  finalize(vmtime);
  outf.Write(pchar(eol)^,length(eol));//writeln(outf);


end; (* Activity.Estimate *)


begin (* unit S_EST *)
  histograms := false;
end.

