{$I+}
                             (* S_ML.PAS *)
Unit s_ml;

(* Siena unit for likelihood based inference. *)

interface

uses classes, s_constants, eight, slib, s_base, s_re, s_matrices, s_mcmc, s_mcmc_II,
     digraph, chains;

(* Unit chains contains the data representation of embedded chains. *)


   type MLModel = object(ModelSpecification)

    errortext: string;
    p1, p2, p_n: byte;
    covariance_parameters, var_a_min, var_n_tot: integer;
    fine_tuning: boolean;
    newrate1, newrate2, newlprob1, newlprob2, oldrate1, oldrate2, detsigma: real;
    aREfile, aCovREfile, aThetafile, adllfile, arPATHfile, arREfile,
      arCovREfile, arThetafile,
      thetafile, covrefile, covfile2, SLRCovREfile: textfile;
    proposalcovre, i_covRE, c_covRE, c_icovRE: pmatrix;
    actuals, mnlactor1, mnlactor2, const_S_v, p1_accepted, ISE_theta: R_Ar;
    lastebre, li, ksi, elimination_li, elimination_ksi, t_elimination_li,
    t_elimination_ksi, ce, dupl, t_dupl, const_S_m, const_H, ISE_RE, ISE_CovRE: rrarray;
    MadeChain: boolean;
(* MadeChain indicates that procedure MakeChains, which initialises the chains
   for all periods, has been carried out.*)
    prnet, prinsdel, prpermute, prdins, prddel, cprpermute,
     prndins, prbdins: real;
(* probabilities of diverse MH steps ; c denotes cumulative probabilities *)
    ninserts, nnoinserts, ncancels, nnocancels, ninsds,npermtots, nnopermtots,
            nnoinsds, ncands, nnocands, nimpcands,
            nimpinsdels, nstep: integer;
    ntrypermutes, npermutes, nnopermutes: array of integer;
    maxpermute,nimppermutes,nnotok: integer;
    numm: array of real;  (* the current number of permuted ministeps in MH steps *)
    (* numm is determined by a loose RM heuristic with the purpose
       to have an acceptance rate close to 0.5 *)
   prelins, promul, promul0: real;
   (* prelins is insertion probability in MH_Insdelpermute;
      promul is the probability of choosing a random single multiple
      for entering MH_InsDelPermute *)
    cov1: pmatrix;
    Chain: array of TChain;
(* Chain[period-1] contains the embedded chain for observations = period to period+1.
   See unit Chains for how type TChain works. *)
(* for communication between S_Est and S_ML for the Geyer-Thompson procedure: *)
    thetas: array of array of real;
    maxlength: integer;
    writechain: boolean;
    chainfile: textfile;
    ww: R_Ar;
    weightversion: byte;  (* version of weighting in multiple geyer procedure *)
(* The thre following arrays are used in calcratelprob,
   but initialised generally to save time. *)
    p_crp, pract_crp: array of real;
    permitted_crp: array of boolean;
// abcdefg  p_crp, pract_crp: array of real;
// abcdefg  permitted_crp: array of boolean;


(* procedures for representing embedded chain: uses unit Chains *)
    procedure MLInitialisePeriods(var outf,lf: TFileStream);
                          (* initialises all periods without any ministeps *)
    procedure MLFinalizePeriods(var outf,lf: TFileStream);

    procedure Cancel(var k: TMiniStep; const per: integer; var outf,lf: TFileStream);
                          (* cancels change k in period per (Chain[per-1]) *)
    procedure Insert(k: TMiniStep;
                     const V0: TAspect; const V1, V2: integer; const V3: byte; const V4: Tdif;
                      const per: integer; var outf,lf: TFileStream);
                          (* inserts a change of this kind in period per (Chain[per-1])
                             immediately before change k *)
    procedure Swap(var k1,k2: TMiniStep; const per: integer; var outf,lf: TFileStream);
                          (* swaps the kinds of changes k1 and k2 in period per (Chain[per-1]) *)
     (* The ministeps k1 and k2 are removed in this procedure, it is unknown
        what will come out! *)
(* end of procedures for representing embedded chain *)

  procedure MLInitialise(var p: byte;
  (*MS: <p> is declared as variable:
    important for ML estimation of random effects models!*)
                               var theta: pVec; var outf, lf: TFileStream);
// TS weggehaald @@@  procedure BurnIn(var outf, lf: textfile; var theta: pvec; var p: byte);
  function  valid(const V0: TAspect; const V1, V2: integer; const V3: byte; const V4: Tdif): boolean;
  procedure SetData(const k: TMiniStep; const per: integer; var outf,lf: TFIleStream);
  (* sets the current data set in Eight to the situation in in period per (Chain[per-1]),
     using the embedded chain in Embedded as it is valid
     immediately after change k *)
  function PR_Step(steptype: integer; k1, k2: TMiniStep; const V0: TAspect; const V1, V2: integer; const V3: byte; const V4: Tdif;
                   const per: integer; var outf,lf: TFileStream; var possible: boolean): real;
(* Log probability ratio (changed / current) for steps of type steptype in period per (Chain[per-1]) *)
(* steptype 1: insert (i,j) immediately before k1 and k2;
            2: cancel k1 and k2;
            3: swap k1 and k2;
            4: InsertDiag, insert (i,i) immediately before k1;
            5: CancelDiag, cancel the diagonal step at k1.
   possible indicates whether the indicated step is possible.   *)

(* The next five functions carry out MH steps,
   the function values are the MH ratios, but for MH_TrySwap it is the
   log of the MH ratio.
   Usually the function values are not important externally
   and the functions can be used just as procedures. *)
  function MH_TryInsertDiag(a: TAspect; i: integer; v: byte; k1: TMiniStep;
                           const per: integer; var outf,lf: TFileStream): real;
                          (* Metropolis Hastings step for inserting one
                             diagonal ministep in period per (Chain[per-1]) *)
  function MH_TryCancelDiag(var k1: TMiniStep; const per: integer; var outf,lf: TFileStream): real;
                          (* Metropolis Hastings step for canceling one
                              "change"  of diagonal kind in period per (Chain[per-1]) *)
  function  MH_InsDelPermute(const insdel: boolean;
                             const per: integer; var outf,lf: TFileStream): real;
  procedure MH_Step(const per: integer; var outf,lf: TFileStream;
                    const stype: integer);
                          (* Metropolis Hastings step in period per (Chain[per-1]) *)
(* stype = 0 gives a randomsteptype, stype between 1 and 6 gives a specified type of step. *)  // @@@@
  procedure burnextra(var outf,lf: TFileStream);
  procedure CalcRatelProb(var therate, thelprob: real; const V0: TAspect; const V1, V2: integer; const V3: byte; const V4: Tdif;
                          var outf,lf: TFileStream);
(* Calculates value of rate of change and log probability of
   changing (i,j) given the current state of the network in thisperiod,
   as represented in object ModelSpecification.
   The rate of change is the value valid before this change is made. *)
  procedure AdaptProposalProbabilities;
(* Adapts probabilities that depend on the period in case of
   structurally determined elements *)
  procedure MakeChains(theta: pvec; var outf, lf: TFileStream);
(* Constructs a chain linking observed digraphs
   for start and end of all periods *)
  function StepProb(const ka: TAspect; const ki, kj: Integer;
               const kv: byte; const kd: TDif;
               const theta: pvec;
               const esta: byte;
               var st: rrarray; var pp: realnvec;
               var prat: pmatrix;
               var cctotn, cctota: integer; var missing, error: boolean;
               var errstr: shortstring; var outf, lf: TFileStream): real;
(* This procedure calculates the probability of making step kk,
   the statistics st,
   the probabilities pp for other steps,
   with outcome 1 if the data is missing.
   If esta >= 30, it also calculates relative probabilities prat
   at distances epsi from theta.
   Then it makes the indicated step.
   This procedure uses the theta parameters, not the alpha parameters. *)
  procedure Geyer(const npar, nchain: integer;
                        var theta: pvec; var cc: pmatrix;
                        var outf,lf: TFileStream);
(* calculates weighted Geyer-Thompson procedure *)
  procedure Scores(p: Byte; var theta: pVec;
                allstats, varmat: boolean;
                var fseed: integer; var ystat: pVec; var d: array of real;
                var dd,relp: pmatrix;
                updatestat: Boolean;
                var feasible, setpar, error: Boolean;  // @@@@@
                var errstr: shortstring; var outf, lf: TFileStream;
                var estage: byte);
(* This procedure is the ML version of FRAN, cf. SimStats in S_Base,
   and it is attempted to have similar functionality.
   This works only for observations >= 2.
   It simulates one chain and stores the outcome in ystat
   and the observed complete data information matrix in dd. *)
  procedure FRAN(p: Byte; var theta: pVec;
                allstats, varmat: boolean;
                var fseed: integer; var ystat: pVec; var d: array of real;
                var dd: pmatrix;
                updatestat: Boolean;
                var feasible, error: Boolean;
                var errstr: shortstring; var outf, lf: TFileStream;
                var estage: byte;
                var pc: boolean);
                (* FRAN = Function RANdom *)

  function logLR(const p: Byte; const theta0, theta1, m0: pVec; var steps: integer;
                 var errstr: shortstring; var error: Boolean;
                 var outf, lf: TFileStream): real;
(* Calculates a MC approximation of twice the log likelihood ratio
   by bridge sampling from theta0 to theta1.
   It is assumed that the process already has burnt in for theta0;
   simulations go in steps steps from theta0 to theta1.
   steps is set to at least 2.*)
  function ConsistencyCheck(var outf,lf: TFileStream; CheckEnd: boolean): Boolean;
  procedure WritePeriod(const per: integer; var outf,lf: TFileStream); overload;
  procedure WritePeriod(const per: integer; var outf,lf: textfile); overload;
  procedure WriteFullPeriod(const per: integer; var outf,lf: TFileStream); overload;
  procedure WriteFullPeriod(const per: integer; var outf,lf: textfile); overload;
  procedure TryOut(var estimates, standard_errors: pvec; var outf, lf: TFileStream);
  procedure ExtraMLResults(var outf: TFileStream);

  (*---------------------------------------------------------------------------------*)
  (*                                 MS: BEGIN...                                    *)
  (*Markov chain Monte Carlo routines applied to conduct likelihood-based inference: *)
  (*- estimate non-hierarchical Bayesian models                                      *)
  (*- estimate frequentist random effects models with actor-dependent random effects.*)
  (*- estimate hierarchical Bayesian models with actor-dependent random effects.     *)
  (*...placed in <S_mcmc_I>. These routines make up only a small part of the total   *)
  (*number of MCMC-related routines, outsourced to <S_mcmc>, <S_mcmc_II>, <S_re>.    *)
  (*---------------------------------------------------------------------------------*)
  procedure InitialiseMS(var theta: pvec; var p: byte; var outf, lf: TFileStream);
  (*MS: Initialises some arrays needed in maximum likelihood and Bayesian estimation of
  models with and without actor-dependent random effects.*)
  procedure SoNIA(var pp: byte; var theta: pvec);
  (*MS: Writes path sampled from the conditional distribution of paths to the ASCII file.*)
  procedure LastPath(var pp: byte; var theta: pvec; const boolean1: boolean);
  (*MS: Writes last path sampled from the conditional distribution of the paths to ASCII file.*)
  procedure GetInitialRE(var outf, lf: TFileStream; const p: byte; const theta: pvec);
  (*MS: regarding the random effects as fixed, latent constants, the augmented-data likelihood function of the random effects is evaluated on a grid,
  and initial guesses of the random effects are computed.*)
  procedure ComputeTotal(var total_t: R_Ar; var total_n: iiarray);
  (*MS: obtain and store the length of the time interval and the total number of changes in the time interval for each dependent variable.*)
  procedure PartitionParameter(const dim1, dim2: byte; var vector1, vector2: pvec; const vector3: pvec);
  (*MS: Partition the parameter vector into rate-function (vector valued) parameter and non-rate function (vector-valued) parameter.*)
  procedure InversePartitionParameter(const dim1, dim2: byte; const vector1, vector2: pvec; var vector3: pvec);
  (*MS: Inverse the operation of partitioning the parameter vector: obtain the parameter vector <vector3> from rate parameter <vector1> and objective function parameter <vector2>.*)
  function Poisson(var theta: pvec): real;
  (*MS: evaluate joint Poisson kernel on logarithmic scale.*)
  procedure EvaluateMultinomial(const sampling: byte; const theta1, theta2: pvec; const actorRE1, actorRE2: rrarray; (*Input.*)
  var mnl_i_1, mnl_i_2: R_Ar; var outf, lf: TFileStream); (*Output.*)
  (*MS: Evaluate joint multinomial choice probability on logarithmic scale twice:
  - <sampling = 1>: once given <theta1> and once given <theta2>.
  - <sampling = 2>: once given <actorRE1> and once given <actorRE2>.
  Note: if <EvaluateMultinomial> is followed by statements manipulating either parameters or random effects or both,
  then it must be followed by <UpdateLProbs>.
  *)
  procedure ImproveMH(var outf, lf: TFileStream; var theta: pvec; var p: byte);
  (*MS: Method to improve efficiency of Metropolis-Hastings algorithm by improving acceptance rate.*)
  procedure UpdateLProbs(const proposal_accepted: boolean);
  (*MS: update the change probabilities <lprob>;
  required when
  - random effects: the random effects have changed.
  - Bayesian framework: the objective function parameter has changed.
  Arguments of <EvaluateMultinomial> must be in accordance with the present procedure: see below.*)
  procedure UpdateLProbsActors;
  (*MS: update the change probabilities <lprob>;
  the present procedure must be called if multi-block M-H algorithms are used to sample actor-dependent random effects.
  Arguments of <EvaluateMultinomial> must be in accordance with the present procedure: see below.*)
  procedure EvaluateEfficiency(var estage: byte; var outf, lf: TFileStream; var p: byte; var theta: pvec);
  (*MS: Evaluate efficiency of Markov chain Monte Carlo algorithm.*)
  procedure MH_P_Step(const p: byte; var theta: pvec; var outf, lf: TFileStream);
  (*MS: Gibbs and Metropolis-Hastings steps to sample parameters: rate and objective function parameters.*)
  procedure MH_PI_Step(const parameter, position: byte; const n_a, tot_n: integer; const tot_t: real; var theta: pvec; var outf, lf: TFileStream);
  (*MS: Gibbs sampler and Metropolis-Hastings step to sample rate parameters corresponding to
  - two or more observations
  - network and behavior models
  assuming constant rate functions; to be extended to non-constant rate functions.*)
  procedure MH_PII_Step(const p: byte; const alpha_constant: pvec; var beta, theta: pvec; var outf, lf: TFileStream);
  (*MS: Metropolis-Hastings step to sample network and behavior objective and gratification function parameter.
  Candidate sampled from
  (1) random walk density.
  (2) autoregressive density.
  (3) independence sampler.*)
  procedure MH_RE_CovRE(const p: byte; const theta: pvec; var outf, lf: TFileStream);
  (*MS: Metropolis-Hastings step to sample random effects and, in the Bayesian case, (inverse) variance-covariance matrices.*)
  procedure Get_Cov(const p: byte; const theta: pvec);
  (*MS: Compute the entities <i_covRE>, <c_covRE>, and <covRE> (in ML framework, not: Bayesian framework),
  which are required for sampling actor-dependent random effects;
  computation of these entities is not strictly necessary on each iteration,
  but is safer while doing little harm in terms of efficiency.*)
  procedure MCMC_initialise_n(const p: byte);
  (*MS: Initialise entities on iteration <n>.*)
  procedure MCMC_cycle_storage(const aim: byte; const iteration: integer; var estage: byte; var p: byte; var theta: pvec);
  (*MS: At the end of each iteration of the MCMC cycle algorithm, some functions of the latent data
  - random effects
  - Bayesian framework: parameters
  are stored.*)
  procedure MCMC_cycle(const setpar: boolean; var outf, lf: TFileStream; var p: byte; var theta: pvec; var estage: byte; const aim: byte);
  (*MS: Cycle through
  (1) M-H step to sample path: network and behavior (ML and Bayesian models)
  (2) M-H step to sample actor-dependent random effects (ML and Bayesian random effects models)
  (3) Gibbs and M-H steps to sample parameters (Bayesian models).*)
  procedure Post_MCMC_cycle(var p: byte; var theta: pvec; var estage: byte; var outf, lf: TFileStream);
  (*MS: When the MCMC cycle algorithm has finished and control is <Siena> is to return to <S_est>,
  some functions of the latent data
  - random effects
  - Bayesian framework: parameters
  are stored.*)
  procedure Choose_MC_SE_Estimate(var ascii: TFileStream; const p: byte; const completed: boolean; var mcRE: rrarray;
  var cov_covRE: pmatrix; const cov_theta: pmatrix; var var_theta: R_Ar);
  (*MS: There are two estimates of the Monte Carlo variance:
  the present procedure selects between them and manipulates the relevant arrays.*)
  procedure CD_Score_Hessian_RE(const p: byte; const evaluate_hessian: boolean; var d: pvec; var dd: pmatrix);
  (*Evaluate the complete-data score and the complete-data Hessian with respect to
  functions of the (co)variance parameters;
  see paper on random effects models.*)
  procedure InitialiseMS_PhaseIII;
  (*MS: when estimating random effects models, phase 3 uses an alternative parametrization;
  here, some constant vectors and matrices are evaluated, which, owing to their constant natures,
  must be evaluated only once.*)
  procedure OpenFiles;
  (*MS: Prepare files for MCMC output.*)
  procedure CloseFiles;
  (*MS: Close MCMC output files.*)
  procedure FinaliseMS;
  (*Assign nil operator to arrays used in random effects models and Bayesian models.*)
  (*---------------------------------------------------------------------------------*)
  (*                                  MS: END...                                     *)
  (*Markov chain Monte Carlo routines applied to conduct likelihood-based inference: *)
  (*- estimate non-hierarchical Bayesian models                                      *)
  (*- estimate frequentist random effects models with actor-dependent random effects.*)
  (*- estimate hierarchical Bayesian models with actor-dependent random effects.     *)
  (*...placed in <S_mcmc_I>. These routines make up only a small part of the total   *)
  (*number of MCMC-related routines, outsourced to <S_mcmc>, <S_mcmc_II>, <S_re>.    *)
  (*---------------------------------------------------------------------------------*)

  private

  ChainVersion: Byte;
  (* determines MH proposal distribution; values 1 - 3; must be 2 or 3 for nza >= 1;
     1 is older and simpler, 2 presumably is better.
     3 is even better. The new FCCP procedure in Chains does not allow 2 any more. *)
  aa1: array of TAspect;
  vv1: array of Byte;
  dd1: array of TDif;
  ii1, jj1: array of integer;
  lp, ra: array of real;
  maxchainlength: integer;
  ppermitted: array of boolean;
  hah: array of array of array of real;

  procedure RenewRateslprobs(k1, k2: TMiniStep; check: boolean;
                              const per: integer; var outf,lf: TFileStream);
(* Renews rates[h] and lprobs[h] for h from k1 to k2 *)
  procedure UpdateRateslprobs(steptype: integer; k1,k2: TMiniStep;
                               const per: integer; var outf,lf: TFileStream);
(* Updates rates[h], lprobs[h] for h from k1 to k2, using newrates, newlprobs *)
  function truncaction(a: real; h: integer):integer; // @@ lapmiddel

 end; (* MLModel *)

implementation

uses SysUtils, Forms, s_dat, rangen, siena_7;



{$I s_mcmc_I} (*MS: Input file <S_mcmc_I>. *)

function MLmodel.truncaction(a: real; h: integer):integer; // @@ lapmiddel
var t: integer;
begin
  if (round(a) < round(zcmin[h])) then t := round(zcmin[h]) else t := round(a);
  if (t > round(zcmax[h])) then t := round(zcmax[h]);
  truncaction := t;
end;

procedure MLModel.MLInitialisePeriods(var outf,lf: TFileStream);
var h: integer;
begin
  FinaliseMS; (*MS: assign nil operator to arrays used in random effects models and Bayesian models.*)
  SetLength(Chain,observations);
(* <<<<< *)
  SetLength(ppermitted,nmax+1);
  maxlength := 0;
(* >>>>> *)
(*[[[*)
  errortext := '';
  maxpermute := 200;
  SetLength(numm,observations);
  for h:=1 to observations-1 do numm[h] := 20;
  maxchainlength := 100;
  SetLength(lp,maxchainlength);
  SetLength(ra,maxchainlength);
  SetLength(aa1,maxchainlength);
  SetLength(ii1,maxchainlength);
  SetLength(jj1,maxchainlength);
  SetLength(vv1,maxchainlength);
  SetLength(dd1,maxchainlength);
(*]]]*)
  for h := 1 to observations-1 do
//  begin
    if (not skipi[h]) then
    Chain[h-1] := TChain.Create(nni[h],nza);
//@@@ weg    initialise_actors(h);
// weg    SetData(Chain[h-1].First,h,outf,lf);
//   end; (* for h *)
  MadeChain := false;
//  assignfile(tf,'testfile.txt');               // @@@@
//  rewrite(tf);                                 // @@@@

end; (* MLInitialisePeriods *)

procedure MLModel.MLFinalizePeriods(var outf,lf: TFileStream);
var h: integer;
x: string;
begin
  if (errortext > '') then begin
    x:= 'Warning: error in unit S_ML: '+errortext+eol;
    outf.Write(pchar(x)^,length(x));
    //writeln(lf,'Warning: error in unit S_ML: ',errortext);
    lf.Write(pchar(x)^,length(x));
   end else
   begin
    if ConsistencyCheck(outf,lf,true)
     then begin
      x := 'Consistency check OK at finalization of chain.'+eol;
      lf.Write(pchar(x)^,length(x));
     end
     else begin
       x := 'Error in consistency check OK at finalization of chain: '+ errortext+eol;
       outf.Write(pchar(x)^,length(x));
     end;
   end;
  FinaliseMS; (*MS: assign nil operator to arrays used in random effects models and Bayesian models.*)
  for h:=0 to observations-2 do
   if (not skipi[h+1]) then  begin
    if (Chain[h].ErrorMes > '') then begin
      x := format('%s%s%s%s',['Warning: error in unit S_ML from unit Chains[',h,']: ',errortext])+eol;
      outf.Write(pchar(x)^,length(x));
     //writeln(lf,'Warning: error in unit S_ML from unit Chains[',h,']: ',errortext);
     lf.Write(pchar(x)^,length(x));
     end;
     Chain[h].Destroy;
   end;
  MadeChain := false;
//  closefile(tf); // @@@@
end; (* MLFinalizePeriod *)


procedure MLModel.RenewRateslprobs(k1, k2: TMiniStep; check: boolean; const per: integer;
                                var outf,lf: TFileStream);
(* Renews mu, sigma2, kappa, and rates[h], lprobs[h] for h from k1 to k2 *)
var i,j: integer;
    k3, k4: TMiniStep;
    thisrate, thislprob, oldkappa, oldmu, oldsigma2: real;
    founderror: boolean;
    Asp: TAspect;
    v: byte;
    d: Tdif;
    x : string;
begin
// writeln(outf,'renewrateslprobs ',k1,' ',k2);
{ // @@@@@ Simple checks, can be omitted, I tried it once and it saved 1% computing time.
  if (not Assigned(k1)) then begin
    errortext := 'RenewRateslprobs called with non-assigned k1';
    writeln(outf,'Error: ',ErrorText);
    exit;
   end;
  if (not Assigned(k2)) then begin
    errortext := 'RenewRateslprobs called with non-assigned k2';
    writeln(outf,'Error: ',ErrorText);
    exit;
   end;

  if (not Thisorder(k1,k2)) then begin
    errortext := 'RenewRateslprobs called with incorrectly ordered k1 = '
                 +inttostr(k1.kindex)+'('+FloatToStrF(k1.time,ffFixed,8,5)+
                 ') , k2 = '+inttostr(k2.kindex)+'('+
                 FloatToStrF(k2.time,ffFixed,8,5)+')';
    writeln(outf,'Error: ',ErrorText);
    exit;
   end;

  if (k2.kindex = 0) then begin
    errortext := 'RenewRateslprobs called with k2 = First';
    writeln(outf,'Error: ',ErrorText);
    exit;
   end;
}
  founderror := false;
  if check then begin
// @@@@<
    if (Chain[per-1].sigma2 < 1e-40) then
      begin
      x := format('%s%20.10f',['Chain[per-1].sigma2 = ',Chain[per-1].sigma2])+eol;
      outf.Write(pchar(x)^,length(x));
      errortext := 'sigma2 is negative in RenewRatesLProbs';
      writefullperiod(per,outf,lf);
      exit;
      end;
// >@@@@
    oldkappa := -0.5*log(Chain[per-1].sigma2) - (sqr(Chain[per-1].totaltime - Chain[per-1].mu)/(2.0*Chain[per-1].sigma2));
    if (abs(Chain[per-1].kappa - oldkappa) > 1e-8) then begin
      x := format('%s%s%20.10f%s%20.10f',['start of RenewRatesLprobs: ',
              '***** old kappa = ',oldkappa,'; kappa = ',Chain[per-1].kappa])+eol;
      outf.Write(pchar(x)^,length(x));
      errortext := 'Error in check at start of RenewRatesLprobs.';
     end;
   end;


  oldkappa := Chain[per-1].kappa; (* given that k1 is in chain[per-1], this is identical to Chain[per-1].kappa *)
  oldmu := Chain[per-1].mu;
  oldsigma2 := Chain[per-1].sigma2;
  Chain[per-1].mu := 0.0;  (* it is not possible to assign a value to Chain[per-1].mu  *)
  Chain[per-1].sigma2 := 0.0;

(* First move from k3 = First to k3 = k1.Predh
   to calculate the partial sums for mu and sigma2. *)
  k3 := Chain[per-1].First;
  if ((k1 <> k3) and (k1 <> k3.Such)) then repeat
    k3 := k3.such;
    if (k3.rate < 1e-50) then begin
     errortext := 'rate = 0.0 in RenewRatesLprobs for k3 = '+inttostr(k3.kindex);
     x := 'Error: '+ErrorText+eol;
     outf.Write(pchar(x)^,length(x));
     exit;
     end;
    thisrate := 1.0/k3.rate;
    Chain[per-1].mu := Chain[per-1].mu + thisrate;
    Chain[per-1].sigma2 := Chain[per-1].sigma2 + sqr(thisrate);
   until (k3.such = k1);
  k3 := k1.predh;
(* Now move from k3 = k1.predh to k2
   to calculate the partial sums for mu and sigma2,
   and update rates and lprobs. *)
  SetData(k3,per,outf,lf);
  k3 := k3.such;
(* k3 := k1 does not work properly in case k1 = First *)
  if k2.Extreme then k4 := Chain[per-1].Last else k4 := k2.such;
  repeat
    asp := k3.Ash;
    i := k3.ih;
    j := k3.jh;
    v := k3.varh;
    d := k3.Difh;
    CalcRatelProb(thisrate,thislprob,asp,i,j,v,d,outf,lf);
    if (errortext > '') then
      begin
      x :='Error in RenewRatesLprobs 1: '+errortext+eol;
      outf.Write(pchar(x)^,length(x));
      exit;
      end;
    if check then begin
//      write(outf,k3:3,' ',thisrate:12:4,' ',rates[k3]:12:4,
//                                       thislprob:12:4,' ',lprobs[k3]:12:4);
//      if (abs(thisrate - rates[k3]) + abs(thislprob - lprobs[k3]) > 1e-6)
//       then writeln(outf,'*****') else writeln(outf);
      if (abs(thisrate - k3.rate) + abs(thislprob - k3.lprob) > 1e-6)
       then begin
         if (not founderror) then begin
           outf.Write(pchar(eol)^,length(eol));//writeln(outf);
           x := '       k    time        rate        oldrate      lprob       oldlprob'+eol;
           outf.Write(pchar(x)^,length(x));
           founderror := true;
           errortext := 'Error found in value checks in RenewRatesLprobs.';
          end;
         x := format('%s%3d%s%8.5f%s%12.4f%s%12.4f%12.4f%s%12.4f',['*****', k3.kindex,' ',k3.time,' ',thisrate,' ',k3.rate,
                 thislprob,' ',k3.lprob])+eol;
         outf.Write(pchar(x)^,length(x));
       end;
     end;
    k3.rate := thisrate;
    k3.lprob := thislprob;
    if k3.NotExtreme then begin
       if (asp = Network) then ChangeTie(i,j) else ChangeBehavior(i,v,d);
      end;
    if (k3.kindex >= 1) then begin
      Chain[per-1].mu := Chain[per-1].mu + 1.0/thisrate;
      Chain[per-1].sigma2 := Chain[per-1].sigma2 + sqr(1.0/thisrate);
     end;
    k3 := k3.Such;
   until (k3 = k4);

   (* last updates still need to be done *)
  if k3.NotExtreme then begin
    asp := k3.Ash;
    i := k3.ih;
    j := k3.jh;
    v := k3.varh;
    d := k3.Difh;
   end else begin (* arbitrary choices for k3 = Last *)
    i := 1;
    j := 2;
    asp := Network;
    v := 1;
    d := 0;
   end;
  CalcRatelProb(thisrate,thislprob,asp,i,j,v,d,outf,lf);
  if (errortext > '') then
    begin
    x := 'Error in RenewRatesLprobs 2: '+errortext+eol;
    outf.Write(pchar(x)^,length(x));
    exit;
    end;
  k3.rate := thisrate;
  if k3.NotExtreme then begin
    k3.lprob := thislprob;
    if (asp = Network) then ChangeTie(i,j) else ChangeBehavior(i,v,d);
   end;
  Chain[per-1].mu := Chain[per-1].mu + 1.0/thisrate;
  Chain[per-1].sigma2 := Chain[per-1].sigma2 + sqr(1.0/thisrate);

(* Finally move from k3 = k2 to Last
   to finalize the partial sums for mu and sigma2. *)
  if k3.NotExtreme then repeat
    k3 := k3.such;
    thisrate := 1.0/(k3.rate);
    Chain[per-1].mu := Chain[per-1].mu + thisrate;
    Chain[per-1].sigma2 := Chain[per-1].sigma2 + sqr(thisrate);
   until (k3 = Chain[per-1].Last);
  Chain[per-1].kappa := -0.5*log(Chain[per-1].sigma2)
    - (sqr(Chain[per-1].totaltime - Chain[per-1].mu)/(2.0*Chain[per-1].sigma2));
  if check then begin
    if (abs(Chain[per-1].kappa - oldkappa) > 1e-8) then begin
   x := format('%s%20.10f%s%20.10f',
     ['***** old kappa = ',oldkappa,'; kappa = ',Chain[per-1].kappa])+eol;
     outf.Write(pchar(x)^,length(x));
    end;
    if (abs(Chain[per-1].mu - oldmu) > 1e-8) then begin
    x := format('%s%20.10f%s%20.10f',
     ['***** old mu = ',oldmu,'; mu = ',Chain[per-1].mu])+eol;
     outf.Write(pchar(x)^,length(x));
    end;
   if (abs(Chain[per-1].sigma2 - oldsigma2) > 1e-8) then begin
   x := format('%s%20.10f%s%20.10f',
     ['***** old sigma2 = ',oldsigma2,'; sigma2 = ',Chain[per-1].sigma2])+eol;
     outf.Write(pchar(x)^,length(x));
   end;
   end;
  if (errortext > '') then begin
    x := 'error: '+ErrorText+'; called from RenewRateslprobs'+eol;
    outf.Write(pchar(x)^,length(x));
    exit;
   end;
//  if check and (errortext = '') then writeln(outf,'renewrateslprobs OK'); 
end; (* RenewRateslprobs *)


procedure MLModel.UpdateRateslprobs(steptype: integer; k1, k2: TMiniStep;
                                const per: integer; var outf,lf: TFileStream);
(* Updates rates[h], lprobs[h] for h from k1 to k2, including bounds, using newrates, newlprobs *)
(*  steptype 1: insert; 2: cancel; 3: swap.
    For steptypes 1 and 3, also updates k1.predh and k2.such *)
var k: TMiniStep;
    thismu, thissigma2: real;
    x: string;
begin
  if (not (Thisorder(k1,k2) or (k1 = k2))) then begin
    errortext := 'UpdateRateslprobs called with incorrectly ordered k1, k2';
    x := 'Error: '+ErrorText+eol;
    outf.Write(pchar(x)^,length(x));
    exit;
   end;

  thismu := Chain[per-1].mu;
  thissigma2 := Chain[per-1].sigma2;

    k := k1;
    while (k <> k2) do
      begin
      if k.Notextreme then
        begin
if (k.newrate < 1e-20) then begin
    x := format('%s%4d%s%12.4f',['error !newrates[k]! ',k.kindex,' ',k.newrate])+eol;
    lf.Write(pchar(x)^,length(x));
  end
 else
          begin
          thismu := thismu + 1.0/k.newrate - 1.0/k.rate;
          thissigma2 := thissigma2 + sqr(1.0/k.newrate) - sqr(1.0/k.rate);
          k.rate := k.newrate;
          k.lprob := k.newlprob;
          end;
        end; (* if *)
      k := k.such;
      end;

  case steptype of // @@@@
    1: begin
if (k1.predh.extreme) or (k2.extreme) then begin
  errortext := 'next to extreme insertion in UpdateRateslprobs';
  exit;
 end;
       k1.predh.lprob := newlprob1;
       k1.predh.rate := newrate1;
       k2.lprob := newlprob2;
       k2.rate := newrate2;
       thismu := thismu + 1.0/newrate1 + 1.0/newrate2;
       thissigma2 := thissigma2 + 1.0/sqr(newrate1) + 1.0/sqr(newrate2);
       end;
    2: begin
       thismu := thismu - 1.0/oldrate1 - 1.0/oldrate2;
       thissigma2 := thissigma2 - 1.0/sqr(oldrate1) - 1.0/sqr(oldrate2);
       end;
    3: begin
if (k1.extreme) or (k2.extreme) then begin
  errortext := 'next to extreme insertion in UpdateRateslprobs';
  exit;
 end;
       k1.predh.lprob := newlprob1;
       k1.predh.rate := newrate1;
       k2.lprob := newlprob2;
       k2.rate := newrate2;
       thismu := thismu + 1.0/newrate1 + 1.0/newrate2
                        - 1.0/oldrate1 - 1.0/oldrate2;
       thissigma2 := thissigma2 + 1.0/sqr(newrate1) + 1.0/sqr(newrate2)
                               - 1.0/sqr(oldrate1) - 1.0/sqr(oldrate2);
       end;
    end; (* case *)

  Chain[per-1].kappa := -0.5*log(thissigma2) -
        (sqr(Chain[per-1].totaltime - thismu)/(2.0*thissigma2));
  Chain[per-1].mu := thismu;
  Chain[per-1].sigma2 := thissigma2;
  if (errortext > '') then begin
    x := 'error: '+ErrorText+'; called from UpdateRateslprobs'+eol;
    outf.Write(pchar(x)^,length(x));
    exit;
   end;
end; (* UpdateRateslprobs *)

procedure MLModel.Cancel(var k: TMiniStep; const per: integer; var outf,lf: TFileStream);
(* cancels change with index k *)
var x:string;
begin
  if k.Extreme then begin
    errortext := 'Cancel called for unchangeable index.';
    x := 'Error: '+ErrorText+eol;
    outf.Write(pchar(x)^,length(x));
    exit;
   end;
  if (not Assigned(k)) then begin
    errortext := 'Cancel called for non-assigned index.';
    x := 'Error: '+ErrorText+eol;
    outf.Write(pchar(x)^,length(x));
    exit;
   end;

  Chain[per-1].Delete(k);

  if (Chain[per-1].ErrorMes > '') then
    Errortext := ErrorText + chain[per-1].ErrorMes;
  if (errortext > '') then begin
    x := 'error: '+ErrorText+'; called from Cancel'+eol;
    outf.Write(pchar(x)^,length(x));
    exit;
   end;
end; (* Cancel *)

procedure MLModel.Insert(k: TMiniStep; const V0: TAspect; const V1, V2: integer;
                         const V3: byte; const V4: Tdif;
                         const per: integer; var outf,lf: TFileStream);
(* inserts a change of element i,j immediately before the change with index k *)
var x: string;
begin
  if (not Assigned(k)) then begin
    errortext := 'Insert called for non-assigned ministep.';
    x := 'Error: '+ErrorText+eol;
    outf.Write(pchar(x)^,length(x));
    exit;
   end;
  if (k.kindex <= 0) then begin
    errortext := 'Insert called for initial ministep k = First.';
    x := 'Error: '+ErrorText+eol;
    outf.Write(pchar(x)^,length(x));
    exit;
   end;
  if (not valid(V0,V1,V2,V3,V4)) then begin
    errortext := 'Insert called for non-existing ministep.';
    x := format('%s%s%s',['Error: ',ErrorText,Kindsstring(V0,V1,V2,V3,V4)])+eol;
    outf.Write(pchar(x)^,length(x));
    exit;
   end;

  Chain[per-1].InsertBefore(k,V0,V1,V2,V3,V4);

  if Chain[per-1].ErrorMes > '' then
    ErrorText := ErrorText + Chain[per-1].Errormes;
  if (errortext > '') then begin
    x := 'error: '+ErrorText+'; called from Insert'+eol;
    outf.Write(pchar(x)^,length(x));
    exit;
   end;
end; (* Insert *)


procedure MLModel.Swap(var k1,k2: TMiniStep; const per: integer; var outf,lf: TFileStream);
(* swaps the coordinates of the elements with index k1 and k2 *)
(* The ministeps k1 and k2 are removed in this procedure, it is unknown
   what will come out! *)
var k3,k4: TMiniSTep;
    i1, j1, i2, j2: integer;
    a1, a2: TAspect;
    v1, v2: Byte;
    d1, d2: TDif;
    x: string;
begin
  if SameRkind(k1,k2) and (k1.Difh = k2.Difh) then exit;
  if (k1.Extreme or k2.Extreme) then begin
    errortext := 'Swap called for extreme index.';
    x := 'Error: '+ErrorText+eol;
    outf.Write(pchar(x)^,length(x));
    exit;
   end;
  if (k1.Such = k2) or (k1.Predh = k2) then begin
(* immediately after each other *)
    if ThisOrder(k1,k2) then begin
      k3 := k2.Such;
      a1 := k1.ash;
      i1 := k1.ih;
      j1 := k1.jh;
      v1 := k1.varh;
      d1 := k1.difh;
      a2 := k2.ash;
      i2 := k2.ih;
      j2 := k2.jh;
      v2 := k2.varh;
      d2 := k2.difh;
     end else begin
      k3 := k1.Such;
      a1 := k2.ash;
      i1 := k2.ih;
      j1 := k2.jh;
      v1 := k2.varh;
      d1 := k2.difh;
      a2 := k1.ash;
      i2 := k1.ih;
      j2 := k1.jh;
      v2 := k1.varh;
      d2 := k1.difh;
     end;
     with Chain[per-1] do begin
       Delete(k1);
       Delete(k2);
       InsertBefore(k3,a2,i2,j2,v2,d2);
       InsertBefore(k3,a1,i1,j1,v1,d1);
      end;
   end else (* not immediately after each other *) begin
     a1 := k1.ash;
     i1 := k1.ih;
     j1 := k1.jh;
     v1 := k1.varh;
     d1 := k1.difh;
     a2 := k2.ash;
     i2 := k2.ih;
     j2 := k2.jh;
     v2 := k2.varh;
     d2 := k2.difh;
     k3 := k1.such;
     k4 := k2.Such;
     with Chain[per-1] do begin
       Delete(k1);
       Delete(k2);
       InsertBefore(k4,a1,i1,j1,v1,d1);
       InsertBefore(k3,a2,i2,j2,v2,d2);
      end;
    end; (* not immediately after each other *)

  if Chain[per-1].ErrorMes > '' then ErrorText := ErrorText + Chain[per-1].Errormes;
  if (errortext > '') then begin
    x := 'error: '+ErrorText+'; called from Swap'+eol;
    outf.Write(pchar(x)^,length(x));
    exit;
   end;
end; (* Swap *)


function MLModel.PR_Step(steptype: integer; k1, k2: TMiniStep;
                 const V0: TAspect; const V1, V2: integer; const V3: byte;
                 const V4: Tdif;
                 const per: integer; var outf,lf: TFileStream;
                 var possible: boolean): real;
(* Log probability ratio (changed / current) for steps of type steptype.
  steptype 1: insert (V0,V1,V2,V3,V4) immediately before k1 and also before k2;
              for V0 = behavior, V4 applies before k1 and -V4 before k2;
              if V0 = behavior and k1=k2, the first change is V4 and the second -V4.
           2: cancel k1 and k2;
           3: swap k1 and k2;
           4: InsertDiag, insert (V1,V1) or (V3,V4) immediately before k1;
           5: CancelDiag, cancel the diagonal step at k1.
   possible indicates whether the indicated step is possible.        *)
var lpr, mu_new, sigma2_new, thisrate, thislprob: real;
 i1, j1: integer;
 k, k11, k22, k4: TMiniStep;
 a1: TAspect;
 va1: byte;
 d1, d11: TDif;
 x: string;
begin
  Result := 0.0;
  possible := true;
  if (errortext > '') then begin
    x := 'error: '+ErrorText+'; discovered when entering PR_Step'+eol;
    outf.Write(pchar(x)^,length(x));
    exit;
   end;

  if ((k1.kindex = 0) or (k2.kindex = 0)) then begin
    errorText := 'PR_step '+inttostr(steptype)+' called for First ministep.';
    x := 'Error: '+ErrorText+eol;
    outf.Write(pchar(x)^,length(x));
    exit;
   end;

  if (((k1.kindex = 1) or (k2.kindex = 1)) and (steptype in [2,3])) then begin
    errorText := 'PR_step '+inttostr(steptype)+' called for Last ministep:'
            +inttostr(k1.kindex)+', '+inttostr(k2.kindex);
    x := 'Error: '+ErrorText+eol;
    outf.Write(pchar(x)^,length(x));
    exit;
   end;


  if (not valid(V0,V1,V2,V3,V4)) and (steptype = 1) then begin
    ErrorText := 'PR_step '+inttostr(steptype)+' called for invalid kind ('
                 +KindsString(V0,V1,V2,V3,V4)+').';
    x := 'Error: '+ErrorText+eol;
    outf.Write(pchar(x)^,length(x));
    exit;
   end;

  lpr := 0.0;
  mu_new := Chain[per-1].mu;
  sigma2_new := Chain[per-1].sigma2;

  k11 := k1;
  k22 := k2;
  d11 := V4;

  case Steptype of
    1: begin
       if ThisOrder(k2,k1) then
         begin
         k11 := k2;
         k22 := k1;
         d11 := -V4;
         end;
       end;
    2: begin
       if not ((k1.ash = k2.ash) and (k1.ih = k2.ih) and (k1.jh = k2.jh)
              and (k1.varh = k2.varh) and (k1.difh + k2.difh = 0) ) then
         begin
         ErrorText := 'PR_step 2 called for two non-canceling changes.';
         x := 'Error: '+ErrorText+eol;
         outf.Write(pchar(x)^,length(x));
         exit;
         end;
       if ThisOrder(k2,k1) then
         begin
         k11 := k2;
         k22 := k1;
         d11 := -V4;
         end;
       if (k1 = k2) then
         begin
         errorText := 'PR_Step 2 called for the same changes.';
         x := 'Error: '+ErrorText+eol;
         outf.Write(pchar(x)^,length(x));
         exit;
         end;
       end;
    3: begin
       if ThisOrder(k2,k1) then
         begin
         k11 := k2;
         k22 := k1;
         end;
       if (k1 = k2) then
         begin
         errorText := 'PR_Step 3 called for the same changes.';
         x := 'Error: '+ErrorText+eol;
         outf.Write(pchar(x)^,length(x));
         exit;
         end;
       end;
    4: begin
       end;
    5: begin
       if (not k1.Diagonal) then
         begin
         ErrorText := 'PR_step 5 called for invalid change ('
                      +k1.KindString+').';
         x := 'Error: '+ErrorText+eol;
         outf.Write(pchar(x)^,length(x));
         exit;
         end;
       if k1.Extreme then
         begin
         errorText := 'PR_step '+inttostr(steptype)+' called for Extreme ministep.';
         x := 'Error: '+ErrorText+eol;
         outf.Write(pchar(x)^,length(x));
         exit;
         end;
       end;
    else begin
         errorText := 'PR_step called for impossible steptype '+inttostr(steptype)+'.';
         x := 'Error: '+ErrorText+eol;
         outf.Write(pchar(x)^,length(x));
         exit;
         end;
     end; (* case StepType *)

  k4 := nil;
  k := k11.predh;
  if (steptype < 5) then SetData(k,per,outf,lf);

  case steptype of
(* k4 is the change where the process will stop *)
    1: begin
(* new step before k11 *)
(* [[[]]] *)       if ((Chainversion >= 2) and (k.ash = Behavior)) then if
       ((Minimalya(V1,V3) and (d11 = -1)) or (Maximalya(V1,V3) and (d11 = 1)))
          then
           begin
           possible := false;
           exit;
           end;
         CalcRatelprob(thisrate,thislprob,V0,V1,V2,V3,d11,outf,lf);
         if (errortext > '') then
         begin
         x := 'Error in CalcRatelProb 1: '+errortext+eol;
         outf.Write(pchar(x)^,length(x));
         exit;
         end;
         if (thisrate < -1.0) then begin possible :=false; exit; end;
         (* This happens only when V0 = Behavior, and d11 implies a
            non-permitted change. *)
         newrate1 := thisrate;
         newlprob1 := thislprob;
         lpr := lpr + thislprob;
         mu_new := mu_new + 1.0/thisrate;
         sigma2_new := sigma2_new + 1.0/sqr(thisrate);
         if (V0 = Network) then ChangeTie(V1,V2) else ChangeBehavior(V1,V3,d11);
         k4 := k22.predh;
        end;
    2: begin
         k := k11; (* k11 is skipped *)
         lpr := lpr - k11.lprob - k22.lprob;
         mu_new := mu_new - (1.0/k11.rate) - (1.0/k22.rate);
         sigma2_new := sigma2_new  - (1.0/sqr(k11.rate)) - (1.0/sqr(k22.rate));
         k4 := k22.predh;
        end;
    3: begin
         k4 := k22;
        end;
    4: begin
(* new diagonal step before k11 *)
         CalcRatelprob(thisrate,thislprob,V0,V1,V1,V3,0,outf,lf);
         if (errortext > '') then
         begin
         x := 'Error in CalcRatelProb 2: '+errortext+eol;
         outf.Write(pchar(x)^,length(x));
         exit;
         end;
         newrate1 := thisrate;
         newlprob1 := thislprob;
         lpr := lpr + thislprob;
         mu_new := mu_new + 1.0/thisrate;
         sigma2_new := sigma2_new + 1.0/sqr(thisrate);
        end;
    5: begin
(* k11 is skipped *)
         lpr := lpr - k11.lprob;
         mu_new := mu_new - 1.0/k11.rate;
         sigma2_new := sigma2_new  - 1.0/sqr(k11.rate);
        end;
   end; (* case *)

  if (steptype <= 3) then
(* go on with calculations until k4 *)
  while (k <> k4) do begin
    if (k.kindex = 1) then begin
      errortext := 'PR_Step '+inttostr(steptype)+' stopped because k = Last';
      x :=format('%s%s%s%s',['Error: ',ErrorText,' ',k.KindString])+eol;
      outf.Write(pchar(x)^,length(x));
      exit;
     end;
    k := k.such;
(* a change will be made in element (i1,j1) *)
    a1 := k.ash;
    i1 := k.ih;
    j1 := k.jh;
    va1 := k.varh;
    d1 := k.difh;
    if (Steptype = 3) then begin
      if (k = k1) then begin
        a1 := k2.ash;
        i1 := k2.ih;
        j1 := k2.jh;
        va1 := k2.varh;
        d1 := k2.difh;
       end;
      if (k = k2) then begin
        a1 := k1.ash;
        i1 := k1.ih;
        j1 := k1.jh;
        va1 := k1.varh;
        d1 := k1.difh;
       end;
      end; (* if steptype = 3 *)
    CalcRatelprob(thisrate,thislprob,a1,i1,j1,va1,d1,outf,lf);
    if (errortext > '') then
      begin
      x := 'Error in CalcRatelProb 3: '+errortext+eol;
      outf.Write(pchar(x)^,length(x));
      x := format('%s%d',['Steptype ',steptype])+eol;
      outf.Write(pchar(x)^,length(x));
      x := format('%s%d%s%12.8f',['k1  : ',k1.kindex,k1.kindstring,k1.time])+eol;
      outf.Write(pchar(x)^,length(x));
      x := format('%s%d%s%12.8f',['k2  : ',k2.kindex,k2.kindstring,k2.time])+eol;
      outf.Write(pchar(x)^,length(x));
      x := format('%s%d%s%12.8f',['k11 : ',k11.kindex,k11.kindstring,k11.time])+eol;
      outf.Write(pchar(x)^,length(x));
      x := format('%s%d%s%12.8f',['k22 : ',k22.kindex,k22.kindstring,k22.time])+eol;
      outf.Write(pchar(x)^,length(x));
      x := format('%s%d%s%12.8f',['k   : ',k.kindex,k.kindstring,k.time])+eol;
      outf.Write(pchar(x)^,length(x));
      x := kindsstring(a1,i1,j1,va1,d1)+eol;
      outf.Write(pchar(x)^,length(x));
      exit;
      end;
    if (thisrate < -1.0) then begin possible := false; exit; end;
    (* This happens only when the change is not permitted. *)
//writeln(outf,'thisrate = ',thisrate:12:4, '; thislprob = ',thislprob:12:4);
//writeln(outf,'rates[k] = ',rates[k]:12:4, '; lprobs[k] = ',lprobs[k]:12:4);
    lpr := lpr + thislprob - k.lprob;
    if (a1 = Network) then ChangeTie(i1,j1) else ChangeBehavior(i1,va1,d1);
    k.newrate := thisrate;
    k.newlprob := thislprob;
    if (steptype = 3) then
      begin
      if (k = k11) then
        begin
        newrate1 := thisrate;
        newlprob1 := thislprob;
        end
       else
        if (k = k22) then
        begin
        newrate2 := thisrate;
        newlprob2 := thislprob;
        end;
      end;
    mu_new := mu_new + 1.0/thisrate - 1.0/k.rate;
    sigma2_new := sigma2_new + 1.0/sqr(thisrate) - sqr(1.0/k.rate);
   end; (* if (steptype <= 3) while *)

  if (steptype = 1) then begin
(* insert extra change before k22; for behavior this is a canceling change, hence -d11 *)
    CalcRatelprob(thisrate,thislprob,V0,V1,V2,V3,-d11,outf,lf);
    if (errortext > '') then
      begin
      x := 'Error in CalcRatelProb 4: '+errortext+eol;
      outf.Write(pchar(x)^,length(x));
      exit;
      end;
    if (thisrate < -1.0) then begin possible := false; exit; end;
    (* This happens only when the change is not permitted. *)
    newrate2 := thisrate;
    newlprob2 := thislprob;
    lpr := lpr + thislprob;
    mu_new := mu_new + 1.0/thisrate;
    sigma2_new := sigma2_new + 1.0/sqr(thisrate);
    if (V0 = Network) then ChangeTie(V1,V2) else ChangeBehavior(V1,V3,-d11);
   end;

  PR_Step := lpr  - Chain[per-1].kappa
               - 0.5*log(sigma2_new) - (sqr(Chain[per-1].totaltime - mu_new)/(2.0*sigma2_new));

  if (errortext > '') then begin
    x := 'Error : '+errortext+'; called from PR_Step'+eol;
    outf.Write(pchar(x)^,length(x));
    exit;
   end;
end; (* PR_Step *)


function MLModel.MH_TryInsertDiag(a: TAspect; i: integer; v: byte;
                                  k1: TMiniStep; const per: integer;
                                  var outf,lf: TFileStream): real;
(* Metropolis Hastings step for inserting one "change"  of element i,i
   immediately before k1 *)
var r: real;
    insd,poss: boolean;
    k: TMiniStep;
    x: string;
label EndOfProcedure;
begin
//if (a = Network) then
//writeln(outf,'Try InsertDiag' ,k1.Kindstring,k1.kindex:4,' ',kindsstring(a,i,i,v,0))
//else
//writeln(outf,'Try InsertDiag' ,k1.Kindstring,k1.kindex:4,' ',kindsstring(a,i,0,v,0));
  r := PR_Step(4,k1,k1,a,i,i,v,0,per,outf,lf,poss);
  if (a = Network) then
    r := (expo(r)*(Chain[per-1].TotNumber)*prddel)/
          ((Chain[per-1].DiagNumber+1)*prndins)
   else
    r := (expo(r)*(Chain[per-1].TotNumber)*prddel)/
          ((Chain[per-1].DiagNumber+1)*prbdins);
  Result := r;
//write(outf,'r = ',r:12:4);
  if (not poss) then ErrorText := 'PR_Step reports a non-permitted diagonal insert step.';
    (* which is itself impossible *)
  if (errortext > '') then begin
    goto EndOfProcedure;
   end;
  if (r >= 1.0) then begin
    insd := true;
   end else begin
    insd := (rando(version) < r);
   end;
  if insd then begin
    newpath := true; (*MS: monitor acceptance rate of candidate paths.*)
    inc(ninsds);
//writeln(outf,' ; accepted ');
    Insert(k1,a,i,i,v,0,per,outf,lf);
    if (errortext > '') then begin
      goto EndOfProcedure;
     end;
//    writeln(lf,'     1  ',nstep:6,TotNumber:6,r:12:8);
    k := k1.predh;
    if (newrate1 < 1e-20) then begin
      errortext := 'newrate1 = 0 in MH_TryInsertDiag';
      goto EndOfProcedure;
      exit;
     end;
    k.rate := newrate1;
    k.lprob := newlprob1;
    Chain[per-1].mu := Chain[per-1].mu + 1.0/newrate1;
    Chain[per-1].sigma2 := Chain[per-1].sigma2 + 1.0/sqr(newrate1);
    Chain[per-1].kappa := -0.5*log(Chain[per-1].sigma2)
      - (sqr(Chain[per-1].totaltime - Chain[per-1].mu)/(2.0*Chain[per-1].sigma2));
   end else begin
    inc(nnoinsds);
//    writeln(lf,'     0  ',nstep:6,TotNumber:6,r:12:8);
//writeln(outf,' ; not accepted ');
    end;
EndOfProcedure:
   if (errortext > '') then begin
    x := 'error : '+errortext+eol;
    outf.Write(pchar(x)^,length(x));
    x := 'called from MH_TryInsertDiag'+eol;
    outf.Write(pchar(x)^,length(x));
    exit;
   end;
end; (* MH_TryInsertDiag *)

function MLModel.MH_TryCancelDiag(var k1: TMiniStep;
                                  const per: integer; var outf,lf: TFileStream): real;
                          (* Metropolis Hastings step for canceling one
                              "change"  of element i,i *)
var r,rat: real;
    cand,poss: boolean;
    a: TAspect;
    x: string;
label EndOfProcedure;
begin
//writeln(outf,'Try CancelDiag' ,k1.Kindstring,k1.kindex:4);
  Result := 1.0;
  if (not k1.Diagonal) then begin
    ErrorText := 'MH_TryCancelDiag called for non-diagonal element.';
    exit;
   end;
   a := k1.ash;

//  r := PR_Step(5,k1,k1,Network,1,1,1,0,per,outf,lf,poss); @@@@
  r := PR_Step(5,k1,k1,a,1,1,1,0,per,outf,lf,poss);
(* Network,1,1,1,0 is used because it is not important which kind it is. *)
  if (a = Network) then
    r := (expo(r)*Chain[per-1].DiagNumber*prndins)
        /((Chain[per-1].TotNumber-1)*prddel)
   else
    r := (expo(r)*Chain[per-1].DiagNumber*prbdins)
        /((Chain[per-1].TotNumber-1)*prddel);
  Result := r;
//write(outf,'r = ',r:12:4);
  if (not poss) then ErrorText := 'PR_Step reports a non-permitted diagonal cancel step.';
     (* which is itself impossible *)
  if (errortext > '') then begin
    goto EndOfProcedure;
   end;
  if (r >= 1.0) then begin
    cand := true;
   end else
    begin
    cand := (rando(version) < r);
    end;
  if cand then begin
    newpath := true; (*MS: monitor acceptance rate of candidate paths.*)
    inc(ncands);
//writeln(outf,' ; accepted ');
    rat := k1.rate;
    Cancel(k1,per,outf,lf);
    if (errortext > '') then begin
      goto EndOfProcedure;
     end;
    Chain[per-1].mu := Chain[per-1].mu - 1.0/rat;
    Chain[per-1].sigma2 := Chain[per-1].sigma2 - (1.0/sqr(rat));
    Chain[per-1].kappa := -0.5*log(Chain[per-1].sigma2)
      - (sqr(Chain[per-1].totaltime - Chain[per-1].mu)/(2.0*Chain[per-1].sigma2));
//    writeln(lf,'     1  ',nstep:6,TotNumber:6,r:12:8);
   end else begin
    inc(nnocands);
//    writeln(lf,'     0  ',nstep:6,TotNumber:6,r:12:8);
//writeln(outf,' ; not accepted ');
    end;
EndOfProcedure:
   if (errortext > '') then begin
    x := 'error : '+errortext+eol;
    outf.Write(pchar(x)^,length(x));
    x := 'called from MH_TryCancelDiag'+eol;
    outf.Write(pchar(x)^,length(x));
    exit;
   end;
end; (* MH_TryCancelDiag *)


function  MLModel.MH_InsDelPermute(const insdel: boolean;
                                   const per: integer; var outf,lf: TFileStream): real;
(* This procedure still contains a lot of extra checks,
   which later on will have to be deleted. @@@@ *)
var  V0, a: TAspect; v: byte; d: Tdif;
   mu_new, mu_newk, sigma2_new, sigma2_newk,
     thisrate, nwrate1, nwrate2, thislprob, nwlprob1, nwlprob2: real;
   pstep: array of TMiniStep;
   k0, k1, k2, k, kk, kkk, kist, knum, knums: Tministep;
   V1, V2, num, numtot, numstep, numav1, numav2, i, j, ist, h, hh,
   kno, ctn: integer;
   prb, pp, t1, pr, pa, slp, smu, smu2: real;
   accept, OK, stopit, pluspossible, minpossible: boolean;
   nc, ncnew, nk, nmuls: integer;
   ins, truncated, extrasteps: boolean;
   V3: byte;
   V4, v4check: Tdif;
(* for checking only: *)
   k1check, k2check,k11,k22: TMiniStep;
   prob1: real;
   del: boolean;
   x: string;
//   q1,q2,qq1,qq2,ppr1,ppr2: real;

label EndOfProcedure;

procedure Reporterror(var ka,kb: TMinistep);
begin
  x := 'error : '+errortext+eol;
  outf.Write(pchar(x)^,length(x));
  x := format('%s%d',['period ',per])+eol;
  outf.Write(pchar(x)^,length(x));
  outf.write(pchar(string('ka = '))^,5); ka.WriteKind(outf); outf.Write(pchar(eol)^,length(eol));//writeln(outf);
  outf.write(pchar(string('kb = '))^,5); kb.WriteKind(outf); outf.Write(pchar(eol)^,length(eol));//writeln(outf);
  if insdel then
    begin
    x := 'insdel'+eol;
    outf.Write(pchar(x)^,length(x));
    if ins then begin
      x := 'ins'+eol;
      outf.Write(pchar(x)^,length(x));
      end
       else begin
        x := 'del'+eol;
        outf.Write(pchar(x)^,length(x));
        end;
    end;
  x := format('%s%d',['num = ',num])+eol;
  outf.Write(pchar(x)^,length(x));
  writeperiod(per,outf,lf);
end;
begin (* MH_InsDelPermute *)

(* Determine a random permitted permutation of num ministeps;
  if insdel, insert or delete a CCP of rkind (V0,V1,V2,V3,1);
  the permuted part does not contain any elements of this rkind.
  the permuted part starts immediately after the first element of this CCP.
  Note that for nza <= 0, all permutations are permitted.
  The permutation and associated probabilities and rates
  are stored in the arrays aa1, ii1, jj1, vv1, dd1, lp, ra.
   *)

  result := 0.0;
(* determine interval length num that will be permuted *)
  ctn := chain[per-1].TotNumber;
  V4 := 0;
  if madechain then
    begin
    if (numm[per] >= ctn) then numm[per] := ctn-1;
    if (numm[per] >= 40) then numm[per] := 40;
    if (numm[per] < 2) then numm[per] := 2;
    num := trunc(numm[per]);
    end
    else num := 0;

(*if madechain then begin
  write(outf,'MH_InsDelPermute called for ');
  writeln(outf,'period = ',per);
  writeln(outf,'insdel = ', insdel);
  writeln(outf,'num = ',num);
end;
*)


(* Choose between insertion and deletion.
   Determine the first element k1 (and if insdel then also the last element k2)
   of the interval of ministeps that will be considered.
   For insert, a canceling pair of ministeps will be inserted before k1 and k2;
   for delete, k1.predh and k2 will be deleted.
    *)
  if insdel then
    begin
(* determine random rkind of ins/del *)
    if (Chain[per-1].MultipleSNumber >= 1) and (rando(version) < promul) then
      begin
(* a random single multiple *)
        repeat
        k1 := Chain[per-1].RandomMultipleS(version);
        if assigned(k1) then
          begin
          V0 := k1.ash;
          V1 := k1.ih;
          V2 := k1.jh;
          V3 := k1.varh;
          end;
        until ((V0 = Network) and (V1 <> V2) and (not Struct(V1,V2,per)))
           or ((V0 = Behavior) and Structactive[V1]);
      end
     else
      begin
(* a random active element *)
      if (nza <= 0) then V0 := Network else
        begin
        if (rando(version) < prnet) then V0 := Network else V0 := Behavior;
        end;
      if (V0 = Network) then
        begin
          repeat
          V1 := randomd(version,n)+1;
          V2 := randomd(version,n)+1;
          V3 := 0;
          until ((V1 <> V2) and (not Struct(V1,V2,per)));
        end
       else
        begin
          repeat V1 := randomd(version,n)+1 until StructActive[V1];
          V2 := 1;
          if (nza <= 1) then V3 := 1 else V3 := randomd(version,nza)+1;
        end;
      end;

    nc := Chain[per-1].NCCPOfRkind(V0,V1,V2,V3,1);
    nk := Chain[per-1].Numberrkind(V0,V1,V2,V3,1);  // nc and nk could be calculated more efficiently together
    nmuls := Chain[per-1].MultipleSNumber;
    if (nc <= 0) then ins := true else ins := (rando(version) < prelins);

//if madechain then begin
//  if ins then write(outf,nc,' ins ') else write(outf,nc, ' del ');
//end;

//writeperiod(per,outf,lf);
    if ins then
      begin
(* choose random k1 and k2 in the same "interval".
   numav1, numav2 (av for available) are
   the numbers of possibilities for k1, k2. *)
      if assigned(Chain[per-1].FirstOfRkind(V0,V1,V2,V3,1)) then
        begin
        repeat k1 := Chain[per-1].RandomElement(version)
         until not SameRkind(k1,V0,V1,V2,V3,1);
        numav1 := ctn - Chain[per-1].NumberRkind(V0,V1,V2,V3,1);
(* determine k and kk which are the last ministep before k1,
   and the first after k1, of this rkind *)
        k := Chain[per-1].First;
        kk := Chain[per-1].FirstOfRkind(V0,V1,V2,V3,1);
        while (k1.Time > kk.Time) do
          begin
          k := kk;
          kk := Chain[per-1].Sucrkind(kk);
          end;
(* and move k up one step: *)
         k := k.Such;
         if k.extreme then
(* this is possible if k1 = Last and k1.Predh is of this rkind;
   then an insertion of a CCP is impossible. *)
           begin
           inc(nimpinsdels);
           exit;
           end;
(* now k.Time <= k1.Time <= kk.Time; k1 = kk is possible only if k1 = Last.
   choose random k2 in interval from k to kk, including boundaries, k2 <> k1. *)

(*
if (k = kk) then
  begin
  errortext := 'kk = k';
  Reporterror(kk,k);
  writeln(outf,'numav1 = ',numav1,' ');
  writeln(outf,'kno = ',kno);
  goto endofprocedure;
  end;
*)
        knum := k;
        numav2 := 1;
(* numstep will be the index number of k1 in the interval from k to kk;
   numav2 will be the total length of the interval minus 1 *)
        numstep := 0;
(*
if (kk.Time < knum.time) then
  begin
  errortext := 'time reversal';
  goto endofprocedure;
  end;
*)
        while knum <> kk do
          begin
            if (knum = k1) then numstep := numav2;
            (* this is true once if kk <> k1, and else never *)
          knum := knum.Such;
          inc(numav2);
          end;
        dec(numav2);
        if (kk = k1) then numstep := numav2+1;

(* random choice for k2: *)
        kno := 1+randomd(version,numav2);
(* define k2 as the ministep in the interval from k to kk with index number kno;
   but if this is k1, which has index number numstep, then let k2 := kk *)
        if (kno = numstep) then k2 := kk
         else
          begin  // dit kan iets efficienter, terugtellen bij kno > numav2/2. @@@
          knum := k;
          numstep := 1;
          while (kno > numstep) do
            begin
            knum := knum.Such;
            inc(numstep);
            end;
          k2 := knum;
          end;

(*
if (k2.Time > kk.Time + 1e-10) then
  begin
  errortext := 'k2 too late';
  goto endofprocedure;
  end;
*)
        end
       else (* there are no ministeps of this rkind *)
        begin
        k1 := Chain[per-1].RandomElement(version);
        numav1 := ctn;
        repeat k2 := Chain[per-1].RandomElement(version)
         until (k1 <> k2);
        numav2 := ctn-1;
        end;
      if (k1.Time > k2.Time) then  (* swap *)
        begin
        kk := k1;
        k1 := k2;
        k2 := kk;
        end;
(* Now k1 and k2 have been chosen for the ins case*)
      end
     else (* not ins: del *)
      begin
      k1 := Chain[per-1].RandomCCPrkind(version,V0,V1,V2,V3);
      k2 := k1.Suchrkind;
      k := Chain[per-1].Predrkind(k1);
      kk := Chain[per-1].Sucrkind(k2);

(* determine numav1 and numav2,
   the number of available possibilities for k1 and k2
   in the case of the inverse ins step *)
      numav1 := ctn - Chain[per-1].Numberrkind(V0,V1,V2,V3,1);
      if (k = Chain[per-1].First) and (kk = Chain[per-1].Last)
       then numav2 := numav1 - 1
       else
        begin
        knum := k.Such;
        numav2 := -2;
        repeat
          inc(numav2);
          knum := knum.Such;
         until (knum = kk);
        end;

(*
if (numav2 <= 0) then
  begin
  errortext := 'numav2 <= 0';
  Reporterror(k1,k2);
  goto endofprocedure;
  end;
*)

      end; (* del *)
    end (* if insdel *)
   else
    begin
    if (num <= 1) then exit;
    ins := false;
    k2 := Chain[per-1].Last;
    repeat k1 := Chain[per-1].RandomElement(version)
      until k1.Such <> k2;
    end;

(*
if (k2 = k1) then
  begin
  errortext := 'k1 = k2';
  writeln(outf,'numav = ',numav1,' ',numav2);
  writeln(outf,'kno = ',kno);
  ReportError(k1,k2);
  goto endofprocedure;
  end;
*)


  k1check := k1.predh;
  k2check := k2;
  if insdel and (not ins) then k2check := k2.Such;
  v4check := k1.difh;

(*
if madechain then
begin
if insdel then begin
  write(outf,'insdel ');
  if ins then write(outf,' ins ') else write(outf,' del ');
  writeln(outf,' ',kindsstring(V0,V1,V2,V3,1));
  end else write(outf,' permutation only ');
end;
*)

  k0 := k1.Predh;
  SetData(k0,per,outf,lf);
  slp := 0.0;
  smu := 0.0;
  smu2 := 0.0;

  if insdel then
    begin
    pr := 1.0;
    if (V0 = Behavior) then
      begin
      minpossible  := (not Minimalya(V1,V3));
      pluspossible := (not Maximalya(V1,V3));
      end;
    if ins then
      begin
(* Carry out the insertion *)
      if (V0 = Behavior) then
        begin
        if (minpossible and pluspossible) then
          begin
          pr := 0.5;
          if (rando(version) < 0.5) then V4 := -1 else V4 := 1;
          end
         else (* exactly one of minpossible and pluspossible must be true *)
          begin
          if pluspossible then V4 := 1 else V4 := -1;
          end;
        end;
      CalcRatelprob(nwrate1,nwlprob1,V0,V1,V2,V3,V4,outf,lf);
      slp := slp + nwlprob1;
//abc if madechain then writeln(outf,'slp nwrate1 = ',slp:12:6);
//writeln(outf,'nwrate1 = ',nwrate1:12:5);

      smu := smu + 1.0/nwrate1;
      smu2 := smu2 + 1.0/sqr(nwrate1);
      if (V0 = Behavior) then ChangeBehavior(V1,V3,V4) else ChangeTie(V1,V2);
      end
     else
      begin (* insdel and (not ins): delete *)
      if ((V0 = Behavior) and minpossible and pluspossible) then pr := 0.5;
(* k1 is the first to-be-deleted ministep: *)
        if (k1.rate <= 0.0) then begin
          x := 'error in MH_InsDelPermute: k1 = ';
          outf.Write(pchar(x)^,length(x));
          x := k1.kindstring+eol;
          outf.Write(pchar(x)^,length(x));
          errortext := 'k1.rate = '+floattostr(k1.rate);
          Reporterror(k1,k2);
          goto endofprocedure;
          end; 
      slp := - k1.lprob;
      smu := -1/k1.rate;
      smu2 := -1/sqr(k1.rate);
      k1 := k1.Such;
      end;
    end; (* if insdel *)

(* Now the data is set for (not insdel) as SetData(k1.Predh,per,outf,lf),
   for ins as SetData(k1.Predh,per,outf,lf) with additionl insertion,
   and for del as SetData(k1.Predh.Predh,per,outf,lf) *)

(* Determine a random permutation of an interval of num ministeps
  starting with k1, and not including k2 or later ministeps,
  under the condition that it is permitted ("OK").
  Note that for nza <= 0, all permutations are permitted.
  The permutation and associated probabilities and rates
  are stored in the arrays aa1, ii1, jj1, vv1, dd1, lp, ra.
  if there are not enough ministeps available from k1 to k2,
  then num is truncated.  *)

  numstep := 1;
  k0 := k1;

(*
if (k2 = k1) then
  begin
  errortext := 'k1 = k2 (2nd)';
  writeln(outf,'numav = ',numav1,' ',numav2);
  writeln(outf,'kno = ',kno);
  ReportError(k1,k2);
  goto endofprocedure;
  end;
*)

  if (num >= 2) then
    begin
(* determine number of ministeps numstep which is the minimum of num
   and the number of ministeps from k1 until and excluding k2;
   truncate num if necessary.
   define knum as the last ministep of the permuted interval;
   knum has number num, when counting from k1. *)
    stopit := false;
     repeat
      k0 := k0.Such;
      if (k0 = k2) then stopit := true else inc(numstep);
     until (stopit or (numstep >= num));
    if (numstep < num) then
      begin
      num := numstep;
      knum := k0.Predh;
      truncated := true;
      extrasteps := false;
      end
     else
      begin
      truncated := false;
      knum := k0;
      knums := k0.Such;
      extrasteps := (knums <> k2);
      end;
(* extrasteps indicates that there are ministeps between knum and k2;
   then the first of these is knums = knum.such *)

(*
 if madechain then
  begin
  if ins then writeln(outf,'ins') else
    begin
    if insdel then writeln(outf,'del') else writeln(outf,'perm only');
    end;
  if truncated then writeln(outf,'truncated') else writeln(outf,'not truncated');
  writeln(outf,'num = ',num);
   if (V1 = 1) and (V2 = 22) then
    begin
    (k1.Predh).WriteKind(outf); writeln(outf);
    (k2).WriteKind(outf); writeln(outf);
    write(outf,'knum = '); knum.writekind(outf); writeln(outf);
    writeperiod(per,outf,lf);
    end;
  end;
*)

    if (not (truncated or insdel)) then
      begin
      if (num < maxpermute) then inc(ntrypermutes[num]) else inc(ntrypermutes[pred(maxpermute)]) ;
      end;

(* define the sequence pstep of ministeps that is to be permuted *)
    if (num >= 2) then begin
      SetLength(pstep,num+1);
      repeat
      i := 0;
      k := k1.Predh;
        repeat
        inc(i);
        k := k.Such;
        pstep[i] := k;
        until (i >= num);

(* determine a permitted permutation *)
      OK := true;
      hh := num + 1;
      ist := 0;
        repeat
        inc(ist);
   (* Generate the new ist'th ministep denoted by k.
   The condition OK is to ensure that the step is allowed. *)
        dec(hh);
        if (ist < num) then
// not for checking @@@
          h := ist + randomd(version,hh)         (* random value from ist to num *)
// h := ist  // identity
// h := num+1-ist (* reversal *)
         else
          h := num;
          k := pstep[h];
        if (nza >= 1) then
          begin
          a := k.Ash;
          if (a = Network) then
            begin
            i := k.ih;
            j := k.jh;
            end
           else
            begin
            i := k.ih;
            v := k.varh;
            d := k.Difh;
            if ((d <= -1) and Minimalya(i,v)) or
                  ((d >= 1) and Maximalya(i,v))
              then OK := false;
            end;
          end;

        if OK then
          begin
          pstep[h] := pstep[ist];
          pstep[ist] := k;
          if (nza >=1) then
            begin
            if (a = Network) then ChangeTie(i,j) else ChangeBehavior(i,v,d);
            end;
          end
         else inc(nnotok);
        until (ist >= num) or (not OK);
//until (2*ist >= num-2) or (not OK); // reversal
//if not ok then begin
//  if madechain then writeln(outf,'no luck');
// // exit;
//  end;

(* restore data; see below for alternative *)
       if OK then h := succ(num) else h := ist;
      if (nza >= 1) then while (h >= 2) do
          begin
          dec(h);
          k := pstep[h];
          a := k.Ash;
          i := k.ih;
          j := k.jh;
          v := k.varh;
          d := k.Difh;
          if (a = Network) then ChangeTie(i,j) else ChangeBehavior(i,v,-d);
          end;


// equivalent alternative :
(* if insdel then
   begin
   if ins then
     begin
    SetData(k1.Predh,per,outf,lf);
    if (V0 = Behavior) then ChangeBehavior(V1,V3,V4) else ChangeTie(V1,V2);
     end else
     SetData(k1.Predh.Predh,per,outf,lf);
   end else
     SetData(k1.Predh,per,outf,lf);
*)

       if (not OK) then
(* This is one possibility for avoiding having to look too long for a permitted permutation.
   Other, more efficient, possibilities may exist. @@@ *)
         begin
         if (num >= 11) then
           begin
           num := num-5;
           if (num <= 9) then num := 10;
           extrasteps := true;
           truncated := true;
           k := k1;
           h := 1;
           while (h < num) do
             begin
             inc(h);
             k := k.Such;
             end;
           knum := k;
           knums := knum.Such;
//abc if madechain then writeln(outf,'not OK; num reduced to ',num);
           end;
         end;
     until OK;

(* Now a permitted permutation has been drawn. *)

(* calculate probabilities and rates for this permutation *)

      if (num+1 > maxchainlength) then
        begin
        maxchainlength := num+5;
        SetLength(lp,maxchainlength);
        SetLength(ra,maxchainlength);
        SetLength(aa1,maxchainlength);
        SetLength(ii1,maxchainlength);
        SetLength(jj1,maxchainlength);
        SetLength(vv1,maxchainlength);
        SetLength(dd1,maxchainlength);
        end;

(* Calculate probabilities under this sequence. *)
      k := k1.Predh;
      for ist :=1 to num do
        begin
        k := k.Such;
        kk := pstep[ist];
        a := kk.Ash;
        i := kk.ih;
        j := kk.jh;
        v := kk.varh;
        d := kk.Difh;
        CalcRatelprob(thisrate,thislprob,a,i,j,v,d,outf,lf);
        if (errortext > '') then
          begin
         x := 'Error in CalcRatelProb 1: '+errortext+eol;
          outf.Write(pchar(x)^,length(x));
          goto EndofProcedure;
          end;
        aa1[ist] := a;
        ii1[ist] := i;
        jj1[ist] := j;
        vv1[ist] := v;
        dd1[ist] := d;
        lp[ist] := thislprob;
        ra[ist] := thisrate;
        slp := slp  - k.lprob + thislprob;
(*if madechain  then
begin
if (ist = 1) then writeln(outf,'num = ',num);
//writeln(outf,k.kindex:4,' ',kk.kindex:4);
writeln(outf,'slp[',ist:2,'] = ',slp:12:6);
if (not insdel) and (k.lprob <> thislprob) then begin
  writeln(outf,k.lprob:12:8,' ',thislprob:12:8); k.WriteKind(outf); writeln(outf);
  writeln(outf,kindsstring(a,i,j,v,d));
  end;
end;
*)
        smu := smu - (1.0/k.rate) + (1.0/thisrate);
        smu2 := smu2 - (1.0/sqr(k.rate)) + (1.0/sqr(thisrate));
        if (a = Network) then ChangeTie(i,j) else ChangeBehavior(i,v,d);
        end;
      if (errortext > '') then goto EndOfProcedure;
      end; (* if num >= 2 *)
    end; (* if num >= 2 *)

  if (num <= 1) then
    begin
    extrasteps := true;
    end;

(*
if madechain then begin
if extrasteps then writeln(outf,'extrasteps') else writeln(outf,'no extrasteps');
end;
*)

  if insdel then
(* also the probabilities of further ministeps until k2 are changed;
   note that (num <= 1) implies insdel.  *)
    begin
    if extrasteps then
      begin
      if (num >= 2) then kk := knum.Such else kk := k1;

(*
if (kk.Time > k2.Time) then begin
  errortext := 'insdel, before the permutation, wrong order:';
  write('knum = '); knum.WriteKind(outf); writeln(outf);
  ReportError(kk,k2);
  goto endofprocedure;
  end;
*)

      while (kk <> k2) do
        begin
        a := kk.Ash;
        if (a = Network) then
          begin
          i := kk.ih;
          j := kk.jh;
          CalcRatelprob(thisrate,thislprob,Network,i,j,0,0,outf,lf);
          end
         else
          begin
          i := kk.ih;
          v := kk.varh;
          d := kk.Difh;
          CalcRatelprob(thisrate,thislprob,Behavior,i,1,v,d,outf,lf);
          end;
        if (errortext > '') then
          begin
          x := 'Error in CalcRatelProb 2: '+errortext+eol;
          outf.Write(pchar(x)^,length(x));
          goto EndofProcedure;
          end;
        kk.newrate := thisrate;
        kk.newlprob := thislprob;
        slp := slp  - kk.lprob + thislprob;
//if madechain  then writeln(outf,'slp newrate = ',slp:12:6);
        if (kk.rate <= 0.0) then begin
          x := 'error in MH_InsDelPermute: kk = ';
          outf.Write(pchar(x)^,length(x));
          x := kk.kindstring+eol;
          outf.Write(pchar(x)^,length(x));
          errortext := 'kk.rate = '+floattostr(kk.rate);
          outf.write(pchar(string('k0 = '))^,5); k0.WriteKind(outf); outf.Write(pchar(eol)^,length(eol));//writeln(outf);
          Reporterror(k1,k2);
          goto endofprocedure;
          end;
        smu := smu - (1.0/kk.rate) + (1.0/thisrate);
        smu2 := smu2 - (1.0/sqr(kk.rate)) + (1.0/sqr(thisrate));
        if (a = Network) then ChangeTie(i,j) else ChangeBehavior(i,v,d);
        kk := kk.Such;
        end; (* while kk <> k2 *)
      end; (* if extrasteps *)
    if (errortext > '') then goto EndOfProcedure;

    if ins then
      begin
      CalcRatelprob(nwrate2,nwlprob2,V0,V1,V2,V3,-V4,outf,lf);
//abc if madechain then writeln(outf,'slp nwlprob2 = ',nwlprob2:12:6);
      slp := slp + nwlprob2;
//abc if madechain  then writeln(outf,'slp nwrate2 = ',slp:12:6);
//writeln(outf,'nwrate2 = ',nwrate2:12:5);
      smu := smu + 1.0/nwrate2;
      smu2 := smu2 + 1.0/sqr(nwrate2);
(* To have an easy calculation for the new number of CCPs of this rkind,
   necessary for the proposal probability of the inverse proposal,
   the insertions are carried out already now: *)
      Insert(k1,V0,V1,V2,V3,V4,per,outf,lf);
      Insert(k2,V0,V1,V2,V3,-V4,per,outf,lf);
      end
     else (* del *)
      begin
      slp := slp - k2.lprob;
//abc if madechain  then writeln(outf,'slp k2 = ',slp:12:6);
      smu := smu - 1.0/k2.rate;
      smu2 := smu2 - 1.0/sqr(k2.rate);
(* To have an easy calculation for the new number of CCPs of this rkind,
   necessary for the proposal probability of the inverse proposal,
   the deletions are carried out already now: *)
      k0 := k1.Predh;
      V4 := k0.Difh;
      k11 := k1;
      nwrate1 := k0.rate;
      nwlprob1 := k0.lprob;
      Cancel(k0,per,outf,lf);
      k0 := k2.Such;
      k22 := k2.Such;
      nwrate2 := k2.rate;
      nwlprob2 := k2.lprob;
      Cancel(k2,per,outf,lf);
      end;
    ncnew := Chain[per-1].NCCPOfRkind(V0,V1,V2,V3,V4);
    end; (* if insdel *)
  if (errortext > '') then begin
    x := 'here an error : '+eol;
    outf.Write(pchar(x)^,length(x));
    goto EndOfProcedure;
    end;


(* now slp, smu and smu2 are the differences in sumlpr, mu, and sigma2 *)
   mu_new := Chain[per-1].mu + smu;
   sigma2_new := Chain[per-1].sigma2 + smu2;

//writeln(lf,slp:10:4,' ',mu_new:10:4,' ',sigma2_new:10:4);
(* log ratio of probabilities under target distribution: *)
   prb := slp - Chain[per-1].kappa
             - 0.5*log(sigma2_new)
                - (sqr(Chain[per-1].totaltime - mu_new)/(2.0*sigma2_new));
   prb := expo(prb);
//q1 := prb;

(*
if madechain then begin
write(outf,'; nk = ',nk);
write(outf,'; nc = ',nc);
write(outf,'; ncnew = ',ncnew);
writeln(outf,';  nmuls = ', nmuls);
end;
*)

   if insdel then
     begin
     if (V0 = Network) then  pa := (1.0-promul)*prnet/NStructActiveElements
       else pa := (1.0-promul)*(1.0-prnet)/(n_act*nza);
     if ins then
       begin
(* for ratio of propoposal probabilities:
   note that two insertions took place *)
       if (nk <= 1) then
          pp := (((promul/(nmuls+1)) + pa)*(1.0-prelins))/pa
         else
          pp := 1.0-prelins;
       if (nc >= 1) then pp := pp/prelins;
//if madechain then write(outf,'ins A. pp = ',pp:10:6,' prb = ',log(prb):10:6);
       prb := (prb*pp*numav1*numav2)/(2.0*pr*ncnew);
//qq1 := (pp*numav1*numav2)/(2.0*pr*ncnew);
//ppr1 := pr;
       end
      else (* del *)
       begin
       if (nk <= 3) then
         pp := pa/(((promul/nmuls) + pa)*(1.0-prelins))
        else pp := 1.0/(1.0-prelins);
       if (ncnew >= 1) then pp := pp*prelins;
//if madechain then write(outf,'del A. pp = ',pp:10:6,' prb = ',log(prb):10:6);
       prb := (prb*pp*2.0*pr*nc)/(numav1*numav2);
//qq1 :=  (pp*2.0*pr*nc)/(numav1*numav2);
//ppr1 := pr;
       end;
//if madechain then writeln(outf,' numav1 = ',numav1,'; numav2 = ',numav2)
     end; (* if insdel *)

   Result := prb;
//write(outf,'prob = ',prob:12:4);
//   if (prob >= 0.0) then accept:= true
  //  else accept := (rando(version) < expo(prob));
  accept := (rando(version) < prb);

// if madechain then writeln(outf,prb:8:4,' ',accept);
//writeln(outf,'k2.predh = ',k2.predh.kindstring);
//writeln(outf,nwrate2:10:4,nwlprob2:10:4);

  newpath := accept; (*MS: monitor acceptance rate of candidate paths.*)

  if accept then
    begin
    if insdel then
      begin
      if ins then
        begin
        inc(ninserts);
(* the insertions before k1 and k2 were already carried out *)
        (k1.Predh).rate := nwrate1;
        (k1.Predh).lprob := nwlprob1;
        (k2.Predh).rate := nwrate2;
        (k2.Predh).lprob := nwlprob2;
        k0 := k2.Predh;
// writeln(outf,'k2.predh = ',(k2.predh).kindstring,nwrate2:10:4,nwlprob2:10:4);
        end
       else
        begin
        inc(ncancels);
        end;
      end
     else (* not insdel *)
      begin
      if (not truncated) then
        begin
        inc(npermtots);
        if (num < maxpermute) then inc(npermutes[num]) else inc(npermutes[pred(maxpermute)]) ;
        end;
      end;
//writeln(outf,' ; accepted ');
(* Carry out the permutation of the ministeps:
   delete old ministeps, insert new ones  *)

    if (num >= 2) then
      begin
      kist := k1;
      for ist := 1 to num do
        begin
        kk := kist.Such;
        Chain[per-1].Delete(kist);
        kist := kk;
        end;
(* Now knum also has been deleted.
   kk now is the first ministep after the deleted interval.
   k1 and knum have been deleted. Of the new num ministeps created below,
   the first will be called k1 and the last knum. *)
      Chain[per-1].InsertBefore(kk,aa1[1],ii1[1],jj1[1],vv1[1],dd1[1]);
      k1 := kk.Predh;
      k1.lprob := lp[1];
      k1.rate := ra[1];
      for ist := 2 to num do
        begin
        Chain[per-1].InsertBefore(kk,aa1[ist],ii1[ist],jj1[ist],vv1[ist],dd1[ist]);
        kist := kk.Predh;
        kist.lprob := lp[ist];
        kist.rate := ra[ist];
        end;
(* The new ministep that is the last of the permuted interval: *)
      knum := kist;
      end;

//if insdel and (knum = k0) then
//   begin
//  writeln(outf,'num = ',num);
//  write(outf,'knum = '); knum.WriteKind(outf); writeln(outf);
//  write(outf,'k0 = '); k0.WriteKind(outf); writeln(outf);
//  writeperiod(per,outf,lf);
//   end;

    if insdel and extrasteps then
      begin
      if (num >= 2) then kk := knums else kk := k1;

(*
if (kk.Time > k0.Time) then begin
  writeln(outf,' insdel, after the permutation, wrong order:');
  write(outf,'knums = '); knums.WriteKind(outf); writeln(outf);
  write(outf,'k1 = '); k1.WriteKind(outf); writeln(outf);
  if ins then write(outf,'k2 = '); k2.WriteKind(outf); writeln(outf);
  ReportError(kk,k0);
  goto endofprocedure;
  end;
*)

      while (kk <> k0) do
        begin
//if (V1 = 1) and (V2 = 22) then writeln(outf,'kk.newlprob = ',kk.newlprob:12:6);        
        kk.rate := kk.newrate;
        kk.lprob := kk.newlprob;
        kk := kk.Such;
        end;
      end;

    Chain[per-1].mu := mu_new;
    Chain[per-1].sigma2 := sigma2_new;
    Chain[per-1].kappa := -0.5*log(sigma2_new) -
           (sqr(Chain[per-1].totaltime - mu_new)/(2.0*sigma2_new));
    if (not (truncated or insdel)) then numm[per] := numm[per] + 0.5;
    end
   else (* not accept *)
    begin
    if insdel then
      begin
      if ins then
        begin
        inc(nnoinserts);
(* The insertions before k1 and k2 were already carried out,
   and now must be canceled again. *)
        k0 := k1.Predh;
        Cancel(k0,per,outf,lf);
        k0 := k2.Predh;
        Cancel(k0,per,outf,lf);
        end
       else
        begin
(* The cancels before k11 and k22 were already carried out,
   and now must be restored. *)
        inc(nnocancels);
        Insert(k11,V0,V1,V2,V3,V4,per,outf,lf);
        k11.Predh.rate := nwrate1;
        k11.Predh.lprob := nwlprob1;
        Insert(k22,V0,V1,V2,V3,-V4,per,outf,lf);
        k22.Predh.rate := nwrate2;
        k22.Predh.lprob := nwlprob2;
        end;
      end
     else
      begin
      inc(nnopermtots);
      if (num < maxpermute) then inc(nnopermutes[num]) else inc(nnopermutes[pred(maxpermute)]) ;
      end;
//writeln(outf,' ; not accepted ');
    if (not (truncated or insdel)) then numm[per] := numm[per] - 0.5;
    end;

(* @@@@ only for checking: *)
(*  if madechain then begin
  writeln(outf,' end 1 ');
  renewrateslprobs(Chain[per-1].First, Chain[per-1].Last,true,per,outf,lf);
  end;
*)

{  exit;

(*
kk := Chain[per-1].First.such;
k2 := Chain[per-1].Last;
while (kk <> k2) do begin
  if (kk.rate <= 0.0) then
        begin
        write(outf,'final check: error in MH_InsDelPermute: kk = ');
        writeln(outf,kk.kindstring);
        errortext := 'kk.rate = '+floattostr(kk.rate);
        write(outf,'k0 = '); k0.WriteKind(outf); writeln(outf);
        write(outf,'knum = '); knum.WriteKind(outf); writeln(outf);
        write(outf,'knums = '); knums.WriteKind(outf); writeln(outf);
        for ist:=1 to num do writeln(outf,' ist , ra[ist] ',ist:4,' ',
                   ii1[ist]:3,jj1[ist]:3,ra[ist]:10:4);
        ReportError(k1,k2);
        goto endofprocedure;
        end;
  kk := kk.Such;
  end;
*)


(* For checking: the remainder of this procedure MH_InsDelPermute until EndOfProcedure
   must be bracketed away for normal operations.
   Checks only for identity or reverse pemutations.
   Use num as above; invert del-ins
    *)

//exit;

if madechain then writeln(outf);
  if not (madechain and accept) then

   begin  exit; end;
  if (rando(version) < 0.5) then
   begin writeln(outf); exit; end;
  prob1 := prb;
  ctn := chain[per-1].TotNumber;
  V4 := 0;
  del := not ins;

(* Choose between insertion and deletion.
   Determine the first element k1 (and if insdel then also the last element k2)
   of the interval of ministeps that will be considered.
   For insert, a canceling pair of ministeps will be inserted before k1 and k2;
   for delete, k1.predh and k2 will be deleted.
    *)
  if insdel then
    begin
    nc := Chain[per-1].NCCPOfRkind(V0,V1,V2,V3,1);
    nk := Chain[per-1].Numberrkind(V0,V1,V2,V3,1);  // nc and nk could be calculated more efficiently together
    nmuls := Chain[per-1].MultipleSNumber;
//if ins then writeln(outf,'ins') else writeln(outf,'del');
    if del then
      begin
(* choose random k1 and k2 in the same "interval".
   numav1, numav2 (av for available) are
   the numbers of possibilities for k1, k2. *)
      k1 := k1check.such;
      k2 := k2check;

//if madechain then begin
//k1.WriteKind(outf);k2.WriteKind(outf);writeln(outf);
//end;
      if assigned(Chain[per-1].FirstOfRkind(V0,V1,V2,V3,1)) then
        begin
        numav1 := ctn - Chain[per-1].NumberRkind(V0,V1,V2,V3,1);
(* determine k and kk which are the last ministep before k1,
   and the first after k1, of this rkind *)
        k := Chain[per-1].First;
        kk := Chain[per-1].FirstOfRkind(V0,V1,V2,V3,1);
        while (k1.Time > kk.Time) do
          begin
          k := kk;
          kk := Chain[per-1].Sucrkind(kk);
          end;
         k := k.Such;
         if k.extreme then
(* this is possible if k1 = Last and k1.Predh is of this rkind;
   then an insertion of a CCP is impossible. *)
           begin
           inc(nimpinsdels);
           exit;
           end;
(* now k.Time <= k1.Time <= kk.Time; k1 = kk is possible only if k1 = Last.
   choose random k2 in interval from k to kk, including boundaries, k2 <> k1. *)

if (k = kk) then
  begin
  errortext := 'check; kk = k';
  Reporterror(kk,k);
  writeln(outf,'numav1 = ',numav1,' ');
  writeln(outf,'kno = ',kno);
  goto endofprocedure;
  end;

        knum := k;
        numav2 := 1;
(* numstep will be the index number of k1 in the interval from k to kk;
   numav2 will be the total length of the interval minus 1 *)
        numstep := 0;
if (kk.Time < knum.time) then
  begin
  errortext := 'check; time reversal';
  goto endofprocedure;
  end;
        while knum <> kk do
          begin
          knum := knum.Such;
          inc(numav2);
          end;
        dec(numav2);
(* The following if - then - else option gives two branches
   that yields equivalent results; the choice is for efficiency purposes.
   The multiplicative number in the next condition must be
   close to ratio (time for "randomelement") to (time for "knum:=knum.Such") *)
        end
       else (* there are no ministeps of this rkind *)
        begin
        numav1 := ctn;
        numav2 := ctn-1;
        end;
      if (k1.Time > k2.Time) then  (* swap *)
        begin
        kk := k1;
        k1 := k2;
        k2 := kk;
        end;
(* Now k1 and k2 have been chosen for the ins case*)
      end
     else (* not ins: del *)
      begin
      k1 := k1check.such;
      k2 := k2check.Predh;
if madechain then begin
writeln(outf,'c  del:: ');
k1.WriteKind(outf);writeln(outf,'  '); k2.WriteKind(outf);writeln(outf);
end;
      k := Chain[per-1].Predrkind(k1);
      kk := Chain[per-1].Sucrkind(k2);

(* determine numav1 and numav2,
   the number of available possibilities for k1 and k2
   in the case of the inverse ins step *)
      numav1 := ctn - Chain[per-1].Numberrkind(V0,V1,V2,V3,1);
      if (k = Chain[per-1].First) and (kk = Chain[per-1].Last)
       then numav2 := numav1 - 1
       else
        begin
        knum := k.Such;
        numav2 := -2;
        repeat
          inc(numav2);
          knum := knum.Such;
         until (knum = kk);
        end;
if (numav2 <= 0) then
  begin
  errortext := 'check numav2 <= 0';
  Reporterror(k1,k2);
  goto endofprocedure;
  end;
      end; (* del *)
    end (* if insdel *)
   else
    begin
    k1 := k1check.such;
    k2 := k2check;
    end;

if (k2 = k1) then
  begin
  errortext := 'check k1 = k2';
  writeln(outf,'numav = ',numav1,' ',numav2);
  writeln(outf,'kno = ',kno);
  ReportError(k1,k2);
  goto endofprocedure;
  end;

(*
if madechain then begin
  writeln(outf,'k1 = ',k1.kindstring);
  writeln(outf,'k2 = ',k2.kindstring);
end;
*)

  k0 := k1.Predh;
  SetData(k0,per,outf,lf);
  slp := 0.0;
  smu := 0.0;
  smu2 := 0.0;

  if insdel then
    begin
    pr := 1.0;
    if (V0 = Behavior) then
      begin
      minpossible  := (not Minimalya(V1,V3));
      pluspossible := (not Maximalya(V1,V3));
      end;
    if del then
      begin
(* Carry out the insertion *)
      if (V0 = Behavior) then
        begin
        if (minpossible and pluspossible) then
          begin
          pr := 0.5;
          V4 := V4check;
//          if (rando(version) < 0.5) then V4 := -1 else V4 := 1;
          end
         else (* exactly one of minpossible and pluspossible must be true *)
          begin
          if pluspossible then V4 := 1 else V4 := -1;
          end;
        end;
        CalcRatelprob(nwrate1,nwlprob1,V0,V1,V2,V3,V4,outf,lf);
        slp := slp + nwlprob1;
//abc if madechain  then writeln(outf,'c slp nwrate1 = ',slp:12:6);
        smu := smu + 1.0/nwrate1;
        smu2 := smu2 + 1.0/sqr(nwrate1);
        if (V0 = Behavior) then ChangeBehavior(V1,V3,V4) else ChangeTie(V1,V2);
      end
     else
      begin (* insdel and (not ins): delete *)
      if ((V0 = Behavior) and minpossible and pluspossible) then pr := 0.5;
(* k1 is the first to-be-deleted ministep: *)
      slp := - k1.lprob;
//abc if madechain  then writeln(outf,'c slp ins = ',slp:12:6);
      smu := -1/k1.rate;
      smu2 := -1/sqr(k1.rate);
      k1 := k1.Such;
      end;
    end; (* if insdel *)

(* Now the data is set as SetData(k1.Predh,per,outf,lf)
   and for del as SetData(k1.Predh.Predh,per,outf,lf) *)

(* Determine a random permutation of an interval of num ministeps
  starting with k1, and not including k2 or later ministeps,
  under the condition that it is permitted ("OK").
  Note that for nza <= 0, all permutations are permitted.
  The permutation and associated probabilities and rates
  are stored in the arrays aa1, ii1, jj1, vv1, dd1, lp, ra.
  if there are not enough ministeps available from k1 to k2,
  then num is truncated.  *)

  numstep := 1;
  k0 := k1;
if (k2 = k1) then
  begin
  errortext := 'check k1 = k2 (2nd)';
  ReportError(k1,k2);
  goto endofprocedure;
  end;

  if (num >= 2) then
    begin
(* determine number of ministeps numstep which is the minimum of num
   and the number of ministeps from k1 until and excluding k2;
   truncate num if necessary.
   define knum as the last ministep of the permuted interval;
   knum has number num, when counting from k1. *)
    stopit := false;
     repeat
      k0 := k0.Such;
      if (k0 = k2) then stopit := true else inc(numstep);
     until (stopit or (numstep >= num));
    if (numstep < num) then
      begin
      num := numstep;
      knum := k0.Predh;
      truncated := true;
      extrasteps := false;
      errortext := 'check: extra truncation';
      ReportError(k1,k2);
      goto EndOfProcedure;
      end
     else
      begin
      truncated := false;
      knum := k0;
      knums := k0.Such;
      extrasteps := (knums <> k2);
      end;
(* extrasteps indicates that there are ministeps between knum and k2;
   then the first of these is knums = knum.such *)
if madechain then begin
if extrasteps then writeln(outf,'c extrasteps') else writeln(outf,'c no extrasteps');
end;

(* define the sequence pstep of ministeps that is to be permuted *)
    if (num >= 2) then begin
      SetLength(pstep,num+1);
      i := 0;
      k := k1.Predh;
        repeat
        inc(i);
        k := k.Such;
        pstep[i] := k;
        until (i >= num);

(* determine a permitted permutation *)
(* identity *)

(* reversal *)
      ist := 0;
        repeat
        inc(ist);
   (* Generate the new ist'th ministep denoted by k.
   The condition OK is to ensure that the step is allowed. *)
        if (ist < num) then
// not for checking @@@
//          h := ist + randomd(version,hh)         (* random value from ist to num *)
// h := ist  // identity
 h := num+1-ist (* reversal *)
         else
          h := num;
          k := pstep[h];
          pstep[h] := pstep[ist];
          pstep[ist] := k;
until (2*ist >= num-2) or (not OK); // reversal


(* restore data *)
 if insdel then
   begin
   if del then
     begin
    SetData(k1.Predh,per,outf,lf);
    if (V0 = Behavior) then ChangeBehavior(V1,V3,V4) else ChangeTie(V1,V2);
     end else
     SetData(k1.Predh.Predh,per,outf,lf);
   end else
     SetData(k1.Predh,per,outf,lf);


(* Calculate probabilities under this sequence. *)
      k := k1.Predh;
      hh := num+1;
      for ist :=1 to num do
        begin
        dec(hh);
        k := k.Such;
        kk := pstep[ist];
        a := kk.Ash;
        i := kk.ih;
        j := kk.jh;
        v := kk.varh;
        d := kk.Difh;
        CalcRatelprob(thisrate,thislprob,a,i,j,v,d,outf,lf);
        if (errortext > '') then
          begin
          writeln(outf,'check Error in CalcRatelProb 1: ',errortext);
          goto EndofProcedure;
          end;
        aa1[ist] := a;
        ii1[ist] := i;
        jj1[ist] := j;
        vv1[ist] := v;
        dd1[ist] := d;
        lp[ist] := thislprob;
        ra[ist] := thisrate;
        slp := slp  - k.lprob + thislprob;
//abc if madechain  then writeln(outf,'c slp[',ist,'] = ',slp:12:6);
        smu := smu - (1.0/k.rate) + (1.0/thisrate);
        smu2 := smu2 - (1.0/sqr(k.rate)) + (1.0/sqr(thisrate));
        if (a = Network) then ChangeTie(i,j) else ChangeBehavior(i,v,d);
        end;
      if (errortext > '') then goto EndOfProcedure;
      end;
    end; (* if num >= 2 *)

  if (num <= 1) then
    begin
    extrasteps := true;
    end;

  if insdel then
(* also the probabilities of further ministeps until k2 are changed;
   note that (num <= 1) implies insdel.  *)
    begin
      if extrasteps then
        begin
      if (num >= 2) then kk := knum.Such else kk := k1;
if (kk.Time > k2.Time) then begin
  writeln(outf,' insdel, before the permutation, wrong order:');
  writeln(outf,'num = ',num);
  write('knum = '); knum.WriteKind(outf); writeln(outf);
  write('kk = '); kk.WriteKind(outf); writeln(outf);
  write('k2 = '); k2.WriteKind(outf); writeln(outf);
  writeperiod(per,outf,lf);
  goto endofprocedure;
  end;
      while (kk <> k2) do
        begin
        a := kk.Ash;
        if (a = Network) then
          begin
          i := kk.ih;
          j := kk.jh;
          CalcRatelprob(thisrate,thislprob,Network,i,j,0,0,outf,lf);
          end
         else
          begin
          i := kk.ih;
          v := kk.varh;
          d := kk.Difh;
          CalcRatelprob(thisrate,thislprob,Behavior,i,1,v,d,outf,lf);
          end;
        if (errortext > '') then
          begin
          writeln(outf,'check Error in CalcRatelProb 2: ',errortext);
          goto EndofProcedure;
          end;
        kk.newrate := thisrate;
        kk.newlprob := thislprob;
        slp := slp  - kk.lprob + thislprob;
//abc if madechain  then writeln(outf,'c slp newrate = ',slp:12:6);
        if (kk.rate <= 0.0) then begin
          write(outf,'check error in MH_InsDelPermute: kk = ');
          writeln(outf,kk.kindstring);
          errortext := 'check kk.rate = '+floattostr(kk.rate);
          write(outf,'k0 = '); k0.WriteKind(outf); writeln(outf);
          Reporterror(k1,k2);
          goto endofprocedure;
          end;
        smu := smu - (1.0/kk.rate) + (1.0/thisrate);
        smu2 := smu2 - (1.0/sqr(kk.rate)) + (1.0/sqr(thisrate));
        if (a = Network) then ChangeTie(i,j) else ChangeBehavior(i,v,d);
        kk := kk.Such;
        end; (* while kk <> k2 *)
      end; (* if extrasteps *)
    if (errortext > '') then goto EndOfProcedure;

    if del then
      begin
      CalcRatelprob(nwrate2,nwlprob2,V0,V1,V2,V3,-V4,outf,lf);
      slp := slp + nwlprob2;
//abc if madechain  then writeln(outf,'c slp nwrate2 = ',slp:12:6);
      smu := smu + 1.0/nwrate2;
      smu2 := smu2 + 1.0/sqr(nwrate2);
(* To have an easy calculation for the new number of CCPs of this rkind,
   necessary for the proposal probability of the inverse proposal,
   the insertions are carried out already now: *)

      Insert(k1,V0,V1,V2,V3,V4,per,outf,lf);
      Insert(k2,V0,V1,V2,V3,-V4,per,outf,lf);
      end
     else (* del *)
      begin
      slp := slp - k2.lprob;
//abc if madechain  then writeln(outf,'c slp k2 = ',slp:12:6);
      smu := smu - 1.0/k2.rate;
      smu2 := smu2 - 1.0/sqr(k2.rate);
(* To have an easy calculation for the new number of CCPs of this rkind,
   necessary for the proposal probability of the inverse proposal,
   the deletions are carried out already now: *)
      k0 := k1.Predh;
      V4 := k0.Difh;
      k11 := k1;
      nwrate1 := k0.rate;
      nwlprob1 := k0.lprob;
      Cancel(k0,per,outf,lf);
      k0 := k2.Predh;
      k22 := k2.Such;
      nwrate2 := k2.rate;
      nwlprob2 := k2.lprob;
      Cancel(k2,per,outf,lf);
      end;
    ncnew := Chain[per-1].NCCPOfRkind(V0,V1,V2,V3,V4);
    end; (* if insdel *)
  if (errortext > '') then begin
    writeln(outf,'check here an error : ');
    goto EndOfProcedure;
    end;


(* now slp, smu and smu2 are the differences in sumlpr, mu, and sigma2 *)
   mu_new := Chain[per-1].mu + smu;
   sigma2_new := Chain[per-1].sigma2 + smu2;

//writeln(lf,slp:10:4,' ',mu_new:10:4,' ',sigma2_new:10:4);
(* log ratio of probabilities under target distribution: *)
   prb := slp - Chain[per-1].kappa
             - 0.5*log(sigma2_new)
                - (sqr(Chain[per-1].totaltime - mu_new)/(2.0*sigma2_new));
   prb := expo(prb);

q2 := prb;

   if insdel then
     begin
     if (V0 = Network) then  pa := (1.0-promul)*prnet/NStructActiveElements
       else pa := (1.0-promul)*(1.0-prnet)/(n_act*nza);
     if del then
       begin
(* for ratio of propoposal probabilities:
   note that two insertions took place *)
       if (nk <= 1) then
          pp := (((promul/(nmuls+1)) + pa)*(1.0-prelins))/pa
         else
          pp := 1.0-prelins;
       if (nc >= 1) then pp := pp/prelins;
writeln(outf,'ins; B. pp = ',pp:10:6,' prb = ',log(prb):10:6);
       prb := (prb*pp*numav1*numav2)/(2.0*pr*ncnew);
qq2 := (pp*numav1*numav2)/(2.0*pr*ncnew);
ppr2 := pr;
       end
      else (* del *)
       begin
       if (nk <= 3) then
         pp := pa/(((promul/nmuls) + pa)*(1.0-prelins))
        else pp := 1.0/(1.0-prelins);
       if (ncnew >= 1) then pp := pp*prelins;
writeln(outf,'del; B. pp = ',pp:10:6,' prb = ',log(prb):10:6);
       prb := (prb*pp*2.0*pr*nc)/(numav1*numav2);
qq2 := (pp*2.0*pr*nc)/(numav1*numav2);
ppr2 := pr;
       end;
if madechain then writeln(outf,'c numav1 = ',numav1,'; numav2 = ',numav2)
     end; (* if insdel *)


   if (abs(prob1*prb - 1.0) < 1e-6) then
     begin
     writeln(outf,'prb       = ',prb :12:8);
     writeln(outf,'prob1     = ',prob1:12:8);
     writeln(outf,'prb*prob1 = ',prb*prob1:12:8);
     writeln(outf,'OK')
     end
     else
     begin
     errortext := 'check: wrong product';
     writeln(outf,'wrong product');
     writeln(outf,'prb       = ',prb :12:8);
     writeln(outf,'prob1     = ',prob1:12:8);
     writeln(outf,'prb*prob1 = ',prb*prob1:12:8);
     writeln(outf,'q1*q2 = ',q1*q2:12:8);
     if insdel then begin
       writeln(outf,'qq1*qq2 = ',qq1*qq2:12:8);
       writeln(outf,'qq1 = ',qq1:12:8);
       writeln(outf,'qq2 = ',qq2:12:8);
       writeln(outf,'ppr1 = ',ppr1:12:8);
       writeln(outf,'ppr2 = ',ppr2:12:8);
       end;
     writeln(outf,'nk = ',nk);
     writeln(outf,'nc = ',nc);
     writeln(outf,'ncnew = ',ncnew);
     writeln(outf,';  nmuls = ', nmuls);
     ReportError(k1,k2);
     goto EndOfProcedure;
     end;

  accept := (rando(version) < prb);

//writeln(outf,prb:8:4,' ',accept);
//writeln(outf,'k2.predh = ',k2.predh.kindstring);
//writeln(outf,nwrate2:10:4,nwlprob2:10:4);

  if accept then
    begin
    if insdel then
      begin
      if del then
        begin
        inc(ninserts);
(* the insertions before k1 and k2 were already carried out *)
        (k1.Predh).rate := nwrate1;
        (k1.Predh).lprob := nwlprob1;
        (k2.Predh).rate := nwrate2;
        (k2.Predh).lprob := nwlprob2;
        k0 := k2.Predh;
// writeln(outf,'k2.predh = ',(k2.predh).kindstring,nwrate2:10:4,nwlprob2:10:4);
        end
       else
        begin
        inc(ncancels);
        end;
      end
     else (* not insdel *)
      begin
      if (not truncated) then
        begin
        inc(npermtots);
        if (num < maxpermute) then inc(npermutes[num]) else inc(npermutes[pred(maxpermute)]) ;
        end;
      end;
//writeln(outf,' ; accepted ');
(* Carry out the permutation of the ministeps:
   delete old ministeps, insert new ones  *)

    if (num >= 2) then
      begin
      kist := k1;
      for ist := 1 to num do
        begin
        kk := kist.Such;
        Chain[per-1].Delete(kist);
        kist := kk;
        end;
(* Now knum also has been deleted.
   kk now is the first ministep after the deleted interval.
   k1 and knum have been deleted. Of the new num ministeps created below,
   the first will be called k1 and the last knum. *)
      Chain[per-1].InsertBefore(kk,aa1[1],ii1[1],jj1[1],vv1[1],dd1[1]);
      k1 := kk.Predh;
      k1.lprob := lp[1];
      k1.rate := ra[1];
      for ist := 2 to num do
        begin
        Chain[per-1].InsertBefore(kk,aa1[ist],ii1[ist],jj1[ist],vv1[ist],dd1[ist]);
        kist := kk.Predh;
        kist.lprob := lp[ist];
        kist.rate := ra[ist];
        end;
(* The new ministep that is the last of the permuted interval: *)
      knum := kist;
      end;

//if insdel and (knum = k0) then
//   begin
//  writeln(outf,'num = ',num);
//  write(outf,'knum = '); knum.WriteKind(outf); writeln(outf);
//  write(outf,'k0 = '); k0.WriteKind(outf); writeln(outf);
//  writeperiod(per,outf,lf);
//   end;

    if insdel and extrasteps then
      begin
      if (num >= 2) then kk := knums else kk := k1;
if (kk.Time > k0.Time) then begin
  writeln(outf,' check: insdel, after the permutation, wrong order:');
  write(outf,'knums = '); knums.WriteKind(outf); writeln(outf);
  write(outf,'k1 = '); k1.WriteKind(outf); writeln(outf);
  if ins then write(outf,'k2 = '); k2.WriteKind(outf); writeln(outf);
  ReportError(kk,k0);
  goto endofprocedure;
  end;
      while (kk <> k0) do
        begin
        kk.rate := kk.newrate;
        kk.lprob := kk.newlprob;
        kk := kk.Such;
        end;
      end;

    Chain[per-1].mu := mu_new;
    Chain[per-1].sigma2 := sigma2_new;
    Chain[per-1].kappa := -0.5*log(sigma2_new) -
           (sqr(Chain[per-1].totaltime - mu_new)/(2.0*sigma2_new));
    if (not (truncated or insdel)) then numm[per] := numm[per] + 0.5;
    end
   else (* not accept *)
    begin
    if insdel then
      begin
      if del then
        begin
        inc(nnoinserts);
(* The insertions before k1 and k2 were already carried out,
   and now must be canceled again. *)
        k0 := k1.Predh;
        Cancel(k0,per,outf,lf);
        k0 := k2.Predh;
        Cancel(k0,per,outf,lf);
        end
       else
        begin
(* The cancels before k11 and k22 were already carried out,
   and now must be restored. *)
        inc(nnocancels);
        Insert(k11,V0,V1,V2,V3,V4,per,outf,lf);
        k11.Predh.rate := nwrate1;
        k11.Predh.lprob := nwlprob1;
        Insert(k22,V0,V1,V2,V3,-V4,per,outf,lf);
        k22.Predh.rate := nwrate2;
        k22.Predh.lprob := nwlprob2;
        end;
      end
     else
      begin
      inc(nnopermtots);
      if (num < maxpermute) then inc(nnopermutes[num]) else inc(nnopermutes[pred(maxpermute)]) ;
      end;
//writeln(outf,' ; not accepted ');
    if (not (truncated or insdel)) then numm[per] := numm[per] - 0.5;
    end;


  writeln(outf,' end 2 ');

//check:   renewrateslprobs(Chain[per-1].First, Chain[per-1].Last,true,per,outf,lf);

}

EndOfProcedure:
  if (errortext > '') then
    begin
    x := 'Error in MH_InsDelPermute: '+errortext+eol;
    outf.Write(pchar(x)^,length(x));
    exit;
    end;
end; (* MH_InsDelPermute *)



procedure MLModel.MH_Step(const per: integer; var outf,lf: TFileStream;
                          const stype: integer);
(* Metropolis Hastings step in period per *)
(* stype = 0 gives a random steptype, stype between 1 and 4 gives a specified type of step.
   ctype < 0 gives a random varh, ctype = 0 gives varh = network, ctype > 0 gives varh = ctype. *)  // @@@@
var p,pr: real;
 step,tnm,i,j,hh,h1,h2,ex,b1,b2: integer;
 k0,k1,k2: TMiniStep;
 a: TAspect; v: byte; d: Tdif;
 t0, t1, t2: real;
 between: boolean;
 k22: TMiniStep;
 x: string;
label EndOfProcedure;
begin
  inc(nstep);
if (per >= observations) or (per <= 0) then begin
  x := format('%s%d',['error in MH_Step: per = ',per])+eol;
  outf.Write(pchar(x)^,length(x));
  errortext := 'wrong per ';
  exit;
  end;
  tnm := Chain[per-1].TotNumber;

  if (stype >= 1) then step := stype
   else
    begin
    p := rando(version);
    if (p < prinsdel) then step := 1
     else
      begin
      if (p < cprpermute) then step := 2
       else step := 3;
      end;
    end;

    case step of
(* determine proposal; refer to other procedure
   to see if it is accepted and then carry it out. *)
     1: begin
    (* insert  delete step *)
        MH_InsDelPermute(true,per,outf,lf);
        end; (* case 1 *)
     2: begin
    (* permutation step *)
        MH_InsDelPermute(false,per,outf,lf);
        end; (* case 2 *)
     3: begin
    (* combined independent insert and delete diagonal steps *)
    (* determine proposal *)
        if (nza <= 0) then
          begin
          if (StaySameType <= 1) then exit;
          a := Network;
          end
         else
          begin
          if (rando(version) < prnet) then a := Network else a := Behavior;
          end;
         repeat i := randomd(version,n)+1 until StructActive[i];
         if (a = Behavior) then
           begin
           if (nza <= 1) then v := 1 else v := randomd(version,nza)+1;
           end
          else v := 0;
         k1 := Chain[per-1].RandomElement(version);
//writeln(outf,'Insert diagonal step ',i:4,'  ',k1.kindstring);
         MH_TryInsertDiag(a,i,v,k1,per,outf,lf);
       (* delete diagonal step *)
          if (Chain[per-1].DiagNumber <= 0) then begin
            inc(nimpcands);
            exit;
           end;
          k1 := Chain[per-1].RandomDiag(version);
          if k1.Extreme then
            begin
            errortext := 'diagonal pointer to extreme element '+k1.KindString;
            exit;
            end;
          if (not k1.Diagonal) then begin
//            writeln(outf,'*****  Error : random element diagonal sequence '+
//                         ' has kind = ',k1.Kindstring);
            errortext := 'diagonal pointer to nondiagonal element '+k1.KindString;
            exit;
           end;
//writeln(outf,'Diagonal Cancel step, k1 = ',k1.kindstring);  // @@@@
            MH_TryCancelDiag(k1,per,outf,lf);
        end; (* case 3 *)
       end; (* cases *)

EndOfProcedure:
  if (Chain[per-1].ErrorMes > '') then                       // @@@@
    errortext := errortext + ' '+Chain[per-1].ErrorMes;      // @@@@
  if (errortext > '') then begin
    x := 'Error: '+errortext+'; called from MH_Step'+eol;
    outf.Write(pchar(x)^,length(x));
    exit;
   end;
end; (* MH_Step *)



procedure MLModel.burnextra(var outf,lf: TFileStream);
var iteration,period: integer;
x: string;
begin
     for period := 1 to observations-1 do
       if (not skipi[period]) then begin
(* There is only one data structure. This is set to period. *)
       initialise_actors(period);
       initialise_y(period);
       AdaptProposalProbabilities;
       RenewRateslprobs(Chain[period-1].First,Chain[period-1].Last,false,period,outf,lf); // @@@@ added

       for iteration := 10*nrunMH downto 1 do
         begin
         MH_Step(period,outf,lf,0);
         if (ErrorText > '') or (Chain[period-1].ErrorMes > '') then begin
           x := 'Called from MCMC_cycle from burnextra.'+eol;  // @@@@
           outf.Write(pchar(x)^,length(x));
           exit;
          end;
        end;
      end;
end;


function MLModel.ConsistencyCheck(var outf,lf: TFileStream; CheckEnd: boolean): Boolean;
var i,j,numberij, numbers, dn, per: integer;
    OK: Boolean;
    k,k1: TMiniStep;
    a: Taspect;
    d: Tdif;
    v: byte;
    mn, msn: integer;
    x: string;
begin
 OK := true;
 ErrorText := '';
//  writeln(outf,'cs: 1 thisperiod = ',thisperiod);

for per := 1 to observations-1 do
 if (not skipi[per]) then
 begin
 n := nni[per];
 for i:=1 to n do for j:=1 to n do begin
   k := Chain[per-1].firstofrkind(Network,i,j,1,0);
   if assigned(k) then begin
     numberij := k.nrkind;
     numbers := 1;
     mn := k.mindex;
     msn := k.msindex;
     if (numberij <= 1) then
       begin
       if (mn >= 1) then
         begin
         x := format('%s%s%s%d%s%d%s',['Multiplicity index[',k.KindString,'] = ',mn,' but it reports ',
                      numberij,' elemens of its rkind.'])+eol;
         lf.Write(pchar(x)^,length(x));
         ErrorText := 'Wrong multiplicity number, see logfile ';
         OK := false;
         end;
       if (msn >= 1) then
         begin
         x := format('%s%s%s%d%s%d%s',['Single Multiplicity index[',k.KindString,'] = ',msn,' but it reports ',
                      numberij,' elemens of its rkind.']);
         lf.Write(pchar(x)^,length(x));
         ErrorText := 'Wrong single multiplicity number, see logfile ';
         OK := false;
         end;
      end
     else
       begin
       if (mn <= 0) then
         begin
         x := format('%s%s%s%d%s%d%s',['Multiplicity index[',k.KindString,'] = ',mn,' but it reports ',
                      numberij,' elements of its rkind.'])+eol;
         lf.Write(pchar(x)^,length(x));
         ErrorText := 'Wrong multiplicity number, see logfile ';
         OK := false;
         end;
       if (msn <= 0) and (not k.Diagonal)  then
         begin
         x := format('%s%s%s%d%s%d%s',['Single Multiplicity index[',k.KindString,'] = ',msn,' but it reports ',
                      numberij,' elements of its rkind.']);
         lf.Write(pchar(x)^,length(x));
         ErrorText := 'Wrong single multiplicity number, see logfile ';
         OK := false;
         end;
      end;
     if (k.kindex = 1) then begin
       x := 'First['+k.KindString+'] = 1. '+eol;
       lf.Write(pchar(x)^,length(x));
       ErrorText := 'Wrong first, see logfile ';
       OK := false;
      end;
     if k.NotExtreme then begin
       repeat
        k1 := k;
        k := k.Suchrkind;
        if not assigned(k) then begin
          if (numbers <> numberij) then begin
           x := format('%s%3d%s%3d%s',['Wrong numbers for (',i,',',j,').'])+eol;
            lf.Write(pchar(x)^,length(x));
            ErrorText := 'Wrong numbers, see logfile ';
            OK := false;
           end;
         end else begin
           Inc(numbers);
           if (k.nrkind <> numberij) then begin
             x := format('%s%3d%s%3d%s%d%s%d',['Not the same numbers for (',i,',',j,'): ',
                        k.nrkind,' ',numberij])+eol;
             lf.Write(pchar(x)^,length(x));
             ErrorText := 'Not same multiplicity numbers, see logfile ';
             OK := false;
            end;
           if (k.predhrkind <> k1) then begin
             x := format('%s%3d%s%3d%s',['Wrong sequence for Network (',i,',',j,').'])+eol;
             lf.Write(pchar(x)^,length(x));
             ErrorText := 'Wrong sequence, see logfile ';
             OK := false;
            end;
           if ((k.ih <> k1.ih) or (k.jh <> k1.jh)) then begin
            x := format('%s%3d%s%3d%s%3d%s%3d%s',['Wrong indices for Network (',i,',',j,') with ',k1.kindex,',',k.kindex,'.'])+eol;
            lf.Write(pchar(x)^,length(x));
            ErrorText := 'Wrong indices, see logfile ';
            OK := false;
           end;
           if ((k.cindex > 0) and (not k.FCCP)) or ((k.cindex < 0) and k.FCCP) then begin
            x := format('%s%3d%s%3d%s',['Wrong CCF index for Network (',i,',',j,').'])+eol;
            lf.Write(pchar(x)^,length(x));
            ErrorText := 'Wrong CCF index, see logfile ';
            OK := false;
           end;
         end;
        until (not assigned(k));
       if (numbers <> numberij) then begin
         x := format('%s%3d%s%3d%s',['Not all numbers found for Network(',i,',',j,').'])+eol;
         lf.Write(pchar(x)^,length(x));
         ErrorText := 'Not all found, see logfile ';
         OK := false;
        end;
      end; (* if notextreme(k) *)
    end; (* if assigned *)
  end; (* for i,j *)

 for i:=1 to n do for j:=1 to nza do for d:=0 to 1 do begin
   k := Chain[per-1].firstofrkind(Behavior,i,0,j,d);
   if assigned(k) then begin
     numberij := k.nrkind;
     numbers := 1;
     mn := k.mindex;
     msn := k.msindex;
     if (numberij <= 1) then
       begin
       if (mn >= 1) then
         begin
         x := format('%s%s%d%s%d%s',['Multiplicity index[',k.KindString,'] = ',mn,' but it reports ',
                      numberij,' elemens of its rkind.'])+eol;
                      lf.Write(pchar(x)^,length(x));
         ErrorText := 'Wrong multiplicity number, see logfile ';
         OK := false;
         end;
       if (msn >= 1) then
         begin
         x := format('%s%s%d%s%d%s',['Single Multiplicity index[',k.KindString,'] = ',msn,' but it reports ',
                      numberij,' elemens of its rkind.'])+eol;
                      lf.Write(pchar(x)^,length(x));
         ErrorText := 'Wrong single multiplicity number, see logfile ';
         OK := false;
         end;
      end
     else
       begin
       if (mn <= 0) then
         begin
         x := format('%s%s%d%s%d%s',['Multiplicity index[',k.KindString,'] = ',mn,' but it reports ',
                      numberij,' elements of its rkind.'])+eol;
                      lf.Write(pchar(x)^,length(x));
         ErrorText := 'Wrong multiplicity number, see logfile ';
         OK := false;
         end;
       if (msn <= 0) and (not k.Diagonal) then
         begin
         x := format('%s%s%d%s%d%s',['Single Multiplicity index[',k.KindString,'] = ',msn,' but it reports ',
                      numberij,' elements of its rkind.'])+eol;
                      lf.Write(pchar(x)^,length(x));
         ErrorText := 'Wrong single multiplicity number, see logfile ';
         OK := false;
         end;
      end;
     if (k.kindex = 1) then begin
       x := 'First['+k.KindString+'] = 1. '+eol;
       lf.Write(pchar(x)^,length(x));
       ErrorText := 'Wrong first, see logfile ';
       OK := false;
      end;
     if k.NotExtreme then begin
       repeat
        k1 := k;
        k := k.Suchrkind;
        if not assigned(k) then begin
          if (numbers <> numberij) then begin
            x := format('%s%3d%s%3d%s%3d%s',['Wrong numbers for Behavior(',i,',',j,',',d,').'])+eol;
            lf.Write(pchar(x)^,length(x));
            ErrorText := 'Wrong numbers, see logfile ';
            OK := false;
           end;
         end else begin
           Inc(numbers);
           if (k.nrkind <> numberij) then begin
             x := format('%s%3d%s%3d%s%3d%s%d%s%d',['Not the same numbers for Behavior(',i,',',j,',',d,'): ',
                        k.nrkind,' ',numberij])+eol;
                        lf.Write(pchar(x)^,length(x));
             ErrorText := 'Not same multiplicity numbers, see logfile ';
             OK := false;
            end;
           if (k.predhrkind <> k1) then begin
              x := format('%s%3d%s%3d%s%3d%s',['Wrong sequence for Behavior(',i,',',j,',',d,').'])+eol;
             lf.Write(pchar(x)^,length(x));
             ErrorText := 'Wrong sequence, see logfile ';
             OK := false;
            end;
           if ((k.ih <> k1.ih) or (k.varh <> k1.varh) or (abs(k.difh) <> abs(k1.difh))) then begin
            x := format('%s%3d%s%3d%s%3d%s%3d%s%3d%s',['Wrong indices for Behavior(',i,',',j,',',d,') with ',k1.kindex,',',k.kindex,'.'])+eol;
            lf.Write(pchar(x)^,length(x));
            ErrorText := 'Wrong indices, see logfile ';
            OK := false;
           end;
           if ((k.cindex > 0) and (not k.FCCP)) or ((k.cindex < 0) and k.FCCP) then begin
            x := format('%s%3d%s%3d%s%3d%s',['Wrong CCF index for Behavior(',i,',',j,',',d,').'])+eol;
            lf.Write(pchar(x)^,length(x));
            ErrorText := 'Wrong CCF index, see logfile ';
            OK := false;
           end;
         end;
        until (not assigned(k));
       if (numbers <> numberij) then begin
         x := format('%s%3d%s%3d%s%3d%s',['Not all numbers found for Behavior(',i,',',j,',',d,').'])+eol;
         lf.Write(pchar(x)^,length(x));
         ErrorText := 'Not all found, see logfile ';
         OK := false;
        end;
      end; (* if notextreme(k) *)
    end; (* if assigned *)
  end; (* for i,j,d *)


 k := CHain[per-1].First;
 dn := 0;
 repeat
  k1 := k;
  k := k.such;
  if k.Diagonal then inc(dn);
  if (k.predh <> k1) then begin
    x := format('%s%3d%s%3d%s',['Wrong succession for ',k1.kindex,' to ',k.kindex,').'])+eol;
    lf.Write(pchar(x)^,length(x));
    ErrorText := 'Wrong succesion, see logfile ';
    x := Errortext+eol;
    outf.Write(pchar(x)^,length(x));
    OK := false;
   end;
  if not ThisOrder(k1,k) then begin
    x := format('%s%9.6f%s%3d%s%9.6f%s%3d%s',['Wrong time order: ',k1.time,' for ',k1.kindex,' to ',
                                    k.time,' for ',k.kindex,'.'])+eol;
    lf.Write(pchar(x)^,length(x));
    ErrorText := 'Wrong time order, see logfile ';
    x := Errortext+eol;
    outf.Write(pchar(x)^,length(x));
    OK := false;
   end;
  if (k <> Chain[per-1].FindexStep[k.kindex]) then begin
    x := format('%s%3d%s',['Wrong inversion of kindex for ',k.kindex,'.'])+eol;
    lf.Write(pchar(x)^,length(x));
    ErrorText := 'Wrong inversion of kindex, see logfile ';
    x := Errortext+eol;
    outf.Write(pchar(x)^,length(x));
    OK := false;
   end;

  if k.Diagonal and (k.msindex >= 0) then
    begin
    x := format('%s%3d%s',['Single multiple indicator for diagonal element ',k.kindex,'.'])+eol;
    lf.Write(pchar(x)^,length(x));
    ErrorText := 'Single multiple indication for diagonal ministep, see logfile ';
    x := Errortext+eol;
    outf.Write(pchar(x)^,length(x));
    OK := false;
    end;

  if (k.nrkind >= 2) then begin
    if (k <> Chain[per-1].FMultIndexStep[k.mindex]) then
      begin
      x := format('%s%3d%s%d',['Incorrect multiplicity indicator ',k.mindex,' for ',k.kindex])+eol;
      lf.Write(pchar(x)^,length(x));
      ErrorText := 'Incorrect multiplicity indicator, see logfile ';
      x := Errortext+eol;
      outf.Write(pchar(x)^,length(x));
      OK := false;
      end;

    a := k.ash;
    i := k.ih;
    j := k.jh;
    v := k.varh;
    d := k.Difh;
    if (k.msindex >= 0) then
      begin
      if (Chain[per-1].firstofrkind(a,i,j,v,d) <> Chain[per-1].FMultSIndexStep[k.msindex])
       then
        begin
        x := format('%s%3d%s%d',['Incorrect single multiplicity indicator ',k.msindex,' for ',k.kindex])+eol;
        lf.Write(pchar(x)^,length(x));
        ErrorText := 'Incorrect single multiplicity indicator, see logfile ';
        x := Errortext+eol;
        outf.Write(pchar(x)^,length(x));
        OK := false;
        end;
      end;
   end;

  if ((k.FCCP) and (k.cindex > 0))
  then if (k <> Chain[per-1].FCCPIndexStep[k.cindex])
  then begin
    x := format('%s%3d',['Incorrect FCCP indicator ',k.cindex])+eol;
    lf.Write(pchar(x)^,length(x));
    ErrorText := 'Incorrect FCCP indicator, see logfile ';
    x := Errortext+eol;
    outf.Write(pchar(x)^,length(x));
    OK := false;
   end;

  if ((k.FCCP) and (k.cindex < 0))
  then begin
    x := format('%s%3d',['Incorrect FCCP indicator ',k.cindex])+eol;
    lf.Write(pchar(x)^,length(x));
    ErrorText := 'Incorrect FCCP indicator, see logfile ';
   x := Errortext+eol;
    outf.Write(pchar(x)^,length(x));
    OK := false;
   end;


  if (k.Diagonal)
  then if (k <> Chain[per-1].FDiagIndexStep[k.dindex])
  then begin
    x := format('%s%3d',['Incorrect diagonal indicator ',k.dindex])+eol;
    lf.Write(pchar(x)^,length(x));
    ErrorText := 'Incorrect diagonal indicator, see logfile ';
    x := Errortext+eol;
    outf.Write(pchar(x)^,length(x));
    OK := false;
   end;
  until k.Extreme;

  if CheckEnd then begin
    k := CHain[per-1].Last.predh;
    initialise_actors(per);
    SetData(k,per,outf,lf);
    for i:=1 to n do for j := 1 to n do
     if ((i <> j) and (yn(i,j) <> ynm(i,j,per+1)) and (not Struct(i,j,per))) then begin
       OK := false;
       x := format('%s%d%s%d%s%d%s',['Wrong Network data for this chain at end of period ',per,
               ' for values ',i,', ',j,'.'])+eol;
       lf.Write(pchar(x)^,length(x));
       ErrorText := 'Chain does not lead to the correct end situation ';
       x := Errortext+eol;
       outf.Write(pchar(x)^,length(x));
      end;
    if (nza >= 1) then
    for i:=1 to n do if StructActive[i] then for v := 1 to nza do
      if (ya(i,v) <> truncaction(zc[i,v,per+1],v)) then begin
       OK := false;
       x := format('%s%d%s%d%s%d%s',['Wrong Behavior data for this chain at end of period ',per,
               ' for values ',i,', ',v,'.'])+eol;
       lf.Write(pchar(x)^,length(x));
       ErrorText := 'Chain does not lead to the correct end situation ';
       x := Errortext+eol;
       outf.Write(pchar(x)^,length(x));
      end;
    end; (* if *)

  if (dn <> Chain[per-1].DiagNumber) then begin
    x := format('%s%d%s%d%s',['Chain.DiagNumber = ', CHain[per-1].DiagNumber, ', but only ', dn,
               ' diagonal elements were found.'])+eol;
    lf.Write(pchar(x)^,length(x));
    ErrorText := 'Incorrect value for DiagNumber.';
    x := Errortext+eol;
    outf.Write(pchar(x)^,length(x));
    writefullperiod(per,outf,lf);
    OK := false;
   end; (* if *)
 end; (* for per *)

  ConsistencyCheck := OK;

//  writeln(outf,'cs: 2 thisperiod = ',thisperiod);
end; (* ConsistencyCheck *)
(*--------------------------------------------------------------------------------------------------*)
procedure MLModel.WritePeriod(const per: integer; var outf,lf: text);
var num: integer;
    it: TMiniStep;
    k: TMiniSTep;
begin
    if (per <= 0) or (per >= observations) then begin
      errortext := 'Called WritePeriod for per = '+inttostr(per);
      exit;
     end;
    writeln(outf,'Period          = ',per:6);
    writeln(outf,'TotNumber       = ',Chain[per-1].TotNumber:6);
    writeln(outf,'MultipleNumber  = ',Chain[per-1].MultipleNumber:6);
    writeln(outf,'MultipleSNumber = ',Chain[per-1].MultipleSNumber:6);
    writeln(outf,'CCPNumber       = ',Chain[per-1].CCPNumber:6);
    writeln(outf,'DiagNumber      = ',Chain[per-1].DiagNumber:6);
    writeln(outf);
    writeln(outf,'Chain in order of occurrence.');
    writeln(outf,
    '      Aspect    ih  jh varh difh predh such nrkind     time       rate     lprob  mi  msi   di   ci'); // @@@@
    k := Chain[per-1].First;
    num := 0;
    repeat
      inc(num);
      k.Writekind(outf);
      if k.Diagonal then begin
        it := Chain[per-1].FDIagIndexstep[k.dindex];
        if (it = nil) then writeln(outf,
           ' ***** error: diagonal element without diagonal pointer  *****')
          else writeln(outf,' diagonal pointer to ',it.kindex);
       end (* if *) else writeln(outf);
      k := k.such;
     until (k.Extreme or (num >= 30000));
    k.Writekind(outf);
    writeln(outf);
    writeln(outf);
end; (* WritePeriod *)

procedure MLModel.WritePeriod(const per: integer; var outf,lf: TFileStream);
var num: integer;
    it: TMiniStep;
    k: TMiniSTep;
    x: string;
begin
    if (per <= 0) or (per >= observations) then begin
      errortext := 'Called WritePeriod for per = '+inttostr(per);
      exit;
     end;
    x := format('%s%6d',['Period          = ',per])+eol;
    outf.Write(pchar(x)^,length(x));
    x := format('%s%6d',['TotNumber       = ',Chain[per-1].TotNumber])+eol;
    outf.Write(pchar(x)^,length(x));
    x := format('%s%6d',['MultipleNumber  = ',Chain[per-1].MultipleNumber])+eol;
    outf.Write(pchar(x)^,length(x));
    x := format('%s%6d',['MultipleSNumber = ',Chain[per-1].MultipleSNumber])+eol;
    outf.Write(pchar(x)^,length(x));
    x := format('%s%6d',['CCPNumber       = ',Chain[per-1].CCPNumber])+eol;
    outf.Write(pchar(x)^,length(x));
    x := format('%s%6d',['DiagNumber      = ',Chain[per-1].DiagNumber])+eol;
    outf.Write(pchar(x)^,length(x));
    outf.Write(pchar(eol)^,length(eol));//riteln(outf);
    x := 'Chain in order of occurrence.'+eol;
    outf.Write(pchar(x)^,length(x));
    x :=
    '      Aspect    ih  jh varh difh predh such nrkind     time       rate     lprob  mi  msi   di   ci'+eol; // @@@@
    outf.Write(pchar(x)^,length(x));
    k := Chain[per-1].First;
    num := 0;
    repeat
      inc(num);
      k.Writekind(outf);
      if k.Diagonal then begin
        it := Chain[per-1].FDIagIndexstep[k.dindex];
        if (it = nil) then begin
          x :=
           ' ***** error: diagonal element without diagonal pointer  *****'+eol;
          outf.Write(pchar(x)^,length(x));
        end
          else begin
            x := format('%s%d',[' diagonal pointer to ',it.kindex])+eol;
            outf.Write(pchar(x)^,length(x));
          end;
       end (* if *) else outf.Write(pchar(eol)^,length(eol));//writeln(outf);
      k := k.such;
     until (k.Extreme or (num >= 30000));
    k.Writekind(outf);
    outf.Write(pchar(eol)^,length(eol));//writeln(outf);
    outf.Write(pchar(eol)^,length(eol));//writeln(outf);
end; (* WritePeriod *)
 //------------------------
 procedure MLModel.WriteFullPeriod(const per: integer; var outf,lf: text);
var i,j,num: integer;
    v: byte;
    d: Tdif;
    k: TMiniSTep;
begin
    n := nni[per];
    WritePeriod(per,outf,lf);
    writeln(outf,'Chain in order of used indices:');
    writeln(outf,'            Aspect    ih  jh var dif   predh such  nrkind    time       rate     lprob  mi  msi   di   ci'); // @@@@
    writeln(outf);
    for num:=1 to Chain[per-1].TotNumber do begin  // was 0 @@@@
      k := Chain[per-1].FIndexSTep[num];
      write(outf,num:5,' ');
      k.WriteKind(outf);
      writeln(outf);
     end;
   writeln(outf);
{    writeln(outf,'First index of changes : ');
    write(outf,'      ');
    for j:=1 to n do write(outf,j:4);
    writeln(outf);
    for i:=1 to n do begin
      write(outf,i:6);
      for j:=1 to n do
        if assigned(Chain[per-1].firstofrkind(i,j)) then write(outf,Chain[per-1].firstofrkind(i,j).kindex:4)
          else write(outf,'    ');
      writeln(outf);
     end;
    writeln(outf);
    writeln(outf,'Number of changes : ');
    write(outf,'      ');
    for j:=1 to n do write(outf,j:4);
    writeln(outf);
    for i:=1 to n do begin
      write(outf,i:6);
      for j:=1 to n do
        if assigned(Chain[per-1].firstofrkind(i,j)) then write(outf,Chain[per-1].firstofrkind(i,j).nkind:4)
          else write(outf,'    ');
      writeln(outf);
     end;
 }
    writeln(outf);

    writeln(outf,'Occurrences of changes ');
    writeln(outf);
    writeln(outf,'              Aspect    ih  jh var dif   predh such  nrkind    time       rate     lprob  mi  msi   di   ci'); // @@@@
    for i:=1 to n do for j:=1 to n do if assigned(Chain[per-1].firstofrkind(Network,i,j,1,0)) then begin
      k := Chain[per-1].firstofrkind(Network,i,j,1,0);
      repeat
        write(outf,i:3,' ',j:3,' ');
        k.WriteKind(outf);
        writeln(outf);
        k := k.suchrkind;
      until not Assigned(k);
      writeln(outf);
     end;
   writeln(outf);
   if (nza >= 1) then begin
    writeln(outf, '  i  v  d      Aspect    ih  jh var dif   predh such  nrkind    time       rate     lprob  mi  msi   di   ci');
      for i:=1 to n do for v := 1 to nza do for d := -1 to 1 do
        if assigned(Chain[per-1].firstofrkind(Behavior,i,1,v,d)) then begin
      k := Chain[per-1].firstofrkind(Behavior,i,1,v,d);
      repeat
        write(outf,i:3,v:3,d:3);
        k.WriteKind(outf);
        writeln(outf);
        k := k.suchrkind;
      until not Assigned(k);
      writeln(outf);
     end;
   writeln(outf);
    end; (* if nza *)
    writeln(outf);
end; (* WriteFullPeriod *)

procedure MLModel.WriteFullPeriod(const per: integer; var outf,lf: TFileStream);
var i,j,num: integer;
    v: byte;
    d: Tdif;
    k: TMiniSTep;
    x: string;
begin
    n := nni[per];
    WritePeriod(per,outf,lf);
    x := 'Chain in order of used indices:'+eol;
    outf.Write(pchar(x)^,length(x));
    x := '            Aspect    ih  jh var dif   predh such  nrkind    time       rate     lprob  mi  msi   di   ci'+eol; // @@@@
    outf.Write(pchar(x)^,length(x));
    outf.Write(pchar(eol)^,length(eol));//writeln(outf);
    for num:=1 to Chain[per-1].TotNumber do begin  // was 0 @@@@
      k := Chain[per-1].FIndexSTep[num];
      x := format('%4d%s',[num,' ']);
      outf.Write(pchar(x)^,length(x));
      k.WriteKind(outf);
      outf.Write(pchar(eol)^,length(eol));//writeln(outf);
     end;
   outf.Write(pchar(eol)^,length(eol));//writeln(outf);
{    writeln(outf,'First index of changes : ');
    write(outf,'      ');
    for j:=1 to n do write(outf,j:4);
    writeln(outf);
    for i:=1 to n do begin
      write(outf,i:6);
      for j:=1 to n do
        if assigned(Chain[per-1].firstofrkind(i,j)) then write(outf,Chain[per-1].firstofrkind(i,j).kindex:4)
          else write(outf,'    ');
      writeln(outf);
     end;
    writeln(outf);
    writeln(outf,'Number of changes : ');
    write(outf,'      ');
    for j:=1 to n do write(outf,j:4);
    writeln(outf);
    for i:=1 to n do begin
      write(outf,i:6);
      for j:=1 to n do
        if assigned(Chain[per-1].firstofrkind(i,j)) then write(outf,Chain[per-1].firstofrkind(i,j).nkind:4)
          else write(outf,'    ');
      writeln(outf);
     end;
 }
    outf.Write(pchar(eol)^,length(eol));//writeln(outf);

    x := 'Occurrences of changes '+eol;
    outf.Write(pchar(x)^,length(x));
   outf.Write(pchar(eol)^,length(eol));// writeln(outf);
    x := '              Aspect    ih  jh var dif   predh such  nrkind    time       rate     lprob  mi  msi   di   ci'+eol; // @@@@
    outf.Write(pchar(x)^,length(x));
    for i:=1 to n do for j:=1 to n do if assigned(Chain[per-1].firstofrkind(Network,i,j,1,0)) then begin
      k := Chain[per-1].firstofrkind(Network,i,j,1,0);
      repeat
        x := format('%3d%s%3d%s',[i,' ',j,' ']);
        outf.Write(pchar(x)^,length(x));
        k.WriteKind(outf);
        outf.Write(pchar(eol)^,length(eol));//writeln(outf);
        k := k.suchrkind;
      until not Assigned(k);
      outf.Write(pchar(eol)^,length(eol));//writeln(outf);
     end;
   outf.Write(pchar(eol)^,length(eol));//writeln(outf);
   if (nza >= 1) then begin
    x := '  i  v  d      Aspect    ih  jh var dif   predh such  nrkind    time       rate     lprob  mi  msi   di   ci'+eol;
    outf.Write(pchar(x)^,length(x));
      for i:=1 to n do for v := 1 to nza do for d := -1 to 1 do
        if assigned(Chain[per-1].firstofrkind(Behavior,i,1,v,d)) then begin
      k := Chain[per-1].firstofrkind(Behavior,i,1,v,d);
      repeat
        x := format('%3d%3d%3d',[i,v,d]);
        outf.Write(pchar(x)^,length(x));
        k.WriteKind(outf);
        outf.Write(pchar(eol)^,length(eol));//writeln(outf);
        k := k.suchrkind;
      until not Assigned(k);
      outf.Write(pchar(eol)^,length(eol));//writeln(outf);
     end;
   outf.Write(pchar(eol)^,length(eol));//writeln(outf);
    end; (* if nza *)
    outf.Write(pchar(eol)^,length(eol));//writeln(outf);
end; (* WriteFullPeriod *)


procedure MLModel.MLInitialise(var p: byte; (*MS: <p> is declared as variable: important for ML estimation of random effects models!*)
                               var theta: pVec; var outf, lf: TFileStream);
(* p = number of parameters *)
var halter: boolean;
  per: integer;
  x: string;
begin
  errortext := '';
  halter := false;
  InitialiseBase2(theta,outf,lf);
  SetAlpha(theta);
  if exo then begin
    errortext := 'ML not yet implemented for changing composition';
    exit;
   end;
  MLInitialisePeriods(outf,lf);
  if (randomeffects or bayesian) then
    InitialiseMS(theta,p,outf,lf); (*MS: initialises some arrays needed in ML / Bayesian estimation of models (with actor-dependent random effects).*)
  (* The following two probabilities may be chosen differently and perhaps optimized. *)
  prinsdel := 0.6;
  prpermute := 0.3;
  if (nza <= 0) then prnet := 1.0 else prnet := 0.5;
  prelins := 0.1;
  promul0 := 0.5;
  SetLength(ntrypermutes,maxpermute);
  SetLength(npermutes,maxpermute);
  SetLength(nnopermutes,maxpermute);
  (*MS: important note: <MakeChains> must not precede <InitialiseMS>:
  otherwise, in random effects and multilevel models, the <lprob>'s are wrong.*)
  SetLength(permitted_crp,nmax+1);
  SetLength(p_crp,nmax+1);
  SetLength(pract_crp,nza+1);
  MakeChains(theta,outf,lf);

  if (errortext > '') then begin
    x := 'MLInitialise: errortext 1 = '+errortext+eol;
    outf.Write(pchar(x)^,length(x));
    x := 'Error noted in MakeChains called from MLInitialise.'+eol;
    outf.Write(pchar(x)^,length(x));
    //writeln(lf,'Error noted in MakeChains called from MLInitialise.');
    lf.Write(pchar(x)^,length(x));
    x := 'errortext = '+errortext+eol;
    outf.Write(pchar(x)^,length(x));
    for per := 1 to observations-1 do if (not skipi[per]) then WriteFullPeriod(per,outf,lf);
    exit;
   end else begin
      x := 'MakeChains worked properly.'+eol;
      lf.Write(pchar(x)^,length(x));
   end;
  if (errortext > '') then begin
    x := 'MLInitialise: errortext 2 = '+errortext+eol;
    outf.Write(pchar(x)^,length(x));
  end;
  x := 'Probabilities for types of MH step: '+eol;
  if (nza > 0) then begin
    x := format('%s%6.3f',['Network                   ',prnet])+eol;
    lf.Write(pchar(x)^,length(x));
  end;
  x := format('%s%6.3f',['Insert - delete           ',prinsdel])+eol;
  lf.Write(pchar(x)^,length(x));
  x := format('%s%6.3f',['Permutation               ',prpermute])+eol;
  lf.Write(pchar(x)^,length(x));
  x := format('%s%6.3f',['Diagonal insert - delete  ',1.0 - cprpermute])+eol;
  lf.Write(pchar(x)^,length(x));
  x := format('%s%6.3f',['Relative insertion        ',prelins])+eol;
  lf.Write(pchar(x)^,length(x));
  x := format('%s%6.3f',['Extra multiples           ',promul])+eol;
  lf.Write(pchar(x)^,length(x));
  if ConsistencyCheck(outf,lf,true)
  then begin
    x := 'Consistency check in MLInitialise OK.'+eol;
    lf.Write(pchar(x)^,length(x));
  end
  else begin
    x := 'Consistency check in MLInitialise for chain not OK. ErrorText = '+errortext+eol;
    outf.Write(pchar(x)^,length(x));
    for per := 1 to observations-1 do if (not skipi[per]) then WriteFullPeriod(per,outf,lf);
  end;

  ninserts := 0;
  nnoinserts := 0;
  ninsds := 0;
  nnoinsds := 0;
  ncancels := 0;
  npermtots := 0;
  nnopermtots := 0;
  ncands := 0;
  nnocands := 0;
  nimpcands := 0;
  nimpinsdels := 0;
  nimppermutes := 0;
  nnotok := 0;
  nstep := 0;
end; (* MLModel.MLInitialise *)

{procedure MLModel.BurnIn(var outf, lf: TFileStream; var theta: pvec; var p: byte);
var i: integer;
(* aantal nrunMH is laag, kan verbeterd *)
(* Deze procedure wordt helemaal niet gebruikt, kan weg;
   is opgenomen in Polrup *)
begin
  for i:=1 to nrunMH do MCMC_Cycle(outf,lf,theta,p);
end; (* MLModel.BurnIn *)}


function MLModel.valid(const V0: TAspect; const V1, V2: integer; const V3: byte; const V4: Tdif): boolean;
(* This procedure may depend only on variables that depend on the current period
   but not on the current moment within the period. *)
begin
  if (V0 = Network) then
  valid := ((V1 >= 1) and (V1 <= n) and (V2 >= 1) and (V2 <= n) and
            ((V1 <> V2) or (StaySameType >= 2)))
  else
  valid := ((V1 >= 1) and (V1 <= n) and (V3 >= 1) and (V3 <= nza));
end;


procedure MLModel.SetData(const k: TMiniStep; const per: integer;
                          var outf,lf: TFileStream);
  (* sets the current data set in Eight to the situation in period,
     using the embedded chain in Embedded as it is valid
     immediately after the change represented with index k.
     initialise_actors must have been called for this period. *)
var h: TMinistep;
begin
  if (k = Chain[per-1].Last) then begin
    ErrorText := 'SetData called for unreachable index.';
    exit;
   end;
//  if (k.time < 0.5) then
//    begin
    initialise_y(per);
    h:= Chain[per-1].first;
    while (h <> k) do begin
      h := h.such;
      if (h.ash = Network) then ChangeTie(h.ih,h.jh) else ChangeBehavior(h.ih,h.varh,h.Difh);
     end; (* while *)
{    end else
    begin
// @@@@@@ gebruik hier de eindsituatie van elke periode, gemaakt in MakeChains
    initialise_y_end(per);
    h:= Chain[per-1].last.predh;
    while (h <> k) do begin
      if (h.ash = Network) then ChangeTie(h.ih,h.jh) else ChangeBehavior(h.ih,h.varh,h.Difh);
      h := h.predh;
     end; (* while *)
    end;
}
end; (* MLModel.SetData*)

procedure MLModel.CalcRatelProb(var therate, thelprob: real;
        const V0: TAspect; const V1, V2: integer; const V3: byte; const V4: Tdif;
        var outf,lf: TFileStream);
(* This is, with SetData, the crucial linking procedure between
   the data structures in Chains and the probability model in S_Base.
   It corresponds to S_Base.Runstep.
   Eventually it would be better to let CalcRatelProb and Runstep
   use the same constituents. *)
(* CalcRatelProb calculates value of rate of change and probability of
   making the ministep (V0,V1,V2,V3,V4) given the current state of the network.
   The rate of change is the value valid before this change is made.
   If the change is not permitted, then therate is set to -9.0. *)
var ii,jj,s,effectnumber,red: integer;
  lnp, lami, lamsum, probi, lxi, sp, spplus, spmin, pii, pij, plusprob, coefficient: real;
  pp, pm, lnpp, lnpm, trate: real;
  plus, halterr, error, pluspossible, minpossible: boolean;
// abcdefg  p, pract: array of real;
// abcdefg  permitted: array of boolean;
  thecomponents: array_0_20;
begin
(* copied and simplified from runstep;
   but since in S_ML.CalcRatelProb only one probability needs to be calculated,
   for model specifications such as type 3 it is more efficient
   not to use a common procedure to calculate the rate and
   full probability vectors. *)
(* This procedure was checked only for ModelType = 1. @@@ *)

  thelprob := 0.0;
  therate := 1.0;

  if (not valid(V0,V1,V2,V3,V4)) then
    begin
    ErrorText := 'CalcRatelProb called for invalid kind = ('
                 +Kindsstring(V0,V1,V2,V3,V4)+').';
    exit;
    end;


  if (not StructActive[V1]) then begin
     therate := -9.0;
     exit;
    end;

  if (V0 = Behavior) then
    begin
    minpossible  := (not Minimalya(V1,V3));
    pluspossible := (not Maximalya(V1,V3));
    if ((V4 = -1) and (not minpossible)) then
      begin
// @@@ writeln(outf,'not min  ',V1:3,ya(V1,V3):3,V4:3,zcmin[v3]:3:0);
      therate := -9.0;
      thelprob := -1e100;
      exit;
      end;
    if ((V4 = 1) and (not pluspossible)) then
      begin
// @@@ writeln(outf,'not plus ',V1:3,ya(V1,V3):3,V4:3,zcmax[v3]:3:0);
      therate := -9.0;
      thelprob := -1e100;
      exit;
      end;
    end;

// for constant rate:         @@@ moet veranderd voor algemene rate
// Note that in S_Base for Bayesian and Maxlike in Networklambda and Actionlambda,
// non-active actors still get a positive rate;
// this makes it possible to use actor 1 here.
//ccc
  trate := NetworkLambda(1,halterr,outf,lf);
  if (nza > 0) then
    begin
    pract_crp[0] := trate;
    for jj := nza downto 1 do
      begin
      pract_crp[jj] := ActionLambda(1,jj,halterr,outf,lf);
      trate := trate + pract_crp[jj];
      end;
    for jj := nza downto 0 do pract_crp[jj] := pract_crp[jj]/trate;
(* pract_crp[jj] = probability that behavior jj is changed;
   pract_crp[0] =   probability that network is changed. *)
    end;
  therate := n_act*trate;

{
// dit kan versneld voor puur exogene lambda
    lamsum := 0.0;
    for ii:= n downto 1 do begin
      if (StructActive[ii]) then lami := NetworkLambda(ii,halterr,outf,lf)
       else lami := 0.0;
      if halterr then exit;
      lamsum := lamsum + lami;
      p_crp[ii] := lami;
      //sis[ii] := lami; (*MS*)
     end; (* for ii *)

    if (lamsum < 1e-40) then begin
       ErrorText := 'nonpositive lamsum in CalcRateLprob, lamsum = '+floattostr(lamsum);
       exit;
      end;
    probi := p_crp[i]/lamsum;
    therate := lamsum;
    }

  if (V0 = Network) then
    begin
    for ii:= nmax downto 0 do p_crp[ii] := 0.0;

    if (ModelType <> 3) then
      begin
(* probi = probability that i makes the next change *)
      probi := 1.0/n_act;  (* constant rate. *)

      case ModelType of
      1: begin
           if checkchange then
             begin
             PermitChange(n,V1,thisperiod,maxdegree,permitted_crp);
             permitted_crp[V1] := (StaySameType >= 2);
             if (not permitted_crp[V2]) then
               begin
               therate := -9.0;
               exit;
               end;
             end
            else for ii := n downto 1 do permitted_crp[ii] := true;
         end; (* case 1 *)
      2: begin
// of dit klopt weet ik niet.  permitted_crp moet nog!
         spmin := 0.0;
         spplus := 0.0;
         s := outd_yn(V1);
         plus :=  (s <= 0);
         if (not (plus or (s >= (n-1)))) then
           begin
           lxi :=  lambda2[0,V1]*xi(s);
           plusprob := nu(s)*lxi*(1.0+lambda2[0,V1]*xi(s-1))/(nu(s-1)*(1.0+lxi));
           end; (* if *)
(* @@@ -9.0 indien onmogelijk of not permitted_crp *)
         end; (*  case 2 *)
       end; (* case ModelType *)

      ChoiceProbabilities(p_crp,V1,permitted_crp); (*MS: note: present random effects <RE> are used.*)

      sp := 0.0;
      for jj := n downto 1 do if permitted_crp[jj] then
        begin
          case ModelType of
          1: sp:=sp+expo(p_crp[jj]);
          2: if yn(V1,jj) then   // @@@
              spmin := spmin + expo(p_crp[jj]) else spplus := spplus + expo(p_crp[jj]);
          end; (* case *)
        end; (* for jj *)

        case ModelType of
        1: if (nza > 0) then  thelprob := p_crp[V2] + log(probi*pract_crp[0]/sp)
            else thelprob := p_crp[V2] + log(probi/sp);
        2: if (nza > 0) then
             begin
             if yn(V1,V2) then thelprob := p_crp[V2] + log(probi*(1.0-plusprob)*pract_crp[0]/spmin)
                        else thelprob := p_crp[V2] + log(probi*plusprob*pract_crp[0]/spplus);
             end
            else
             begin
             if yn(V1,V2) then thelprob := p_crp[V2] + log(probi*(1.0-plusprob)/spmin)
                        else thelprob := p_crp[V2] + log(probi*plusprob/spplus);
             end;
        end; (* case *)
      end (* if ModelType <> 3 *)
     else
      begin (* ModelType = 3 *)
      sp := 0.0;
      for ii:=n downto 1 do for jj:=n downto 1 do if (ii <> jj) then
        begin
        lnp := UtilityGain(ii,jj,thisperiod,symmetric,false,NewNetworkEffects);
        probi := expit(lnp);
        if ((ii = V1) and (jj = V2)) then pij := probi;
        sp := sp + probi;
        end; (* for ii, jj if *)
      therate := networklambda(1,halterr,outf,lf)*sp;  // this is not OK for networks and behavior
      if (nza > 0) then thelprob := log(pij*pract_crp[0]/sp)
       else thelprob := log(pij/sp);
      end; (* if ModelType = 3 *)
    end (* if V0 = Network *)
   else
    begin (* V0 = Behavior *)
    lnpm := 0.0;
    lnpp := 0.0;
    probi := 1.0/n_act;  (* constant rate *)
    if minpossible then
      begin
      lnpm := 0.0;
      CalcComponents_fa(V1,NetworkEffects_f,ActionEffects_f,nz,V3,thisperiod,-1,false,
                         Effects_s,thecomponents,outf);
      for jj:= 1 to ActionEffects_f do
        begin
        effectnumber := NetworkEffects_f + (V3-1)*ActionEffects_f + jj;
        if Effects_f[effectnumber] then
          begin
          coefficient := alpa_f[effectnumber] + GetRE(1,effectnumber,V1); (*MS: compute coefficient: objective function.*)
          lnpm := lnpm + coefficient*Contr_fa(V1,jj,V3,-1,thecomponents,halterr,outf,lf);
          end;
        end;
      for jj:= 1 to ActionEffects_g do
        begin
        effectnumber := NetworkEffects_g + (V3-1)*ActionEffects_g + jj;
        if Effects_g[effectnumber] then
          begin
          coefficient := alpa_g[effectnumber] + GetRE(2,effectnumber,V1); (*MS: compute coefficient: gratification function.*)
          lnpm := lnpm + coefficient*Contr_fa(V1,jj,V3,-1,thecomponents,halterr,outf,lf);
          end;
        end;
      if halterr then
        begin
        errortext := 'error in Contr_fa- '+st(V3);
        exit;
        end; (* if halterr *)
      pm := expo(lnpm);
    end (* if minpossible*)
   else pm := 0.0;

  if pluspossible then
    begin
    lnpp := 0.0;
    CalcComponents_fa(V1,NetworkEffects_f,ActionEffects_f,nz,V3,thisperiod,1,false,
                         Effects_f, thecomponents,outf);
    for jj:= 1 to ActionEffects_f do
      begin
      effectnumber := NetworkEffects_f + (V3-1)*ActionEffects_f + jj;
      if Effects_f[effectnumber] then
        begin
        coefficient := alpa_f[effectnumber] + GetRE(1,effectnumber,V1); (*MS: compute coefficient: objective function.*)
        lnpp := lnpp + coefficient*Contr_fa(V1,jj,V3,1,thecomponents,halterr,outf,lf);
        end;
      end;
   if halterr then
     begin
     errortext := 'error in Contr_fa+ '+st(V3);
     exit;
     end; (* if halterr *)
    pp := expo(lnpp);
    end (* if pluspossible*)
   else pp := 0.0;

 (* pract_crp[V3] is the contribution of the constant rate functions *)
 (* It is already excluded that ((V4 = -1) and (not minpossible))
   or ((V4 = 1) and (not pluspossible)) by earlier exits. *)
    case V4 of
(* the 1.0 is expo(0),
   corresponding to zero utility change when not changing behaviour V3 *)
     -1: thelprob := lnpm + log(probi*pract_crp[V3]/(pm+1.0+pp));
      0: thelprob := log(probi*pract_crp[V3]/(pm+1.0+pp));
      1: thelprob := lnpp + log(probi*pract_crp[V3]/(pm+1.0+pp));
    end; (* case *)
  end; (* if V0 = Behavior *)
end; (* CalcRatelprob *)

procedure MLModel.AdaptProposalProbabilities;
begin
  if madechain then promul := promul0 else promul := 0.1;
  if (nza <= 0) then prnet := 1.0;
  if (prinsdel + prpermute >= 1.0) then
    begin
    prinsdel := prinsdel/(prinsdel + prpermute + 0.01);
    prpermute := prpermute/(prinsdel + prpermute + 0.01);
    end;

  cprpermute := prinsdel + prpermute;
  prddel := 0.5*(1.0 - cprpermute);
  prdins := prddel;

  if (nza > 0) then
    begin
    prndins := prdins*prnet/n_act;
    prbdins := prdins*(1.0-prnet)/(n_act*nza);
    end
   else
    begin
    prnet := 1.0;
    prndins := prdins/n_act;
    prbdins := 0.0;
    end;
  if (StaySameType = 1) then
    begin
    prndins := 0.0;
    if (nza <= 0) then prbdins := 0.0 else prbdins := prdins/(n_act*nza); // @@@@
    end;
end; (* AdaptProposalProbabilities *)


procedure MLModel.MakeChains(theta: pvec; var outf, lf: TFileStream);
(* Constructs a chain linking observed digraphs for start and end of thisperiod *)
var i,j,h,num,per,avl1,avl2,nummax,nrep,maxrep: integer;
  k: TMiniStep;
  sp: real;
  totdif: integer;
  a1,a2: integer;
  x: string;
begin
  if MadeChain then begin
    x := 'MakeChains was called, but there already was a chain. Exit.'+eol;
    lf.Write(pchar(x)^,length(x));
    exit;
   end;
  x := 'MakeChains.'+eol;
  lf.Write(pchar(x)^,length(x));
  SetAlpha( theta);


  for per := 1 to observations-1 do
    if (not skipi[per]) then begin
    initialise_actors(per);
    initialise_y(per);
    AdaptProposalProbabilities;
    for i:=1 to n do for j:=1 to n do
// @@@@@@ maak hier de eindsituatie van elke periode
     if ((i <> j) and ((ynm(i,j,per) <> ynm(i,j,per+1))) and (not Struct(i,j,per))) then
      begin
      if (chain[per-1].TotNumber <= 1) then k := Chain[per-1].Last
        else k := Chain[per-1].RandomElement(version);
      Insert(k,Network,i,j,1,0,per,outf,lf);
      if (errortext > '') then
        begin
        x := 'MakeChains: errortext 1 = '+errortext+eol;
        outf.Write(pchar(x)^,length(x));
        x := format('%s%d%s%d%s%d%s%d%s',['Error in MakeChain for period = ',per,' i = ',i,', j = ',j,', k = ',k.kindex,'.'])+eol;
        lf.Write(pchar(x)^,length(x));
        x := ErrorText+eol;
        lf.Write(pchar(x)^,length(x));
        exit;
        end;
      end; (* for i,j if i <> j and ynm <> *)

    for i:=1 to n do if StructActive[i] then for h := 1 to nza do
      begin
      a1 := truncaction(zc[i,h,per],h);
      a2 := truncaction(zc[i,h,per+1],h);
      if (a1 <> a2) then
        begin
        totdif := a2-a1;
        if (totdif > 0) then for j := 1 to totdif do
          begin
          if (chain[per-1].TotNumber <= 1) then k := Chain[per-1].Last
            else k := Chain[per-1].RandomElement(version);
          Insert(k,Behavior,i,0,h,1,per,outf,lf);
          end
         else for j := 1 to -totdif do
          begin
          if (chain[per-1].TotNumber <= 1) then k := Chain[per-1].Last
            else k := Chain[per-1].RandomElement(version);
          Insert(k,Behavior,i,0,h,-1,per,outf,lf);
          end;
        end; (* if a1 <> a2 *)
      end; (* for i, h *)
//wwriteln(outf,'puep');      
    RenewRateslprobs(Chain[per-1].First,Chain[per-1].Last,false,per,outf,lf);
//    RenewRateslprobs(Chain[per-1].First,Chain[per-1].Last,true,per,outf,lf);
//if errortext > '' then exit; // abcdefg
//wwriteln(outf,'poep');
// exit; // abcdefg
//acac := true;
//writeln(outf,'nstep = ', nstep);
    (* A bit of burn in: burn in will increase average chain length;
  avl1 is sum of chain lengths in earlier batch of nummax MH steps,
  avl2 is the same in the last batch;
  stop if (avl2 < 1.1 * avl1) *)
    nummax := 500;
    maxrep := 10;
    nrep := 0;
    avl2 := nummax * Chain[per-1].TotNumber;
      repeat
      inc(nrep);
      avl1 := avl2;
      avl2 := 0;
      for num := 1 to nummax do
        begin
        MH_Step(per,outf,lf,0);
        Inc(avl2,Chain[per-1].TotNumber);
        if (errortext > '') then
          begin
          x := format('%s%d',['MC  num = ',num])+eol;
          outf.Write(pchar(x)^,length(x));
          x := 'Error in MakeChains: '+errortext+eol;
          outf.Write(pchar(x)^,length(x));
          exit;
          end;
        end;
      until ((avl2 < trunc(1.1*avl1)) or (nrep >= maxrep));
    if (avl2 > trunc(1.1*avl1)) then
      begin
      x := format('%s%d%s%d%s%d%s',['Burnin during procedure MakeChains for period ',per,
                 ' stopped after ',nrep*nummax,' steps with average chain length ',
                 avl2 div nummax ,','])+eol;
      lf.Write(pchar(x)^,length(x));
      x := format('%s%d%s',['while average chain length in preceding batch was ',avl1 div nummax,'.'])+eol;
      lf.Write(pchar(x)^,length(x));
      end
     else begin
      x := format('%s%d%s%d%s%d%s',['Burnin for period ',per,' was finished after ',nummax*nrep,
              ' MH steps, leading to average chain length of ',avl2 div nummax,'.'])+eol;
      lf.Write(pchar(x)^,length(x));
     end;
//wwriteln(outf,'piep');
    RenewRateslprobs(Chain[per-1].First,Chain[per-1].Last,false,per,outf,lf);
//    RenewRateslprobs(Chain[per-1].First,Chain[per-1].Last,true,per,outf,lf);
//if errortext > '' then exit; // abcdefg
//wwriteln(outf,'paep');
  MadeChain := true;
  nummax := 100;
      for num := 1 to nummax do
        begin
        MH_Step(per,outf,lf,0);
        if (errortext > '') then
          begin
          x := format('%s%d',['MC  num = ',num])+eol;
          outf.Write(pchar(x)^,length(x));
          x := 'Error in MakeChains: '+errortext+eol;
          outf.Write(pchar(x)^,length(x));
          exit;
          end;
        end;
//wwriteln(outf,'plep');
    RenewRateslprobs(Chain[per-1].First,Chain[per-1].Last,true,per,outf,lf);
//wwriteln(outf,'pkep');
    if (errortext > '') then exit;
    end; (* for per *)

//  for per :=1 to observations -1 do Writeperiod(per,outf,lf);

  if (not consistencycheck(outf,lf,true)) then
   begin
    errortext := 'Error: Last consistency check in MakeChains not OK.';
    x := 'MakeChains: errortext 5 = '+errortext+eol;
    outf.Write(pchar(x)^,length(x));
    x := 'Error in MakeChains.'+eol;
    lf.Write(pchar(x)^,length(x));
    x := ErrorText+eol;
    lf.Write(pchar(x)^,length(x));
    exit;
   end
    else begin
      x := 'Last consistency check in MakeChains OK.'+eol;
      lf.Write(pchar(x)^,length(x));
    end;

  MadeChain := true;
end; (* MakeChains *)

function MLModel.StepProb(const ka: TAspect; const ki, kj: Integer;
               const kv: byte; const kd: TDif;
               const theta: pvec;
               const esta: byte;
               var st: rrarray; var pp: realnvec;
               var prat: pmatrix;
               var cctotn, cctota: integer; var missing, error: boolean;
               var errstr: shortstring; var outf, lf: TFileStream): real;
(* This procedure calculates the probability of making step kk,
   the statistics st,
   the probabilities pp for other steps,
   with outcome 1 if the data is missing.
   If esta >= 30, it also calculates relative probabilities prat
   at distances epsi from theta.
   Then it makes the indicated step.
   This procedure uses the theta parameters, not the alpha parameters. *)
var j,hh,maxh: integer;
    probj, spp, coefficient: real;
    PermDir: B3array;
    m: integer;
    halterr: boolean;
    x: string;
begin
if (ki <= 0) then
begin
x := format('%s%d',['ki = ',ki,'!!!'])+eol;
outf.Write(pchar(x)^,length(x));
exit;
end;
  halterr := false;
  var_n_tot := ActiveNetworkEffects_l + ActiveNetworkEffects_f +
               ActiveNetworkEffects_g;
  if (ka = Network) then
(* This is about a choice among n possibilities j = 1..n *)
    begin
    if StructActive[ki] then inc(cctotn);               // how to handle rate parameters for missings??? @@@
    if ((not (mis(ki,kj,thisperiod) or mis(ki,kj,thisperiod+1) or Struct(ki,kj,thisperiod+1)))
      and StructActive[ki]) then
(* calculate pr[j] = probability that change in this ministep is w.r.t. j;
   for kk.jh = ki, (not Structactive[ki]) does not imply Struct(ki,kk.jh,thisperiod+1)) *)
       begin
       missing := false;
       for hh := NumberParameters downto 1 do for j:=n downto 1 do st[j,hh] := 0.0;
       if CheckChange then
       PermitChange(n,ki,thisperiod,maxdegree,ppermitted)
        else for j:=n downto 1 do ppermitted[j] := true;
       if anynetworkmissings then ForbidMissings(ki,thisperiod,ppermitted);
(* This means that conditional probabilities among the nonmissings are calculated *)
       ppermitted[ki] := (StaySameType >= 2);
       UtilityComponents_yn(ki,thisperiod,ActiveNetworkEffects_l,n,st,ppermitted,NewNetworkEffects);
(* now st[j, ActiveNetworkEffects_l + kk] = s_{2ijk} for all j and kk. *)
(*Evaluate multinomial logit probability.*)
       spp := 0.0;
       for j:=1 to n do if ppermitted[j] then  // dit wordt anders voor j :=n downto 1 ????
         begin
         probj := 0.0;
         for hh := (ActiveNetworkEffects_l + 1) to var_n_tot do
           begin
           coefficient := ComputeCoefficient(theta,ki,hh,RE); (*MS: Compute coefficient, which may be fixed or random.*)
           probj := probj + (coefficient * st[j,hh]);
           end;
         pp[j] := expo(probj);
         spp := spp + pp[j];
         end (* if ppermitted *)
       else pp[j] := 0.0;
       for j:=n downto 1 do pp[j] := pp[j]/spp;
       StepProb := pp[kj];
(* Now pr[j] = \bar r_(2\beta)(x,ih[kk],j) *)
       if (estage >= 30) then
         begin
         end;
       end (* if not missing *)
      else
       begin
       missing := true;
       StepProb := 1.0;
       for j:=n downto 1 do pp[j] := 0.0;
       pp[kj] := 1.0;
       end;
     ChangeTie(ki,kj);
     end (* if ka = network *)
    else
     begin (* if ka = Behavior *)
(* This is about a choice among 3 possibilities j = 1, 2, 3
   corresponding to differences -1, 0, +1. *)
     if ((not (miszc(ki,kv,thisperiod) or miszc(ki,kv,thisperiod+1)))
      and Structactive[ki]) then
       begin
(* calculate pr[j] = probability that change in this ministep is j *)
       inc(cctota);
       missing := false;
       for hh := NumberParameters downto 1 do for j:=1 to 3 do st[j,hh] := 0.0;
       PermDir[1] := (not Minimalya(ki,kv));  (* change -1 permitted *)
       PermDir[2] := true;                  (* change  0 permitted *)
       PermDir[3] := (not Maximalya(ki,kv));  (* change +1 permitted *)
       UtilityComponents_ya(ki,kv,BeforeActionParameters_f[kv],
                            st,PermDir,halterr,outf,lf);
       if halterr then
         begin
         error := true;
         errstr := 'Error in StepProb: Action Utility Components.';
         exit;
         end;
(* now st[d+2, BeforeActionParameters_f[kv] + k] = s_{2ihdk} for all d,k *)
(*Evaluate multinomial logit probability.*)
       spp := 0.0;
       for j:=1 to 3 do if PermDir[j] then
         begin
         probj := 0.0;
         maxh := BeforeActionParameters_f[kv+1];
         for hh := (BeforeActionParameters_f[kv] + 1) to maxh do
           begin
           coefficient := ComputeCoefficient(theta,ki,hh,RE); (*MS: Compute coefficient, which may be fixed or random.*)
           probj := probj + (coefficient * st[j,hh]);
           end;
         pp[j] := expo(probj);
         spp := spp + pp[j];
         end (* if permDir *)
       else pp[j] := 0.0;
       for j:=1 to 3 do pp[j] := pp[j]/spp;
(* Now pr[j] = \bar r_(2\beta) *)
       StepProb := pp[kd+2];
       end (* not miszc *)
      else
       begin
       missing := true;
       StepProb := 1.0;
       for j:=1 to 3 do pp[j] := 0.0;
       pp[kd+2] := 1.0;
       end;
     ChangeBehavior(ki,kv,kd);
     end; (* if ka = Behavior *)
end; (* StepProb *)


procedure MLModel.Geyer(const npar, nchain: integer;
                        var theta: pvec; var cc: pmatrix;
                        var outf,lf: TFileStream);
var i, j, h, per, k: integer;
  errstr: shortstring;
  kv: array of shortint;
  ki, kj: array of integer;
  kd: array of shortint;
  np: array of integer;  (* cumulative number of ministeps per period *)
  thetaj: pvec;
  lprob, rate, s1, cs, scs, adif: real;
  dumm, rep: integer;
  aah: real;
  pr: realnvec;
  s: rrarray;
  b, w, w0: array of real;
  swhah: array of array of real;
  x: string;
procedure readnextchain;
var ii,jj,hh,kk,per,kih,kjh: integer;
    a: char;
    endofchain: boolean;
    kvarh: shortint;
    kdifh: TDif;
begin
(* Read the observations-1 chains from ChainFile *)
    readln(ChainFile); (* A *)
    for jj:=numberparameters downto 1 do readln(ChainFile); (* parameter, is known already *)
    kk := 0;
    np[0] := 0;
    readln(ChainFile); (* P 1 *)
    for per := 1 to observations-1 do
      if (not skipi[per]) then begin
      repeat
        read(ChainFile, a);
        case a of
        'N': begin
             readln(ChainFile,kih,kjh,kvarh);
             inc(kk);
             kv[kk] := -kvarh;
             ki[kk] := kih;
             kj[kk] := kjh;
             endofchain := false;
             end;
        'B': begin
             readln(ChainFile,kih,kjh,kvarh,kdifh);
             inc(kk);
             kv[kk] := kvarh;
             ki[kk] := kih;
             kj[kk] := kjh;
             kd[kk] := kdifh;
             endofchain := false;
             end;
          else endofchain := true; (* ' P <per+1> ' or 'L' *)
         end; (* case *)
        until endofchain;
        np[per] := kk;
      end; (* per *)
    readln(ChainFile,lprob); (* the lprob is not used, but this checks that you don't hit a letter *)
end; (* readnextchain *)
procedure Calchah(const thetaj: pvec; var ah: real);
(* calculates ah as a constant (depending not on the parameter)
   plus the log of the joint probability
   for  the current chain under parameter vector thetaj *)
var ii,jj,hh,kk,per: integer;
  ctota: array[0..nzmax] of integer;
  ctotn, dumm: integer;
  missing, error: boolean;
  rp: pmatrix; (* dummy, only for parameter correspondence *)
begin
  setlength(pr,n+1);
  ah := 0.0;
  for per := 1 to observations-1 do
    if (not skipi[per]) then begin
    initialise_actors(per);
    Initialise_y(per);
    ctotn := 0;
    for kk:= 0 to nza do ctota[kk] := 0;
    SetData(Chain[per-1].First,per,outf,lf);
(* for per >= 2, this can change only missings and structurally determined values *)
    for kk:= np[per-1]+1 to np[per] do
      begin
      if (kv[kk] <= 0) then
        begin
        ah := ah +
             log(StepProb(Network, ki[kk], kj[kk], 0,0,
             thetaj, 0, s, pr, rp, ctotn, dumm,
             missing, error, errstr, outf, lf));
        end
       else
        begin
        ah := ah +
          log(StepProb(Behavior, ki[kk], 0, kv[kk], kd[kk],
             thetaj, 0, s, pr, rp, dumm, ctota[kv[kk]],
             missing, error, errstr, outf, lf));
        end;
      end; (* kk *)
(* constant rates: *)
      rate := //networklambda(1,error,outf,lf);  this requires SetAlpha!
        thetaj[ActiveObservationRank[per]];
      ah := ah - n_act*rate + ctotn*log(rate);
      for kk :=1 to nza do
        begin
        rate := //actionlambda(1,kk,error,outf,lf); this requires SetAlpha.
         thetaj[var_n_tot + (kk-1)*ActivePeriods + ActiveObservationRank[per]];
        ah := ah - n_act*rate + ctota[kk]*log(rate);
        end;
    end; (* for per *)
end; (* Calcah *)
function stepc(const c: real): real;
(* one step in solution of sum_j nchain/(b[j] + c) = 1
   by linear approximation *)
var s0,s1: real;
    jj: integer;
begin
  s0 := 0.0;
  s1 := 0.0;
  for jj:= npar downto 1 do
    begin
    s0 := s0 + nchain/(b[jj] + c);
    s1 := s1 + nchain/sqr(b[jj] + c);
    end;
  stepc := (s0-1.0)/s1;
end; (* stepc *)
function obj: real;
(* the objective function maximized  *)
var s0,s1: real;
    ii,jj,hh: integer;
begin
  s0 := 0.0;
  s1 := 0.0;
  for ii:= npar downto 1 do for hh := nchain downto 1 do
    begin
    s1 :=0.0;
    for jj := npar downto 1 do s1:= s1 + w[jj]*hah[ii,jj,hh];
    s0 := s0 + log(w[ii]*hah[ii,ii,hh]/s1);
    end;
  obj := s0;
end; (* obj *)
function simloglik(const th: pvec): Real;
(* simulated approximation to minus loglikelihood *)
var s0,s1: real;
    ii,jj,hh: integer;
    x : string;
begin
//  s1 := 0.0;
//  for ii := 1 to numberparameters do s1 := s1 + sqr(th[ii])*(10 + sin(th[i]));
//  simloglik := s1;
//  exit;

  reset(ChainFile);
  s1 := 0.0;
(* calculate probabilities of the npar*nchain chains for parameter vector th *)
  for ii:=1 to npar do
    begin
    s0 := 0.0;
    for hh:=1 to nchain do
      begin
(* Read the next observations-1 chains from ChainFile *)
      readnextchain;
(* Calculate the probabilities of the chain according to parameter th *)
(* This procedure uses theta directly, and does not use the Alpha parameters.
   Therefore SetAlpha is not used. *)
      Calchah(th,aah);
(* aah is a constant (depending not on the parameter)
   plus the log of the joint probability
   for  chain (ii,hh) under parameter vector th *)
      s0 := s0 + expo(aah)/hah[ii,0,hh];
      end; (* hh *)
    case weightversion of
    1: s1 := s1 + s0/nchain;  (* geyer *)
    2: s1 := s1 + s0/w[ii]; (* snijders *)
    end; (* case *)
    end; (* ii *)
  simloglik := - log(s1);
  outf.Write(pchar(eol)^,length(eol));//writeln(outf);
for jj := 1 to numberparameters do begin
  x := format('%6.4f%s',[th[jj],' ']);
  outf.Write(pchar(x)^,length(x));
  end;
outf.Write(pchar(eol)^,length(eol));//writeln(outf);
x := format('%s%12.8f',['- loglik = ',-log(s1)])+eol;
outf.Write(pchar(x)^,length(x));
end; (* simloglik *)
(* =========== *)
procedure NRMinimize(p, mode: byte; var theta: pvec; var Cov: pmatrix;
                     var cf: TFileStream);
(* Mini !!! Maximization of function simloglik *)
(* Mode 0: one Quasi-Newton Raphson step;
   Mode 1: iterated Quasi-Newton Raphson steps,
           but in each step, if NR is not indicated,
           a steepest descent step is made;
   Mode 2: iterated steepest descent steps;
   Mode 3: first multi steepest descent steps,
           then NR steps;
           in each step where NR is not indicated,
           multi steepest descent steps are made. *)
 (* theta is the input (initial) and output parameter vector *)
const epsi = 1e-12; (* Convergence criterion *)
      eps = 0.001;  (* Argument difference for approximating derivatives *)
      nstepmax = 2; (* maximum number of iteration steps *)
type
    pvecplus = array[1..pmax] of double;
    pmatrixplus = array[1..pmax,1..pmax] of double;
var funct0,thefunction: double;
  ste: Boolean;
  i,j,multi,nsteps: integer;
  thet: pvec;     (* the working parameter *)
  insensitive: array[1..pmax] of Boolean;
  x : string;
procedure ThetaIncrement(i: byte; r: real);
begin
  thet[i] := thet[i] + r;
end; (* ThetaIncrement *)
procedure VecSum(a,b: pvecplus; var c: pVec; r: real);
var i: byte;
begin
  for i:=1 to p do c[i] := a[i]+r*b[i];
end; (* VecSum *)
procedure NormLength(var a: pVec; var s: real);
var i: byte;
begin
  s := 0.0;
  for i:=1 to p do s := s + sqr(a[i]);
  s := root(s);
  if (s > 0.0) then for i:=1 to p do a[i]:=a[i]/s;
end; (* NormLength *)
procedure NewtonRaphsonStep(option: byte; var stepp: boolean);
(* stepp indicates whether a step has been made *)
var i,j: byte;
    theta0: pVecplus;
    d1,step: pvecplus;
    d2: pmatrixplus;
    d2r: pmatrix;
    d0: double;
    s: real;
    ert: shortstring;
    outs: boolean;
    x : string;
procedure NoStepNoCov;
var i,j: byte;
begin
  stepp := false;
  for i:=1 to p do for j:=1 to p do Cov[i,j] := 9999.0;
end;
begin (* NewtonRaphsonStep *)
  inc(nsteps);
  stepp := false;
  for i:=1 to pmax do theta0[i] := thet[i];
  d0 := simloglik(thet);
(* approximate first and second derivatives by difference quotients *)
  for i:=1 to p do //if (not fix[i]) then
    begin
    ThetaIncrement(i,eps);
    d1[i] := simloglik(thet);
    ThetaIncrement(i,-eps);
   end;           //  else d1[i] := d0;
  for i:=1 to p do for j:=1 to i do
//   if (not(fix[i] or fix[j])) then
    begin
    ThetaIncrement(i,eps); ThetaIncrement(j,eps);
    d2[i,j] := (simloglik(thet) - d1[i] - d1[j] + d0)/sqr(eps);
    if (j < i) then d2[j,i] := d2[i,j];
    ThetaIncrement(i,-eps); ThetaIncrement(j,-eps);
    end (* for i, j if *);
//     else begin
//     d2[i,j] := 0.0;
//     if (j < i) then d2[j,i] := 0.0 else d2[i,i] := 1/9999.0;
//    end;
  for i:=1 to p do d1[i] := (d1[i]-d0)/eps;
(* Now d1 is vector of 1st, d2 matrix of 2nd derivatives. *)
  for i:=1 to pmax do for j:=1 to pmax do d2r[i,j] := d2[i,j];
  for i:=1 to p do if (d2[i,i] <= 0.0) then begin
    insensitive[i] := true;
    x := format('%s%d%s%8.6f',['Function insensitive for parameter ',i,', ',d2[i,i]])+eol;
    cf.Write(pchar(x)^,length(x));
    d2r[i,i] := 0.1;
   end (* for i if *) else insensitive[i] := false;
  pinverse(p,d2r,Cov,ert);
  if (ert > '') then begin
(*   Matrix inversion error in Quasi-Newton Raphson procedure. *)
    (* add a ridge *)
    x := 'Error in matrix inversion: '+ert+eol;
    cf.Write(pchar(x)^,length(x));
    for j:=1 to p do d2r[j,j] := d2r[j,j] + 1.0;
    pinverse(p,d2r,Cov,ert);
    if (ert > '') then begin
      x := 'Still error in matrix inversion: '+ert+eol;
      cf.Write(pchar(x)^,length(x));
      NoStepNoCov; exit;
     end;
   end; (* if *)
(* If one of the diagonal elements of Cov is negative, the thefunction function
   is not locally convex and a Quasi-Newton Raphson step will not be made: *)
  outs := false;
  for i:=1 to p do if (//(not fix[i]) and
                     (Cov[i,i] <= 0)) then
    begin outs := true; end;
  if outs then begin (* NoStepNoCov; *) exit; end;

(* Calculate the step to be made and make it: *)
  for i:=1 to p do //if fix[i] then step[i] := 0.0 else
    begin
    step[i] := 0.0;
    for j:=1 to p do step[i] := step[i] - Cov[i,j]*d1[j];
    end; (* for i if *)
  VecSum(theta0,step,thet,1.0);
  thefunction := simloglik(thet);

(* Check whether the step is an improvement;
   if not, take action depending on option. *)
  case option of
  1: begin
       s := 1.0;
       while ((thefunction >= d0) and (s > 1e-20)) do begin
    (* halve the step *)
         s := 0.5*s;
         VecSum(theta0,step,thet,s);
         thefunction := simloglik(thet);
        end; (* while *)
       stepp := (thefunction < d0);
     end; (* case 1 *)
   2: if ((thefunction-d0) > 0.0) then begin
(* skip the step that was set *)
        for i:=1 to pmax do thet[i] := theta0[i];
        stepp := false;
        thefunction := d0;
       end else stepp := true;
    end; (* case *)
  x := format('%s%20.10f',['End Quasi-Newton Raphson step, function value = ', thefunction])+eol;
  cf.Write(pchar(x)^,length(x));
end; (* NewtonRaphsonStep *)

procedure Descentstep(var step: boolean);
(* step indicates whether a step has been made *)
var i: byte;
  d0,d1,d2: double;
  length: real;
  de1,theta0: pVecplus;
  de1r: pvec;
  x: string;
begin
  inc(nsteps);
  step := false;
  d0 := simloglik(thet);
(* approximate first derivatives *)
  for i:=1 to p do //if not fix[i] then
    begin
    ThetaIncrement(i,eps);
    de1[i] := (simloglik(thet)-d0)/eps;
    ThetaIncrement(i,-eps);
    end;          // else de1[i] := 0.0;
  for i:=1 to p do de1r[i] := de1[i];
  NormLength(de1r,length);
  if (length < 1e-10) then exit;
(* de1 is the direction of steepest ascent.
   Calculate first and second derivatives in this direction,
   and then calculate length := stepsize according to
   onedimensional Quasi-Newton Raphson method. *)
  for i:=1 to pmax do theta0[i] := thet[i];
  VecSum(theta0,de1,thet,eps);
  d1 := simloglik(thet);
  VecSum(theta0,de1,thet,2.0*eps);
  d2 := simloglik(thet);
  if (abs(d0 - 2.0*d1 + d2) < 1e-20) then exit else
    length := (eps*(d1-d0))/(d0 - 2.0*d1 + d2);
(* if length < 0.0 then the 1-dimensional function is not locally convex,
   and a Newton-Raphson step size makes no sense. *)
  if (length <= 0.0) then begin length := 1.0; end;
  i:= 0;
  repeat
(* Make the step and halve the step size if the step yields no improvement. *)
    inc(i);
    VecSum(theta0,de1,thet,-length);
    length := 0.5*length;
    thefunction := simloglik(thet);
   until ((thefunction <= d0) or (i >= 20));
  if (thefunction >= d0) then begin
    for i:=1 to pmax do thet[i] := theta0[i];
    thefunction := d0; step := false;
   end else step := true;
  x := format('%s%20.10f',['Descent step, function value = ', thefunction])+eol;
  cf.Write(pchar(x)^,system.length(x));
end; (* Descentstep *)

begin (* NRMinimize *)
  nsteps := 0;
  thet := theta;
  case mode of
  0: NewtonRaphsonstep(1,ste);
  1: begin
      thefunction := simloglik(thet);
        x := format('%s%20.10f',['Initial function value = ', thefunction])+eol;
        cf.Write(pchar(x)^,length(x));
      for i:=1 to 5 do NewtonRaphsonstep(1,ste);
      repeat
        funct0 := thefunction;
        NewtonRaphsonstep(1,ste);
        if (not ste) then DescentStep(ste);
        thefunction := simloglik(thet);
      until ((abs(thefunction - funct0) < epsi) or (nsteps > nstepmax));
      NewtonRaphsonstep(2,ste);
     end; (* case 1 *)
  2: begin
      thefunction := simloglik(thet);
      repeat
        funct0 := thefunction;
        DescentStep(ste);
        thefunction := simloglik(thet);
      until ((abs(thefunction - funct0) < epsi) or (nsteps > nstepmax));
      NewtonRaphsonstep(2,ste);
     end; (* case 2 *)
   3: begin
       multi := p;
       if (multi < 5) then multi := 5;
       thefunction := simloglik(thet);
       for i:=1 to multi do Descentstep(ste);
       repeat
         funct0 := thefunction;
         NewtonRaphsonStep(1,ste);
         thefunction := simloglik(thet);
         if (not ste) then begin
         (* no NR step was set, therefore some DC steps are made. *)
           i := 0;
           repeat
             funct0 := thefunction;
             inc(i);
             Descentstep(ste);
             thefunction := simloglik(thet);
            until (i >= multi);
          end; (* if not ste *)
        until ((abs(thefunction - funct0) < epsi) or (nsteps > nstepmax));
      end; (* case 3 *)
  end; (* case *)
  for i:=1 to p do for j:=1 to p do
    if (insensitive[i] or insensitive[j]) then Cov[i,j] := -1.0;
  theta := thet;
  if (mode >= 1) then begin
    x := format('%s%d%s',['Number of steps = ',nsteps,'.'])+eol;
    cf.Write(pchar(x)^,length(x));
    end;
end; (* NRMinimize *)
(* ============ *)

begin (* Geyer *)
  weightversion := 2;
  x := format('%s%d',['Maximal chain length ',maxlength])+eol;
  lf.Write(pchar(x)^,length(x));
  reset(ChainFile);
  setlength(hah,npar+1,npar+1,nchain+1);
  setlength(kv,maxlength+1);
  setlength(ki,maxlength+1);
  setlength(kj,maxlength+1);
  setlength(kd,maxlength+1);
  setlength(np,observations+1);
  setlength(s,nmax+1,NumberParameters+1); (* note that always n >= 3, which is required for the behavior part *)
  var_n_tot := ActiveNetworkEffects_l + ActiveNetworkEffects_f +
               ActiveNetworkEffects_g;

(* calculate probabilities of the npar*nchain chains
   for all npar parameter vectors *)
  for i:=1 to npar do for h:=1 to nchain do
    begin
(* Read the next observations-1 chains from ChainFile *)
    readnextchain;
(* Calculate the probabilities of the chain according to each parameter vector *)
    for j:=1 to npar do
      begin
      for k := numberparameters downto 1 do thetaj[k] := thetas[j,k];
(* This procedure uses theta directly, and does not use the Alpha parameters.
   Therefore SetAlpha is not used. *)
      Calchah(thetaj,aah);
      hah[i,j,h] := expo(aah);
(* hah[i,j,h] is a constant (depending not on the parameter)
   times the joint probability
   for  chain (i,h) under parameter vector j *)
      x := format('%3d%s%3d%s%3d%s%20.12f',[i,' ',j,' ',h,' ',hah[i,j,h]])+eol;
      outf.Write(pchar(x)^,length(x));
      end; (* j *)
    outf.Write(pchar(eol)^,length(eol));//writeln(outf);
    end; (* i, h *)

(* Calculate weights *)
  SetLength(w,npar+1);
  SetLength(w0,npar+1);
  SetLength(b,npar+1);
  SetLength(swhah,npar+1,nchain+1);
  w[0] := 0.0;
  for i:=1 to npar do
    begin
    w[i] := 1.0/npar;
    w0[i] := 0.0;
    end;
  cs := 0.0;
  rep := 0;
    repeat
(* Solve for w from 1st order equation *)
    inc(rep);
    for i:=1 to npar do for h:=1 to nchain do
      begin
      s1 := 0.0;
      for j:= npar downto 1 do s1 := s1 + w[j]*hah[i,j,h];
      swhah[i,h] := s1;
      end; (* i, h *)
    for j:=1 to npar do
      begin
      s1 := 0.0;
      for i:= npar downto 1 do for h:= nchain downto 1 do
                            s1 := s1 + hah[i,j,h]/swhah[i,h];
      b[j] := s1;
      end; (* j *)
    dumm := 0;
      repeat
(* Find cs such that the sum of the new w is 1;
   The previous cs is the new starting value. *)
      scs := stepc(cs);
      cs := cs + scs;
      inc(dumm);
      until ((abs(scs) < 1e-10) or (dumm > 20));
x := format('%s%d',['rep = ',rep])+eol;
outf.Write(pchar(x)^,length(x));
x := format('%s%f',['cs = ',cs])+eol;
outf.Write(pchar(x)^,length(x));
    adif := 0.0;
    for j:=1 to npar do
      begin
      w[j] := nchain/(b[j] + cs);
      adif := adif + abs(w[j]-w0[j]);
      w0[j] := w[j];
x := format('%3d%s%10.6f',[j,'. ',w[j]]);
outf.Write(pchar(x)^,length(x));
      end;
outf.Write(pchar(eol)^,length(eol));//writeln(outf);
x := format('%s%20.10f',['objective = ',obj])+eol;
outf.Write(pchar(x)^,length(x));
    until ((adif < 1e-10) or (rep >= 20));
(* Now w has been found so that it maximizes obj *)
{
for i:=1 to 100 do
begin
s1 := 0.0;
for j:=1 to npar do
begin
w[j] := rando(version);
s1 := s1+w[j];
end;
writeln(outf,i);
for j:=1 to npar do write(outf,j:3,'. ',w[j]/s1:10:6);
writeln(outf);
writeln(outf,'objective = ',obj:20:10);
end;
}
(* Calculate the weights for the npar*nchain chains *)
  for i:=1 to npar do for h:=1 to nchain do
    begin
    s1 := 0.0;
      case weightversion of
      1: for j:= npar downto 1 do s1 := s1 + hah[i,j,h]*w[j]; (* geyer *)
      2: for j:= npar downto 1 do s1 := s1 + hah[i,j,h]; (* snijders *)
      end; (* case *)
    hah[i,0,h] := s1;
//writeln(outf,'hah0 ',i:3,h:3,s1:20:10);
    end;

  x := format('%s%20.4',['initial deviance ',2*simloglik(theta)])+eol;
  outf.Write(pchar(x)^,length(x));
  NRMinimize(numberparameters, 1, theta, cc, outf);
  x := format('%s%20.4f',['final deviance ',2*simloglik(theta)])+eol;
  outf.Write(pchar(x)^,length(x));
end; (* Geyer *)
(* >>>>> *)

procedure MLModel.Scores(p: Byte; var theta: pVec;
                allstats, varmat: boolean;
                var fseed: integer; var ystat: pVec; var d: array of real;
                var dd,relp: pmatrix;
                updatestat: Boolean;
                var feasible, setpar, error: Boolean;
                var errstr: shortstring; var outf, lf: TFileStream;
                var estage: byte);
(* This procedure is the ML version of FRAN/SimStats,
   and it is attempted to have similar functionality.
   setpar indicates that before entering this procedure, the parameter may have changed,
   and therefore the rates and lprobs in Chain must be renewed.
   When in doubt, set setpar = true.
   It simulates one chain for each period and stores the outcome of the
   complete data score function in ystat,
   and if varmat stores the complete data observed information matrix in dd. *)
(* There will be runtime errors if MakeChain has not been called earlier. *)

var sp,probj: real;
    i,j,iteration,h,hh,hhh,period,aor: integer;
    success, check,halter: Boolean;
    statsum,statk,wmean: pvec;
    ctota: array[1..nzmax] of integer;
    ctotn: integer;
    ddsum,w2mean: pmatrix;
    pr: realnvec;
    k: TMiniSTep;
    pp,cc,ct: byte;
    s: rrarray;
    constantraten: array of real;
    constantratea: array of array of real;
    r1: array[1..100] of real;
    sumlprob: real;
    missing: boolean;
    totlength: integer;
    name: shortstring;
    f: textfile;
begin (* Scores *)

  Error := false; errstr := '';
  if (observations <= 1) then begin
    errstr:= 'observations = '+inttostr(observations)+' in Scores';
    error := true;
    exit;
   end;

  if not MadeChain then begin
    errstr:= 'No chain is available in Scores';
    error := true;
    exit;
   end;

  halter := false;
  pp := NumberParameters;
  Setlength(s,nmax+1,NumberParameters+1); (* note that always n >= 3, which is required for the behavior part *)
  SetLength(Constantraten,observations);
  SetLength(Constantratea,observations,nza+1);
  setlength(pr,nmax+1);

(* Order of the pp parameters in theta:
   For parameters in theta, the order is:
   first ActiveNetworkEffects_l network rate effects;
   then ActiveNetworkEffects_f  network evaluation effects;
   then ActiveNetworkEffects_g  network endowment effects;
   then ActiveActionEffects_l   action  rate effects;
   then ActiveActionEffects_f   action  evaluation effects;
   then the action endowment effects.
   This is determined in SetAlpha and SetTheta. *)

  var_a_min := BeforeActionParameters_f[1];
  var_n_tot := ActiveNetworkEffects_l + ActiveNetworkEffects_f +
               ActiveNetworkEffects_g;

  for i:= 1 to pp do
    begin
    statsum[i] := 0.0;
    for j:= 1 to pp do ddsum[i,j] := 0.0;
    end;

  if setpar then InitialiseAlpha(outf,theta,fseed);

  (*MS*)
  if randomeffects then
    begin
    Get_Cov(p,theta); (*Initialise functions of (co)variance parameters.*)
    if (estage in [11, 21, 31]) then SetProposalRE(false,nmax,nre,re,eb,lastmvn0,reflectionre,sumsigma,nno); (*Prepare some things to be exploited by the candidate-generating density of the actor-dependent random effects.*)
    end;
  if fine_tuning then (*Start of phase 1.*)
    begin
    if (Bayesian or randomeffects) then ImproveMH(outf,lf,theta,p); (*Method to improve efficiency of Metropolis-Hastings algorithm by improving acceptance rate.*)
    fine_tuning := false;
    end;
  if (estage = 31) then (*Start of phase 3.*)
    begin
    if (maxlike and randomeffects) then InitialiseMS_PhaseIII; (*Must be called at the beginning of phase 3, and must be immediately preceded by <Get_Cov_Theta>.*)
    end;

  MCMC_cycle(setpar,outf,lf,p,theta,estage,0);
  (*Cycle through
  (1) TS:   M-H step to sample path: network and behavior (ML and Bayesian: fixed effects models).
  (2) MS:   M-H step to sample actor-dependent random effects (ML and Bayesian: random effects models).
  (3) MS:   Gibbs and M-H steps to sample parameters (Bayesian: fixed effects and random effects models).*)
  if (Bayesian or randomeffects) then Post_MCMC_cycle(p,theta,estage,outf,lf); (*MS: Some functions of the latent data are stored.*)
  (*MS*)

  if ((Bayesian or randomeffects) and (estage = 33)) then LastPath(p,theta,true); (*666*) (*Writes last path sampled from the conditional distribution of the paths to ASCII file.*)

(* Evaluation of complete-data score function ystat
   and complete-data information matrix dd *)

  if maxlike then
    begin
    if writechain then
      begin
      writeln(chainfile,'A ');
      for i:=1 to p do writeln(chainfile,theta[i]:12:8);
      totlength := 0;
      end;
    sumlprob := 0.0;
    for period :=1 to observations-1 do
      if (not skipi[period]) then
      begin
      for i:=1 to nzmax do ctota[i] := 0;  (* total number of steps of each dependent behavior variable *)
      ctotn := 0;                          (* total number of network steps *)
      initialise_actors(period);
      Initialise_y(period);
      AdaptProposalProbabilities;
(* Move ministep k from the first to the last in this period. *)
      k := Chain[period-1].First.Such;
      SetData(Chain[period-1].First,period,outf,lf);
      if writechain then
        begin
        writeln(chainfile,'P ',period);
        inc(totlength,Chain[period-1].TotNumber);
        end;
      repeat
      if writechain then k.writestep(chainfile);
      sumlprob := sumlprob + k.lprob;
      if (k.ash = Network) then
(* This is about a choice among n possibilities j = 1..n *)
        begin
        StepProb(k.ash, k.ih, k.jh, k.varh, k.difh,
                 theta, estage, s, pr, relp, ctotn, ctota[k.varh],
                 missing, error, errstr, outf, lf);
(* Now s are the statistics, pr[j] = \bar r_(2\beta)(x,ih[k],j) *)
        if not missing then
          begin
          for h := ActiveNetworkEffects_l + 1 to var_n_tot do
            begin
            wmean[h] := 0.0;
            for j:=n downto 1 do wmean[h] := wmean[h] + s[j,h]*pr[j];
              if varmat then for hh := h to var_n_tot do
                begin
                w2mean[h,hh] := 0.0;
                for j:=n downto 1 do w2mean[h,hh] := w2mean[h,hh] + s[j,h]*s[j,hh]*pr[j];
                end;
            statsum[h] := statsum[h] + s[k.jh,h] - wmean[h];
            end; (* for h *)
          if varmat then
           for h := ActiveNetworkEffects_l + 1 to var_n_tot do
            for hh := h to var_n_tot do
              ddsum[h,hh] := ddsum[h,hh] + wmean[h]*wmean[hh] - w2mean[h,hh];
          end; (* if not missing *)
        end (* if k.ash = network *)
       else
        begin (* if k.ash = Behavior *)
(* This is about a choice among 3 possibilities j = 1, 2, 3
   corresponding to differences -1, 0, +1. *)
        StepProb(k.ash, k.ih, k.jh, k.varh, k.difh,
                 theta, estage, s, pr, relp, ctotn, ctota[k.varh],
                 missing, error, errstr, outf, lf);
        if error then exit;
(* Now s[d+2, var_a_min + k] = s_{2ihdk} for all d,k, *)
(*     pr[j] = \bar r_(2\beta) *)
        if not missing then
          begin
          for hh := var_a_min + 1 to pp do
            begin
            wmean[hh] := 0.0;
            for j:=1 to 3 do wmean[hh] := wmean[hh] + s[j,hh]*pr[j];
            if varmat then for hhh := hh to pp do
              begin
              w2mean[hh,hhh] := 0.0;
              for j:=1 to 3 do
                w2mean[hh,hhh] := w2mean[hh,hhh] + s[j,hh]*s[j,hhh]*pr[j];
              end;
            statsum[hh] := statsum[hh] + s[k.difh+2,hh] - wmean[hh];
            end; (* for hh *)
          if varmat then for hh := var_a_min + 1 to pp do
            for hhh := hh to pp do
              ddsum[hh,hhh] := ddsum[hh,hhh] + wmean[hh]*wmean[hhh] - w2mean[hh,hhh];
          end; (* not miszc *)
        end; (* if k.ash = Behavior *)
        k := k.such;
      until k.Extreme;
    sumlprob := sumlprob + Chain[period-1].kappa;

(* Requirement of the model specification is that for each period
   there is only one parameter for the rate function,
   being the logarithm of the <<constant>> Poisson parameter,
   and given by  networklambda(1,error,outf,lf) ;
   i.e., ActiveNetworkEffects_l = observations-1. *)
(* Network rate: *)
      aor := ActiveObservationRank[period];
      constantraten[period] := networklambda(1,error,outf,lf);
//      statsum[period] :=  ctotn - n_act*constantraten[period];
      statsum[aor] :=
        statsum[aor] - n_act + ctotn/constantraten[period]; // @@@@@
(* Behavior rates: *)
      for h := 1 to nza do
        begin
        constantratea[period,h] := actionlambda(1,h,error,outf,lf);
        statsum[var_n_tot + ActivePeriods*(h-1) + aor] :=
//          ctota[h] - n_actnonmis[h]*constantratea[period,h];
          statsum[var_n_tot + ActivePeriods*(h-1) + aor] // @@@@@
          - n_actnonmis[h] + ctota[h]/constantratea[period,h];
        end;
//write(outf,period,' ',  ctota[1] - n_actnonmis[h]*constantratea[period,1]:8:4,' ',
//         theta[var_n_tot+period]:8:4,' ', ctota[1]:4); // @@@
//if (period = 1) then write(outf,'   ') else writeln(outf);
//writeln(outf,' <<<<< ',period,ndian:6,ndiab:6,CHain[period-1].DiagNumber:6, Chain[period-1].TotNumber:6); // controle
//if (period = 1) then write(outf,'   ') else writeln(outf);

(* Derivatives of score function for rate parameters: *)
      if varmat then
        begin
        ddsum[aor,aor] := ddsum[aor,aor]   // @@@@@
           - ctotn/sqr(constantraten[period]);
        // - n_act*constantraten[period];      // @@@
        for h:=1 to nza do
          ddsum[var_n_tot + ActivePeriods*(h-1) + aor,
                var_n_tot + ActivePeriods*(h-1) + aor]
//            := -n_actnonmis[h]*constantratea[period,h];                   // @@@
              := ddsum[var_n_tot + ActivePeriods*(h-1) + aor,
                var_n_tot + ActivePeriods*(h-1) + aor]   // @@@@@
              - ctota[h]/sqr(constantratea[period,h]);
        end;
      end; (* for period *)
    if writechain then
      begin
      writeln(chainfile,'L ');
      writeln(chainfile,sumlprob:20:8);
      if (totlength > maxlength) then maxlength := totlength;
      end;

(*
   if not consistencycheck(outf,lf,true) then
     begin
     writeln(lf,'Wrong data for this chain at end of period ',thisperiod,'.');
     error := true;
     ErrStr := 'Chain does not lead to the correct network end situation.';
     exit;
     end;
*)
      
      for h:= 1 to pp do ystat[h] := statsum[h];
      if varmat then for h:=1 to pp do
        begin
        for hh := h to pp do dd[h,hh] := -ddsum[h,hh];
        for hh := 1 to h-1 do dd[h,hh] := dd[hh,h];
        end;

    if randomeffects then CD_Score_Hessian_RE(p,varmat,ystat,dd); (*MS: complete Hessian matrix.*)

    (*MS: temporary code:*) (*666*)
    {
    if randomeffects then
      begin
      name := 'ScoreCovRE.m';
      if (estage = 31) then OpenFile(f,name);
      h := p - covariance_parameters;
      for i := (h + 1) to p do write(f,ystat[i]:10:6);
      writeln(f);
      if (estage = 33) then closefile(f);
      end;
    }
    (*MS*)

    end (*if maxlike*)
   else varmat := false;

  fseed := readstate;

  for period :=1 to observations-1 do
    if (not skipi[period]) then
    begin
    if (Chain[period-1].ErrorMes > '')
     then ErrorText := ErrorText + Chain[period-1].ErrorMes;
    end;
  if (ErrorText > '') then
    begin
    error := true;
    errstr := 'Error in Scores: '+ErrorText;
    end;

  Application.ProcessMessages; (* if Scores takes less time, this need not be called so often *)

end; (* Scores *)


procedure MLModel.FRAN(    p          : Byte;
                        var theta      : pVec;
                            allstats,
                            varmat     : boolean;
                        var fseed      : integer;
                        var ystat      : pVec;
                        var d          : array of real;
                        var dd         : pmatrix;
                            updatestat : Boolean;
                        var feasible,
                            error      : Boolean;
                        var errstr     : shortstring;
                        var outf,
                            lf         : TFileStream;
                        var estage     : byte;
                        var pc: boolean);
var i: integer; relp: pmatrix; setp: boolean;
begin
  if (MaxLike or Bayesian)
  then begin
    setp := true;
    Scores(p, theta, allstats, varmat, fseed, ystat, d, dd, relp,
           updatestat, feasible, setp, error, errstr, outf, lf, estage);
    for i:=1 to p do ystat[i] := -ystat[i];
  end
  else SimStats(p,theta,allstats,varmat,fseed,ystat,d,dd,
                updatestat,feasible,error,errstr,outf,lf);
 (*MS: posterior predictive checks:*)
  if pc then SimStats(p,theta,allstats,varmat,fseed,ystat,d,dd,
                updatestat,feasible,error,errstr,outf,lf);
end; (* FRAN *)



function MLModel.logLR(const p: Byte; const theta0, theta1, m0: pVec; var steps: integer;
                 var errstr: shortstring; var error: Boolean;
                 var outf, lf: TFileStream): real;
(* Calculates a MC approximation of twice the log likelihood ratio
   by bridge sampling from theta0 to theta1.
   It is assumed that the process already has burnt in for theta0;
   simulations go in nstep steps from theta0 to theta1. *)
var i,j,fseed,stepsmade  : integer;
    relp       : pmatrix;
    allstats, varmat     : boolean;
    ystat, theta, dif : pVec;
    d, ls, x   : array of real;
    dd         : pmatrix;
    updatestat : Boolean;
    feasible, lpc: Boolean;
    estage     : byte;
    llr, s     : real;
    estring: shortstring;
    xx: string;
begin
  errstr := '';
  error := false;
  logLR := 0.0;
  lpc := false;
  if (observations >= 2) then
    begin
    if (not MadeChain) then
      begin
      errstr:= 'No chain is available in logLR';
      error := true;
      exit;
      end;
    if not (MaxLike or Bayesian) then
      begin
      errstr:= 'logLR called but estimation is not according to ML or Bayes';
      error := true;
      exit;
     end;
   end;
  varmat := false;
  allstats := false;
  estage := 0;
  fseed := readstate;
  if (steps <= 1) then steps := 2;
  EstForm.EPhase.Text:=' 4';
  EstForm.EPhase.hint:='phase 4 is for estimation of the likelihood ratio';
  EstForm.EIteration.hint := 'iterations in phase 4 run from 1 to '+st(steps);
  EstForm.ESubphase.visible := false;
  EstForm.LSubphase.visible := false;
  EstForm.QAutocor.Visible := false;
  EstForm.Label1.Visible := false;
  EstForm.ParameterValues.Hint := 'parameters in phase 4 form path from estimates to initial values';
  EstForm.Statistics.Hint := 'average of simulated scores times parameter differences is log likelihood ratio';
  EstForm.progressbar.hint := '';

  setlength(ls,2*(steps+1));
  setlength(x,2*(steps+1));
  for i:=0 to steps do
    begin
    x[2*i] := i/(steps);
    x[(2*i)+1] := i/(steps);
    end;
  for i:=1 to p do dif[i] := (theta1[i] - theta0[i]);
  llr := 0.0;
  stepsmade := 0;
  EstForm.EPhase.Text:=' 4';
  EstForm.EIteration.hint := 'iterations in phase 4 run from 1 to '+st(steps);
  EstForm.ESubphase.visible := false;
  EstForm.LSubphase.visible := false;
  EstForm.QAutocor.Visible := false;
  EstForm.Label1.Visible := false;
  EstForm.hint := 'SIENA calculates likelihood ratio';
  repeat
    begin
    j := stepsmade;
    for i:=1 to p do theta[i] := (j*theta1[i] + (steps - j)*theta0[i])/steps;
    InitialiseAlpha(outf,theta,fseed);

    FRAN(p, theta, allstats, varmat, fseed, ystat, d, dd,
         updatestat, feasible, error, errstr, outf, lf, estage, lpc);
    if (observations <=1) then for i:=1 to p do ystat[i] := m0[i]-ystat[i];

    if (observations >= 2) or (stepsmade mod 10 = 0) then
     for i:=1 to p do
      begin
      estring := '  '+inttostr(j);
      if (j < 10) then estring := '  '+estring;
      if (j < 100) then estring := '  '+estring;
      EstForm.EIteration.Text:= estring;
{$if defined(batch)}
      Writeln('Iteration ',j);
{$ifend}
      NiceString(theta[i],estring);
      if (i <= 9) then estring := ' '+estring;
      if (i <= 30)
      then EstForm.ParameterValues.lines[i-1] := st(i)+'. '+estring;
      NiceString(ystat[i],estring);
      if (i <= 9) then estring := ' '+estring;
      if (i <= 30) then
        EstForm.Statistics.lines[i-1] := st(i)+'. '+estring;
      end;
    ls[2*j] := 0.0;
    for i:=1 to p do ls[2*j] := ls[2*j] - dif[i]*ystat[i];
    llr := llr + ls[2*j];

    FRAN(p, theta, allstats, varmat, fseed, ystat, d, dd,
         updatestat, feasible, error, errstr, outf, lf, estage, lpc);
    if (observations <=1) then for i:=1 to p do ystat[i] := m0[i] - ystat[i];

    ls[(2*j)+1] := 0.0;
    for i:=1 to p do ls[(2*j)+1] := ls[(2*j)+1] - dif[i]*ystat[i];
    llr := llr + ls[(2*j)+1];
//    writeln(outf,j:3,'a. ',ls[2*j]:10:2);
//    writeln(outf,j:3,'b. ',ls[2*j+1]:10:2);
    Application.ProcessMessages; (* this may return a changed stopo; see Siena_7 StopKeyClick *)
    if stopo then
      begin
      xx := format('%s%d%s',['User interrupted likelihood ratio calculations after ',j,' steps.'])+eol;
      outf.Write(pchar(xx)^,length(xx));
      xx := 'Therefore, the likelihood ratio is not reliable.'+eol;
      outf.Write(pchar(xx)^,length(xx));
      end;
    inc(stepsmade);
    end;
   until ((stepsmade >= steps) or stopo);
  if (stepsmade >= 10) then
    begin
    logLR := llr/(stepsmade);
    xx := format('%s%d%s',['The estimated log-likelihood ratio is calculated in ',stepsmade,' steps, for parameters '])+eol;
    outf.Write(pchar(xx)^,length(xx));
    writevec(outf,p,theta0);
    xx := 'to'+ eol;
    outf.Write(pchar(xx)^,length(xx));
    writevec(outf,p,theta1);
    outf.Write(pchar(eol)^,length(eol));//writeln(outf);
    lf.Write(pchar(eol)^,length(eol));//writeln(lf);
    xx := 'Simulation outcomes for calculation of log likelihood ratio:'+eol;
    outf.Write(pchar(xx)^,length(xx));
    plot(lf,x,ls,nil,2*(stepsmade),'path','complete-data score');
    lf.Write(pchar(eol)^,length(eol));//writeln(lf);
    end;
  EstForm.hint := 'SIENA';
end; (* logLR *)


procedure MLModel.TryOut(var estimates, standard_errors: pvec;
                         var outf, lf: TFileStream);
var i,j,kk,h, nummax, tnm, per: integer;
  k,k1: TMiniSTep;
  a: TAspect;
  v: byte;
  d: TDif;
  ppar,pfun: byte;
  cf: TFileStream;
  theta: pvec;
  num: longint;
  cmu, csigma2, ckappa, maxdev, de: real;
  pr,prpr: array of real;
  time0: TDateTime;
  x: string;
begin
  Heading(1,outf,'Maximum Likelihood Estimation.');
  x := DateTimeToStr(Now)+eol;
  outf.write(pchar(x)^,length(x));
  outf.write(pchar(eol)^,length(eol));//writeln(outf);
  lf.write(pchar(eol)^,length(eol));//writeln(lf);
  x := 'SIENA10'+eol;
  lf.write(pchar(x)^,length(x));
  x := 'Date and time: '+DateTimeToStr(Now)+eol;
  lf.write(pchar(x)^,length(x));
  Heading(1,lf,'ML Estimation.');
  //assignfile(cf,pname+'.cck');
  //rewrite(cf);
  if (not fileExists(pname+'.cck'))
  then begin
  //rewrite(cf)
    cf := TFileStream.Create(pname+'.cck',fmCreate);
    cf.Free;
    cf := TFileStream.Create(pname+'.cck',fmOpenWrite or fmShareDenyWrite);
  end
  else
  begin
    cf:= TFileStream.Create(pname+'.cck',fmOpenReadWrite or fmShareDenyWrite);
    cf.size := 0;
  end;
  x := 'This file serves to check what happens within the'+eol;
  cf.Write(pchar(x)^,length(x));
  x := 'ML estimation algorithm of SIENA.'+eol;
  cf.Write(pchar(x)^,length(x));
  x := 'It is rewritten in each estimation procedure,'+eol;
  cf.Write(pchar(x)^,length(x));
  x := 'so it contains only the checks for the most recent procedure.'+eol;
  cf.Write(pchar(x)^,length(x));
  x := 'Project '+pname+'; estimation algorithm started '+DateTimeToStr(Now)+' .'+eol;
  cf.Write(pchar(x)^,length(x));
  cf.Write(pchar(eol)^,length(eol));//writeln(cf);
  time0 := now;

  for i:=1 to pmax do begin
    estimates[i] := 999.0;
    standard_errors[i] := 99999.0;
   end;

//  if InitialStandard then InitialEstimate;
  ppar := NumberParameters;
  pfun := NumberFunctions;
  MLInitialise(ppar,theta,outf,lf);

  if (errortext > '') then begin
    x := 'Error noted in MLInitialise called from TryOut.'+eol;
    outf.write(pchar(x)^,length(x));
    x := 'Error noted in MLInitialise called from TryOut.'+eol;
    lf.write(pchar(x)^,length(x));
    x := 'errortext = '+errortext+eol;
    outf.write(pchar(x)^,length(x));
    MLFinalizePeriods(outf,lf);
    cf.free;
    exit;
   end;

  x := format('%s%d%s%s%s',['Model Type ',Modeltype,': ',ModelTypeString,'.'])+eol;
  outf.write(pchar(x)^,length(x));
  x := 'Maximum Likelihood estimation.'+eol;
  outf.write(pchar(x)^,length(x));
  x := 'Distances for simulations are'+eol;
  outf.write(pchar(x)^,length(x));
  x := 'period   : '+eol;
  outf.write(pchar(x)^,length(x));
  for i:=1 to observations-1 do if (not skipi[i]) then begin
    x := format('%4d%s',[i,' ']);
     outf.write(pchar(x)^,length(x));
  end;
  outf.write(pchar(eol)^,length(eol));//writeln(outf);
  x := 'distance : ';
  outf.write(pchar(x)^,length(x));
  for i:=1 to observations-1 do if (not skipi[i]) then begin
    x := format('%s%4d',[' ',ObservedNetworkDistance[1,i]]);
    outf.write(pchar(x)^,length(x));
  end;
  x := ' .'+eol;
  outf.write(pchar(x)^,length(x));
  x := 'Time duration for simulations ';
  outf.write(pchar(x)^,length(x));
  if (observations >= 3) then begin
    x := 'in each period ';
    outf.write(pchar(x)^,length(x));
  end;
  x := 'is 1.0.'+eol; (* see FRAN/SimStats in S_Base *)
  outf.write(pchar(x)^,length(x));

  x := 'Initial parameter values are '+eol;
  outf.write(pchar(x)^,length(x));
  for i:=1 to NumberParameters do begin
    x := format('%4d%s%s%s%10.5f',[i,'. ',alignstring(Description[i],50),' ',theta[i]])+eol;
    outf.write(pchar(x)^,length(x));
//    if fix[i] then writeln(outf,'  (fixed)') else writeln(outf);
   end;
  outf.write(pchar(eol)^,length(eol));//writeln(outf);

  if ConsistencyCheck(outf,lf,true) then begin
    x := 'Consistency check 2 OK.'+eol;
    outf.write(pchar(x)^,length(x));
    for per :=1 to observations-1 do  if (not skipi[per]) then WritefullPeriod(per,outf,lf);
   end else begin
    x := 'Consistency check 2 not OK. ErrorText = '+errortext+eol;
    outf.write(pchar(x)^,length(x));
//    for per :=1 to observations-1 do WriteFullPeriod(per,outf,lf);
   end;


  outf.write(pchar(eol)^,length(eol));//writeln(outf);
  if (errortext > '') then begin
    x := errortext+eol;
    outf.write(pchar(x)^,length(x));
  // writeln(lf,errortext);
    lf.write(pchar(x)^,length(x));
    x := 'Program stops.'+eol;
    outf.write(pchar(x)^,length(x));
    writefullperiod(per, outf,lf);
    cf.free;
    MLFinalizePeriods(outf,lf);
    exit;
   end else begin
    x := 'Final consistency check OK.'+eol;
     outf.write(pchar(x)^,length(x));
   end;
  outf.write(pchar(eol)^,length(eol));//writeln(outf);


(*  FunctionEvaluations(ordgs,halter,outf,lf,targets);
  if halter then begin
    closefile(cf);
    MLFinalizePeriods(outf,lf);
    exit;
   end; *)
//  if VertexPar then InitialiseVertrarrays;


 for per :=1 to observations-1 do  if (not skipi[per]) then
  begin
  initialise_actors(per);
  Renewrateslprobs(Chain[per-1].First,Chain[per-1].Last,false,per,outf,lf);
  x := format('%s%20.6f%s%20.6f%s%20.6f',['kappa  = ',Chain[per-1].kappa ,
     '; mu  = ',Chain[per-1].mu,'; sigma2  = ',Chain[per-1].sigma2])+eol;
  outf.write(pchar(x)^,length(x));
  Renewrateslprobs(Chain[per-1].First,Chain[per-1].Last,true,per,outf,lf);
  k := Chain[per-1].First;
  cmu := 0.0;
  csigma2 := 0.0;
  repeat
    k := k.such;
    cmu := cmu + (1/k.rate);
    csigma2 := csigma2 + (1/sqr(k.rate));
   until k.Extreme;
  ckappa := -0.5*log(csigma2) - sqr(chain[per-1].totaltime - cmu)/(2.0*csigma2);
  x := format('%s%20.6f%s%20.6f%s%20.6f',['ckappa = ',ckappa,'; cmu = ',cmu,'; csigma2 = ',csigma2])+eol;
  outf.write(pchar(x)^,length(x));
  if (abs(ckappa - Chain[per-1].kappa) > 1e-20) then begin
    x := '***** error ! kappa is not correct.'+eol;
    outf.write(pchar(x)^,length(x));
  end;
  x := 'Renewrateslprobs'+eol;
  outf.write(pchar(x)^,length(x));
  x := format('%s%20.6f%s%20.6f%s%20.6f',['kappa  = ',Chain[per-1].kappa,
        '; mu  = ',Chain[per-1].mu,'; sigma2  = ',Chain[per-1].sigma2])+eol;
        outf.write(pchar(x)^,length(x));
  if (abs(ckappa - Chain[per-1].kappa) > 1e-20) then begin
    x  := '***** error ! kappa is not correct.'+eol;
    outf.write(pchar(x)^,length(x));
  end;

(* [[[ *)
  ninserts := 0;
  nnoinserts := 0;
  ninsds := 0;
  nnoinsds := 0;
  nimpinsdels := 0;
  nstep := 0;
  ncancels := 0;
  npermtots := 0;
  nnopermtots := 0;

  outf.write(pchar(eol)^,length(eol));//writeln(outf);
  x := 'Metropolis Hastings steps '+eol;
  outf.write(pchar(x)^,length(x));
  x := 'Probabilities for types of MH step: b'+eol;
  outf.write(pchar(x)^,length(x));
  if (nza > 0) then begin
    x := format('%s%6.3f',['Network               ',prnet])+eol;
    outf.write(pchar(x)^,length(x));
  end;
  x := format('%s%6.3f',['Insertion - deletion      ',prinsdel])+eol;
  outf.write(pchar(x)^,length(x));
  x := format('%s%6.3f',['Permutation               ',prpermute])+eol;
  outf.write(pchar(x)^,length(x));
  x := format('%s%6.3f',['Diagonal insert - delete  ',1.0 - cprpermute])+eol;
  outf.write(pchar(x)^,length(x));
  x := format('%s%6.3f',['Relative insertion        ',prelins])+eol;
  outf.write(pchar(x)^,length(x));
  x := format('%s%6.3f',['Extra multiples           ',promul])+eol;
  outf.write(pchar(x)^,length(x));
(* ]]] *)
  num := 0;
  repeat
    inc(num);
//    write(outf,'MH_Step ',num:4,'. ');
//    if (num mod 20 = 0) then writeln(outf);
    MH_Step(per,outf,lf,0);
//    writeln(outf,Chain[per-1].Totnumber);
   until (num >= 1000) or (errortext > '');
  RenewRateslprobs(Chain[per-1].First,Chain[per-1].Last,true,per,outf,lf);
  outf.write(pchar(eol)^,length(eol));//writeln(outf);
  if (errortext > '') then begin
    x := 'Error: '+errortext+eol;
    outf.write(pchar(x)^,length(x));
  end;
(* [[[ *)
  x := format('%10d%s%10d%s%10d%s',[ninserts,' inserts;          ',nnoinserts,' refused inserts;         ',
               nimpinsdels, ' impossible insdels.'])+eol;
  outf.write(pchar(x)^,length(x));
  x := format('%10d%s%10d%s',[ninsds   ,' diagonal inserts; ',nnoinsds,  ' refused diagonal inserts.'])+eol;
  outf.write(pchar(x)^,length(x));
  x := format('%10d%s%10d%s%10d%s',[ncands   ,' diagonal cancels; ',nnocands,  ' refused diagonal cancels;',
               nimpcands,' impossible diagonal cancels;'])+eol;
  outf.write(pchar(x)^,length(x));
  x := format('%10d%s',[nstep, ' steps'])+eol;
  outf.write(pchar(x)^,length(x));
  x := 'permutations of chain parts of various lengths: '+eol;
  outf.write(pchar(x)^,length(x));
  for i:=0 to (length(npermutes)-1) do if (ntrypermutes[i] > 0) then
    begin
    x := format('%s%3d%s%6.4f%s%10d%s',[' length ',i,': ',npermutes[i]/ntrypermutes[i],' executed out of ',ntrypermutes[i],' trials.'])+eol;
    outf.write(pchar(x)^,length(x));
    end;
(* ]]] *)
  end;

 for per :=1 to observations-1 do if (not skipi[per]) then begin
  initialise_actors(per);
  outf.write(pchar(eol)^,length(eol));//writeln(outf);
  ninserts := 0;
  nnoinserts := 0;
  ninsds := 0;
  nnoinsds := 0;
  ncands := 0;
  nnocands := 0;
  nimpcands := 0;
  nimpinsdels := 0;
  ncancels := 0;
(* [[[ *)
  npermtots := 0;
  nnopermtots := 0;
(* ]]] *)
  nstep := 0;
  outf.write(pchar(eol)^,length(eol));//writeln(outf);
  x := 'Metropolis Hastings steps with checks for inverse steps'+eol;
  outf.write(pchar(x)^,length(x));
  maxdev := 0.0;
  num := 0;
  repeat
    inc(num);
//    writeln(outf,'MH_CheckStep ',num:6,'. ');
    MH_Step(per,outf,lf,0);
    if (de > maxdev) then maxdev := de;
    if (de > 0.0001) then  WritePeriod(per,outf,lf);
    until (num >= 1000) or (errortext > '') or (de > 0.0001);
    RenewRateslprobs(Chain[per-1].First,Chain[per-1].Last,true,per,outf,lf);

 outf.write(pchar(eol)^,length(eol));// writeln(outf);
//WriteFullPeriod(outf,lf);
  x := format('%10d%s%10d%s%10d%s',[ninserts ,' inserts;          ',nnoinserts,' refused inserts;',
               nimpinsdels , ' impossible insdels.'])+eol;
  outf.write(pchar(x)^,length(x));
  x := format('%10d%s%10d%s',[ncancels ,' cancels;          ',nnocancels,' refused cancels;'])+eol;
  outf.write(pchar(x)^,length(x));
  x := format('%10d%s%10d%s',[ninsds   ,' diagonal inserts; ',nnoinsds,  ' refused diagonal inserts.'])+eol;
  outf.write(pchar(x)^,length(x));
  x := format('%10d%s%10d%s%10d%s',[ncands   ,' diagonal cancels; ',nnocands,  ' refused diagonal cancels;',
               nimpcands,' impossible diagonal cancels;'])+eol;
  outf.write(pchar(x)^,length(x));
  x := format('%10d%s%10d%s',[npermtots,' permutations;     ',nnopermtots,' refused permutations;'])+eol;
  outf.write(pchar(x)^,length(x));
  x := format('%s%15.12f',['Maximal deviation is ',maxdev])+eol;
  outf.write(pchar(x)^,length(x));

  outf.write(pchar(eol)^,length(eol));//writeln(outf);
  Renewrateslprobs(Chain[per-1].First,Chain[per-1].last,true,per,outf,lf);
  outf.write(pchar(eol)^,length(eol));//writeln(outf);
  Writeperiod(per,outf,lf);
  outf.write(pchar(eol)^,length(eol));//writeln(outf);


{  writeln(outf,'New placements of the same element a number of times. ');
  kk := 0;
  SetLength(pr,Chain[per-1].TotNumber+1);
  SetLength(prpr,Chain[per-1].TotNumber+1);
  repeat
//    write(outf,kk:4,'.');
    k:= Chain[per-1].RandomElement(version);
    if (k = chain[per-1].last) then k := Chain[per-1].First.such.such.such;
    i := k.ih; j := k.jh;
//    k.Writekind(outf);       writeln(outf);
    num := MH_NewPlace(k,per,outf,lf).kindex;
    if (errortext > '') then begin
      writeln(outf,errortext);
//      Writeperiod(per,outf,lf);
      writeln(outf);
     end;
//    writeln(outf,' newly placed to before ',num);
// the following was used with a version of MH_NewPlace that outputs an array
// called as pr with the cumulative probabilities.
//    if (kk >= 1) then begin
//      maxdev := 0.0;
//      for h:=0 to available do if (inverseused[h] <= totnumber) and (abs(prpr[h] - pr[h]) > maxdev)
//        then begin
//          maxdev := abs(prpr[h] - pr[h]);
//         end;
//      writeln(outf,'maxdev = ',maxdev:20:15);
//     end;
//    for h:=0 to available do prpr[h] := pr[h];
    if (errortext > '') then begin                                                                             writeln(outf,'error during tryout of new placements : ',errortext);
      kk := kk + 1000;
     end else begin
      k1 := CHain[per-1].Findexstep[num].predh;
// writeln(outf,'New index is ',k1.kindex,' (',k1.KindString,')');
     end;
    inc(kk);
  until (kk >= 10);

  if (errortext > '') then Writeperiod(per,outf,lf);
}

 end; (* for per *)

  if ConsistencyCheck(outf,lf,true) then begin
    x := 'Consistency check 3 OK.'+eol;
    outf.write(pchar(x)^,length(x));
   end else begin
    x := '***** Consistency check 3 not OK. ErrorText = '+errortext+eol;
    outf.write(pchar(x)^,length(x));
   end;
  for per := 1 to observations-1 do if (not skipi[per]) then WritePeriod(per,outf,lf);

  outf.write(pchar(eol)^,length(eol));//writeln(outf);
  time0 := 86400*(now - time0);
  x := format('%s%8.2f%s',['Total computation time ',time0,' seconds.'])+eol;
  outf.write(pchar(x)^,length(x));

  cf.free;
  MLFinalizePeriods(outf,lf);
end; (* TryOut *)

procedure MLModel.ExtraMLResults(var outf: TFileStream);
var i : integer;
x: string;
begin
  x := 'Probabilities for types of MH step (latent paths):'+eol;
  outf.Write(pchar(x)^,length(x));
  if (nza > 0) then begin
    x := format('%s%6.3f',['Network                   ',prnet])+eol;
    outf.Write(pchar(x)^,length(x));
  end;
  x := format('%s%6.3f',['Insert - delete           ',prinsdel])+eol;
  outf.Write(pchar(x)^,length(x));
  x := format('%s%6.3f',['Permutation               ',prpermute])+eol;
  outf.Write(pchar(x)^,length(x));
  x := format('%s%6.3f',['Diagonal insert - delete  ',1.0 - cprpermute])+eol;
  outf.Write(pchar(x)^,length(x));
  x := format('%s%6.3f',['Relative insertion        ',prelins])+eol;
  outf.Write(pchar(x)^,length(x));
  x := format('%s%6.3f',['Extra multiples           ',promul])+eol;
  outf.Write(pchar(x)^,length(x));
  outf.Write(pchar(eol)^,length(eol));//writeln(outf);
  x := format('%10d%s',[nstep, ' steps'])+eol;
  outf.Write(pchar(x)^,length(x));
  x := format('%10d%s%10d%s%10d%s',[ninserts ,' inserts;          ',nnoinserts,' refused inserts;         ',
               nimpinsdels, ' impossible insdels.'])+eol;
  outf.Write(pchar(x)^,length(x));
  x := format('%10d%s%10d%s',[ncancels ,' cancels;          ',nnocancels,' refused cancels;         '])+eol;
  outf.Write(pchar(x)^,length(x));
  x := format('%10d%s%10d%s',[ninsds   ,' diagonal inserts; ',nnoinsds,  ' refused diagonal inserts.'])+eol;
  outf.Write(pchar(x)^,length(x));
  x := format('%10d%s%10d%s%10d%s',[ncands   ,' diagonal cancels; ',nnocands,  ' refused diagonal cancels;',
               nimpcands,' impossible diagonal cancels;'])+eol;
  outf.Write(pchar(x)^,length(x));
  x := format('%10d%s%10d%s',[npermtots,' permutations;',nnopermtots,' refused permutations;'])+eol;
  outf.Write(pchar(x)^,length(x));
  x := 'permutations of chain parts of various lengths: '+eol;
  outf.Write(pchar(x)^,length(x));
  for i:=0 to (length(npermutes)-1) do if (ntrypermutes[i] > 0) then
    begin
    x := format('%s%3d%s%6.4f%s%10d%s',[' length ',i,': ',npermutes[i]/ntrypermutes[i],' executed out of ',ntrypermutes[i],' trials.'])+eol;
    outf.Write(pchar(x)^,length(x));
    end;
  DistributionCovRE(randomeffects,sequences,nre,covfile2,covRE,run); (*MS: Completes Matlab source file to plot sampling distribution of the estimate of the variance-covariance matrix of the random effects.*)
end;



begin (* unit S_ML *)
end.

